/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// @doc     Factory
// @module  Parameter List | This module contains the implementation of the
//          ParamList class and any other derivatives. ParamListImpl is the
//          base class for all Parameter list objects and contains all
//          generic parameter list functionality.
//
//          <p>The classes derived from ParamListImpl simply provide convenient
//          constructors that will initialise the list with a default set of
//          parameters.</p>
//
///////////////////////////////////////////////////////////////////////////////

#include <stdio.h>

///////////////////////////////////////////////////////////////////////////////
// Cyrano includes
#include <Factory/ParamList_i.h>
#include <Factory/Factory.h>

///////////////////////////////////////////////////////////////////////////////
// @mfunc   The destructor empties the array of parameters before tidying it
//          up.
ParamListImpl::~ParamListImpl()
{
    // Tidy memory
    m_array.clear();
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method retrieves a requested parameter from the array of
//          parameters.  If the value is not found then an empty string will
//          be returned.
// @rdesc   The stringified parameter value.
// @parm    The name of the paraamter we want out of the array.
char* ParamListImpl::GetValue( const char* name )
{
    return CORBA::string_dup( m_array[name].c_str() );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method adds a new parameter value to the array.  If the
//          parameter value already exists within the array then the value will
//          be replaced.
// @parm    The name of the parameter value to add.
// @parm    The parameter value as a string.
void ParamListImpl::AddValue( const char* name, const char* value )
{
    // If value already exists then dump it
    string key( name );
    ParamValues::iterator i=m_array.find(key);
    if(i==m_array.end())
        m_array.erase( key );

    // Add value to array
    m_array.insert(ParamValues::value_type(key, string(value)) );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This constructor simply calls the base constructor before adding
//          the three default values required by the Factory to create an
//          object.
// @parm    The name of the class to create.
// @parm    The name of the object to register.
// @parm    The number of periods that the object will be kept alive for.
FactoryParamList::FactoryParamList( const char* cls, const char* name,
                                    long period ) : ParamListImpl()
{
    // Add minimum required values to list
    AddValue( CyrFactory::PARAM_CLASS, cls );
    AddValue( CyrFactory::PARAM_NAME, name );

    // Convert integer to a string and add it to minimum required list as well
    char buf[20];
    sprintf( buf, "%d", period );
    AddValue( CyrFactory::PARAM_PERIOD, buf );
}

