/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// @doc     CyrStRepository
// @module  File Iterators | This module contains the implementation of
//          the CyrStFileIt CORBA object.  This CORBA object is a mechanism
//          for transferring file data in small and easy to handle chunks.
//
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// Cyrano includes
#include <Repository/CyrStFileIt_i.h>
#include <CyrException.h>
#ifdef WIN32
#include <io.h>
#else
   #ifndef SUN
	#include <sys/io.h>
   #endif	
#endif
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifdef __VDK002__
#include <Logging/CECLog.h>
extern CECLog* v2logger;
#endif

const char* CyrStFileItImpl::CLASS_NAME = "CyrStFileIt";

///////////////////////////////////////////////////////////////////////////////
// @mfunc   The constructor is responsible for opening the file to be iterated
//          through.  This code is cross-platform due to the use of RogueWave
//          file classes.
// @parm    The full path name of the file to be opened.
CyrStFileItImpl::CyrStFileItImpl( const char* filename )
{
    // And add some tracing
#ifdef __VDK002__
    string msg = "Constructing a CyrStFileItImpl (File Iterator) for: ";
    msg += filename;
    v2logger->AddTrace( CYR_VDK002, 50,  msg.c_str() );
#endif

    // Error information
    const char* where = "CyrStFileItImpl::CyrStFileItImpl";

    //Determine if file exists and check size
    struct stat st;
    int result=::stat(filename,&st);
    // Determine if file exists
    if ( result != 0)
        throw CyrRepositoryException( "File does not exist", where, CYR_FILE_NOT_FOUND );
    // get size of file
    m_size = st.st_size;
    // Create new file
#ifdef WIN32
    m_filedesc = open( filename, _O_BINARY|_O_RDONLY);
#else
    m_filedesc = open( filename, O_RDONLY);
#endif
    if (m_filedesc==-1)
        throw CyrRepositoryException( "Failed to open file", where, CYR_FILE_NOT_FOUND );


}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   The destructor ensures that the file is closed.
CyrStFileItImpl::~CyrStFileItImpl()
{
    // And add some tracing
#ifdef __VDK002__
    v2logger->AddTrace( CYR_VDK002, 50, "Deleting a CyrStFileItImpl (File Iterator)" );
#endif

    // Tidy memory
    close(m_filedesc);
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns a pointer to a sequence of bytes from the
//          file we are currently itearting through.  The maximum size of the
//          sequence is defined and iteration is done by repeated requests
//          to this method.  A sequence that has no elements indicates that the
//          iteration is complete.
// @rdesc   Pointer to the sequence of bytes.
ByteSeq* CyrStFileItImpl::ReadBytes()
{
    // And add some tracing
#ifdef __VDK002__
    v2logger->AddTrace( CYR_VDK002, 100, "Reading the File" );
#endif

    // Determine size of file to read - limiting it to the maximum we
    // have specified.  If we have read to the end of the file then we
    // return an empy sequence so that the iteration can be stopped.
    int count = m_size - tell(m_filedesc);
    if ( count == 0 )
    {
        // Create and return empty sequence
        ByteSeq_var bySeq = new ByteSeq();
        return bySeq._retn();
    }

    if ( count > BYTESEQ_SIZE )
        count = BYTESEQ_SIZE;

    // Allocate new buffer for bytes
    CORBA::Octet* buf = ByteSeq::allocbuf( count );

    // Read characters from file
    read(m_filedesc, buf, count );

    // Create sequence of bytes for transmission
    ByteSeq_var bySeq = new ByteSeq( count, count, buf, 1 );

    // Return reference and inform client that it has responsiblity for
    // deallocating memory.  Code copied from Henning and Vinoski pg 423.
    return bySeq._retn();
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns the total size of the file in bytes.
// @rdesc   The file size in bytes.
long CyrStFileItImpl::GetSize()
{
    return m_size;
}

long CyrStFileItImpl::ReadLong()
{
    long data;
    read(m_filedesc,&data ,sizeof(long));
    return data;
}

short CyrStFileItImpl::ReadShort()
{
    short data;
    read(m_filedesc,&data,sizeof(short) );
    return data;
}

CORBA::Octet CyrStFileItImpl::ReadByte()
{
    CORBA::Octet data;
    read(m_filedesc,&data,sizeof(CORBA::Octet) );
    return data;
}

CORBA::Boolean CyrStFileItImpl::ReadBool()
{
    CORBA::Boolean data;
    read(m_filedesc,&data,sizeof(CORBA::Boolean ) );
    return data;
}

