/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// @doc     CyrStRepository
// @module  File Implementation | This module contains the implementation of
//          the CyrStFile CORBA object.  This CORBA object is simply a CORBA
//          wrapper around the OpenSTARep COM interface IFileEx.
//
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// Cyrano includes
#include <Repository/CyrStFile_i.h>
#include <Repository/CyrStFileIt_i.h>
#include <CyrException.h>

#ifdef __VDK002__
#include <Logging/CECLog.h>
extern CECLog* v2logger;
#endif


///////////////////////////////////////////////////////////////////////////////
// @mfunc   This constructor extracts the relevant parameters from the
//          parameter list and calls the correct initialisation function.
// @parm    Pointer to parent Provider module
// @parm    List of parameters to be used to construct File object
CyrStFileImpl::CyrStFileImpl( CyrStProvider* pProvider, ParamList_ptr params )
        : CyrStLinkImpl( pProvider )
{
    // And add some tracing
#ifdef __VDK002__
    v2logger->AddTrace( CYR_VDK002, 50, "Constructing CyrStFileImpl" );
#endif

    const char* where = "CyrStFileImpl::CyrStFileImpl";

    // Assign member variables
    m_itFile = NULL;

    // Retrieve common parameters
    CORBA::String_var name = params->GetValue( CyrFactory::PARAM_NAME );
    CORBA::String_var folder = params->GetValue( CyrStFile::PARAM_FOLDER );
    CORBA::String_var desc = params->GetValue( CyrStFile::PARAM_DESC );
    if ( strlen( folder ) == 0 )
        throw CyrRepositoryException( "Invalid folder parameter", where, CYR_FOLDER_NOT_FOUND );
    if ( strlen( name ) == 0 )
        throw CyrRepositoryException( "Invalid name parameter", where, CYR_FILE_NOT_FOUND );

    // Locate correct Server and Folder interface
    CECFFFolder* pFolder = (CECFFFolder*)m_provider->GetPtr( folder.in() );
    CECFFServer* pServer = pFolder->GetParentServer();

    // Determine if File already exists and if so retrieve it
    if ( strlen( desc ) > 0 )
    {
        // FUDGE: OpenSTARep no longer allows you to add files that do not exist.
        //        Therefore we simply create an empty file before adding the
        //        new one.
        string filename( pFolder->GetFullPath() );
        filename += name;

        // And add some tracing
#ifdef __VDK002__
        string msg="Trying to create file: " + filename;
        v2logger->AddTrace( CYR_VDK002, 100, msg.c_str() );
#endif

        std::ofstream* pFile = new std::ofstream( filename.c_str());
        delete pFile;
        // FUDGE

        m_com = new CECFFFile( pFolder, name, desc );
    }
    else
    {
        // And add some tracing
#ifdef __VDK002__
        string msg("Trying to retrieve file: ");
        msg+=name;
        v2logger->AddTrace( CYR_VDK002, 100, msg.c_str() );
#endif
        m_com = new CECFFFile( pFolder, name );

#ifdef __VDK002__
        v2logger->AddTrace( CYR_VDK002, 100, "Retrieved COM file" );
#endif

    }

    string user;

    try
    {
        if (pServer == NULL)
        {
#ifdef __VDK002__
            v2logger->AddTrace( CYR_VDK002, 100, "pServer is NULL" );
#endif
        }

        // Calculate current user name
        user = pServer->GetUserName();
    }
    catch (...)
    {
#ifdef __VDK002__
        v2logger->AddTrace( CYR_VDK002, 100, "Caught Exception when obtaining username" );
#endif
    }
#ifdef __VDK002__
    string msg="Current User is: " + user;
    v2logger->AddTrace( CYR_VDK002, 100, msg.c_str() );
#endif

    // Store pointer with parent Provider
    m_regName = BuildItemPath( m_com ) + string((const char*)name);
    m_regName = m_provider->StorePtr( m_com, m_regName.c_str(), user.c_str() );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   The destructor destroys the underlying COM object as part of its
//          tidying up.
CyrStFileImpl::~CyrStFileImpl()
{
    // And add some tracing
#ifdef __VDK002__
    v2logger->AddTrace( CYR_VDK002, 50, "Disposing of CyrStFileImpl" );
#endif

    // Tidy up any allocated memory
    if ( m_itFile != NULL )
        m_itFile->_remove_ref();
    delete m_com;
    m_com = NULL;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns the name of the File as recognised by
//          OpenSTARep .  It should be the same name as the one passed into the
//          constructor and therefore should not be needed.  However, it has
//          been added here for completeness.
// @rdesc   The name of the File.
char* CyrStFileImpl::GetName()
{
    // And add some tracing
#ifdef __VDK002__
    v2logger->AddTrace( CYR_VDK002, 100, "Get File Name" );
#endif

    // Retrieve correct interface from array
    return CORBA::string_dup( m_com->GetName().c_str() );
}

/*long CyrStFileImpl::GetCurrentStatus()
{
    // And add some tracing
#ifdef __VDK002__
    v2logger->AddTrace( CYR_VDK002, 100, "Get Current File Status." );
#endif
 
    return m_com->GetCurrentStatus();
}*/

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method locks the remote File so that no other user can modify
//          it.  This method should be called before updating any changes to
//          the File.
// @rdesc   Returns true if the lock request was successful, otherwise false.
bool CyrStFileImpl::Lock()
{
    // And add some tracing
#ifdef __VDK002__
    v2logger->AddTrace( CYR_VDK002, 100, "Locking the file." );
#endif

    // Retrieve correct COM interface from array
    return m_com->Lock();
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method removes a lock placed previously by this user.
void CyrStFileImpl::Unlock()
{
    // And add some tracing
#ifdef __VDK002__
    v2logger->AddTrace( CYR_VDK002, 100, "Unlocking the file." );
#endif

    // Retrieve correct COM interface from array
    m_com->Unlock();
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method retrieves a file from the local machine and creates a
//          file iterator from it.  This iterator can then be used to transmit
//          the File to a remote machine.
// @rdesc   Reference to a file iterator.
CyrStByteIt_ptr CyrStFileImpl::GetFile( bool lock )
{
    // And add some tracing
#ifdef __VDK002__
    v2logger->AddTrace( CYR_VDK002, 100, "Accessing local file." );
#endif

    // Tidy up an existing iterators
    if ( m_itFile != NULL )
        m_itFile->_remove_ref();

    // Create new file iterator
    m_com->CheckOut( lock );
    m_itFile = new CyrStFileItImpl( m_com->GetFullPath().c_str() );
    m_provider->GetHelper()->ActivateObject( m_itFile );

    // Return reference to new iterator
    return m_itFile->_this();
}

////////////////////////////////////////////////////////////////////////////////
// @mfunc   This method tranfers a sequence of bytes from a remote location and
//          stores them in a file.  The file is then checked into OpenSTARep as
//          the latest version.
// @parm    Reference to a byte sequence iterator.
// @parm    If true then releases an exclusive lock on the file.  Otherwise leaves
//          the lock status unchanged.
// @parm    If true the file is checked into OpenSTARep
void CyrStFileImpl::PutFile( CyrStByteIt_ptr itFile, bool release)
{
    // And add some tracing
#ifdef __VDK002__
    string msg="Creating local file: " + m_com->GetFullPath();
    v2logger->AddTrace( CYR_VDK002, 100, msg.c_str() );
#endif

    // Create temporary file for write access
    fstream* f = new fstream( m_com->GetFullPath().c_str(), std::ios::binary|std::ios::out );
    // Copy entire file contents to a temporary location
    ByteSeq_var bytes = itFile->ReadBytes();
    while( bytes->length() != 0 )
    {
        // Write contents of sequence to file
        f->write( (char*)bytes->get_buffer(), bytes->length() );

        // Get next set of data
        bytes = itFile->ReadBytes();
    }

    // Close file before transferring to correct location
    delete f;

    // Check file into OpenSTARep
    m_com->CheckIn( release );
}

