/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

#ifndef __SOCKET__
#define __SOCKET__

//==================================================================< Pragmas >
#ifdef  WIN32
#  pragma warning(disable : 4251 4710)
#endif  // WIN32


#ifdef WIN32
#pragma warning(disable: 4201)
#include <winsock.h>
#pragma warning(default: 4201)
#else
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#ifdef sun
#include <sys/systeminfo.h>
#endif
#include <sys/types.h>
#include <fcntl.h>
#endif

#ifndef __DEFINITION__
#include "definition.h"
#endif



#ifndef __EXCEPTION__
#include <exception.hpp>
#endif

#ifndef __MTSAFE__
#  include <mtsafe.hpp>
#endif  // __MTSAFE__

#ifndef SUN
#ifndef __LOGFILE__
#  include <logfile.hpp>
#endif
#endif

#include <list>
using std::list;


#ifndef PORT_T_TYPEDEF
#define PORT_T_TYPEDEF
typedef unsigned short port_t;
#endif

class DLLSPEC_CYRTOOLS CTInetAddress : virtual public CTMTSafe
{
public:
    static const int iMaxHostNameLen_;

    CTInetAddress();
    CTInetAddress(const CTInetAddress& address);
    CTInetAddress(const string &host);

    short getFamily() const
    {
        return sFamily_;
    }
    long getAddress() const
    {
        return lAddress_;
    }
    void setFamily(short sFamily)
    {
        sFamily_ = sFamily;
    }
    void setAddress(long lAddress)
    {
        lAddress_ = lAddress;
    }
    void setHostName(const string &host)
    {
        hostName_ = host;
    }

    const string& getHostName();
    string getHostAddress() const;

    bool operator== (const CTInetAddress& address) const;
    void operator= (const CTInetAddress& address);

    static CTInetAddress getByName(const string& host);
    static CTInetAddress getByAddr(const string& address);
    static port_t getPortByName(const string& serviceName);
    static string getLocalHostName();
    static string getHostByAddr(long lAddress);

    //---------------------------------------------------------------------------

private:

    string hostName_;
    long lAddress_;
    short sFamily_;

    CTInetAddress(const string &host, long lAddress, short sFamily)
: hostName_(host), lAddress_(lAddress), sFamily_(sFamily)
    {}

};

DLLSPEC_CYRTOOLS ostream& operator<< (ostream& os, const CTInetAddress& inetAddress);
#ifndef SUN
DLLSPEC_CYRTOOLS CTOLogFileList& operator<< (CTOLogFileList& os, const CTInetAddress& inetAddress);
#endif

#define _S_RECEIVE_     0x1
#define _S_SEND_        0x2
#define _S_EXCEPTION_   0x4

#define _S_TIMEDOUT_    0x8

#define _S_ALL_     (_S_RECEIVE_|_S_SEND_|_S_EXCEPTION_)

class DLLSPEC_CYRTOOLS CTSocket : virtual public CTMTSafe
{
public:
    static const int iMaxListenEntries_;
    enum CTCreationMode { DGRAM, STREAM };
    enum CTTermination  { RECEPTION, TRANSMISSION, BOTH };
    enum CTSelection    { RECEIVE, SEND, EXCEPTION, UNKNOWN, TIMEDOUT };

    // Default Constructor
    CTSocket(CTCreationMode eCreationMode = CTSocket::STREAM, bool bReuseAddress = true);

    // Create a CTSocket (default is stream) and connects it to the specified address on the specified port
    CTSocket(const CTInetAddress& address, port_t tPort, CTCreationMode eCreationMode = CTSocket::STREAM, bool bReuseAddress = true);

    // Create a CTSocket (default is stream) and connects it on the specified port on the specified host
    CTSocket(const string& host, port_t tPort, CTCreationMode eCreationMode = CTSocket::STREAM, bool bReuseAddress = true);

    // Copy Constructor
    CTSocket(const CTSocket& socket);

    // Destructor
    ~CTSocket();

    // Allow the socket to be bound to an address that is already in use.
    void reuseAddress (bool bReuseAddress) ;
    void setLinger (const char *, int) ;

    /* Bind this CTSocket to a known port at the specified address. */
    void bindSocket(const CTInetAddress& inetAddress, port_t tPort);
    /* Bind this CTSocket to a known port. */
    void bindToPort(port_t tPort);
    /* Bind this CTSocket to a local address. */
    void bindAnonymously(const CTInetAddress& inetAddress);

    void setNonBlockingMode();
    void setBlockingMode();
    /* Specify how many connections can attempt to connect to this
        CTSocket at once before getting an error. */
    void listen(int iBacklog = CTSocket::iMaxListenEntries_);

    /* Accept a connection on this CTSocket, returning a new one.
        The old CTSocket is still around for listening for new connections. */
    CTSocket accept();

    CTSocket::CTSelection select(timeval& timeout, int iEvent = _S_ALL_);
    CTSocket::CTSelection select(int iEvent = _S_ALL_);
    CTSocket::CTSelection poll(int iEvent = _S_ALL_);

    /* Connect this CTSocket to the specified destination and port. */
    void connect(const CTInetAddress& dest, port_t tPort);
    void connect(const string& host, port_t tPort);

    // Closes the CTSocket
    void close();

    // Terminate gracefully
    void shutdown(CTTermination eTermination);

    // Gets the address to which the CTSocket is connected
    const CTInetAddress& getInetAddress() const
    {
        return address_;
    }

    // Gets the address to which the peer CTSocket is connected
    const CTInetAddress& getPeerInetAddress() const
    {
        return peerAddress_;
    }

    long send(const string& data);
    long recv(string& data);

    long read(string& data);
    long write(const string& data);

    long sendTo(void *pv, int iLen, const CTInetAddress& address, port_t tPort);
    long recvFrom(void *pv, int iLen, CTInetAddress& address, port_t& tPort);

    // General purpose functions.
    long recvBuffer(void* pv, int iLen, const char* zMessage, bool bUrgent = false);
    long sendBuffer(void* pv, int iLen, const char* zMessage, bool bUrgent = false);

    // Standard version.
    long recvAvailableData(void* pBuffer, int iMaxLen, bool bUrgent = false);

    int getFileDescriptor() const
    {
        return iSock_;
    }

    long getString(char*    , size_t tMaxLen);
    long get(char&);
    long get(unsigned char&);
    long get(char*          , size_t);
    long get(unsigned char* , size_t);
    long get(int*           , size_t);
    long get(unsigned int*  , size_t);
    long get(long*          , size_t);
    long get(unsigned long* , size_t);
    long get(short*         , size_t);
    long get(unsigned short*, size_t);
    long get(double*        , size_t);
    long get(float*         , size_t);

    CTSocket& operator>>(char*&);
    CTSocket& operator>>(char&);
    CTSocket& operator>>(unsigned char&);
    CTSocket& operator>>(int&);
    CTSocket& operator>>(unsigned int&);
    CTSocket& operator>>(long&);
    CTSocket& operator>>(unsigned long&);
    CTSocket& operator>>(short&);
    CTSocket& operator>>(unsigned short&);
    CTSocket& operator>>(float&);
    CTSocket& operator>>(double&);
    CTSocket& operator>>(string& data);
    CTSocket& operator>>(bool&);

    long put(char);
    long put(unsigned char);
    long put(const char*            , size_t);
    long put(const unsigned char*   , size_t);
    long put(const short*           , size_t);
    long put(const unsigned short*  , size_t);
    long put(const int*             , size_t);
    long put(const unsigned int*    , size_t);
    long put(const long*            , size_t);
    long put(const unsigned long*   , size_t);
    long put(const float*           , size_t);
    long put(const double*          , size_t);

    CTSocket& operator<<(const char* zStr);     // !!! NULL terminated string !!!
    CTSocket& operator<<(char);
    CTSocket& operator<<(unsigned char);
    CTSocket& operator<<(int);
    CTSocket& operator<<(unsigned int);
    CTSocket& operator<<(long);
    CTSocket& operator<<(unsigned long);
    CTSocket& operator<<(short);
    CTSocket& operator<<(unsigned short);
    CTSocket& operator<<(float);
    CTSocket& operator<<(double);
    CTSocket& operator<<(const string& data);
    CTSocket& operator<<(bool);

    // Gets the  port to which the CTSocket is connected
    port_t getPort() const
    {
        return tPort_;
    }

    // Gets the port to which the peer CTSocket is connected
    port_t getPeerPort() const
    {
        return tPeerPort_;
    }

    static string getErrorMsg();
    static long      getErrorNum();
    friend class CTSocketSelectorItem;
    friend class CTSocketSelectorList;
    CTSocket(const char *);

    //---------------------------------------------------------------------------

private:

    enum CTBlockedMode {BLOCKED, NONBLOCKED};
    CTBlockedMode eBlockedMode_;
    CTCreationMode eCreationMode_;
    int iRef_;

    int iSock_;
    /* Address this CTSocket is bound to */
    CTInetAddress address_;

    /* Port this CTSocket is bound to in network byte order. */
    port_t tPort_;

    /* Address the peer CTSocket is bound to */
    CTInetAddress peerAddress_;

    /* Port the peer CTSocket is bound to in network byte order. */
    port_t tPeerPort_;

    // Gets the local port to which the CTSocket is connected
    //port_t getLocalPort() const           { return localPort_; }

    /** Create an internet CTSocket.  If isStream is true, creates
        a stream CTSocket, otherwise a DGRAM style CTSocket is created. */
    void create(CTCreationMode eCreationMode, bool bReuseAddress = true);
    void getPortFromConnectedSocket();

    CTSocket(int iNewsock, port_t tPort, CTBlockedMode eBlockedMode, CTCreationMode eCreationMode, const string& host, short sFamily, long lAddress, const string& peerHost, short sPeerFamily, long lPeerAddress, port_t tPeerPort) ;
    int makeStreamSocket();
    int makeDgramSocket();
    int makeInetSocket(int iType);
    CTSocket::CTSelection doSelect(timeval* pTimeOut, int iEvent  = _S_ALL_);
    bool isBlocking(void)
    {
        return eBlockedMode_ == BLOCKED;
    }
    double swap(double d);
    float swap(float d);
};

class CTSocketException : public CTException
{
public:
    CTSocketException(const string& msg, const string& filename, short sLineno, long lNumber = -1L)
: CTException(msg, filename, sLineno, lNumber)
    {}

    void                    operator= (const CTSocketException& exception)
    {
        CTException::operator= (exception) ;
    }

};

DLLSPEC_CYRTOOLS ostream& operator<< (ostream& os, const CTSocket& socket);
DLLSPEC_CYRTOOLS CTOLogFileList& operator<< (CTOLogFileList& os, const CTSocket& socket);

class DLLSPEC_CYRTOOLS CTSocketSelectorItem : virtual public CTMTSafe
{
public:

    CTSocketSelectorItem(CTSocket& socket, int iEvent);

    CTSocketSelectorItem& operator = (const CTSocketSelectorItem& item);
    bool operator == (const CTSocketSelectorItem& item);
    CTSocket& getSocket();
    int getEvent();
    static bool testFunction (CTSocketSelectorItem* selectorItem, CTSocket* pSocket);
    void setEvent(int iEvent)
    {
        iEvent_ = iEvent ;
    }

    //---------------------------------------------------------------------------
private:

    CTSocket& socket_;
    int iEvent_;
};
class DLLSPEC_CYRTOOLS CTSocketSelectorList : virtual public CTMTSafe
{
public:

    CTSocketSelectorList();
    ~CTSocketSelectorList();

    CTSocket & operator()(int& iTypeOfEvent);
    bool operator() (CTSocketSelectorList& resultList, int& iTypeOfEvent);
    void setTimeOut(const timeval *pTimeout);
    void setPollingMode();
    void add(CTSocket& socket, int iTypeOfEvent);
    void remove(CTSocket& socket, int iTypeOfEvent);
    void reset();
    bool contains (CTSocket& socket, int& iTypeOfEvent) ;
    bool isEmpty (void) const;

    //---------------------------------------------------------------------------
private:

    typedef list<CTSocketSelectorItem*> CTSelectorItemList;
    CTSelectorItemList list_;
    fd_set readfds_, writefds_, exceptfds_;
    fd_set *preadfds_, *pwritefds_, *pexceptfds_;
    timeval *pTimeOut_;
};

#endif // __SOCKET__

