/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

/*===========================================================================*/
/*                                       */
/*                **** C O M M O N ****                          */
/*                                       */
/*===========================================================================*/
/*                                       */
/*          Module Name: PSL_ALLOC_RTNS                  */
/*          =============                                */
/*                                       */
/* Description:                                  */
/* ============                                  */
/*                                       */
/*  This module contains memory allocation and deallocation functions    */
/*  for use by PSL code that requires trackling of its memory useage     */
/*                                       */
/*  PSL_MALLOC  - Allocates memory with trace information        */
/*  PSL_FREE    - Deallocates memory, checking trace data        */
/*  PSL_MEM_DEBUG   - Debugging hook to alert the developer of errors    */
/*                                       */
/*===========================================================================*/

#include <psl_macros.h>

/*
 *  Standard library functions.
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

/*
 * Function prototypes
 */
#include <psl_rtndef.h>             /* PSL routines */
#include <trc_rtndef.h>             /* Trace routines */
#include <psl_alloc_rtndef.h>           /* These routines */
/*
 *  Internal static routine(s)
 */
static void OUTPUT_MEM_TEXT (void * start, long length);

/*
 *  Symbols
 */
#include <psl_trc_prmdef.h>         /* Trace parameters */

/*
 *  Structure to contain details of an allocation, where and how much
 *  memory.
 */
struct alloc_details
{
    void * start;
    void * end;
    size_t amount;
    long   ref;
};
/*
 *  Size of array that holds the list of current allocations
 */
#define MAX_ALLOCATIONS 200
static struct alloc_details allocations [MAX_ALLOCATIONS];

/*
 *  Some variables to hold data about usage, can be examined in debugger
 */

static  long    mallocs;            /* Number of malloc calls    */
static  long    frees;              /* Number of free calls      */
static  long    dupfrees;           /* Number of duplicate free's*/
static  long    overwrites;         /* Number of mem overwrites  */
static  long    overlaps;           /* Number of mem overlaps    */
static  long    unknown;            /* Number of unknown free's  */
static  long    lost;               /* Number of lost details    */
static  long    next;               /* Next reference number     */
static  Boolean trace;              /* Trace enabled flag        */
static  char    *why="Unknown error:";      /* Reason for alert          */

/*
 *  How much memory to allocate at the end of the user's block so we can
 *  detect overrun by filling it with a pre-determined pattern.
 */
#define OVERRUN 16


/*===========================================================================*/

/*    PSL_MALLOC - Record a memory allocation for debugging          */
/*                                                                   */

/*    This routine may be used to allocate memory as a replacement for       */
/*    malloc. It records dignostic infomation about the allocation and       */
/*    checks all other know allocations to determine if there is overlapping */

/*    Calling convention:                                                    */

/*  mem  = PSL_MALLOC (amount)                               */

/*  Where:                                                               */

/*  mem         Pointer to void return address of allocated memory   */
/*          NULL    - Allocation failed, insuficient memory      */
/*          !NULL   - Pointer to start of allocated memory       */
/*                                                                       */
/*  ammount     size_t amount of memory to allocate in bytes         */
/*                                                                       */
/*===========================================================================*/
void * PSL_MALLOC (size_t amount)
{
    unsigned long   count;              /* Counter variable          */
    unsigned long   slot;               /* Counter variable          */
    unsigned long   oldest_slot;            /* Counter variable          */
    void   *mem;               /* Ptr to new mem            */
    void   *mem_end;           /* Ptr to end of new mem     */
    char   *temp_end;          /* Temp ptr to end of new mem*/
    void   *test_start;            /* Start of mem to test      */
    void   *test_end;          /* End of mem to test        */
    char   *fill;              /* Ptr used to fill overrun  */
    long   oldest;             /* Temp place holder         */
    Boolean     found;              /* Flag used to find slots   */
    Boolean     overlap;            /* Flag used to test overlap */

    /*
     *  See if memory tracing is enabled for these routines.
     */
    if (next == 0)
    {
        trace = PSL_TEST_TRACE (PTRC_GEN_MEM);
    }
    /*
     *  Attempt to allocate memory
     */
    mem = calloc (amount + (sizeof (char) * OVERRUN),1);
    temp_end  = (char *) mem;
    temp_end += amount + (sizeof (char) * OVERRUN);
    mem_end = temp_end;

    if (mem == NULL)
    {
        why="Insufficient memory available";
        PSL_MEM_DEBUG();
        PSL_OUTPUT_TRACE
        ("MALLOC : Insufficient memory available to allocate %ld bytes\n.",
         amount);
        goto end;
    }

    mallocs +=1;
    overlap = PFALSE;
    count = 0;
    /*
     *  Check allocation details to see if we have been allocated some memory
     *  that overlaps a previous known allocation
     */
    while ( (count < MAX_ALLOCATIONS) && !overlap )
    {
        if (allocations[count].end != NULL)       /* Slot in use ?     */
        {
            test_start = allocations[count].start;
            test_end =  allocations[count].end;

            if ( !((test_end <= mem) || (mem_end <= test_start)) )/* Overlap   */

            {
                overlap = PTRUE;
                why="Allocated memory overlaps known region";
                overlaps +=1;
                PSL_MEM_DEBUG();
                PSL_OUTPUT_TRACE
                ("MALLOC : Overlapping allocation detected, %p + %ld bytes overlaps handle %ld:\n",
                 mem, amount, count);
                PSL_OUTPUT_TRACE
                ("MALLOC : %p to %p length %ld\n", test_start, test_end,
                 allocations[count].amount);
            }
        }
        count +=1;
    }
    /*
     *  If there was an overlap, exit
     */
    if (overlap)
    {
        goto end;
    }
    /*
     *  Find a free slot in the details list, if N/A record the slot number of
     *  the oldest know allocation and replace those details instead.
     */
    slot = 0;
    oldest_slot = 0;
    oldest = next;
    found = PFALSE;

    while ( (slot < MAX_ALLOCATIONS) && (!found) )
    {
        if (allocations[slot].start == NULL)
        {
            found = PTRUE;
        }
        else
        {
            if (allocations[slot].ref < oldest)
            {
                oldest = allocations[slot].ref;
                oldest_slot = slot;
            }
            slot +=1;                   /* Advance position  */
        }
    }
    /*
     *  Replace the oldest know allocation if there are no free slots
     */
    if (!found)
    {
        lost +=1;     /* Increment no of lost allocation details   */
        slot = oldest_slot;   /* Replace the oldest record                 */
        if (trace)
        {
            PSL_OUTPUT_TRACE
            ("MALLOC :  Discarding handle %ld for new allocation\n",
             oldest_slot);
        }
    }
    /*
     *  Record details in list of known allocations, increment no of allocations
     */
    allocations[slot].end    = mem_end;
    allocations[slot].start  = mem;
    allocations[slot].amount = amount;
    allocations[slot].ref    = next;

    /*
     *  Fill overrun section with know data so we may detect writes to it
     *  when it is free'd.
     */
    fill = (char *) mem + amount;

    for (count=0; count < OVERRUN; count++)
    {
        *fill = (char) count;
        fill += 1;
    }

    if (trace)
    {
        PSL_OUTPUT_TRACE
        ("MALLOC :  %ld bytes allocated at %p, details in handle %ld.\n",
         amount, mem, next);
    }
    next += 1;

    /*
     *  end:    Routine exit point
     */
end:
    return (mem);
}

/*===========================================================================*/

/*    PSL_FREE   - Record the deallocation of memory when debugging      */
/*                                                                   */

/*    This routine may be used to free allocated memory as a replacement for */
/*    free. It examines dignostic infomation about known allocations and     */
/*    checks if the memory has been written past or has allready been free'd */

/*    Calling convention:                                                    */

/*  void = PSL_FREE (void * mem)                                 */

/*  Where:                                                               */

/*  mem         Pointer to the memory to deallocate                  */
/*                                                                       */
/*===========================================================================*/
void PSL_FREE (void *mem)
{
    unsigned long   count;              /* Counter variable          */
    unsigned long   slot;               /* Counter variable          */
    char   *fill;              /* Ptr used to fill overrun  */
    Boolean     found;              /* Flag used to find slots   */
    Boolean     overwritten;            /* Flag used to check overrun*/
    Boolean     seen;               /* Flag used to check if free*/

    slot  = 0;
    found = PFALSE;
    seen = PFALSE;
    /*
     *  See if we have allocation details for this memory. Found is set
     *  when a current allocation is found, seen is set when their is evidence
     *  of a previous allocation that has been free'd.
     */
    while ((slot < MAX_ALLOCATIONS) && (!found))
    {
        if((allocations[slot].start == mem) )
        {
            seen = PTRUE;           /* We have details of an alloc*/
            if (allocations[slot].end !=NULL)   /* but is the memory free?   */
            {
                found = PTRUE;            /* It's a live handle        */
            }
            else
            {
                slot +=1;             /* It has already been free'd*/
            }
        }
        else
        {
            slot +=1;               /* No matching details       */
        }
    }
    /*
     *  If never allocated OR details overwritten
     */
    if (!(found) && !(seen))
    {
        free(mem);
        frees +=1;
        unknown +=1;
        if (trace)
        {
            PSL_OUTPUT_TRACE
            ("FREE   :  Deallocating unknown memory at %p\n", mem);
        }

    }
    else
    {
        /*
         *  We are aware of such an allocation, but it has been deallocated
         */
        if (seen && !(found))     /* free'd because end==NULL          */
        {
            why="Allready free'd";
            PSL_MEM_DEBUG();
            dupfrees += 1;
            PSL_OUTPUT_TRACE
            ("FREE   :  Duplicate deallocation of memory at %p\n",mem);
        }
        else
            /*
             *  We are aware of it and it has not been free'd, check for overwriting
             */
        {
            fill  = (char *) allocations[slot].end;
            fill -= OVERRUN;
            count = 0;
            overwritten = PFALSE;

            while (count < OVERRUN && !overwritten)
            {
                if (*fill == count)
                {
                    count += 1;
                    fill  += 1;
                }
                else
                {
                    overwritten = PTRUE;
                }
            }
            /*
             *  If there is evidence of overwriting, notify someone
             */
            if (overwritten)
            {
                why="End of block has been overwritten";
                PSL_MEM_DEBUG();
                overwrites +=1;
                PSL_OUTPUT_TRACE
                ("FREE   :  Memory overwrite at end of handle %ld: %p to %p\n",
                 allocations[slot].ref, mem, allocations[slot].end);
                OUTPUT_MEM_TEXT(fill-16,40);
            }

            /*
             *  Get length of memory region and clear it before deallocation to try and
             *  trip up any further references to it.
             */
            (void) memset (mem,0,allocations[slot].amount + OVERRUN);
            free(mem);
            allocations[slot].end = NULL;       /* Mark it as free'd */
            frees +=1;
            if (trace)
            {
                PSL_OUTPUT_TRACE
                ("FREE   :  Deallocating %ld bytes of known memory at %p handle %ld\n",
                 allocations[slot].amount, mem, allocations[slot].ref);
            }
        }
    }
}

/*===========================================================================*/

/*    PSL_CALLOC - Record a memory allocation for debugging          */
/*                                                                   */

/*    This routine may be used to allocate memory as a replacement for       */
/*    calloc. It records dignostic infomation about the allocation and       */
/*    checks all other know allocations to determine if there is overlapping */

/*    Calling convention:                                                    */

/*  mem  = PSL_CALLOC (nobj, size)                               */

/*  Where:                                                               */

/*  mem         Pointer to void return address of allocated memory   */
/*          NULL    - Allocation failed, insuficient memory      */
/*          !NULL   - Pointer to start of allocated memory       */
/*                                                                       */
/*  nobj        size_t number of objects to allocate mem for         */
/*                                                                       */
/*  size        size_t amount of memory occupied by each object      */
/*                                                                       */
/*===========================================================================*/
void * PSL_CALLOC (size_t nobj, size_t size)
{
    return PSL_MALLOC (nobj * size);
}

/*===========================================================================*/

/*    OUTPUT_MEM_TEXT - Output a the contents of memory to trace             */
/*                                                                   */

/*    This routine may be used to output the contents of a region of memory  */
/*    to trace. It outputs the printing charaters followed by a hex dump.    */

/*    Calling convention:                                                    */

/*  void = OUTPUT_MEM_TEX (start, length)                    */

/*  Where:                                                               */

/*  start       Pointer to the memory to display                     */
/*                                       */
/*  length      Number of bytes to produce output for            */
/*                                                                       */
/*===========================================================================*/
static void OUTPUT_MEM_TEXT (void * start, long length)
{
    long    count;
    char    *mem_ptr;

    mem_ptr = start;
    for (count = 0; count < length; count ++)
    {
        if ((count % 80) == 0)            /* At end of VT line */
        {
            PSL_OUTPUT_TRACE("\n");         /* Go on to next     */
        }
        if (isprint(*mem_ptr))            /* If printable,do it*/
        {
            PSL_OUTPUT_TRACE("%c",*mem_ptr);
        }
        else
        {
            PSL_OUTPUT_TRACE(".");          /* Else print a "."  */
        }
        mem_ptr +=1;
    }
    PSL_OUTPUT_TRACE("\n");


    mem_ptr = start;
    for (count = 0; count < length; count ++)
    {
        if ((count % 16) == 0)            /* Nice place to break*/
        {
            PSL_OUTPUT_TRACE("\n");         /* line in hex mode   */
        }
        PSL_OUTPUT_TRACE("%X,",*mem_ptr);
        mem_ptr +=1;
    }
    PSL_OUTPUT_TRACE("\n");
}

/*===========================================================================*/

/*    PSL_MEM_DEBUG - Debugger hook to allow easy trapping of detected       */
/*            memory allocation, use and dealocation errors.         */

/*    This routine may be used to alert the developer that the PSL_MEM       */
/*    routines have discovered an error in their use. This shell is called   */
/*    by those routines upon an error being detected and is a convenient     */
/*    place to put a breakpoint when using a debugger.               */

/*    A char * :- why, contains the reason for this routine to be called     */
/*    as a NUL terminated character string.                  */


/*===========================================================================*/
void PSL_MEM_DEBUG (void)
{
    ;
}
