/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

/*===========================================================================*/
/*                                       */
/*               **** C O M M O N ****                   */
/*                                       */
/*===========================================================================*/
/*                                       */
/*          Module Name: PSL_DIAG_SUBS               */
/*          ============                         */
/*                                       */
/* Description:                                  */
/* ============                                  */
/*                                       */
/*  This module contains a number of routines related to the use         */
/*  of the diagnostic stack facilities, but which must have access       */
/*  to routines in SVXAPP.                           */
/*                                       */
/*===========================================================================*/


/*   Define useful Include structures   */

#include <psl_macros.h>         /* PSL Standard Macros */

#include <stdlib.h>         /* Standard Library Routines */

#ifdef UNIX
#include <sys/stat.h>           /* File status definitions */
#endif /* UNIX */

/*   Define Function Prototypes for all external routines */
/*   used by the module (and for ourselves) */

#include <psl_rtndef.h>         /* Define ourselves */

#ifdef UNIX
static Boolean STDOUT_EQ_STDERR (void); /* Check if stdout = stderr */
#endif /* UNIX */

/*   Define Data Structures */

#include <diagstkdef.h>         /* Define stack format */
#include <shrmsg.h>         /* Message Statuses */
#include <psldiag_prmdef.h>     /* Diagnostic Routine flags */

/************************************************************************/
/*                                  */
/*  PSL_SETUP_OUDIAG - Set up routine for PSL_OUTPUT_DIAG       */
/*                                  */
/*  This routine may be called to set up the output types to be */
/*  used by PSL_OUTPUT_DIAG                     */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*   iret = PSL_SETUP_OUDIAG (diag_ptr, outtyp, output_ptr, outwid) */
/*                                  */
/*  Where:                              */
/*                                  */
/*      iret       = Unsigned Long PSL Routine Return Status    */
/*      diag_ptr   = Pointer to Diagnostics Stack.          */
/*      outtyp     = Output Type, DIOUTY_xxx as defined in      */
/*           PSLDIAG_PRMDEF.                */
/*      output_ptr = Pointer to Output Type data structure.     */
/*           This may be a pointer to a Listing Context */
/*           or a pointer to a TCB.             */
/*           Passing a NULL pointer switches off the    */
/*           specified output type.             */
/*      outwid     = Width of Output Type               */
/*                                  */
/************************************************************************/

unsigned long   PSL_SETUP_OUDIAG (struct diagstk *diag_ptr,
                                  unsigned long outtyp,
                                  void *output_ptr,
                                  unsigned long outwid)
{

    /*   Local data */

    unsigned long       iret;       /* Return Status */

    /*   Case on the output type (omitting wrap flag) */

    switch (outtyp & ~DIOUTY_WRAP)
    {
    case DIOUTY_LIST:           /* If output to Listing requested */
        {
            (*diag_ptr).diag_outtypl &= ~DIAG_OUTTYP_WRAP_LIST;
            /* Assume no wrapping */
            if ( output_ptr == NULL )   /* If Clearing Information */
            {
                (*diag_ptr).diag_outtypl &= ~DIAG_OUTTYP_LIST;
                /* Clear Output Type */
                (*diag_ptr).diag_list_ctx_ptrps = NULL;
                /* Clear Pointer to Listing Context */
                (*diag_ptr).diag_list_actrtnpf = NULL;
                /* Clear Pointer to Action Routine */
                (*diag_ptr).diag_list_outwidl = 0;
                /* Clear Output Width */
            }

#ifdef UNIX
            else if (!STDOUT_EQ_STDERR())   /* If stdout not same as stderr */
#else
            else                /* Setting Information */
#endif

            {
                (*diag_ptr).diag_outtypl |= DIAG_OUTTYP_LIST;
                /* Store Output Type */
                if ((outtyp & DIOUTY_WRAP) != 0)
                    /* If Wrap flag set */
                {
                    (*diag_ptr).diag_outtypl |= DIAG_OUTTYP_WRAP_LIST;
                    /* Set Wrap flag */
                }
                (*diag_ptr).diag_list_ctx_ptrps =
                    (struct list_ctx *) output_ptr;
                /* Store Pointer to Listing Context */
                (*diag_ptr).diag_list_actrtnpf = LIST_OUTPUT_ERROR_BUFFER;
                /* Store Pointer to Action Routine */
                (*diag_ptr).diag_list_outwidl = outwid;
                /* Store Output Width */
            }
            break;
        }

        /*
         *    Output to TCB not valid for WIN32 platforms.
         */
#ifndef WIN32
    case DIOUTY_TCB:        /* If output to TCB requested */
        {
            (*diag_ptr).diag_outtypl &= ~DIAG_OUTTYP_WRAP_TCB;
            /* Assume no wrapping */
            if ( output_ptr == NULL )   /* If Clearing Information */
            {
                (*diag_ptr).diag_outtypl &= ~DIAG_OUTTYP_TCB;
                /* Clear Output Type */
                (*diag_ptr).diag_tcb_ptrps = NULL;
                /* Clear Pointer to TCB */
                (*diag_ptr).diag_tcb_actrtnpf = NULL;
                /* Clear Pointer to Action Routine */
                (*diag_ptr).diag_tcb_outwidl = 0;
                /* Clear Output Width */
            }
            else                /* Setting Information */
            {
                (*diag_ptr).diag_outtypl |= DIAG_OUTTYP_TCB;
                /* Store Output Type */
                if ((outtyp & DIOUTY_WRAP) != 0)
                    /* If Wrap flag set */
                {
                    (*diag_ptr).diag_outtypl |= DIAG_OUTTYP_WRAP_TCB;
                    /* Set Wrap flag */
                }
                (*diag_ptr).diag_tcb_ptrps = (struct tcb *) output_ptr;
                /* Store Pointer to TCB */
                (*diag_ptr).diag_tcb_actrtnpf = TRM_OUTPUT_ERROR_BUFFER;
                /* Store Pointer to Action Routine */
                (*diag_ptr).diag_tcb_outwidl = outwid;
                /* Store Output Width */
            }
            break;
        }
#endif  /* NOT WIN32 */

    default:                /* Invalid Output Type */
        {
            iret = PSL_BADPARAM;        /* Return Error */
            goto end;           /* And Exit */
        }

    }                   /* End of (switch (outtyp)) */

    iret = PSL_NORMAL;          /* Return Success */

end:
    return iret;
}

#ifdef UNIX
/************************************************************************/
/*                                  */
/*  STDOUT_EQ_STDERR - Check if stdout = stderr         */
/*                                  */
/*  This routine determines whether stdout and stderr are       */
/*  directed to the same place.                 */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*      bret = STDOUT_EQ_STDERR ();                 */
/*                                  */
/*  Where:                              */
/*                                  */
/*      bret = Boolean return status = PTRUE if stdout = stderr */
/*                                  */
/************************************************************************/
static Boolean STDOUT_EQ_STDERR (void)
{
    long    status;         /* Routine status return */
    struct stat stdout_stat_buf;    /* Dev status structure for stdout */
    struct stat stderr_stat_buf;    /* Dev status structure for stderr */

    /*
     *  Get file status for stdout and stderr and compare the directory
     *  and device id's - if they are the same, assume that stdout and
     *  stderr and directed to the same place.
     */
    status = fstat (1, &stdout_stat_buf);
    /* Get file status of stdout */
    if (status == -1) return PFALSE;    /* Assume different if error */
    status = fstat (2, &stderr_stat_buf);
    /* Get file status of stderr */
    if (status == -1) return PFALSE;    /* Assume different if error */
    if ((stdout_stat_buf.st_dev == stderr_stat_buf.st_dev)
            /* If directory id's match */
            && (stdout_stat_buf.st_rdev == stderr_stat_buf.st_rdev))
        /* And file id's match */
    {
        return PTRUE;           /* Then stdout = stderr */
    }
    else
    {
        return PFALSE;          /* stdout <> stderr */
    }
}
#endif /* UNIX */
