/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

/*===========================================================================*/
/*                                       */
/*                                       */
/*===========================================================================*/
/*                                       */
/*          Module Name: PSL_GET_MESSAGE_NTI             */
/*          ============                         */
/*                                       */
/* Description:                                  */
/* ============                                  */
/*                                       */
/*  Return the message text or message prefix for a given message        */
/*  inserting the Utility Name if required                   */
/*                                       */
/*===========================================================================*/

/*
 *  Include PSL macro definitions
 */
#include <psl_macros.h>

/*
 *  THIS MODULE IS WIN32 SPECIFIC AND THEREFORE IS ONLY COMPILED UNDER WIN32
 */
#ifdef WIN32

/*
 *  Include useful standard header files
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#define STRADD_COPY 1
#define STRADD_CONCAT 2
/*
 *  Function prototypes
 */
#include <psl_rtndef.h>            /* PSL Routines */

/*
 *  Include message handling constant definitions and a file defining the
 *  supported Sub-systems and their facilities
 */
#include <pslmsg_prmdef.h>
#include <windows.h>

static void STRADD_MSGTXT(char *msgbuf,
                          unsigned long bufsiz,
                          char* msgtxt,
                          unsigned long flags);

/*
 *  Global data
 */
static char utlnam[21];     /* Utility name */
#ifndef VMS_MSG_FORMAT
static char utlnam_lc[21];      /* Utility name in lower case */
#endif /* VMS_MSG_FORMAT */

/****************************************************************************/
/*                                      */
/*  PSL_GET_MESSAGE - Get a message from the message DLL.           */
/*                                      */
/*  Calling Convention:                             */
/*                                      */
/*  iret = PSL_GET_MESSAGE (msgnum, msgbuf, buflen, msgflg)         */
/*                                      */
/*  Where:                                  */
/*                                      */
/*    iret  =   return status                       */
/*    msgnum    =   windows message number                  */
/*          If = 0, close message file if open              */
/*    msgbuf    =   pointer to message text buffer              */
/*    buflen    =   length of msgbuf                        */
/*    msgflg    =   flags indicating the required parts of the message      */
/*                                      */
/*                                      */
/****************************************************************************/
unsigned long PSL_GET_MESSAGE (
    unsigned long   msgnum,
    char        *msgbuf,
    unsigned long   buflen,
    unsigned long   msgflg)
{
    /*
     * Local Data,
     */
    unsigned long   severity;       /* Message severity */
    unsigned long   status;         /* Routine status */
    char        msg_file[255];      /* Message filename */
    char        msgtxt[255];        /* Message text */
    long        msglen;         /* Message length */
    char        facnam[20];     /* Facility name */
    char        msgnam[20];     /* Message symbolic name */
    char        *chrptr;        /* Useful character pointer */
    int             i;                      /* Loop variable */
    static HINSTANCE dll_handle = NULL; /* Handle to the messagetable DLL */

    /*
     *  If message number = 0, close message file, if open
     */
    if (msgnum == 0)
    {
        if (dll_handle != NULL)         /* Previous msg file to close */
        {
            FreeLibrary (dll_handle);
            dll_handle = NULL;
        }
        return 1;                   /* Return successful */
    }

    /*
     *  Windows message number format:
     *
     
       3 3 2 2 2 2 2 2 2 2 2 2 1 1 1 1 1 1 1 1 1 1
       1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0
      +---+-+-+-----------------------+-------------------------------+
      |Sev|C|R|     Facility          |               Code            |
      +---+-+-+-----------------------+-------------------------------+
     
      where
     
          Sev - is the severity code (0=Success, 1=Informational, 2=Warning, 3=Error)
          C - is the Customer code flag
          R - is a reserved bit
          Facility - is the facility code
          Code - is the facility's message number
     */
    severity = (msgnum >> 30) & 0X3;        /* Extract severity */

    /*
     *  Set the message DLL filename.
     */
#ifdef _DEBUG
    PSL_ADDROOT ("\\msglibd.dll", msg_file, sizeof (msg_file));
#else
    PSL_ADDROOT ("\\msglib.dll", msg_file, sizeof (msg_file));
#endif

    /*
     *  Open the message file if it is not already open
     */
    if (dll_handle == NULL)
    {
        dll_handle = LoadLibraryEx (
                         msg_file,
                         NULL,
                         DONT_RESOLVE_DLL_REFERENCES|LOAD_LIBRARY_AS_DATAFILE);
        if (dll_handle == NULL)         /* Error opening msg file */
        {
            (void) strcpy(msgtxt, "Error opening message file ");
            (void) strcat(msgtxt, msg_file);
            (void) strcat(msgtxt, " - ");
            FormatMessage (FORMAT_MESSAGE_FROM_SYSTEM,
                           NULL,
                           GetLastError(),
                           LANG_NEUTRAL,
                           (LPSTR) msgtxt + strlen(msgtxt),
                           sizeof(msgtxt) - strlen(msgtxt) - 1,
                           NULL);       /* Get system message text */
            goto error;
        }
    } /* if message file NOT open */

    /*
     *  Extract the message name and text
     */
    status = FormatMessage(
                 FORMAT_MESSAGE_FROM_HMODULE,        /* Get the message from the DLL */
                 dll_handle,                 /* Messagetable DLL handle */
                 msgnum,                 /* Message ID */
                 MAKELANGID(LANG_ENGLISH, SUBLANG_ENGLISH_US),
                 /* Language ID */
                 msgtxt,                 /* Message buffer */
                 sizeof(msgtxt),             /* Maximum size of the message buffer */
                 NULL);                  /* No parameters */
    if (status == 0)                /* If error */
    {
        (void) strcpy(msgtxt, "NOMSG,Message Number");
    }
    else
    {
        msglen = strlen (msgtxt);       /* Length of text */
        msgtxt[msglen - 2] = '\0';      /* Remove CRLF */
    }

    /*
     *  Search the recovered text for the first occurrance of ','
     *  The message name and text are divided by a ','
     */
    chrptr = (char*)strchr(msgtxt, ',');
    *chrptr = '\0';
    (void) strcpy(msgnam, msgtxt);
    (void) strncpy(msgtxt, (chrptr + 1), 254);
#ifndef VMS_MSG_FORMAT
    for (i = 0;
            (( i < sizeof(msgnam)) && (msgnam[i] != '\0'));
            i++)
    {
        msgnam[i] = tolower(msgnam[i]);
    }
#endif /* !VMS_MSG_FORMAT */

    /*
     *  If a message prefix was requested then get the facility name
     */
    if ( (msgflg & PSLMSG_PREFIX) != 0)
    {
        /*
         *  Get the facility name from the message file
         */
        status = FormatMessage(
                     FORMAT_MESSAGE_FROM_HMODULE,    /* Get the message from the DLL */
                     dll_handle,             /* Messagetable DLL handle */
                     msgnum,             /* Message ID */
                     MAKELANGID(LANG_ENGLISH, SUBLANG_ENGLISH_US),
                     /* Language ID */
                     facnam,             /* Message buffer */
                     sizeof(facnam),         /* Maximum size of the message buffer */
                     NULL);              /* No parameters */
        if (status == 0)            /* Facility name not found */
        {
            (void) strcpy(facnam, "NONAME");
        }
        else
        {
            msglen = strlen (facnam);       /* Length of name */
            facnam[msglen - 2] = '\0';      /* Remove CRLF */
        }
#ifndef VMS_MSG_FORMAT
        for (i = 0;
                ((i < sizeof(facnam)) && (facnam[i] != '\0'));
                i++)
        {
            facnam[i] = tolower(facnam[i]);
        }
#endif /* !VMS_MSG_FORMAT */
    }

    /*
     *  Construct the message
     */
    msgbuf[0] = '\0';                   /* Start with null string */
    if (((msgflg & PSLMSG_PREFIX) != 0)         /* Construct the prefix if req'd */
            ||  ((msgflg & PSLMSG_CODE) != 0))          /* or if the code required. */
    {
#ifdef VMS_MSG_FORMAT
        STRADD_MSGTXT(msgbuf, buflen, "%", STRADD_COPY);
        /* Add % */
#endif
        if ((msgflg & PSLMSG_UTLNAM) != 0)      /* Add utility_name if requested */
        {
#ifdef VMS_MSG_FORMAT
            STRADD_MSGTXT(msgbuf, buflen, utlnam, STRADD_CONCAT);
#else
            STRADD_MSGTXT(msgbuf, buflen, utlnam_lc, STRADD_COPY);
#endif /* VMS_MSG_FORMAT */
        }

        else if ((msgflg & PSLMSG_PREFIX) != 0)  /* Add facility_name if PREFIX*/
        {
#ifdef VMS_MSG_FORMAT
            STRADD_MSGTXT(msgbuf, buflen, facnam, STRADD_CONCAT);
#else
            STRADD_MSGTXT(msgbuf, buflen, facnam, STRADD_COPY);
#endif
        }
#ifndef VMS_MSG_FORMAT
        STRADD_MSGTXT(msgbuf, buflen, ": ", STRADD_CONCAT);
#endif
        switch (severity)           /* Add the -severity- */
        {
#ifdef VMS_MSG_FORMAT
        case 0:                 /* Success */
            {
                STRADD_MSGTXT(msgbuf, buflen, "-S-", STRADD_CONCAT);
                break;
            }
        case 1:                 /* Informational */
            {
                STRADD_MSGTXT(msgbuf, buflen, "-I-", STRADD_CONCAT);
                break;
            }
        case 2:                 /* Warning */
            {
                STRADD_MSGTXT(msgbuf, buflen, "-W-", STRADD_CONCAT);
                break;
            }
        case 3:                 /* Error */
            {
                STRADD_MSGTXT(msgbuf, buflen, "-E-", STRADD_CONCAT);
                break;
            }
#else
        case 0:             /* Success */
            {
                STRADD_MSGTXT(msgbuf, buflen, "ok", STRADD_CONCAT);
                break;
            }
        case 1:             /* Informational */
            {
                STRADD_MSGTXT(msgbuf, buflen, "info", STRADD_CONCAT);
                break;
            }
        case 2:             /* Warning */
            {
                STRADD_MSGTXT(msgbuf, buflen, "warn", STRADD_CONCAT);
                break;
            }
        case 3:             /* Error */
            {
                STRADD_MSGTXT(msgbuf, buflen, "error", STRADD_CONCAT);
                break;
            }
#endif /* VMS_MSG_FORMAT */
        }
#ifndef VMS_MSG_FORMAT
        STRADD_MSGTXT(msgbuf, buflen, " (", STRADD_CONCAT);
#endif /* !VMS_MSG_FORMAT */
        STRADD_MSGTXT(msgbuf, buflen, msgnam, STRADD_CONCAT);
        /* Add message name */
#ifndef VMS_MSG_FORMAT
        STRADD_MSGTXT(msgbuf, buflen, ")", STRADD_CONCAT);
#endif /* !VMS_MSG_FORMAT */
        if ((msgflg & PSLMSG_TEXT) != 0)    /* If text req'd */
        {
            STRADD_MSGTXT(msgbuf, buflen, ", ", STRADD_CONCAT);
            /* Add , and a space */
            STRADD_MSGTXT(msgbuf, buflen, msgtxt, STRADD_CONCAT);
            /* ... add message text */
        }
    }
    else                    /* Message text only req'd */
    {
        STRADD_MSGTXT(msgbuf, buflen, msgtxt, STRADD_COPY);
    }

    return 1;                       /* Return successful */

    /*
     *  We jump to this label if an error occurs. The error message in msgbuf.
     */
error:
    (void) strcpy(msgbuf, msgtxt);

    return 0;                       /* Return unsuccessful */
}

/****************************************************************************/
/*                                      */
/*  STRADD_MSGTXT - Get a facilities unix SET number and filename       */
/*                                      */
/*  Calling Convention:                         */
/*                                      */
/*      STRADD_MSGTXT (msg_str, buf_len, msg_txt, flags)            */
/*                                      */
/*  Where:                                  */
/*                                      */
/*    msg_str   =   Pointer to message string           */
/*    buf_len   =   Buffer size                 */
/*    msg_txt   =   Pointer to string to add to msg_str     */
/*    flags     =   Flag indicating copy or concatenate     */
/*              STRADD_COPY or STRADD_CONCAT            */
/*                                      */
/****************************************************************************/
static void STRADD_MSGTXT(
    char          *msg_str,
    unsigned long buf_len,
    char          *msg_txt,
    unsigned long flags)
{
    /*
     *  Local Data,
     */

    unsigned long   length;

    /*
     *  Copy or add according to the flags
     */
    if (flags == STRADD_COPY)           /* Copy msg_txt into msg_str */
    {
        /*
         *  If there is not enough room in msg_str to add all of msg_txt then only
         *  add characters until msg_str is full.
         */
        if ( strlen(msg_txt) > (size_t)(buf_len - 1))
            /* Not enough room for msg_txt*/
        {
            length = buf_len - 1 - (unsigned long)strlen(msg_str);
            /* Reduce copy string */
        }
        else                    /* Enough room for msg_txt */
        {
            length = (unsigned long)strlen(msg_txt);
        }
        for (; length != 0; msg_str++, msg_txt++, length--)
            *msg_str = *msg_txt;
        *msg_str = '\0';
    }
    else if (flags == STRADD_CONCAT)        /* Add msg_txt to msg_str */
    {
        /*
         *  If the message string is already full then return immediately
         */
        if (strlen(msg_str) >= (size_t)(buf_len - 1)) return;
        /*
         *  If there is not enough room in msg_str to add all of msg_txt then only
         *  add characters until msg_str is full.
         */
        if ( (strlen(msg_str) + strlen(msg_txt)) > (size_t)(buf_len - 1))
            /* Not enough room for msg_txt*/
        {
            length = buf_len - 1 - (unsigned long)strlen(msg_str);
            /* Reduce add string */
        }
        else                    /* Enough room for msg_txt */
        {
            length = (unsigned long)strlen(msg_txt);
        }
        msg_str = strchr(msg_str, '\0');
        for (; length != 0; msg_str++, msg_txt++, length--)
            *msg_str = *msg_txt;
        *msg_str = '\0';
    }

    return;
}

/****************************************************************************/
/*                                      */
/*  PSL_GET_UTILNAME - Retrieve utility name                    */
/*                                      */
/*  Calling Convention:                             */
/*                                      */
/*  (void) PSL_GET_UTILNAME (utlnam_ptr, max_size);             */
/*                                      */
/*  Where:                                  */
/*                                      */
/*    utlnam_ptr = Pointer to buffer to hold utility name.          */
/*    max_size   = Size of buffer.                      */
/*                                      */
/*  This routine retrieves the utility name used by PSL_GET_MESSAGE     */
/*                                      */
/****************************************************************************/
void    PSL_GET_UTILNAME (char *utlnam_ptr,
                          long max_size)
{
    /*
     *  Copy utility name and ensure it is null-terminated
     */
    (void) strncpy (utlnam_ptr, utlnam, max_size - 1);
    *(utlnam_ptr + max_size - 1) = '\0';
    return;
}


/****************************************************************************/
/*                                      */
/*  PSL_SET_UTILNAME - Set up utility name                  */
/*                                      */
/*  Calling Convention:                             */
/*                                      */
/*  (void) PSL_SET_UTILNAME (util_name)                 */
/*                                      */
/*  Where:                                  */
/*                                      */
/*    util_name =   Pointer to utility name                 */
/*                                      */
/*  This routine stores the utility name for use by PSL_GET_MESSAGE     */
/*                                      */
/****************************************************************************/
void    PSL_SET_UTILNAME (char *util_name)
{
#ifndef VMS_MSG_FORMAT
    int                 i;                      /* loop variable */
#endif /* VMS_MSG_FORMAT */

    /*
     *  Copy utility name and ensure it is null-terminated
     */
    (void) strncpy (utlnam, util_name, (sizeof(utlnam) - 1));
    utlnam[sizeof(utlnam) - 1] = '\0';
#ifndef VMS_MSG_FORMAT
    (void)memcpy(utlnam_lc, utlnam, sizeof(utlnam_lc));
    for (i = 0;
            (i < sizeof(utlnam_lc) &&
             (utlnam_lc[i] != '\0'));
            i++)
    {
        utlnam_lc[i] = tolower(utlnam_lc[i]);
    }
#endif /* VMS_MSG_FORMAT*/
    return;
}

#endif
