/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

/*===========================================================================*/
/*                                       */
/*           **** C O M M O N ****                   */
/*                                       */
/*===========================================================================*/
/*                                       */
/*          Module Name: PSL_VM_SORT                         */
/*          ============                         */
/*                                       */
/* Description:                                  */
/* ============                                  */
/*                                       */
/*  This module contains platform independent routines which provide     */
/*  sorting facilities for that ever popular beast, the PSL linked-      */
/*  list.                                    */
/*                                       */
/*===========================================================================*/

/*
 *  PSL Macros.
 */
#include <psl_macros.h>

/*
 *  Standard includes.
 */
#include <stdlib.h>
#include <string.h>

/*
 *  Function Prototypes.
 */
#include <psl_rtndef.h>         /* PSL routines */

struct globs;
static unsigned long    PSL_TREE_SORT (struct psl_vm_llb *whole_list,
                                       struct psl_vm_skey *key_ptr,
                                       struct globs *gbl_ptr);

static unsigned long    PSL_SORT_COMPARE (void *buff1,
        void *buff2,
        struct psl_vm_skey *key_ptr,
        short *cmpsts,
        struct globs *gbl_ptr);

/*
 *  Structure Definitions and parameters.
 */
#include <shrmsg.h>         /* PSL Messages */
#include <psl_vm_skeydef.h>     /* Key Structure */
#include <psl_vm_prmdef.h>      /* VM routines parameters */

/*
 *  Declare a structure which contains variables that must be global.
 */
struct globs
{
    struct psl_vm_llb *srtd_ent_list;
    /* Sorted entry pointer list */
    struct psl_vm_llb *original_list;
    /* Pointer to users original list */
    struct diagstk *diag_ptr;   /* Diagnostics stack pointer */
    Boolean no_duplicates;  /* No duplicates flag */
    char *first_recptr;     /* First record comparison pointer */
    char *next_recptr;      /* Entry record comparison pointer */
};



/*
 *  Function:  PSL_VM_SORT
 *  =========
 *
 *  This routine can be used to sort a PSL linked list into order.
 *  It supports sorting on multiple keys, each in ascending or 
 *  descending order.
 *
 *
 *  Calling Format:
 *
 *      iret = PSL_VM_SORT (user_list, user_cnt_ptr, ent_size, keyarr, 
 *                  options, diag_ptr)
 *
 *  where:
 *
 *      iret    = PSL return status.
 *
 *      user_list   = Pointer to users linked list to be sorted.
 *            After the sort the pointer is modified to point to
 *            the new sorted list and the users original list is
 *            deleted.  Passed by value.
 *
 *      user_cnt_ptr= Pointer to user list count - may be NULL.
 *
 *      ent_size    = Size of individual entry in users list.  Unsigned
 *            long, passed by value.
 *
 *      keyarr  = Pointer to an array of PSL key structures. 
 *            Note that this routine assumes that the last element
 *            contains a key type of SKTYP_NONE. Passed by value.
 *
 *      options = Sorting options, unsigned long, passed by value.
 *            Currently these include,
 *
 *              PSL_VM_SORT_NODUPS - 
 *              Include only one instance of those
 *              entries that match on all keys.
 *
 *      diag_ptr    = Pointer to diagnostics stack.  Passed by value.
 */
unsigned long PSL_VM_SORT (struct psl_vm_llb **user_list,
                           unsigned long *user_cnt_ptr,
                           unsigned long ent_size,
                           struct psl_vm_skey *key_ptr,
                           unsigned long options,
                           struct diagstk *diagstk)
{
    /*
     *  Declare the local data.
     */
    unsigned long iret;         /* Routine return status */
    struct psl_vm_llb *whole_list;      /* Whole list of pointers */
    char *whocur_ptr;           /* Pointer to full list entry */
    char *whopos_ptr;           /* Pointer to position of */
    /* full entry */
    char *whonext_ptr;          /* Pointer to next full entry */
    char *user_recptr;          /* User list record buffer */
    struct globs pvs_gbl;           /* Global bits */

    /*
     *  Before we do anything, check that there is at least one element to
     *  the sort key.  If there isn't, don't sort anything.
     */
    if (key_ptr->key_typel == SKTYP_NONE)
    {
        iret = PSL_NORMAL;
        goto exit;
    }


    /*
     *  Create the 'whole list'.  This is a list of pointers to the real
     *  linked list entries.
     */
    iret = PSL_VM_INITIALIZE (sizeof(void *), &whole_list);
    if (!psl_test_status(iret)) goto err_init;

    /*
     *  Set up a global pointer to the users list, and other globals.
     */
    pvs_gbl.original_list = *user_list;
    pvs_gbl.diag_ptr = diagstk;
    if ((options & PSL_VM_SORT_NODUPS) != 0)
    {
        pvs_gbl.no_duplicates = PTRUE;
    }
    else
    {
        pvs_gbl.no_duplicates = PFALSE;
    }

    /*
     *  Get the first value from the user list.
     */
    iret = PSL_VM_COPY (pvs_gbl.original_list, NULL, NULL, &whonext_ptr,
                        PSL_VM_MODE_FIRST);

    /*
     *  Now fill in the list of pointers with the pointers to the list
     *  entries.
     */
    while (psl_test_status(iret))
    {
        whopos_ptr = whonext_ptr;
        iret = PSL_VM_INSERT (whole_list, (char *) &whopos_ptr, NULL, NULL,
                              PSL_VM_POS_END);
        if (!psl_test_status(iret)) goto err_insert;

        iret = PSL_VM_COPY (pvs_gbl.original_list, NULL, whopos_ptr,
                            &whonext_ptr, PSL_VM_MODE_NEXT);
    }

    /*
     *  If we didn't stop because we ran out of records, log an error.
     */
    if (iret != PSL_RECNOTFOU) goto err_copy;

    /*
     *  Create the new list which will contain the sorted entry pointers.
     *  Note that this list is global.
     */
    iret = PSL_VM_INITIALIZE (sizeof(void *), &pvs_gbl.srtd_ent_list);

    if (!psl_test_status(iret)) goto err_init;

    /*
     *  Allocate the buffers to hold each individual entry records during
     *  the sort comparison stage.  This is done at this stage with global
     *  pointers, so that it doesn't have to be done at every level of the
     *  sort.  (Hopefully, this will help to keep execution times down).
     */
    pvs_gbl.first_recptr = (char *) malloc (ent_size);
    pvs_gbl.next_recptr = (char *) malloc (ent_size);

    /*
     *  Now perform the sort operation.
     */
    iret = PSL_TREE_SORT (whole_list, key_ptr, &pvs_gbl);
    if (!psl_test_status(iret)) goto exit;

    /*
     *  Now that the sort is done, create a new list for the user and copy
     *  records from the original list in the order given by our sorted list
     *  of entry pointers.
     */
    iret = PSL_VM_INITIALIZE (ent_size, user_list);
    if (!psl_test_status(iret)) goto err_init;
    if (user_cnt_ptr != NULL) *user_cnt_ptr = 0;

    /*
     *  Allocate a buffer to hold each individual entry record.  Note that
     *  we can use space allocated for one of the sort comparison buffers,
     *  since we've finished the actual sort process now.
     */
    user_recptr = pvs_gbl.first_recptr;

    /*
     *  Get the first value form the sorted entry pointer list.
     */
    iret = PSL_VM_COPY (pvs_gbl.srtd_ent_list, (char *) &whocur_ptr,
                        NULL, &whonext_ptr, PSL_VM_MODE_FIRST);

    /*
     *  Now fill in the list of pointers with the pointers to the list
     *  entries.
     */
    while (psl_test_status(iret))
    {
        iret = PSL_VM_COPY (pvs_gbl.original_list, user_recptr, whocur_ptr,
                            NULL, PSL_VM_MODE_SPEC);
        if (!psl_test_status(iret)) goto err_copy;

        iret = PSL_VM_INSERT (*user_list, user_recptr, NULL, NULL,
                              PSL_VM_POS_END);
        if (!psl_test_status(iret)) goto err_insert;
        if (user_cnt_ptr != NULL) *user_cnt_ptr += 1;

        whopos_ptr = whonext_ptr;
        iret = PSL_VM_COPY (pvs_gbl.srtd_ent_list, (char *) &whocur_ptr,

                            whopos_ptr, &whonext_ptr, PSL_VM_MODE_NEXT);
    }

    /*
     *  If we didn't stop because we ran out of records, log an error.
     */
    if (iret != PSL_RECNOTFOU) goto err_copy;

    /*
     *  Free up the first (and user) and next sort comparison buffers.
     */
    free (pvs_gbl.first_recptr);
    free (pvs_gbl.next_recptr);

    /*
     *  Delete the sorted entry linked list.
     */
    iret = PSL_VM_DELETE (pvs_gbl.srtd_ent_list);
    if (!psl_test_status(iret)) goto err_delete;

    /*
     *  Delete the original 'entries' linked list that we have created.
     */
    iret = PSL_VM_DELETE (whole_list);
    if (!psl_test_status(iret)) goto err_delete;


    /*
     *  Delete the users original list.
     */
    iret = PSL_VM_DELETE (pvs_gbl.original_list);
    if (!psl_test_status(iret)) goto err_delete;

    /*
     *  Return Success status.
     */
    iret = PSL_NORMAL;
    goto exit;

    /*
     *  Log a problem with PSL_VM_INITIALIZE
     */
err_init:
    PSL_PUTDIAG (pvs_gbl.diag_ptr, 2, PSL_SYSERVERR,
                 "PSL_VM_INITIALIZE", iret);
    goto exit;

    /*
     *  Log a problem with PSL_VM_COPY
     */
err_copy:
    PSL_PUTDIAG (pvs_gbl.diag_ptr, 2, PSL_SYSERVERR,
                 "PSL_VM_COPY", iret);
    goto exit;

    /*
     *  Log a problem with PSL_VM_INSERT
     */
err_insert:
    PSL_PUTDIAG (pvs_gbl.diag_ptr, 2, PSL_SYSERVERR,
                 "PSL_VM_INSERT", iret);
    goto exit;

    /*
     *  Log a problem with PSL_VM_DELETE
     */
err_delete:
    PSL_PUTDIAG (pvs_gbl.diag_ptr, 2, PSL_SYSERVERR, "PSL_VM_DELETE", iret);

    goto exit;

    /*
     *  Exit with status.
     */
exit:
    return iret;
}


/*
 *  Function:  PSL_TREE_SORT
 *  =========
 *
 *  This routine should only be called from PSL_VM_SORT (and itself
 *  because it's recursive).  It takes as input a linked-list of pointers
 *  to the entries of the list to be sorted.  Output is to a global
 *  linked list of pointers, and this only occurs when the routine reaches
 *  the bottom of each branch in the tree.
 *
 *  Calling Format:
 *
 *      iret = PSL_TREE_SORT (whole_list, key_ptr, gbl_ptr)
 *
 *  where:
 *
 *      iret    = PSL return status.
 *      whole_list  = Pointer to list of entry pointers to be sorted.
 *            Passed by value.
 *      key_ptr = Pointer to a PSL_VM_SKEY structure.
 *      gbl_ptr = Pointer to PSL_VM_SORT global data.
 *
 */
static unsigned long PSL_TREE_SORT (struct psl_vm_llb *whole_list,
                                    struct psl_vm_skey *key_ptr,
                                    struct globs *gbl_ptr)
{


    /*
     *  Declare the local data.
     */
    unsigned long iret;         /* Routine return status */
    struct psl_vm_llb *leq_list;        /* Less than equal to list */
    unsigned long leq_count;        /* 'leq' list count */
    struct psl_vm_llb *grt_list;        /* Greater than list */
    unsigned long grt_count;        /* 'grt' list count */
    char *whocur_ptr;           /* Pointer to full list entry */
    char *whonext_ptr;          /* Pointer to next full entry */
    char *whopos_ptr;           /* Pointer to current
                                   full entry */
    short cmpsts;               /* Compare status */
    Boolean all_equal = PTRUE;      /* All entries match on key */

    /*
     *  Create the 'less than equal to' list.
     */
    iret = PSL_VM_INITIALIZE (sizeof(void *), &leq_list);
    if (!psl_test_status(iret)) goto err_init;
    leq_count = 0;

    /*
     *  Create the 'greater than' list.
     */
    iret = PSL_VM_INITIALIZE (sizeof(void *), &grt_list);
    if (!psl_test_status(iret)) goto err_init;
    grt_count = 0;

    /*
     *  Get the first value from the whole list and use it to start off the
     *  'less than equal to' list.
     */
    iret = PSL_VM_COPY (whole_list, (char *) &whocur_ptr, NULL,
                        &whonext_ptr, PSL_VM_MODE_FIRST);
    if (!psl_test_status(iret)) goto err_copy;

    /*
     *  Start off the 'less than equal to' list.
     */
    iret = PSL_VM_INSERT (leq_list, (char *) &whocur_ptr, NULL, NULL,
                          PSL_VM_POS_END);
    if (!psl_test_status(iret)) goto err_insert;
    leq_count += 1;

    /*
     *  Now get the full value of the first entry by using the pointer from 
     *  the whole list as a position reference into the original list.
     */
    iret = PSL_VM_COPY (gbl_ptr->original_list, gbl_ptr->first_recptr,
                        whocur_ptr, NULL, PSL_VM_MODE_SPEC);
    if (!psl_test_status(iret)) goto err_copy;


    /*
     *  Get the pointer to the next entry in the original list.
     */
    whopos_ptr = whonext_ptr;
    iret = PSL_VM_COPY (whole_list, (char *) &whocur_ptr, whopos_ptr,
                        &whonext_ptr, PSL_VM_MODE_NEXT);

    while (psl_test_status(iret))
    {
        /*
         *      Now get the full value of the next entry by using the pointer from 
         *      the whole list as a position reference into the original list.
         */
        iret = PSL_VM_COPY (gbl_ptr->original_list, gbl_ptr->next_recptr,
                            whocur_ptr, NULL, PSL_VM_MODE_SPEC);
        if (!psl_test_status(iret)) goto err_copy;

        /*
         *      Now compare the two values.
         */
        iret = PSL_SORT_COMPARE (gbl_ptr->next_recptr,
                                 gbl_ptr->first_recptr,
                                 key_ptr, &cmpsts, gbl_ptr);
        if (!psl_test_status(iret)) goto exit;

        /*
         *      If the values match by key, add the whole list entry to the
         *      end of the 'less than equal to' list.
         */
        if (cmpsts == 0)
        {
            iret = PSL_VM_INSERT (leq_list, (char *) &whocur_ptr,
                                  NULL, NULL, PSL_VM_POS_END);
            if (!psl_test_status(iret)) goto err_insert;
            leq_count += 1;
        }

        /*
         *      If the current value was less than the first (by key), add the 
         *      whole list entry to the beginning of the 'less than equal to' list.
         */
        else if (cmpsts < 0)
        {
            iret = PSL_VM_INSERT (leq_list, (char *) &whocur_ptr,
                                  NULL, NULL, PSL_VM_POS_BEG);
            if (!psl_test_status(iret)) goto err_insert;
            leq_count += 1;
            all_equal = PFALSE;
        }

        /*
         *      If the current value was greater than the first (by key), add the 
         *      whole list entry to the end of the 'greater than' list.
         */
        else
        {
            iret = PSL_VM_INSERT (grt_list, (char *) &whocur_ptr,
                                  NULL, NULL, PSL_VM_POS_END);
            if (!psl_test_status(iret)) goto err_insert;
            grt_count += 1;
        }

        /*
         *  Get the pointer to the next entry in the original list.
         */
        whopos_ptr = whonext_ptr;
        iret = PSL_VM_COPY (whole_list, (char *) &whocur_ptr, whopos_ptr,
                            &whonext_ptr, PSL_VM_MODE_NEXT);
    }

    /*
     *  If we didn't stop because we ran out of records, log an error.
     */
    if (iret != PSL_RECNOTFOU) goto err_copy;

    /*
     *  If there was only one entry in the original list, insert the entry
     *  into the list of sorted entries (which is global).
     */
    if ((leq_count == 1) && (grt_count == 0))
    {
        iret = PSL_VM_COPY (leq_list, (char *) &whocur_ptr, NULL, NULL,
                            PSL_VM_MODE_FIRST);
        if (!psl_test_status(iret)) goto err_copy;
        iret = PSL_VM_INSERT (gbl_ptr->srtd_ent_list, (char *) &whocur_ptr,
                              NULL, NULL, PSL_VM_POS_END);
    }

    /*
     *  If the 'greater than' list was empty, we only need to sort the
     *  'less than equal to' list, but look out for a key change!
     */
    else if (grt_count == 0)
    {
        /*
         *      If all in the list are equal, try a new key.
         */
        if (all_equal)
        {
            key_ptr += 1;
            if (key_ptr->key_typel != SKTYP_NONE)
            {
                iret = PSL_TREE_SORT (leq_list, key_ptr, gbl_ptr);
                if (!psl_test_status(iret)) goto exit;
            }
            else
            {
                iret = PSL_VM_COPY (leq_list, (char *) &whocur_ptr, NULL,
                                    &whonext_ptr, PSL_VM_MODE_FIRST);
                while (psl_test_status(iret))
                {
                    iret = PSL_VM_INSERT (gbl_ptr->srtd_ent_list,
                                          (char *) &whocur_ptr,
                                          NULL, NULL, PSL_VM_POS_END);

                    /*
                     *          If we are not to copy key duplicates, force a stop.
                     */
                    if (gbl_ptr->no_duplicates)
                    {
                        iret = PSL_RECNOTFOU;
                    }
                    /*
                     *          Duplicates are allowed so copy the next record.
                     */
                    else
                    {
                        whopos_ptr = whonext_ptr;
                        iret = PSL_VM_COPY (leq_list, (char *) &whocur_ptr,
                                            whopos_ptr, &whonext_ptr,
                                            PSL_VM_MODE_NEXT);
                    }
                }
                /*
                 *          If we didn't stop because we ran out of records, 
                 *          log an error.
                 */
                if (iret != PSL_RECNOTFOU) goto err_copy;
            }
        }
        /*
         *      If all in the list are not equal, keep sorting using the same key
         */
        else
        {
            iret = PSL_TREE_SORT (leq_list, key_ptr, gbl_ptr);
            if (!psl_test_status(iret)) goto exit;
        }
    }

    /*
     *  We have two lists with something in each, sort both.
     */
    else
    {
        iret = PSL_TREE_SORT (leq_list, key_ptr, gbl_ptr);
        if (!psl_test_status(iret)) goto exit;
        iret = PSL_TREE_SORT (grt_list, key_ptr, gbl_ptr);
        if (!psl_test_status(iret)) goto exit;
    }

    /*
     *  Delete the linked lists that we have created.
     */
    iret = PSL_VM_DELETE (leq_list);
    if (!psl_test_status(iret)) goto err_delete;

    iret = PSL_VM_DELETE (grt_list);
    if (!psl_test_status(iret)) goto err_delete;

    /*
     *  Return Success status.
     */
    iret = PSL_NORMAL;
    goto exit;

    /*
     *  Log a problem with PSL_VM_INITIALIZE
     */
err_init:
    PSL_PUTDIAG (gbl_ptr->diag_ptr, 2, PSL_SYSERVERR,
                 "PSL_VM_INITIALIZE", iret);
    goto exit;

    /*
     *  Log a problem with PSL_VM_COPY
     */
err_copy:
    PSL_PUTDIAG (gbl_ptr->diag_ptr, 2, PSL_SYSERVERR,
                 "PSL_VM_COPY", iret);
    goto exit;

    /*
     *  Log a problem with PSL_VM_INSERT
     */
err_insert:
    PSL_PUTDIAG (gbl_ptr->diag_ptr, 2, PSL_SYSERVERR,
                 "PSL_VM_INSERT", iret);
    goto exit;

    /*
     *  Log a problem with PSL_VM_DELETE
     */
err_delete:
    PSL_PUTDIAG (gbl_ptr->diag_ptr, 2, PSL_SYSERVERR,
                 "PSL_VM_DELETE", iret);
    goto exit;

    /*
     *  Exit with status.
     */
exit:
    return iret;
}


/*
 *  Function:  PSL_SORT_COMPARE
 *  =========
 *
 *  This routine is used to compare two entities using a specific key
 *  and return a status indicating the result of the comparison.
 *
 *  The key details are in the form of a PSL_VM_SKEY structure.
 *  
 *  If entity1 matches entity2 zero is returned (via a parameter).
 *
 *  If entity1 compares 'less than' entity2 a negative status is 
 *  returned (via a parameter), or positive if the reverse order flag 
 *  for the key is set.
 *
 *  If entity1 compares 'greater than' entity2 a positive status is 
 *  returned (via a parameter), or negative if the reverse order flag 
 *  for the key is set.
 *
 *  Calling Format:
 *
 *      iret = PSL_SORT_COMPARE (buff1, buff2, key_ptr, cmpsts_ptr, gbl_ptr)
 *
 *  where:
 *
 *      iret    = PSL return status.
 *
 *      buff1       = Pointer to users buffer #1.  Passed by value.
 *
 *      buff2       = Pointer to users buffer #2.  Passed by value.
 *
 *      key_ptr = Pointer to a PSL key structure.  Passed by value.
 *
 *      cmpsts  = Comparison return status. Passed by reference.
 *
 *      gbl_ptr = Pointer to (sort) global data.
 *
 *  NOTE: If any additional key types are added, support for them
 *        must be included in this routine.  It shouldn't be necessary
 *        to modify any other routines.
 */
static unsigned long PSL_SORT_COMPARE (void *buff1,
                                       void *buff2,
                                       struct psl_vm_skey *key_ptr,
                                       short *cmpsts_ptr,
                                       struct globs *gbl_ptr)
{
    /*
     *  Declare the local data.
     */
    unsigned long iret;     /* Routine return status */
    char *txt1_ptr;         /* Text pointer #1 */
    char *txt2_ptr;         /* Text pointer #2 */
    unsigned int *uint1_ptr;    /* unsigned int pointer #1 */
    unsigned int *uint2_ptr;    /* unsigned int pointer #2 */
    unsigned long *ulong1_ptr;  /* unsigned long pointer #1 */
    unsigned long *ulong2_ptr;  /* unsigned long pointer #2 */
    void *void_ptr;         /* Pointer to void */

    /*
     *  Perform this match when key is text type.
     */
    if (key_ptr->key_typel == SKTYP_TEXT)
    {
        txt1_ptr = (char *) buff1 + key_ptr->key_offsetl;
        txt2_ptr = (char *) buff2 + key_ptr->key_offsetl;
        *cmpsts_ptr = memcmp (txt1_ptr, txt2_ptr, key_ptr->key_lengthl);
    }

    /*
     *  Perform this match when key is unsigned long type.
     */
    else if (key_ptr->key_typel == SKTYP_ULONG)
    {
        void_ptr = ((char *) buff1) + key_ptr->key_offsetl;
        ulong1_ptr = (unsigned long *) void_ptr;

        void_ptr = ((char *) buff2) + key_ptr->key_offsetl;
        ulong2_ptr = (unsigned long *) void_ptr;

        if (*ulong1_ptr > *ulong2_ptr)
        {
            *cmpsts_ptr = 1;
        }
        else if (*ulong1_ptr < *ulong2_ptr)
        {
            *cmpsts_ptr = -1;
        }
        else
        {
            *cmpsts_ptr = 0;
        }
    }
    /*
     *  Perform this match when key is unsigned int type.
     */
    else if (key_ptr->key_typel == SKTYP_UINT)
    {
        void_ptr = ((char *) buff1) + key_ptr->key_offsetl;
        uint1_ptr = (unsigned int *) void_ptr;
        void_ptr = ((char *) buff2) + key_ptr->key_offsetl;
        uint2_ptr = (unsigned int *) void_ptr;
        if (*uint1_ptr > *uint2_ptr)
        {
            *cmpsts_ptr = 1;
        }
        else if (*uint1_ptr < *uint2_ptr)
        {
            *cmpsts_ptr = -1;
        }
        else
        {
            *cmpsts_ptr = 0;
        }
    }

    /*
     *  The key type is invalid, log an error and exit.
     */
    else
    {
        PSL_PUTDIAG (gbl_ptr->diag_ptr, 0, PSL_BADPARAM);
        (void) PSL_ADD_STATDIAG (gbl_ptr->diag_ptr,
                                 "Invalid sort key type specified");
        iret = PSL_BADPARAM;
        goto exit;
    }

    /*
     *  Ordering by default is ascending, so if the user requested descending
     *  order reverse the comparison status value.
     */
    if (key_ptr->rev_ordern)
    {
        if (*cmpsts_ptr < 0)
        {
            *cmpsts_ptr = 1;
        }
        else if (*cmpsts_ptr > 0)
        {
            *cmpsts_ptr = -1;
        }
    }

    /*
     *  Exit with success
     */
    iret = PSL_NORMAL;
    goto exit;

    /*
     *  Return with status.
     */
exit:
    return iret;
}
