/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

/*===========================================================================*/
/*                                       */
/*               **** C O M M O N ****                   */
/*                                       */
/*===========================================================================*/
/*                                       */
/*          Module Name: STR_ROUTINES                */
/*          ============                         */
/*                                       */
/* Description:                                  */
/* ============                                  */
/*                                       */
/*  String handling routines.                        */
/*                                       */
/*===========================================================================*/

/*
 *  Include PSL Macros.
 */
#include <psl_macros.h>

/*
 *  Standard ANSI header files.
 */
#include <ctype.h>
#include <stddef.h>
#include <string.h>

/*
 *  Function Prototypes.
 */
#include <psl_rtndef.h>         /* PSL Function Prototypes */


/*
 *  Local function prototypes.
 */
static  void    CNV_CHR_CMD (char cmdchr,
                             char ctlchr,
                             long ctlmnc,
                             char inpchr,
                             char *outbuf_ptr,
                             long *outlen_ptr,
                             long flags);

static  void    CNV_CHR_DSP (char ctlchr,
                             long ctlmnc,
                             char inpchr,
                             char *outbuf_ptr,
                             long *outlen_ptr);

static  void    CNV_FKEY    (char *inp_ptr,
                             long inplen,
                             char *out_ptr,
                             long *outlen_ptr,
                             long *esclen_ptr);

/*
 *  VMS-Specific Structures
 */
#ifdef VMS
#include <descrip.h>            /* VAX Descriptor Format */
#endif /* VMS */


/*
 *  Include Common Data.
 */
#include <shrmsg.h>         /* Message Statuses */
#include <str_prmdef.h>         /* String routine param's */


/*
 *  Define control character values for convenience.
 */
#define CTRL_BELL       0X07    /* Bell control value */
#define CTRL_BS         0X08    /* Backspace control value */
#define CTRL_HT         0X09    /* H. Tab control value */
#define CTRL_LF         0X0A    /* Line feed control value */
#define CTRL_VT         0X0B    /* V. Tab control value */
#define CTRL_FF         0X0C    /* Form Feed control value */
#define CTRL_CR         0X0D    /* Carriage control control value */
#define CTRL_ESC        0X1B    /* Escape control value */
#define CTRL_SP         0X20    /* Space value */
#define CTRL_RUBOUT     0X7F    /* Delete/rubout control value */
#define CTRL_SS3        0X8F    /* SS3 control value */
#define CTRL_CSI        0X9B    /* CSI control value */

#define CTRL_LOWPRINT           0X20    /* Lowest printable ASCII value */
#define CTRL_HIGHPRINT          0X7E    /* Highest printable ASCII value */


/*======================================================================*/
/*  STR_EXPAND_CTL - Expand Control Characters in ASCII String  */
/*                                  */
/*  This routine converts the supplied ASCII string, which may  */
/*  contain non-printable characters, into an ASCII string in   */
/*  which all non-printable characters and 8-bit characters are */
/*  represented by a string of printable characters. The format */
/*  of the output string is determined by the specified control */
/*  character mnemonic and flags as follows:            */
/*                                  */
/*  STR_MNC_HEX                         */
/*    All control characters will be represented using the      */
/*    hexadecimal ASCII code. This mode will always be used for */
/*    8-bit characters.                     */
/*                                  */
/*  STR_MNC_CONTROL                         */
/*    All 7-bit control characters will be represented in the   */
/*    form "^x", and the rubout character will be represented   */
/*    by <DEL>.                         */
/*                                  */
/*  STR_MNC_ASCII                           */
/*    Where possible, control characters will be represented by */
/*    their ASCII mnemonic. Control characters without an ASCII */
/*    mnemonic will be represented as for "Control".        */
/*                                  */
/*  STR_ECF_CMDM                            */
/*    The command character, control character and characters   */
/*    represented using the hexadecimal ASCII code or ASCII     */
/*    mnemonic will be represented by character commands, e.g.  */
/*    the escape character will be represented as "~<ESC>" or   */
/*    "~<1B>" rather than "<ESC>" or "<1B>".            */
/*                                  */
/*  STR_ECF_FKYM                            */
/*    Character strings corresponding to function key escape    */
/*    sequences will be represented using the function key name,    */
/*    e.g. the string or sub-string "ESC[B" will be represented */
/*    by "<DOWN>".                          */
/*                                  */
/*  STR_ECF_PASCMDM                         */
/*    If STR_ECF_CMDM is also specified a command character     */
/*    will be returned unchanged.                   */
/*                                  */
/*  STR_ECF_WLDCMDM                         */
/*    If STR_ECF_CMDM is also specified single character and    */
/*    multi-character wildcard characters will be represented by    */
/*    character commands, i.e. "*" will be represented by "~*"  */
/*    and "%" by "~%" (on VMS). Note that these commands are    */
/*    currently only supported by the V-TIMER timer definition  */
/*    language.                         */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*  STR_EXPAND_CTL (cmdchr, ctlchr, ctlmnc, inpstr_ptr,     */
/*          inpstrlen, outstr_ptr, outbuflen,       */
/*          uncnvcnt_ptr, flags);               */
/*                                  */
/*      Where:                              */
/*                                  */
/*  cmdchr       = (char) Command character             */
/*  ctlchr       = (char) Control character             */
/*  ctlmnc       = (long) Control character mnemonic:       */
/*           STR_MNC_ASCII   - ASCII string, e.g. <ESC> */
/*           STR_MNC_CONTROL - Control specifier, e.g. ^[   */
/*           STR_MNC_HEX     - Hexadecimal code, e.g. <1B>  */
/*  inpstr_ptr   = (char *) Pointer to input ASCII string       */
/*  inpstrlen    = (long) Input ASCII string length         */
/*  outstr_ptr   = (char *) Pointer to output string        */
/*  outbuflen    = (long) Output string buffer length       */
/*  uncnvcnt_ptr = (long *) Pointer to variable to receive count    */
/*             of characters not converted (may be NULL).   */
/*  flags        = (long) Expansion flags:              */
/*           STR_ECF_CMDM    - Return character commands,   */
/*                     e.g. return ~<ESC> not <ESC> */
/*           STR_ECF_FKYM    - Convert function key     */
/*                     sequences, e.g. <DOWN>   */
/*           STR_ECF_PASCMDM - Pass command characters  */
/*                     through unchanged.       */
/*           STR_ECF_WLDCMDM - Return character commands    */
/*                     for wildcard characters, */
/*                     e.g. ~* and ~% on VMS.   */
/*                                  */
/*======================================================================*/
void STR_EXPAND_CTL (char cmdchr,
                     char ctlchr,
                     long ctlmnc,
                     char *inpstr_ptr,
                     long inpstrlen,
                     char *outstr_ptr,
                     long outbuflen,
                     long *uncnvcnt_ptr,
                     long flags)
{
    /*
     *  Local data
     */
    char       wrk_buffer[40]; /* Buffer to hold converted character */
    long       escsize;        /* Number of bytes in escape sequence */
    long       copysize;       /* Number of bytes to copy */
    long       loop;       /* Loop control variable */
    long       outlen;     /* Output string length */

    /*
     *  Loop through the input string converting each character at a time until
     *  we reach the end of the string.
     */
    for (loop = 0, outlen = 0, copysize = 0;
            loop < inpstrlen;
            loop++, outlen += copysize)
    {
        if ((flags & STR_ECF_FKYM) != 0)
        {
            CNV_FKEY (          /* Look for function key sequence */
                (inpstr_ptr + loop),    /* look here */
                (inpstrlen - loop), /* length of input string remaining */
                wrk_buffer,     /* return string here */
                &copysize,      /* return its length here */
                &escsize);      /* return escape sequence length here */
        }
        else
        {
            copysize = 0;       /* No function key string */
        }

        if (copysize == 0)      /* If not function key sequence */
        {
            if ((flags & STR_ECF_CMDM) != 0)
            {               /* If function keys wanted */
                CNV_CHR_CMD (       /* Convert to command string */
                    cmdchr,     /* command character */
                    ctlchr,     /* control character */
                    ctlmnc,     /* control character mnemonic */
                    *(inpstr_ptr + loop),
                    /* this character */
                    wrk_buffer,     /* return string here */
                    &copysize,      /* length of returned string */
                    flags);     /* flags */
            }
            else
            {
                CNV_CHR_DSP (       /* Convert to display string */
                    ctlchr,     /* control character */
                    ctlmnc,     /* control character mnemonic */
                    *(inpstr_ptr + loop),
                    /* this character */
                    wrk_buffer,     /* return string here */
                    &copysize);     /* length of returned string */
            }
        }
        else
        {
            loop += escsize - 1;    /* Skip escape sequence */
        }

        if ((outlen + copysize + 1) > outbuflen) goto cnv_done;
        /* If string too long, all done */

        /*
         *  Copy the string to the output buffer.
         */
        (void) memcpy (         /* Copy string */
            (outstr_ptr + outlen),  /* to the output buffer */
            wrk_buffer,     /* from the work buffer */
            copysize);      /* this many characters */

    }                   /* End of for loop */

    /*
     *  CNV_DONE:
     *  =========
     *
     *  Come here when the string has been set up. Null terminate the output
     *  string and return the count of unconverted characters (if requested).
     */
cnv_done:
    *(outstr_ptr + outlen) = '\0';  /* Null terminate output string */

    if (uncnvcnt_ptr != NULL)       /* If unconverted character count req */
    {
        *uncnvcnt_ptr = inpstrlen - loop;
    }

    return;
}                   /* End STR_EXPAND_CTL routine */


#ifdef VMS

/************************************************************************/
/*                                  */
/*  STR_LOWCASE - Convert string to lower case          */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     STR_LOWCASE (src, dst);                  */
/*                                  */
/*  Where:                              */
/*                                  */
/*     src  = source string, passed by Descriptor           */
/*     dst  = destination string, passed by Descriptor      */
/*                                  */
/************************************************************************/
void    STR_LOWCASE (
    struct dsc$descriptor_s *src,           /* Source String */
    struct dsc$descriptor_s *dst)           /* Destination String */

{
    /*
     *  Local Data.
     */
    char *src_ptr;              /* Pointer to Source String */
    char *dst_ptr;              /* Pointer to Dest. String */
    /*
     *  Initialize Pointers.
     */
    src_ptr = (*src).dsc$a_pointer;
    dst_ptr = (*dst).dsc$a_pointer;
    /*
     *  Loop through the source string copying/translating characters
     *  into the destination.
     */
    while ( (src_ptr < ((*src).dsc$a_pointer + (*src).dsc$w_length)) &&
            (dst_ptr < ((*dst).dsc$a_pointer + (*dst).dsc$w_length)) )

    {
        /*
         *  Copy/translate character.
         */
        *dst_ptr    = tolower (*src_ptr);
        /*
         *  Increment Pointers.
         */
        src_ptr++;
        dst_ptr++;
    }
    return;
}
#endif


/************************************************************************/
/*                                  */
/*  STR_REPLACE_NONPRINTING - Replace non-printing characters   */
/*                                  */
/*    This routine replaces all non-printing characters in a buffer     */
/*    with a specified character.                   */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*  iret    = STR_REPLACE_NONPRINTING (buffer, count, filler);  */
/*                                  */
/*  Where:                              */
/*    iret   = Return Status (SS$_NORMAL)               */
/*    buffer = Pointer to buffer to be modified.            */
/*    count  = Count of bytes to be checked.            */
/*    filler = Replacement character.               */
/*                                  */
/************************************************************************/
unsigned long   STR_REPLACE_NONPRINTING (char *buffer,
        unsigned long count,
        char filler)
{
    /*
     *  Local Data.
     */
    char *buff_ptr;         /* Pointer to Buffer */

    /*
     *  Initialize Pointer
     */
    buff_ptr = buffer;


    /*
     *  Loop through the source string replacing characters as required.
     *  By default we just use the ISPRINT macro, but this is inadequate
     *  for VMS where we want to handle 8-bit characters (ISPRINT only
     *  supports 7-bit characters) so we must do it by hand.  Note that
     *  VMS treats char as signed so we must use decimal values rather
     *  than the more legible hexadecimal ones (as the latter are always
     *  unsigned).
     */
    while (buff_ptr < (buffer + count))
    {
        /*
         *  Check character and replace if non-printing
         */
#ifdef VMS
        if ((*buff_ptr <= -96)
                ||  ((*buff_ptr >= -1) && (*buff_ptr <= 31))
                ||  (*buff_ptr == 127))
#else
        if (!isprint(*buff_ptr))
#endif
        {
            *buff_ptr = filler;
        }

        /*
         *  Increment Pointer
         */
        buff_ptr += 1;
    }
    return PSL_NORMAL;
}

/*======================================================================*/
/*                                  */
/*                                  */
/*    PSL_MEMSTR - Check for string1 contains string2           */
/*    ==========                            */
/*                                  */
/*    Calling Convention:                       */
/*  char_ptr = PSL_MEMSTR (str1_ptr, str1_len, str2_ptr, str2_len)  */
/*                                  */
/*    Inputs:                               */
/*  string1_ptr - Pointer to char main string           */
/*  string1_len - UL length of string1              */
/*  string2_ptr - Pointer to char sub-string            */
/*  string2_len - UL length of string2              */
/*                                  */
/*    Outputs:                              */
/*  None                                */
/*                                  */
/*    Return Status:                            */
/*  Returns a pointer to the first occurence in STRING1 of STRING2  */
/*  or NULL if not present.                     */
/*                                  */
/*    This routine is called to check for STRING1 containing STRING2    */
/*    where either string may contain null characters.          */
/*                                  */
/*    If string1 does contain string2, a pointer to the first occurence */
/*    of that string in string1 is returned.                */
/*                                  */
/*======================================================================*/

char        *PSL_MEMSTR (         char *string1_ptr
                                  ,unsigned long string1_len
                                  ,         char *string2_ptr
                                  ,unsigned long string2_len)

{
    /*
     *    Local data
     */
    unsigned long   iret;           /* Status return variable   */
    unsigned long   string1_remlen;     /* Remaining length of string1  */
    char   *firstchr_ptr;      /* Pointer to first character found */
    char   *return_ptr;        /* return pointer       */
    long   offset;         /* offset of first character    */

    return_ptr = NULL;          /* Assume failure in this routine */


    /*
     *    Validate the parameters. 
     *
     *    1> If STRING1 or STRING2 is NULL return not contains
     *    2> If STRING1 length is zero return not contains.
     *    3> If STRING2 length is zero return pointer to start of STRING1
     */
    if ((string1_ptr == NULL) || (string2_ptr == NULL)
            ||  (string1_len == 0))
    {
        goto psl_memstr_end;        /* ...branch to return      */
    }

    if (string2_len == 0)       /* If string2 length zero   */
    {
        return_ptr = string1_ptr;   /* Return pointer to start of STRING1 */
        goto psl_memstr_end;        /* ...branch to return      */
    }

    /*
    *    Find the first instance in STRING1 of the first character in STRING2
    *    If we cannot find an instance then STRING1 cannot contain STRING2 and
    *    we should return failure.
    */
    firstchr_ptr = string1_ptr;     /* Point to start of string1    */
    string1_remlen = string1_len;   /* Set initial length of string1   */

    do
    {
        firstchr_ptr = (char *) memchr (firstchr_ptr
                                        ,*string2_ptr
                                        ,string1_remlen);
        /* Locate first character in STRING1 */

        if (firstchr_ptr == NULL)   /* If not found the first character */
        {
            goto psl_memstr_end;    /* ...branch to return      */
        }


        /*
         *    We found an instance of the first character of string2 in string1
         *    now scan forward for the rest of the string.
         */
        offset = (char *) firstchr_ptr - string1_ptr;
        /* Offset of the located character */
        string1_remlen -= offset;   /* calculate remaining length   */

        /*
         *    Check to see if there is sufficient room left in string1 for string2 to
         *    be present and return if there is not.
         */
        if (string1_remlen < string2_len)
        {               /* If string1 length < string2  */
            goto psl_memstr_end;    /* ...branch to return      */
        }

        /*
         *    Check to see if string2 is contained within string1 at the position
         *    of the first character.
         */
        iret = memcmp (firstchr_ptr, string2_ptr, string2_len);
        /* Compare string1 part and string2 */
        if (iret == 0)          /* If string2 found in string1  */
        {
            return_ptr = firstchr_ptr;  /* Set to return pointer to it  */
        }

        else                /* If string2 not found here    */
        {
            firstchr_ptr += 1;      /* Point past first character   */
            string1_remlen -= 1;    /* ...and decrement length left */
            string1_ptr = firstchr_ptr; /* Move forwrd base string poiinter */
        }
    }

    while ((string1_remlen != 0) && (return_ptr == NULL));

    /*
     *    PSL_MEMSTR_END
     *    ---------------
     *
     *    We branch to this label when we have completed all processing and are
     *    about to return (with the return pointer) to the caller. The return
     *    pointer must be set up at this point.
     */
psl_memstr_end:
    return return_ptr;      /* Return pointer to contains string */
}

/*======================================================================*/
/*                                  */
/*                                  */
/*    PSL_MEMRCHR - Find last occurrence of a character in string   */
/*    ===========                           */
/*                                  */
/*    Calling Convention:                       */
/*  char_ptr = PSL_MEMRCHR (str1_ptr, str2, str1_len)       */
/*                                  */
/*    Inputs:                               */
/*  string1_ptr - Pointer to string to locate character in      */
/*  string2     - Character to locate               */
/*  string1_len - UL length of string1              */
/*                                  */
/*    Outputs:                              */
/*  None                                */
/*                                  */
/*    Return Status:                            */
/*  Returns a pointer to the last occurence in STRING1 of STRING2   */
/*  or NULL if not present.                     */
/*                                  */
/*    This routine is called to return a pointer to the last occurence  */
/*    of the character specified as STRING2 in the buffer specified as  */
/*    STRING1 where STRING1 may contain null characters.        */
/*                                  */
/*    If string1 does not contain string2, a NULL pointer is returned.  */
/*                                  */
/*======================================================================*/

char    *PSL_MEMRCHR    (         char *string1_ptr
                                  ,     char string2
                                  ,unsigned long string1_len)

{
    /*
     *    Local data
     */
    unsigned long   isub;           /* Loop control variable    */
    char   *wrk_ptr;       /* Work pointer to string   */
    char   *return_ptr = NULL; /* Return pointer       */

    /*
     *    Validate the parameters. 
     *
     *    1> If STRING1 is NULL return as if NOT FOUND
     *    2> If STRING1 length is 0 return as if NOT FOUND
     */
    if ((string1_ptr == NULL) || (string1_len == 0))
    {
        goto psl_memrchr_end;       /* ...branch to return      */
    }

    /*
     *    Simply loop backwards through the string until we reach the end or
     *    find an occurence of the required character.
     */
    wrk_ptr = string1_ptr + string1_len - 1;
    /* Point to end of string   */
    for (isub = string1_len; isub != 0; isub--, wrk_ptr--)
    {
        if (*wrk_ptr == string2)    /* If we found a match      */
        {
            return_ptr = wrk_ptr;   /* Set return pointer to char   */
            goto psl_memrchr_end;   /* ...and branch to exit    */
        }
    }

    /*
     *    PSL_MEMRCHR_END
     *    ---------------
     *
     *    We branch to this label when we have completed all processing and are
     *    about to return (with the return pointer) to the caller. The return
     *    pointer must be set up at this point.
     */
psl_memrchr_end:
    return return_ptr;      /* Return pointer to contains string */
}

/*======================================================================*/
/*                                  */
/*    PSL_SUBSTR - Locate string2 in string1 & replace with string3 */
/*    ==========                            */
/*                                  */
/*    Calling Convention:                       */
/*  iret = PSL_SUBSTR (string1_ptr, string1_size, string2_ptr,  */
/*              string3_ptr)                */
/*                                  */
/*    Inputs:                               */
/*  string1_ptr  - Pointer to char main string (null-terminated)    */
/*  string1_size - UL maximum size of string1           */
/*  string2_ptr  - Pointer to char locate string (null-terminated)  */
/*  string3_ptr  - Pointer to char replacement string (null-term.)  */
/*          May be NULL pointer             */
/*                                  */
/*    Outputs:                              */
/*  Updated string1 if replacement made             */
/*                                  */
/*    Return Status:                            */
/*  PSL Return Status: PSL_NORMAL    - String Replaced      */
/*                     PSL_BADPARAM  - Invalid Parameter Specified  */
/*                     PSL_NOMTCHFOU - String not found     */
/*                     PSL_NOROOMSTR - No room for replacement  */
/*                                  */
/*    This routine will attempt to locate STRING2 in STRING1.           */
/*    If found, STRING3 will be inserted in it's place (if it will fit).*/
/*    STRING3 may be NULL to allow STRING2 just to be removed       */
/*                                  */
/*======================================================================*/

unsigned long   PSL_SUBSTR  (char *string1_ptr,
                             unsigned long string1_size,
                             const char *string2_ptr,
                             const char *string3_ptr)

{
    /*
     *  Local data
     */
    unsigned long   iret;           /* Status return variable */
    char   *chr_ptr;       /* Pointer to string found */
    unsigned long   strsize;        /* Size of 'cut down' string */
    unsigned long   movsize;        /* Size of string to move */

    /*
     *  Validate the parameters - PSL_BADPARAM if fails
     *
     *  1> If STRING1 or STRING2 is NULL
     *  2> If STRING1 length is zero
     */
    if ((string1_ptr == NULL) || (string2_ptr == NULL) || (string1_size == 0))
    {
        iret = PSL_BADPARAM;
        goto end;
    }

    /*
     *  Assume that we can not find the string.
     */
    iret = PSL_NOMTCHFOU;

    /*
     *  Locate the required string (as many times as it exists).
     */
    chr_ptr = strstr (string1_ptr, string2_ptr);
    while (chr_ptr != NULL)
    {

        /*
         *  We have found the string - have we enough room to replace it with the
         *  substitution string?
         *
         *  Take the current length of STRING1 and remove the size of STRING2.
         *  This will give us the size left for the insertion of STRING3.
         *    - take into account null-termination.
         */
        strsize = string1_size - strlen(string1_ptr) + strlen(string2_ptr) - 1;

        /*
         *  Have we enough room to insert STRING3?
         */
        if ((string3_ptr == NULL) || (strlen (string3_ptr) <= strsize))
        {
            movsize = string1_size - (chr_ptr - string1_ptr);
            if (movsize != 0)
            {
                (void) memmove (chr_ptr,
                                chr_ptr + strlen (string2_ptr),
                                movsize);
            }

            /*
             *      We have removed the string: now insert the new string (if necessary)
             */
            if (string3_ptr != NULL)
            {
                movsize = strlen (string1_ptr) - (chr_ptr - string1_ptr) + 1;
                (void) memmove (chr_ptr + strlen (string3_ptr),
                                chr_ptr,
                                movsize);

                /*
                 *      We have the gap in the string - now insert the new string.
                 */
                (void) memcpy (chr_ptr, string3_ptr, strlen(string3_ptr));

                /*
                 *      We have successfully done the replacement - change the status.
                 */
                iret = PSL_NORMAL;
            }
        }
        else if (string3_ptr != NULL)
        {
            iret = PSL_NOROOMSTR;
            goto end;
        }
        /*
         *    Advance search pointer to first character after substitution
         */
        chr_ptr += strlen (string3_ptr);

        /*
         *  Can we find the string again?
         */
        chr_ptr = strstr (chr_ptr, string2_ptr);


    }

    /*
     *    END:
     *    ---
     *
     *    At this point we have done all we can and will be returning the
     *    final status to the user.
     */
end:
    return iret;
}



/************************************************************************/
/*                                  */
/*  STROLD_REPL_NOPRINT - Old-style replace non-printing characters */
/*                                  */
/*    This routine replaces all non-printing characters in a buffer     */
/*    with a specified character.  It is simply a shell around      */
/*    STR_REPLACE_NONPRINTING to keep pre-V2.3 COMMON users happy.  */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*  iret    = STROLD_REPL_NOPRINT (buffer, count_ptr, filler_ptr);  */
/*                                  */
/*  Where:                              */
/*    iret   = Return Status (SS$_NORMAL)               */
/*    buffer = Pointer to buffer to be modified.            */
/*    count  = Pointer to count of bytes to be checked.     */
/*    filler = Pointer to replacement character.            */
/*                                  */
/************************************************************************/
unsigned long   STROLD_REPL_NOPRINT (char *buf_ptr,
                                     unsigned long *count_ptr,
                                     char *filler_ptr)
{
    return STR_REPLACE_NONPRINTING (buf_ptr, *count_ptr, *filler_ptr);
}



/************************************************************************/
/*  CNV_CHR_CMD - Convert ASCII Character to Command Format     */
/*                                  */
/*  This routine takes as input an ASCII character and returns a    */
/*  character string. If the input character is a 7-bit printable   */
/*  the character is returned unchanged. However, all control   */
/*  characters and all 8-bit characters will be represented by  */
/*  character commands or control commands as summarised in the */
/*  following table.                        */
/*                                  */
/*      Input Character     Output Strings          */
/*      ---------------     --------------          */
/*                                  */
/*      Command character       ~~ (unless pass through)    */
/*      Control character       ~^              */
/*      "*"             ~* (if literal command required */
/*                      for wildcards)      */
/*      "%" (VMS)           ~% (if literal command required */
/*                      for wildcards)      */
/*      "?" (UNIX)          ~? (if literal command required */
/*                      for wildcards)      */
/*      Printable ASCII Character   ASCII Character         */
/*      Line Feed           ~<LF>,  ^J, ~<0A>       */
/*      Carriage Return     ~<CR>,  ^M, ~<0D>       */
/*      Escape          ~<ESC>, ^[, ~<1B>       */
/*      Bell            ~<BEL>, ^G, ~<07>       */
/*      Backspace           ~<BS>,  ^H, ~<08>       */
/*      Horizontal Tab      ~<HT>,  ^I, ~<09>       */
/*      Vertical Tab        ~<VT>,  ^K, ~<0B>       */
/*      Del             ~<DEL>,     ~<7F>       */
/*      Other 7-bit control             ^x, ~<xx>       */
/*        characters                        */
/*      Other characters                ~<xx>       */
/*                                  */
/*      where   ~ is assumed to be the command character    */
/*          x  is the control character, e.g. ^Z        */
/*              xx is the character's ANSI code, e.g. ~<1A> */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*  CNV_CHR_CMD (cmdchr, ctlchr, ctlmnc, inpchr, outbuf_ptr,    */
/*       outlen_ptr, flags);                    */
/*                                  */
/*      Where:                              */
/*                                  */
/*  cmdchr     =  (char) Command character              */
/*  ctlchr     =  (char) Control character              */
/*  ctlmnc     =  (long) Control character mnemonic         */
/*  inpchr     =  (char) Character to be converted          */
/*  outbuf_ptr =  (char *) Pointer to buffer to hold the returned   */
/*            string                        */
/*  outlen_ptr =  (long *) Pointer to longword holding length of    */
/*            converted string. It will be updated with the */
/*            length of the copy.               */
/*  flags        = (long) Expansion flags:              */
/*           STR_ECF_PASCMDM - Pass command characters  */
/*                     through unchanged.       */
/*           STR_ECF_WLDCMDM - Return character commands    */
/*                     for wildcard characters, */
/*                     e.g. ~* and ~% on VMS.   */
/*                                  */
/************************************************************************/
static void CNV_CHR_CMD (char cmdchr,
                         char ctlchr,
                         long ctlmnc,
                         char inpchr,
                         char *outbuf_ptr,
                         long *outlen_ptr,
                         long flags)
{
    /*
     *  Local data.
     */
    unsigned char       value;      /* Work value */
    int        outlen;     /* Length of converted text */

    /*
     *    Command Character Literal Specified?
     *    ------------------------------------
     *
     *    Return a buffer containing the command character literal command "~~"
     *    (if command character is tilde - the default), unless command characters
     *    are to be passed through unchanged.
     */
    if ((inpchr == cmdchr) &&           /* If command character  AND */
            ((flags & STR_ECF_PASCMDM) == 0))   /* NOT pass unchanged */
    {
        (void) sprintf (
            outbuf_ptr,
            "%c%c%n",
            cmdchr,
            cmdchr,
            &outlen);       /* Format the output */
    }

    /*
     *    Single Character Wildcard Literal Specified?
     *    --------------------------------------------
     *
     *    Return a buffer containing the single character wildcard literal
     *    command, i.e. "~%" (VMS) and "~?" (UNIX) (if command character is tilde).
     */
#ifdef VMS
    else if ((inpchr == '%') &&         /* If character is "%" AND */
             ((flags & STR_ECF_WLDCMDM) != 0))  /* return literal command */
    {
        (void) sprintf (
            outbuf_ptr,
            "%c%%%n",
            cmdchr,
            &outlen);       /* Format the output */
    }
#else
    else if ((inpchr == '?') &&         /* If character is "?" AND */
             ((flags & STR_ECF_WLDCMDM) != 0))  /* return literal command */
    {
        (void) sprintf (
            outbuf_ptr,
            "%c?%n",
            cmdchr,
            &outlen);       /* Format the output */
    }
#endif

    /*
     *    Multi-Character Wildcard Literal Specified?
     *    -------------------------------------------
     *
     *    Return a buffer containing the multi-character wildcard literal
     *    command "~*" (if command character is tilde).
     */
    else if ((inpchr == '*') &&         /* If character is "*" AND */
             ((flags & STR_ECF_WLDCMDM) != 0))  /* return literal command */
    {
        (void) sprintf (
            outbuf_ptr,
            "%c*%n",
            cmdchr,
            &outlen);       /* Format the output */
    }

    /*
     *    Control character specified?
     *    ----------------------------
     *
     *    Return a buffer containing the control character prefixed by the command
     *    character.
     */
    else if (inpchr == ctlchr)      /* If control character */
    {
        (void) sprintf (
            outbuf_ptr,
            "%c%c%n",
            cmdchr,
            ctlchr,
            &outlen);       /* Format the output */
    }

    /*
     *    Line Feed Specified ?
     *    ---------------------
     *
     *    Return a buffer containing the line feed mnemonic "~<LF>" "<LF>" or "^J"
     *    or "~<0A> depending on the format requested by the user.
     */
    else if (inpchr == CTRL_LF)
    {
        if (ctlmnc == STR_MNC_ASCII)
        {               /* If ASCII representation */
            (void) sprintf (outbuf_ptr, "%c<LF>%n"
                            ,cmdchr
                            ,&outlen);      /* Format the output */
        }
        else if (ctlmnc == STR_MNC_CONTROL)
        {               /* If CONTROL representation */
            (void) sprintf (outbuf_ptr, "%cJ%n"
                            ,ctlchr
                            ,&outlen);      /* Format the output */
        }
        else                /* ...else hex representation */
        {
            (void) sprintf (outbuf_ptr, "%c<0A>%n"
                            ,cmdchr
                            ,&outlen);      /* Format the output */
        }
    }

    /*
     *    Carriage Return Specified ?
     *    ---------------------------
     *
     *    Return a buffer containing the carriage return mnemonic "~<CR>" or "^M"
     *    or "~<0A> depending on the format requested by the user.
     */
    else if (inpchr == CTRL_CR)
    {
        if (ctlmnc == STR_MNC_ASCII)
        {               /* If ASCII representation */
            (void) sprintf (outbuf_ptr, "%c<CR>%n"
                            ,cmdchr
                            ,&outlen);      /* Format the output */
        }
        else if (ctlmnc == STR_MNC_CONTROL)
        {               /* If CONTROL representation */
            (void) sprintf (outbuf_ptr, "%cM%n"
                            ,ctlchr
                            ,&outlen);      /* Format the output */
        }
        else                /* ...else hex representation */
        {
            (void) sprintf (outbuf_ptr, "%c<0D>%n"
                            ,cmdchr
                            ,&outlen);      /* Format the output */
        }
    }

    /*
     *    Escape Specified ?
     *    ------------------
     *
     *    Return a buffer containing the line feed mnemonic "~<ESC>" or "^["
     *    or "~<1B> depending on the format requested by the user.
     */
    else if (inpchr == CTRL_ESC)
    {
        if (ctlmnc == STR_MNC_ASCII)
        {               /* If ASCII representation */
            (void) sprintf (outbuf_ptr, "%c<ESC>%n"
                            ,cmdchr
                            ,&outlen);      /* Format the output */
        }
        else if (ctlmnc == STR_MNC_CONTROL)
        {               /* If CONTROL representation */
            (void) sprintf (outbuf_ptr, "%c[%n"
                            ,ctlchr
                            ,&outlen);      /* Format the output */
        }
        else                /* ...else hex representation */
        {
            (void) sprintf (outbuf_ptr, "%c<1B>%n"
                            ,cmdchr
                            ,&outlen);      /* Format the output */
        }
    }

    /*
     *    Bell character Specified ?
     *    --------------------------
     *
     *    Return a buffer containing the BELL menmonic "~<BEL>" or "^G" or "~<07>
     *    depending on the format requested by the user.
     */
    else if (inpchr == CTRL_BELL)
    {
        if (ctlmnc == STR_MNC_ASCII)
        {               /* If ASCII representation */
            (void) sprintf (outbuf_ptr, "%c<BEL>%n"
                            ,cmdchr
                            ,&outlen);      /* Format the output */
        }
        else if (ctlmnc == STR_MNC_CONTROL)
        {               /* If CONTROL representation */
            (void) sprintf (outbuf_ptr, "%cG%n"
                            ,ctlchr
                            ,&outlen);      /* Format the output */
        }
        else                /* ...else hex representation */
        {
            (void) sprintf (outbuf_ptr, "%c<07>%n"
                            ,cmdchr
                            ,&outlen);      /* Format the output */
        }
    }

    /*
     *    Backspace character Specified ?
     *    -------------------------------
     *
     *    Return a buffer containing the Backspace menmonic "~<BS>" or "^H"
     *    or "~<08> depending on the format requested by the user.
     */
    else if (inpchr == CTRL_BS)
    {
        if (ctlmnc == STR_MNC_ASCII)
        {               /* If ASCII representation */
            (void) sprintf (outbuf_ptr, "%c<BS>%n"
                            ,cmdchr
                            ,&outlen);      /* Format the output */
        }
        else if (ctlmnc == STR_MNC_CONTROL)
        {               /* If CONTROL representation */
            (void) sprintf (outbuf_ptr, "%cH%n"
                            ,ctlchr
                            ,&outlen);      /* Format the output */
        }
        else                /* ...else hex representation */
        {
            (void) sprintf (outbuf_ptr, "%c<08>%n"
                            ,cmdchr
                            ,&outlen);      /* Format the output */
        }
    }

    /*
     *    Horizontal Tab character Specified ?
     *    ------------------------------------
     *
     *    Return a buffer containing the Horizontal tab menmonic "~<HT>" or "^I"
     *    or "~<09> depending on the format requested by the user.
     */
    else if (inpchr == CTRL_HT)
    {
        if (ctlmnc == STR_MNC_ASCII)
        {               /* If ASCII representation */
            (void) sprintf (outbuf_ptr, "%c<HT>%n"
                            ,cmdchr
                            ,&outlen);      /* Format the output */
        }
        else if (ctlmnc == STR_MNC_CONTROL)
        {               /* If CONTROL representation */
            (void) sprintf (outbuf_ptr, "%cI%n"
                            ,ctlchr
                            ,&outlen);      /* Format the output */
        }
        else                /* ...else hex representation */
        {
            (void) sprintf (outbuf_ptr, "%c<09>%n"
                            ,cmdchr
                            ,&outlen);      /* Format the output */
        }
    }

    /*
     *    Vertical Tab character Specified ?
     *    ----------------------------------
     *
     *    Return a buffer containing the vertical tab menmonic "~<VT>" or "^K"
     *    or "~<0B> depending on the format requested by the user.
     */
    else if (inpchr == CTRL_VT)
    {
        if (ctlmnc == STR_MNC_ASCII)
        {               /* If ASCII representation */
            (void) sprintf (outbuf_ptr, "%c<VT>%n"
                            ,cmdchr
                            ,&outlen);      /* Format the output */
        }
        else if (ctlmnc == STR_MNC_CONTROL)
        {               /* If CONTROL representation */
            (void) sprintf (outbuf_ptr, "%cK%n"
                            ,ctlchr
                            ,&outlen);      /* Format the output */
        }
        else                /* ...else hex representation */
        {
            (void) sprintf (outbuf_ptr, "%c<0B>%n"
                            ,cmdchr
                            ,&outlen);      /* Format the output */
        }
    }

    /*
     *    Delete character Specified ?
     *    ----------------------------
     *
     *    Return a buffer containing the delete menmonic "~<DEL>"
     *    or "~<7F> depending on the format requested by the user.
     */
    else if (inpchr == CTRL_RUBOUT)
    {
        if ((ctlmnc == STR_MNC_ASCII) ||
                (ctlmnc == STR_MNC_CONTROL))
        {               /* If ASCII or CONTROL representation */
            (void) sprintf (outbuf_ptr, "%c<DEL>%n"
                            ,cmdchr
                            ,&outlen);      /* Format the output */
        }
        else                /* ...else hex representation */
        {
            (void) sprintf (outbuf_ptr, "%c<7F>%n"
                            ,cmdchr
                            ,&outlen);      /* Format the output */
        }
    }

    /*
     *    Printable ASCII character ?
     *    ---------------------------
     *
     *    If the character is a printable ASCII character, simply return the
     *    character itself.
     */
    else if ((inpchr >= CTRL_LOWPRINT)
             &&       (inpchr <= CTRL_HIGHPRINT))
    {
        *outbuf_ptr = inpchr;       /* Copy character into buffer */
        outlen = 1;         /* Return length of 1 */
    }

    /*
     *    Any other 7-bit control character
     *    ---------------------------------
     *
     *    If the input character is any other 7-bit control character
     *    and ASCII or CONTROL format were selected return a control
     *    character command.
     */
    else if ((inpchr >= 0) &&
             (inpchr <= 0X1F) &&
             ((ctlmnc == STR_MNC_ASCII) ||
              (ctlmnc == STR_MNC_CONTROL)))
    {
        (void) sprintf (outbuf_ptr, "%c%c%n"
                        ,ctlchr
                        ,(inpchr + 0X40)    /* Convert to letter equiv of control */
                        ,&outlen);      /* Format the output */
    }

    /*
     *    Any other value
     *    ---------------
     *
     *    For all other cases return a character command using the characters
     *    hexadecimal ANSI code, i.e. ~<xx> where xx is the character's
     *    hexadecimal ANSI code.
     */
    else
    {
        value = (unsigned char) inpchr;
        /* Copy value to an unsigned char */
        (void) sprintf (        /* Format the output */
            outbuf_ptr
            ,"%c<%02.2X>%n"
            ,cmdchr
            ,value
            ,&outlen);
    }

    *outlen_ptr = outlen;       /* Length of output string */

    /*
     *  Return to caller.
     */
    return;
}                   /* End CNV_CHR_CMD routine */


/************************************************************************/
/*  CNV_CHR_DSP - Convert ASCII Character to Display Format     */
/*                                  */
/*  This routine takes as input an ASCII character and returns a    */
/*  character string. If the input character is a 7-bit printable   */
/*  the character is returned unchanged. However, all control   */
/*  characters and all 8-bit characters will be represented by  */
/*  a string of printable 7-bit ASCII characters as summarised  */
/*  in the following table.                     */
/*                                  */
/*      Input Character     Output Strings          */
/*      ---------------     --------------          */
/*                                  */
/*      Printable ASCII Character   ASCII Character         */
/*      Line Feed           <LF>,  ^J, <0A>         */
/*      Carriage Return     <CR>,  ^M, <0D>         */
/*      Escape          <ESC>, ^[, <1B>         */
/*      Bell            <BEL>, ^G, <07>         */
/*      Backspace           <BS>,  ^H, <08>         */
/*      Horizontal Tab      <HT>,  ^I, <09>         */
/*      Vertical Tab        <VT>,  ^K, <0B>         */
/*      Del             <DEL>,     <7F>         */
/*      Other 7-bit control             ^x,<xx>         */
/*        characters                        */
/*      Other characters               <xx>         */
/*                                  */
/*      where   x  is the control character, e.g. ^Z        */
/*              xx is the character's ANSI code, e.g. <1A>  */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*  CNV_CHR_DSP (ctlchr, ctlmnc, inpchr, outbuf_ptr, outlen_ptr);   */
/*                                  */
/*      Where:                              */
/*                                  */
/*  ctlchr     =  (char) Control character              */
/*  ctlmnc     =  (long) Control character mnemonic         */
/*  inpchr     =  (char) Character to be converted          */
/*  outbuf_ptr =  (char *) Pointer to buffer to hold the returned   */
/*            string                        */
/*  outlen_ptr =  (long *) Pointer to longword holding length of    */
/*            converted string. It will be updated with the */
/*            length of the copy.               */
/*                                  */
/************************************************************************/
static void CNV_CHR_DSP (char ctlchr,
                         long ctlmnc,
                         char inpchr,
                         char *outbuf_ptr,
                         long *outlen_ptr)
{
    /*
     *  Local data.
     */
    unsigned char       value;      /* Work value */
    int        outlen;     /* Length of converted text */

    /*
     *    Line Feed Specified ?
     *    ---------------------
     *
     *    Return a buffer containing the line feed mnemonic "<LF>" or "^J"
     *    or "<0A> depending on the format requested by the user.
     */
    if (inpchr == CTRL_LF)
    {
        if (ctlmnc == STR_MNC_ASCII)    /* If ASCII representation */
        {
            (void) sprintf (        /* Format the output */
                outbuf_ptr,
                "<LF>%n",
                &outlen);
        }
        else if (ctlmnc == STR_MNC_CONTROL)
        {               /* If CONTROL representation */
            (void) sprintf (        /* Format the output */
                outbuf_ptr,
                "%cJ%n"
                ,ctlchr
                ,&outlen);
        }
        else                /* ...else hex representation */
        {
            (void) sprintf (        /* Format the output */
                outbuf_ptr,
                "<0A>%n",
                &outlen);
        }
    }

    /*
     *    Carriage Return Specified ?
     *    ---------------------------
     *
     *    Return a buffer containing the carriage return mnemonic "<CR>" or "^M"
     *    or "<0A> depending on the format requested by the user.
     */
    else if (inpchr == CTRL_CR)
    {
        if (ctlmnc == STR_MNC_ASCII)
        {               /* If ASCII representation */
            (void) sprintf (        /* Format the output */
                outbuf_ptr,
                "<CR>%n"
                ,&outlen);
        }
        else if (ctlmnc == STR_MNC_CONTROL)
        {               /* If CONTROL representation */
            (void) sprintf (        /* Format the output */
                outbuf_ptr,
                "%cM%n",
                ctlchr,
                &outlen);
        }
        else                /* ...else hex representation */
        {
            (void) sprintf (        /* Format the output */
                outbuf_ptr,
                "<0D>%n",
                &outlen);
        }
    }

    /*
     *    Escape Specified ?
     *    ------------------
     *
     *    Return a buffer containing the line feed mnemonic "<ESC>" or "^["
     *    or "<1B> depending on the format requested by the user.
     */
    else if (inpchr == CTRL_ESC)
    {
        if (ctlmnc == STR_MNC_ASCII)
        {               /* If ASCII representation */
            (void) sprintf (        /* Format the output */
                outbuf_ptr,
                "<ESC>%n",
                &outlen);
        }
        else if (ctlmnc == STR_MNC_CONTROL)
        {               /* If CONTROL representation */
            (void) sprintf (        /* Format the output */
                outbuf_ptr,
                "%c[%n",
                ctlchr,
                &outlen);
        }
        else                /* ...else hex representation */
        {
            (void) sprintf (        /* Format the output */
                outbuf_ptr,
                "<1B>%n",
                &outlen);
        }
    }

    /*
     *    Bell character Specified ?
     *    --------------------------
     *
     *    Return a buffer containing the BELL menmonic "<BEL>" or "^G" or "<07>
     *    depending on the format requested by the user.
     */
    else if (inpchr == CTRL_BELL)
    {
        if (ctlmnc == STR_MNC_ASCII)
        {               /* If ASCII representation */
            (void) sprintf (        /* Format the output */
                outbuf_ptr,
                "<BEL>%n",
                &outlen);
        }
        else if (ctlmnc == STR_MNC_CONTROL)
        {               /* If CONTROL representation */
            (void) sprintf (        /* Format the output */
                outbuf_ptr,
                "%cG%n",
                ctlchr,
                &outlen);
        }
        else                /* ...else hex representation */
        {
            (void) sprintf (        /* Format the output */
                outbuf_ptr,
                "<07>%n",
                &outlen);
        }
    }

    /*
     *    Backspace character Specified ?
     *    -------------------------------
     *
     *    Return a buffer containing the Backspace menmonic "<BS>" or "^H"
     *    or "<08> depending on the format requested by the user.
     */
    else if (inpchr == CTRL_BS)
    {
        if (ctlmnc == STR_MNC_ASCII)
        {               /* If ASCII representation */
            (void) sprintf (        /* Format the output */
                outbuf_ptr,
                "<BS>%n",
                &outlen);
        }
        else if (ctlmnc == STR_MNC_CONTROL)
        {               /* If CONTROL representation */
            (void) sprintf (        /* Format the output */
                outbuf_ptr,
                "%cH%n",
                ctlchr,
                &outlen);
        }
        else                /* ...else hex representation */
        {
            (void) sprintf (        /* Format the output */
                outbuf_ptr,
                "<08>%n",
                &outlen);
        }
    }

    /*
     *    Horizontal Tab character Specified ?
     *    ------------------------------------
     *
     *    Return a buffer containing the Horizontal tab menmonic "<HT>" or "^I"
     *    or "<09> depending on the format requested by the user.
     */
    else if (inpchr == CTRL_HT)
    {
        if (ctlmnc == STR_MNC_ASCII)
        {               /* If ASCII representation */
            (void) sprintf (        /* Format the output */
                outbuf_ptr,
                "<HT>%n",
                &outlen);
        }
        else if (ctlmnc == STR_MNC_CONTROL)
        {               /* If CONTROL representation */
            (void) sprintf (        /* Format the output */
                outbuf_ptr,
                "%cI%n",
                ctlchr,
                &outlen);
        }
        else                /* ...else hex representation */
        {
            (void) sprintf (        /* Format the output */
                outbuf_ptr,
                "<09>%n",
                &outlen);
        }
    }

    /*
     *    Vertical Tab character Specified ?
     *    ----------------------------------
     *
     *    Return a buffer containing the vertical tab menmonic "<VT>" or "^K"
     *    or "<0B> depending on the format requested by the user.
     */
    else if (inpchr == CTRL_VT)
    {
        if (ctlmnc == STR_MNC_ASCII)
        {               /* If ASCII representation */
            (void) sprintf (        /* Format the output */
                outbuf_ptr,
                "<VT>%n",
                &outlen);
        }
        else if (ctlmnc == STR_MNC_CONTROL)
        {               /* If CONTROL representation */
            (void) sprintf (        /* Format the output */
                outbuf_ptr,
                "%cK%n" ,
                ctlchr,
                &outlen);
        }
        else                /* ...else hex representation */
        {
            (void) sprintf (        /* Format the output */
                outbuf_ptr,
                "<0B>%n",
                &outlen);
        }
    }

    /*
     *    Delete character Specified ?
     *    ----------------------------
     *
     *    Return a buffer containing the delete menmonic "<DEL>"
     *    or "<7F> depending on the format requested by the user.
     */
    else if (inpchr == CTRL_RUBOUT)
    {
        if ((ctlmnc == STR_MNC_ASCII) ||
                (ctlmnc == STR_MNC_CONTROL))
        {               /* If ASCII or CONTROL representation */
            (void) sprintf (        /* Format the output */
                outbuf_ptr,
                "<DEL>%n",
                &outlen);
        }
        else                /* ...else hex representation */
        {
            (void) sprintf (        /* Format the output */
                outbuf_ptr,
                "<7F>%n",
                &outlen);
        }
    }

    /*
     *    Printable ASCII character ?
     *    ---------------------------
     *
     *    If the character is a printable ASCII character, simply return the
     *    character itself.
     */
    else if ((inpchr >= CTRL_LOWPRINT)
             &&       (inpchr <= CTRL_HIGHPRINT))
    {
        *outbuf_ptr = inpchr;       /* Copy character into buffer */
        outlen = 1;         /* Return length of 1 */
    }

    /*
     *    Any other 7-bit control character
     *    ---------------------------------
     *
     *    If the input character is any other 7-bit control character
     *    and ASCII or CONTROL format were selected return a control
     *    string.
     */
    else if ((inpchr >= 0) &&
             (inpchr <= 0X1F) &&
             ((ctlmnc == STR_MNC_ASCII) ||
              (ctlmnc == STR_MNC_CONTROL)))
    {
        (void) sprintf (
            outbuf_ptr,
            "%c%c%n",
            ctlchr,
            (inpchr + 0X40),    /* convert to letter equiv of control */
            &outlen);       /* format the output */
    }

    /*
     *    Any other value
     *    ---------------
     *
     *    For all other cases return a string using the characters
     *    hexadecimal ANSI code, i.e. <xx> where xx is the character's
     *    hexadecimal ANSI code.
     */
    else
    {
        value = (unsigned char) inpchr;
        /* Copy value to an unsigned char */
        (void) sprintf (        /* Format the output */
            outbuf_ptr,
            "<%02.2X>%n",
            value,
            &outlen);
    }

    *outlen_ptr = outlen;       /* Length of output string */

    /*
     *  Return to caller.
     */
    return;
}                   /* End CNV_CHR_DSP routine */


/************************************************************************/
/*  CNV_FKEY - Convert Function Key Sequence to Mnemonic        */
/*                                  */
/*  If the supplied buffer begins with an escape sequence the   */
/*  function key mnemonic and to add this mnemonic to the buffer    */
/*  the other data as required to fit.              */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*  CNV_FKEY (inp_ptr, inplen, out_ptr, outlen_ptr, esclen_ptr);    */
/*                                  */
/*  Where:                              */
/*                                  */
/*  inp_ptr    = (char *) Pointer to inputs ASCII string        */
/*  inplen     = (long) Input string length             */
/*  out_ptr    = (char *) Pointer to buffer to hold output      */
/*           string                     */
/*  outlen_ptr = (long *) Pointer to variable to hold output    */
/*           string length                  */
/*  esclen_ptr = (long *) Pointer to variable to hold escape    */
/*           sequence length                    */
/*                                  */
/************************************************************************/
static void CNV_FKEY (char *inp_ptr,
                      long inplen,
                      char *out_ptr,
                      long *outlen_ptr,
                      long *esclen_ptr)
{
    /*
     * Local data.
     */
    unsigned long   iret;           /* PSL routine return status */
    long   esclen;         /* Escape sequence length */
    unsigned long   fkylen;         /* Length of function key text */
    char   escbuf [12];        /* Buffer to hold escape sequence */
    char   fkybuf [12];        /* Buffer to hold function key */

    /*
     *  Assume that there is no function key sequence.
     */
    *outlen_ptr = 0;            /* Zero output string length */
    *esclen_ptr = 0;            /* Zero escape sequence length */

    /*
     *  Check the first character of the input string. This must be ESC, CSI or SS3.
     *  We ignore all other characters.
     */
    if ((*inp_ptr == CTRL_ESC) ||   /* If first character escape */
            (*inp_ptr == CTRL_CSI) ||   /* ...or CSI... */
            (*inp_ptr == CTRL_SS3))     /* ...or SS3... */
    {
        for (esclen = 2;
                ((esclen < 6) && (esclen <= inplen));
                esclen++)
        {
            (void) memcpy (escbuf, inp_ptr, esclen);
            /* Copy escape sequence into buffer */
            escbuf [esclen] = '\0'; /* ...and null terminate it */
            /*
             *  See if this is a function key. If so copy the function key mnemonic
             *  into the output buffer enclosed in angle brackets, e.g. "<HELP>".
             */
            iret = PSL_DECODE_FKEY (escbuf, esclen, fkybuf, &fkylen);

            if (psl_test_status (iret)) /* If a function key located */
            {
                *out_ptr = '<';     /* Prefix fkey with "<" */
                (void) memcpy ((out_ptr + 1), fkybuf, fkylen);
                /* Copy function key mnemonic */
                *(out_ptr + fkylen + 1) = '>';
                /* Terminate fkey with ">" */
                *outlen_ptr = fkylen + 2;
                /* Output string length */
                *esclen_ptr = esclen;   /* Escape sequence length */
                goto end;       /* All done */
            }               /* If function key found */
        }               /* End of for loop */
    }                   /* End if 'escape' character */

    /*
     *  END:
     *  ====
     *
     *  Routine end point.
     */
end:
    return;
}                   /* End CNV_FKEY routine */

