/* Copyright (C) 2001 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

// SCLCompiler.cpp: implementation of the CSCLCompiler class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "TestPlugin.h"
#include "SCLCompiler.h"
#include "RegFunctions.h"
#include "OstaDefs.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CSCLCompiler::CSCLCompiler()
{
	m_strOutputFolder = GetCompilerOutputDir();
	CreateDirectory(m_strOutputFolder,NULL);
}

CSCLCompiler::~CSCLCompiler()
{

}

BOOL CSCLCompiler::CompileScript(const CString& strScript,BOOL bCompileDependencies)
{
	m_strErrorMsg.Empty();

	//Get output name
	CString strOutputPath,strDependPath;
	GenerateOutputPath(strScript,strOutputPath,strDependPath);
	
	//Get command line
	CString strCmdLine;
	GenerateCommandLine(strScript,strOutputPath,strCmdLine);

	//GetCurrent directory so we can restore it
	CString strCurDir;
	GetCurrentDirectory(MAX_PATH,strCurDir.GetBuffer(MAX_PATH));
	strCurDir.ReleaseBuffer();

	//Set current directory to src path
	int nSlashPos = strScript.ReverseFind('\\');
	if(nSlashPos!=-1)
		SetCurrentDirectory( strScript.Left(nSlashPos) );

	//Run compiler
    SECURITY_ATTRIBUTES saAttr;
    saAttr.nLength = sizeof(SECURITY_ATTRIBUTES);
    saAttr.bInheritHandle = TRUE;
    saAttr.lpSecurityDescriptor = NULL;

	HANDLE hChildWrite=INVALID_HANDLE_VALUE;
	HANDLE hChildRead=INVALID_HANDLE_VALUE;

    // Create a pipe for the child's STDOUT.
    if(!CreatePipe (&hChildRead, &hChildWrite, &saAttr, 0))
	{
		m_strErrorMsg.Format("Failed to create compiler pipe (Error %d)",GetLastError());
		return false;
	}

    // Now create the child process to execute the command
	PROCESS_INFORMATION ProcInfo;
	STARTUPINFO StartInfo = {sizeof(StartInfo)};
	StartInfo.dwFlags = STARTF_USESTDHANDLES|STARTF_USESHOWWINDOW;
	StartInfo.hStdOutput = hChildWrite;
	StartInfo.hStdError = hChildWrite;
	StartInfo.wShowWindow = SW_HIDE;

	if (!CreateProcess (NULL
		                ,strCmdLine.GetBuffer(0)            // Command Line
                        ,NULL                            // Process Security attributes
                        ,NULL                            // Primary thread security attributes
                        ,TRUE							 // Inherit Handles
                        ,CREATE_NEW_CONSOLE              // Creation flags
                        ,NULL                            // Use Parents environment
                        ,NULL                            // Use Parents current directory
                        ,&StartInfo                      // STARTUPINFO pointer
                        ,&ProcInfo))                     // Receives Process Information
    {
		m_strErrorMsg.Format("Failed to start compiler. (Error %d)",GetLastError());
		return FALSE;
    }
    
    // Close the write end of the pipe before reading
    CloseHandle(hChildWrite);

    //Read from the pipe
	const DWORD dwBufSize = 4096;
    char szBuf[dwBufSize];

    // Read output from child, and write it to the output window edit box.
    DWORD dwRead=0;
	while(1)
    {
        if((!ReadFile(hChildRead, szBuf, dwBufSize-1, &dwRead, NULL)) || (dwRead == 0)) 
			break;
        
        m_strErrorMsg += CString(szBuf,dwRead);
    }

	//Restore directory
	SetCurrentDirectory(strCurDir);

	//Check if compile succeeded 
	return IsUptodate(strScript,strOutputPath);
}

BOOL CSCLCompiler::GenerateCommandLine(const CString& strScript,const CString& strOutput,CString& strCmdLine)
{
	CString strExt;
    int nDotPos = strScript.ReverseFind('.');
	if(nDotPos != -1)
		strExt = strScript.Mid(nDotPos+1);

	CString strType("script");	//default type to script
	if(strExt.CompareNoCase("tss") == 0)
        strType = "test";
    else 
	if(strExt.CompareNoCase("ths") == 0)
        strType = "thread";
    
	CString strIncludeDir( GetRepositoryPath());
	strIncludeDir += strIncludeFolder;

	strCmdLine.Format("%sscl.exe -I \"%s\" -o \"%s\" -t %s \"%s\"",GetCompilerPath(),strIncludeDir,strOutput,strType,strScript);

	return TRUE;
}	

BOOL CSCLCompiler::IsUptodate(const CString& strSrcPath,const CString& strOutputPath)
{
	//Compare date of source and output files
	//If source is newer than output file then 
	//the output is not uptodate
	CFileStatus fsSrc,fsOut;
	if(!CFile::GetStatus(strOutputPath,fsOut))
		return false;

	if(!CFile::GetStatus(strSrcPath,fsSrc))
		return false;

	return (fsSrc.m_mtime <= fsOut.m_mtime);
}

BOOL CSCLCompiler::GenerateOutputPath(const CString& strSrcPath,CString& strOutputPath,CString& strDependPath)
{
	strOutputPath = m_strOutputFolder;

	//Get the name of the src file (ie strip the path off)
	CString strSrcFile(strSrcPath);
	int nSlashPos = strSrcPath.ReverseFind('\\');
	if(nSlashPos != -1)
		strSrcFile.Delete(0,nSlashPos+1);

	//Get the src file extension
	CString strSrcExt;
	int nDotPos = strSrcFile.ReverseFind('.');
	if(nDotPos != -1)
		strSrcExt = strSrcFile.Mid(nDotPos+1);

	//Remove last char from extension
	int nExtLen = strSrcExt.GetLength();
	CString strTrimmedExt(strSrcExt,nExtLen-1);
	
	//Append file title to output
	strOutputPath += strSrcFile.Left(nDotPos+1);
	strDependPath = strOutputPath;
	strDependPath += strTrimmedExt + "d";
    
	//Get output extention from src extension
	if(strSrcExt.CompareNoCase("HTP")==0 ||
		strSrcExt.CompareNoCase("SCL")==0)
		strOutputPath += "tof";
	else
		strOutputPath += strTrimmedExt + "o";

	return false;
}

CString CSCLCompiler::GetCompilerPath()
{
    // Get path to scl.exe
    CString strExePath(getenv("OPENSTA_HOME"));
	strExePath += "Engines\\";
	return strExePath;
}

void CSCLCompiler::SetOutputFolder(const CString& strFolder)
{
	if(CreateDirectory(strFolder,NULL))
		m_strOutputFolder = strFolder;
}
