/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

//========================================================================
//
// HTTPconn.h
// ----------
//=========================================================================

#ifndef HTTPCONN_H
#define HTTPCONN_H

#include "socket.hpp"
#include "ctime.hpp"
#include <cstdlib>

#define modulus SSLmodulus
#define list SSLlist
// SSL includes
#include <openssl/rsa.h> // SSLeay stuff
#include <openssl/pem.h>
#include <openssl/ssl.h>
#include <openssl/err.h>
#include <openssl/rand.h>
#include <openssl/md5.h>
/*#include <openssl/e_os.h>
#include <openssl/buffer.h>
#include <openssl/bio.h>
#include <openssl/pkcs7.h>
#include <openssl/pkcs12.h>
#include <openssl/crypto.h>
#include <openssl/x509.h>
*/

#undef modulus
#undef list


#include <string>
#include <vector>
#include <map>

#include "TConnect.h"
#include "mdbmtx.h"
#include <TContext.h>

#include "phparse.hpp"
#include "Address.hpp"

typedef PhLexer<unsigned char*> Lexer;
typedef PhParser<unsigned char*> Parser;
typedef Address Address;
typedef Parser::AddressValue AddressValue;

typedef map<AddressQuery *, string> constAddressMap ;
typedef map<string, TtChrExpr *> variableAddressMap ;
typedef list<string> stringList ;


#if defined(WIN32)
# if _MSC_VER >= 1100   // MSC version 5 and above
using namespace std;
# endif
#endif
#ifdef SOLARIS
using namespace std;
#endif


typedef map<unsigned int, TConnect*, less<unsigned int> > cons_map_t;


// Define connflags bitmask values
#define CF_DEFAULT          0           // Initial state
#define CF_CONNECTED        1           // Connected
#define CF_SEND             2           // Sending a request
#define CF_RECV             4           // Receiving a request


#define DEFAULT_RECEIVE_SIZE    2048    // Default size for receiving datas
#define DEFAULT_HEADER_SIZE     512     // Reserve this size when CHTMLDocument is created
#define TIMEOUT                 1       // Timeout (in second) for select
#define MAX_STATLIST_SIZE       10      // Maximum number of element in statistic list
#define DEFAULT_REPLY_CODE      999     //
#define REQUEST_TIMEOUT         "REQUEST_TIMEOUT"
#define ON_TIMOUT_MESSAGE       "MESSAGE"
#define STAT_LEVEL              "CYR_HTTP-STAT-LV"

#define TOF_REQOPT_CERTFIL 4


enum enumRequestType
{
    HTTP_REQUEST_GET,
    HTTP_REQUEST_HEAD,
    HTTP_REQUEST_POST,
    HTTP_REQUEST_NONE
};

enum enumTransferType
{
    HTTP_TRANSFER_NONE,					// Transfer mode is unknown (header not already parsed ?)
    HTTP_TRANSFER_CLOSE_SOCKET,			// Response ends when socket is closed
    HTTP_TRANSFER_LENGTH_SPECIFIED,		// Response's length is specified by a header field
    HTTP_TRANSFER_CHUNK					// Response is chunked
};

enum enumSecureType
{
    HTTP_SECURE,
    HTTP_NOT_SECURE
};


// Structures for SSL connections
typedef struct
{
    SSL_CTX     *ctx;
    SSL         *ssl;
    SSL_METHOD  *meth;
}
SSLStructure;

typedef struct
{
    X509		*client_cert;
    EVP_PKEY	*pkey;
	string		certFile;
	string		certPasswd;
}
CERTStructure;



//////////////////////////////////////////////////////////////////////////////
// class CStatistics                                                        //
// (stores statistics for each request to web server)                       //
//////////////////////////////////////////////////////////////////////////////
class CStatistics
{
    time_t          timeStamp;
    string          Url;            // Page name
    CTTime          requestTime;
    CTTime          responseTime;
    short           replyCode;
    unsigned int    replySize;      // response size in bytes

    // Constructors
public:
    CStatistics(string pageName, time_t stamp, const CTTime &rqstTime, const CTTime &respTime = 0, int rplCode = DEFAULT_REPLY_CODE, long rplSize = 0)
:timeStamp(stamp), requestTime(rqstTime), responseTime(respTime), replyCode(rplCode), replySize(rplSize)
    {
        Url = pageName;
    }

    // Readers + Writers
public:
    void            setResponseTime(const CTTime &respTime)
    {
        responseTime = respTime;
    }
    void            setReplyCode(short rplCode)
    {
        replyCode = rplCode;
    }
    void            setReplySize(unsigned int rplSize)
    {
        replySize = rplSize;
    }

    time_t          getTimeStamp(void) const
    {
        return timeStamp;
    }
    string          getUrl(void) const
    {
        return Url;
    }
    const CTTime&   getRequestTime(void) const
    {
        return requestTime;
    }
    const CTTime&   getResponseTime(void) const
    {
        return responseTime;
    }
    short           getReplyCode(void) const
    {
        return replyCode;
    }
    unsigned int    getReplySize(void) const
    {
        return replySize;
    }
};




//////////////////////////////////////////////////////////////////////////////
// class CHTMLDocument                                                      //
// (stores response from web server)                                        //
//////////////////////////////////////////////////////////////////////////////
class CHTMLDocument
{
public:
CHTMLDocument():lastRequestType(HTTP_REQUEST_NONE), transferType(HTTP_TRANSFER_NONE),
    bHeaderLoaded(false), bResponseLoaded(false), bLastChunk(false),
    nbBytesLoaded(0), lTotalLength(-1), sReplyCode(DEFAULT_REPLY_CODE),
    bSaveDocument(true), lastReceiveTime(0)
    {}
    ~CHTMLDocument()
    {}

    void                cleanup(void);
    void                append(const char* cs, int nData = 0);
    void                setLastRequestType(enumRequestType requestType)
    {
        lastRequestType = requestType;
    }
    void                setResponseLoaded(bool b)
    {
        bResponseLoaded = b;
    }
    void                setHeaderLoaded(bool b);
    void                setTransferType(void);
    void                setTotalLength(int length)
    {
        lTotalLength = length;
    }
    void                incTotalLength(int length)
    {
        lTotalLength += length;
    }
    void                setLastChunk(bool b)
    {
        bLastChunk = b;
    }
    int                 removeChunkInformation(void);
    void                remove1XXResponse(void);
    void                setLastReceiveTime(const CTTime &ttime)
    {
        lastReceiveTime = ttime;
    }
    enumTransferType    getTransferType(void) const
    {
        return transferType;
    }
    enumRequestType     getLastRequestType(void) const
    {
        return lastRequestType;
    }
    unsigned int        getNbBytesLoaded(void) const
    {
        return nbBytesLoaded;
    }
    int                 getTotalLength(void) const
    {
        return lTotalLength;
    }
    bool                isResponseLoaded(void) const
    {
        return bResponseLoaded;
    }
    bool                isHeaderLoaded(void) const
    {
        return bHeaderLoaded;
    }
    bool                isLastChunk(void) const
    {
        return bLastChunk;
    }
    int                 find(char *c, int pos = 0) const
    {
        return szData.find(c, pos);
    }

    short               getReplyCode(void) const
    {
        return sReplyCode;
    }
    int                 getChunkSize(void) const;
    string              getHeaderValue(unsigned int size) const;
    string              getHeaderValue(const string& field, unsigned int index = 1) const;
    string              getHeaderValue(const string& field, const string& subfield) const;
    bool				getBodyValue(AddressQueryList & aql, Parser::AddressValueList & avl);
    string              getBodyValue(unsigned int size) const;
    const CTTime&       getLastReceiveTime(void) const
    {
        return lastReceiveTime;
    }

	void				setSaveDocument(bool bSaveDoc) { bSaveDocument = bSaveDoc; }


protected:
    string              szData;             // HTTP response
    enumRequestType     lastRequestType;    // If HEAD, receive only the header
    enumTransferType    transferType;
    unsigned int        nbBytesLoaded;      // Number of bytes already loaded
    int                 lTotalLength;       // Total length of the response (-1 if unknown)
    bool                bHeaderLoaded;      // Indicates header is completely loaded
    bool                bResponseLoaded;    // Indicates response is completely loaded
    bool                bLastChunk;         // Indicates we have receive the last chunk
    short               sReplyCode;         // Last http reply code (100 <= sReplyCode < 600)
    bool                bSaveDocument;      // Indicates if we have to save the document (if its type is HTML)
    // (it is set to true in constructor
    // and will be modify after the http header is received)
    CTTime              lastReceiveTime;    // Time of each receive (cancel the request if it is too long)

};




//////////////////////////////////////////////////////////////////////////////
// class CHTTPconn                                                          //
// (stores informations about connections)                                  //
//////////////////////////////////////////////////////////////////////////////
class CHTTPconn : public TConnect
{
public:
    enum enumReceiveCode
    {
        HTTP_RECEIVE_END,
        HTTP_RECEIVE_EXIT_ON_LOCAL_TIMER,
        HTTP_RECEIVE_EXIT_ON_GLOBAL_TIMER,
        HTTP_RECEIVE_ERROR
    };
    static enumReceiveCode receiveOnAllConnections(TContext &ctxt, unsigned long timeout = 0, CHTTPconn* pConnection = NULL, int location = 1, bool dontSleep = false);
    static void printAllStatistics(TContext &ctxt);
	static void initHTMLParser() { InitTagParsing(); }

public:

    // Constructor/Destructor
    CHTTPconn(unsigned conid, string thread, string statLevel);
    virtual ~CHTTPconn();

    // Public member functions
    int         connect(const string& host, int port=80);
    int         request(int nType, const string &szHeader, const string &szBody, bool bSaveDoc = false);
    int         receive(int size = DEFAULT_RECEIVE_SIZE);   // receive size bytes
    int         disconnect(void);


    // Do any cleanup on this connection. This should be called by
    // the thread that opened the connection, otherwise code may hang.
    void        cleanup(void)
    {
        if (isConnected()) disconnect();
    }

    // Helper function
protected:
    int         normalRequest(int nType, const string &szHeader, const string &szBody);
    int         SSLRequest(int nType, const string &szHeader, const string &szBody);
	int			SSLUseCertificate(SSL *, const char *, const char *);
	int			SSLUseCertificate(SSL_CTX *, const char *);
    int         SSLInit(void);
    int         SSLRand();

    int         receiveHeader(int);
    int         receiveUntilSocketClose(int);
    int         receiveUntilSpecifiedLength(int);
    int         receiveUntilAllChunkLoaded(int);
    int         receiveData(int size = DEFAULT_RECEIVE_SIZE);

    void        setHost(const string &name)
    {
        szHost = name;
    }
    void        setPort(int number)
    {
        nPort = number;
    }
    void        setError(const string &text)
    {
        szErrorText = text;
    }
    void        responseLoaded(void);

    void        addStatistic(string, time_t, const CTTime&);
    void        updateStatistics(void);
    void        printStatistics(void);

    // Readers + Writers
public:
	void		setCertificate(const string &certF, const string &pass= "");
    void        setProtocol(enumSecureType sec)
    {
        secureType = sec;
    }
    void        setProxy(bool b)
    {
        bProxy = b;
    }
    void        setInfo(const string &text)
    {
        szInfo = text;
    }
    void        setTimerName(const string &tName)
    {
        szTimerName = tName;
    }
    void        setStatusName(const string &sName)
    {
        szStatusName = sName;
    }
    void        setCodeName(const string &cName)
    {
        szCodeName = cName;
    }


    unsigned    getID(void) const
    {
        return nID;
    }
    string      getLastError(void) const
    {
        return szErrorText;
    }
    string      getHost(void) const
    {
        return szHost;
    }
    int         getPort(void) const
    {
        return nPort;
    }
    string      getInfo(void) const
    {
        return szInfo;
    }
    bool        isProxy(void) const
    {
        return bProxy;
    }
    bool        isConnected(void) const
    {
        return ((nConnectionFlags & CF_CONNECTED) && (connection != NULL) ) ? true : false;
    }
    bool        isReceiveFinished(void) const
    {
        return ((!(nConnectionFlags & CF_RECV)) ? true : false );
    }
    bool        isStatusCodeLoaded(void) const
    {
        return ((Document.getReplyCode() != 0) ? true : false );
    }
    bool        isHeaderLoaded(void) const
    {
        return Document.isHeaderLoaded();
    }
    CTSocket*   getSocket(void) const
    {
        return connection;
    }
    string      getHTTPReplyText(int nCode) const;
    string      getHeaderResponse(string szIdent) const;
	bool		getBodyResponse(AddressQueryList & aql, Parser::AddressValueList & avl) ;
    string      getHttpResponse(int nType, unsigned int iSize) const ;
    enumSecureType getProtocol(void) const
    {
        return secureType;
    }
    string      getTimerName(void) const
    {
        return szTimerName;
    }
    string      getStatusName(void) const
    {
        return szStatusName;
    }
    string      getCodeName(void) const
    {
        return szCodeName;
    }
    int         getReplyCode(void) const
    {
        return Document.getReplyCode();
    }
    int         getStatus(void) const
    {
        return ( (Document.getReplyCode() < 400) ? 1 : 0 );
    }

    bool        hasTimedOut(const CTTime&); // Cancel the request if time since last receive is too long


protected:
    CTSocket* connection;                   // HTTP Connection

    // Protected generic variables
protected:
    unsigned                nID;
    unsigned                nConnectionFlags;
    string                  szHost;
    string                  szInfo;
    unsigned                nPort;
    string                  szErrorText;

#define HEADER_VAR  0
#define BODY_VAR    1
    CHTMLDocument           Document;   // Entire response from server (Header + [Body])

    SSLStructure            *SSLStruct;
    CERTStructure			*certStruct;

    enumSecureType          secureType;
    bool                    bProxy;

    list<CStatistics>       statisticList;
    string                  statFileID;
    string                  urlFileID;
    struct
    {
        unsigned long ip;
        int pid;
        short user;
    }
    userID;
    bool                    bCollectStat;

    // Variables name for Timer, Status and Code
    string                  szTimerName;
    string                  szStatusName;
    string                  szCodeName;

public:
    static CTMutex*			statmtx; // mutex for access to stat file (one mutex for ten threads)
}
;

int Ascii2Binary(const string &, string &);

unsigned short swaps(unsigned short s);

unsigned long swapl(unsigned long l);

unsigned int swapi(unsigned int i);

unsigned short littles(unsigned short s);

unsigned long littlel(unsigned long l);

unsigned int littlei(unsigned int i);

//extern "C" int SSLUseCertificate(SSL *, const char *, const char *);

#endif  // HTTPCONN_H
