/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

#include "apps.h"
#include <openssl/asn1.h>
#include <openssl/crypto.h>
#include <openssl/err.h>
#include <openssl/pem.h>
#include <openssl/pkcs12.h>
#include <openssl/ssl.h>
#include <openssl/err.h>



int setCertsKeysP12 (SSL* ssl, PKCS12* p12, const char * pass, X509 **, EVP_PKEY **) ;
int setCertsKeysP7 (SSL* ssl, PKCS7* p7, const char * pass) ;
int setCertsPkeysBags (SSL* ssl, STACK *bags, const char *pass, int passlen, X509 **, EVP_PKEY **) ;
int setCertsPkeysBag (SSL* ssl, PKCS12_SAFEBAG *bag, const char *pass, int passlen, X509 **, EVP_PKEY **) ;


int setCertsPEM(SSL_CTX *ctx, const char *cert_file, X509 **client_cert, EVP_PKEY **pkey)
{
	if (	(SSL_CTX_use_certificate_file(ctx, cert_file, SSL_FILETYPE_PEM) > 0)
		&&	(SSL_CTX_use_PrivateKey_file(ctx, cert_file, SSL_FILETYPE_PEM) > 0)
		&&	(SSL_CTX_check_private_key(ctx))
		)
	{
		BIO *cert = BIO_new(BIO_s_file());
		if (BIO_read_filename(cert, cert_file) > 0)
		{
			*pkey = PEM_read_bio_PrivateKey(cert, NULL, NULL, NULL);
			*client_cert = PEM_read_bio_X509(cert, NULL, NULL, NULL);
		}
		BIO_free(cert);
		
		return 0;
	}
	else
		return -1;
}


int setCertsPKCS(SSL* ssl, const char * fileName, const char * pass, X509 **client_cert, EVP_PKEY **pkey)
{
	PKCS12		*p12 ;
	BIO			*in;

	in = BIO_new_file(fileName, "rb");
	if (in == NULL)
		return -2;

	// if a PKCS12 file
	if ( p12 = d2i_PKCS12_bio (in, NULL) ) 
	{
		BIO_free(in);
		return setCertsKeysP12 (ssl, p12, pass, client_cert, pkey);
	}
	
	BIO_free(in);

	return -1;
}


int setCertsKeysP12 (SSL* ssl, PKCS12* p12, const char * pass, X509 **client_cert, EVP_PKEY **pkey)
{
	STACK		*asafes, *bags;
	PKCS7		*p7;
	int			i, bagnid, err;
	
	apps_startup();

	if (!( asafes = M_PKCS12_unpack_authsafes (p12))) 
		return -1;
	
	for (i = 0; i < sk_num (asafes); i++) 
	{
		p7 = (PKCS7 *) sk_value (asafes, i);
		bagnid = OBJ_obj2nid (p7->type);
		if (bagnid == NID_pkcs7_data) 
		{
			bags = M_PKCS12_unpack_p7data (p7);
		} 
		else if (bagnid == NID_pkcs7_encrypted) 
		{
			bags = M_PKCS12_unpack_p7encdata (p7, pass, -1);
		} 
		else 
			continue;

		if (!bags) 
			return -1;
		
		if ( (err = setCertsPkeysBags (ssl, bags, pass, -1 , client_cert, pkey)) != 1) 
		{
			sk_pop_free (bags, PKCS12_SAFEBAG_free);
			return err;
		}
		
		sk_pop_free (bags, PKCS12_SAFEBAG_free);
	}

	sk_pop_free (asafes, PKCS7_free);
	
	return 0;
}


int setCertsKeysP7 (SSL* ssl, PKCS7* p7, const char * pass)
{
	int					i, bagnid;
	STACK_OF(X509)		*certs;
	STACK_OF(X509_CRL)	*crls;
	
	certs = NULL;
	crls = NULL;

	apps_startup();

	bagnid = OBJ_obj2nid(p7->type);

	switch (bagnid)
	{
		case NID_pkcs7_signed:
			certs=p7->d.sign->cert;
			crls=p7->d.sign->crl;
			break;
		case NID_pkcs7_signedAndEnveloped:
			certs=p7->d.signed_and_enveloped->cert;
			crls=p7->d.signed_and_enveloped->crl;
			break;
		default:
			break;
	}

	if (certs != NULL)
	{
		X509 *x;
		int ret;

		for (i=0; i<sk_X509_num(certs); i++)
		{
			x = sk_X509_value(certs,i);

			ret = SSL_use_certificate(ssl, x);
			X509_free(x);
		}
	}

	if (crls != NULL)
	{
		X509_CRL *crl;

		for (i=0; i<sk_X509_CRL_num(crls); i++)
		{
			crl=sk_X509_CRL_value(crls,i);
		}
	}

	PKCS7_content_free(p7);

	return 0;
}




int setCertsPkeysBags (SSL* ssl, STACK *bags, const char *pass, int passlen, X509 **client_cert, EVP_PKEY **pkey)
{
	int i;
	int err;
	for (i = 0; i < sk_num (bags); i++) 
	{
		if ( (err = setCertsPkeysBag (ssl, (PKCS12_SAFEBAG *)sk_value (bags, i), pass, passlen, client_cert, pkey)) != 1 ) 
			return err;
	}

	return 1;
}


int setCertsPkeysBag (SSL* ssl, PKCS12_SAFEBAG *bag, const char *pass, int passlen, X509 **client_cert, EVP_PKEY **pkey)
{
	PKCS8_PRIV_KEY_INFO *p8;
	int err ;
	
	switch (M_PKCS12_bag_type(bag))
	{
		case NID_keyBag:
			p8 = bag->value.keybag;
			if (!(*pkey = EVP_PKCS82PKEY (p8))) 
				return 0;
			err = SSL_use_PrivateKey(ssl, *pkey);
			break;
		
		case NID_pkcs8ShroudedKeyBag:
			if (!(p8 = M_PKCS12_decrypt_skey (bag, pass, passlen)))
				return -3;
			if (!(*pkey = EVP_PKCS82PKEY (p8))) 
				return 0;
			err = SSL_use_PrivateKey(ssl, *pkey);
			PKCS8_PRIV_KEY_INFO_free(p8);
			break;
			
		case NID_certBag:
			PKCS12_get_attr(bag, NID_localKeyID) ;
			if (M_PKCS12_cert_bag_type(bag) != NID_x509Certificate )
				return 0;

			if (!(*client_cert = M_PKCS12_certbag2x509(bag))) 
				return 0;

			err = SSL_use_certificate(ssl, *client_cert);
			break;
			
		case NID_safeContentsBag:
			return setCertsPkeysBags (ssl, bag->value.safes, pass, passlen, client_cert, pkey);
			
		default:
			return -1;
			break;
	}

	return 1;
}
