/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/


#include "stdafx.h"

#include "TEMainStructure.hpp"
#include "TagParsing.hpp"
#include "CyrHelper.inl"
#include "mutex.hpp"
#include "CORBARepository.h"
#include "Common.hpp"           // For BeginThreadEx.

#ifdef USE_PERF_DATA
#   include "PerfCounters.hpp"
    using namespace OpenSTA_Performance;
#   include "PerfCounterMap.hpp"
#endif

#ifndef CYRANO_TP_STANDALONE
#	include <GlobalVariable.h>  // For CyrGlobalVariableException
#endif

#include "TestExecuter_i.h"

#define	ODLFILE "TEHttp.odl"
const string TPTEXT	= "TestExecuter_web";


#ifdef USE_PERF_DATA
OpenSTA_Performance::CPerfSource* g_pPerformance = NULL;
#endif

static TestExecuterImpl *g_pTExecuterImpl = NULL;
#ifndef CYRANO_TP_STANDALONE
    //But just in case we do need to access it elsewhere here is a _ptr ptr
    TestExecuter_ptr	TExecuter_ptr = TestExecuter::_nil();
#endif

extern CTSemaphore blocking_semaphore;

string exedir;      // TODO: change this! For use as an external variable in CORBARepository, CECScriptFile
string filevardir;  // TODO: change this! For use as an external variable in CORBARepository.cpp
CAsyncVirtualUserController*    _threadPoolController=NULL;  // TODO: change this! For use in TContext and TestExecuter_i.

const char g_successMsg[] = "Script Completed Successfully.";
const char g_failureMsg[] = "Script Execution Failed.";

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

CMainExecuterStructure::CMainExecuterStructure()
    : m_pServer(NULL), m_pThreadPoolController(NULL), m_pRule(NULL), m_bInitialised(FALSE), m_pTrace(NULL)
#ifndef __DLLSA__
    , m_cmdLine(0, NULL, "")
#else
    , m_hThreadPoolHandlerThread(NULL)
#endif
{
    m_configFile.SetFile("TestExecuter_web.ini");
}


// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

CMainExecuterStructure::~CMainExecuterStructure()
{
    if(m_pThreadPoolController)
        delete m_pThreadPoolController;
    if(_threadPoolController)
        _threadPoolController = NULL;

    if(m_bInitialised)
        UnInit();

    if(m_pTrace)
        delete m_pTrace;
}


// ----------------------------------------------------------------------------
// Initialising methods. Should be called before each Run.
// ----------------------------------------------------------------------------

#ifdef __DLLSA__

BOOL CMainExecuterStructure::Init(HWND hWnd, char *TofFile, char *OdlDir)
{
    if(m_bInitialised == FALSE)
    {
        if(hWnd == NULL)
        {
            // Detected error while checking args
            if(logr) logr->AddTrace(STARTUP, ERRORLEVEL, "Handle for window is NULL!");
            return FALSE;
        }

        TCHAR szDirAndName [255];
        GetModuleFileName(NULL, szDirAndName, 255);
        exedir = filename_dir(szDirAndName);

        if(CommonInit() == FALSE)
            return FALSE;

        m_hWnd = hWnd;
        gHwnd = m_hWnd; // Init the global HWND pointer (see TPLog.hpp).

        m_zTofFileNameAndPath = TofFile;
        m_zTofFileDir = TofFile;

        // Remove TOF file name
        string::size_type lastSlash = m_zTofFileDir.find_last_of('\\');
        if(lastSlash != string::npos)
            m_zTofFileDir.erase(lastSlash);
        else
            m_zTofFileDir = "";

        filevardir = m_zTofFileDir;

        m_zCurrentDir = OdlDir;

        m_bInitialised = TRUE;
    }

    return TRUE;
}

#else   // __DLLSA__ not defined.

BOOL CMainExecuterStructure::Init(int argc, char* argv[])
{
    if(m_bInitialised == FALSE)
    {
        if(ParseCommandLine(argc, argv) == FALSE)
            return FALSE;

        // Install/uninstall the performance counters. Must be done before most of other classes' InitClass.
        if(m_cmdLine.option_supplied("-u"))
        {
#ifdef USE_PERF_DATA
            return InitPerfCounters(PrfCtrsUninstall);
#else
            TraceMsg("Performance counters not supported.");
            cout << "Performance counters not supported." << endl;
            return FALSE;
#endif  // USE_PERF_DATA
        }
        else
        {
            if(m_cmdLine.option_supplied("-i"))
            {
#ifdef USE_PERF_DATA
                return InitPerfCounters(PrfCtrsInstall);
#else
                TraceMsg("Performance counters not supported.");
                cout << "Performance counters not supported." << endl;
                return FALSE;
#endif  // USE_PERF_DATA
            }
        }

#ifdef USE_PERF_DATA
        if(InitPerfCounters(PrfCtrsUse) == FALSE)
            return FALSE;
#endif

        TCHAR szDirAndName [255];
        GetModuleFileName(NULL, szDirAndName, 255);
        exedir = m_zCurrentDir = filename_dir(szDirAndName);
        m_zExeName = filename_base(szDirAndName);

        if(CommonInit() == FALSE)
            return FALSE;

        if(InitialiseCORBA(argc, argv) == FALSE)
            return FALSE;

        // Build the process name.
        char strProcess[64]; sprintf(strProcess, "%s_%8.8lX", m_zUnderscoredHostIpAddress.c_str(), m_dwProcessId);
        m_zStrProcessName = strProcess;

        m_bInitialised = TRUE;
    }

    return TRUE;
}

#endif  // __DLLSA__

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

BOOL CMainExecuterStructure::CommonInit()
{
    // Get the process ID.
    m_dwProcessId = GetCurrentProcessId();

    CFastLock::Init();

    string zTraceFileName;
    m_configFile.Get("TraceFileName", zTraceFileName, (exedir + '\\' + "Trace.txt").c_str(), "FILES");

#ifdef __DLLSA__
    // Use max trace options.
    int nTraceVal = DETAILLEVEL;
#else
    long nTraceVal;
    m_configFile.Get("TraceLevel", nTraceVal, 0, "FILES");
#endif

    CreateTrace(nTraceVal, zTraceFileName.c_str());

    if(TRecord::InitClass() == FALSE)
    {
        TraceMsg("ERROR initialising the TRecord class.");
        return FALSE;
    }
    if(CORBARepository::InitClass() == FALSE)
    {
        TraceMsg("ERROR initialising the CORBARepository class.");
        return FALSE;
    }
    if(CAsyncVirtualUser::InitClass() == FALSE)
    {
        TraceMsg("ERROR initialising the CAsyncVirtualUser class.");
        return FALSE;
    }

    assert(logr == NULL);
    if(!logr) logr = new TPLog();

#ifndef __DLLSA__
    try
    {
        m_zUnderscoredHostIpAddress = GetHostIPAddressAsUnderscoreString();
    }
    catch(...)
    {
        if(logr) logr->AddTrace(STARTUP, ERRORLEVEL, "Caught Exception on startup. Could not resolve host name." );
        return FALSE;
    }

    m_zStatBaseFileName = strip_extension(CyrGetStatFileName(m_zUnderscoredHostIpAddress.c_str(), TRUE));
#endif

    // Init class internals.
    long lMaxDataBuff, lGrowingBuffCount, lMaxConcurrentSSLReq, lSSLGrowingBuffCount;

    m_configFile.Get("MaxSocketDataBuffersCount",       lMaxDataBuff, 64000,            "SOCKET");
    m_configFile.Get("SocketDataBuffersGrowingCount",   lGrowingBuffCount, 2000,        "SOCKET");
    m_configFile.Get("MaxSSLConcurrentReq",             lMaxConcurrentSSLReq, 2000*4,   "SOCKET");
    m_configFile.Get("SSLGrowingBuffCount",             lSSLGrowingBuffCount, 1000,     "SOCKET");

    if(CAsyncTCPSocket::Init() != 0 
        || CAsyncSecureSocket::Init() != NO_ERROR 
        || CGlobalDataBuffHandler::Init(lMaxDataBuff, lGrowingBuffCount) == FALSE  // Use default values.
#ifndef __DLLSA__
        || CStatistics::Init(m_zStatBaseFileName + ".stat", m_zStatBaseFileName + ".urls") == FALSE
#else
        || CStatistics::Init("", "") == FALSE
#endif
        || CAsyncHTTPRequest::InitClass(lMaxConcurrentSSLReq, lSSLGrowingBuffCount) == FALSE)   // Max number of users times max requests per user.
    {
        TraceMsg("Error initialising structures.");
        return FALSE;
    }

    // Initialize the parser.
    InitTagParsing();

    return TRUE;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

void CMainExecuterStructure::UnInit()
{
#ifndef __DLLSA__
    CStatistics::Release();
#endif
    CAsyncVirtualUser::UninitClass();
    CAsyncHTTPRequest::UninitClass();

    CGlobalDataBuffHandler::UnInit();
    CAsyncSecureSocket::Uninit();
    CAsyncTCPSocket::Uninit();

    if(logr)
    {
        delete logr;
        logr = NULL;
    }

    if(m_pServer)
    {
        delete m_pServer; // End of CORBA communication.
        m_pServer = NULL;
    }

    CORBARepository::UninitClass();
    TRecord::UninitClass();

    m_bInitialised = FALSE;
}


void CMainExecuterStructure::CreateTrace(int nTraceVal, LPCSTR zTraceFileName)
{
    LPCSTR zTraceList   = NULL;
    DWORD dwViewSize    = 10485760U;

    if(nTraceVal < ERRORLEVEL)
        zTraceList = NULL;
    else
    {
        if(nTraceVal < WARNINGLEVEL)
            zTraceList = "errors";
        else
        {
            if(nTraceVal < DEBUGLEVEL)
                zTraceList = "errors connection request sync SecureSocket";
            else
            {
                if(nTraceVal < DETAILLEVEL)
                    zTraceList = "errors connection request sync SecureSocket full";
                else
                    zTraceList = "errors connection request sync SecureSocket full Startup Shutdown ScriptCommand Debug-Generic Note Audit Repository";
            }
        }
    }

    if(zTraceFileName != NULL && zTraceList != NULL)
    {
        string s;
        istringstream is(zTraceList);
        BOOL bContinue = TRUE;

        m_pTrace = new CTrace(zTraceFileName, dwViewSize);
        if(m_pTrace)
        {
            GetTraceStream(m_pTrace) << "Trace started." << "\r\n";
            ReleaseTraceStream();

            while(bContinue)
            {
                is >> s;

                if(is.fail() == FALSE)
                {
                    (*m_pTrace).AddTrace(s.c_str());
                }
                else
                {
                    bContinue = FALSE;
                }
            }
        }
    }
}


// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

BOOL CMainExecuterStructure::Run()
{
    TraceMsg("OpenSTA starting");

    try
    {
        if(RunThreadPoolController() == FALSE)
            return FALSE;
    }
    catch(const string &es)
    {
        // Detected error
        TraceMsg(es); if(logr) logr->AddTrace(STARTUP, ERRORLEVEL, es.c_str());
        return FALSE;
    }
    catch(const mdbIOException& e)
    {
        TraceMsg(e.What() + ": " + e.ioname()); if(logr) logr->AddTrace(STARTUP, ERRORLEVEL, (e.What() + ": " + e.ioname()).c_str());
        return FALSE;
    } 
    catch(const mdbParseException& e) 
    {
        TraceMsg(e.What()); if(logr) logr->AddTrace(STARTUP, ERRORLEVEL, e.What().c_str());
        return FALSE;
    }
    catch(const CyrRepositoryException& e) 
    {
        TraceMsg(e.what()); if(logr) logr->AddTrace(STARTUP, ERRORLEVEL, e.what().c_str());
        return FALSE;
    }
    catch(const exception& e)
    {
        TraceMsg(e.what()); if(logr) logr->AddTrace(STARTUP, ERRORLEVEL, e.what());
        return FALSE;
    }
#ifndef CYRANO_TP_STANDALONE
    catch(const CORBA::Exception& e)
    {
        TraceMsg("Caught CORBA::" + CORBAExString(e)); if(logr) logr->AddTrace(STARTUP, ERRORLEVEL, ("Caught CORBA::" + CORBAExString(e)).c_str());
        return FALSE;
    }
#endif
    catch(...)
    {
        TraceMsg("Caught exception while executing the thread pool."); if(logr) logr->AddTrace(STARTUP, ERRORLEVEL, "Caught exception");
        return FALSE;
    }

    return TRUE;
}

#ifndef __DLLSA__

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

BOOL CMainExecuterStructure::ParseCommandLine(int argc, char* argv[])
{
    // Parse the command line.
    try
    {
        m_cmdLine.init(argc, argv, "TestExecuter: ");

        m_cmdLine.define_option("-u",  CLO_NOVAL);      // uninstall
        m_cmdLine.define_option("-i",  CLO_STRINGVAL);  // install DLL full path
        m_cmdLine.define_option("-v",  CLO_INTVAL);     // trace level
        m_cmdLine.define_option("-d",  CLO_NOVAL);      // debug mode

	    m_cmdLine.validate();
    }
    catch(const string &err)
    {
        TraceMsg("Parameter error: " + err);
        return FALSE;
    }

    return TRUE;
}

#   ifdef USE_PERF_DATA

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

BOOL CMainExecuterStructure::InitPerfCounters(ePrfCtrsAction installType) const
{
    g_pPerformance = new OpenSTA_Performance::CPerfSource(L"OpenSTA_TE", L"OpenSTA_Open", L"OpenSTA_Collect", L"OpenSTA_Close");
    if(g_pPerformance == NULL)
    {
        TraceMsg("ERROR: could not access performance counters.");
        return FALSE;
    }

    if(g_pPerformance->ParseMap(OpenSTAPerf, MAP_SIZE(OpenSTAPerf)) != ERROR_SUCCESS)
    {
        TraceMsg("ERROR: failed to parse performance map.");
        delete g_pPerformance;
        g_pPerformance = NULL;
        return FALSE;
    }

    switch(installType)
    {
    case PrfCtrsUse:
        {
            if(g_pPerformance->AdjustTitleIndexes() != ERROR_SUCCESS)
            {
                TraceMsg("ERROR: failed to adjust performance counter numbers.");
                delete g_pPerformance;
                g_pPerformance = NULL;
                return FALSE;
            }

            DWORD dwRes = g_pPerformance->Initialize();
            if(dwRes != ERROR_SUCCESS && dwRes != ERROR_ALREADY_EXISTS)
            {
                TraceMsg("ERROR: failed to initialise performance source.");
                delete g_pPerformance;
                g_pPerformance = NULL;
                return FALSE;
            }

            TraceMsg("Performance counters successfully initialized.");
        }

        break;

    case PrfCtrsInstall:
        {
            // Install the counters.
            WCHAR lpMultiByteStr [256];
            MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, m_cmdLine.get_value_for("-i").c_str(), -1, lpMultiByteStr, 256);
            if(g_pPerformance->Install(lpMultiByteStr) != ERROR_SUCCESS)
            {
                TraceMsg("Failed installing the performance counters.");
                cout << "Failed installing the performance counters." << endl;
                return FALSE;
            }
            else
            {
                TraceMsg("Performance counters installed.");
                cout << "Performance counters installed." << endl;
                return TRUE;
            }
        }
        break;

    case PrfCtrsUninstall:
        {
            g_pPerformance->Uninstall();
            TraceMsg("Performance counters removed.");
            cout << "Performance counters removed." << endl;
        }
        break;
    };

    return TRUE;
}

void CMainExecuterStructure::ReleasePerfCounters() const
{
    delete g_pPerformance;
    g_pPerformance = NULL;
}

#   endif  // USE_PERF_DATA


// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

BOOL CMainExecuterStructure::InitialiseCORBA(int argc, char* argv[])
{
    // Initialise CORBA for logging purposes
    try
    {
        assert(m_pServer == NULL);
        m_pServer = new CORBAServer(argc, argv);
    }
    catch(const CORBA::Exception &e)
    {
        TraceMsg("Caught CORBA::" + CORBAExString(e));
        return FALSE;
    }
    catch(...)
    {
        // Note TPLog::Setup was not called yet.
        TraceMsg("Error connecting to the CORBA server.");
        return FALSE;
    }

#ifndef CYRANO_TP_STANDALONE
    // Initialise Global Logging Object.
    try
    {
        int nLogLevel = m_cmdLine.get_int_value_for("-v");  // Returns 0 if not set.
        string zRegisteredLogName = string(TPTEXT) + '_' + m_zUnderscoredHostIpAddress + '_' + int_to_string((unsigned)m_dwProcessId);

        if(logr) logr->Setup(m_pServer, zRegisteredLogName.c_str(), nLogLevel);
    }
    catch(const CORBA::Exception &e)
    {
        TraceMsg("Caught CORBA::" + CORBAExString(e));

        if(logr) logr->AddTrace(STARTUP, ERRORLEVEL, ("Caught CORBA::" + CORBAExString(e)).c_str());
        return FALSE;
    }
    catch(...)
    {
        TraceMsg("Caught Exception on startup, is Nameservice Running?");

        if(logr) logr->AddTrace(STARTUP, ERRORLEVEL, "Caught Exception on startup, is Nameservice Running?" );
        return FALSE;
    }
#endif

    return TRUE;
}

BOOL CMainExecuterStructure::ReleaseCORBA()
{
    if(logr) logr->AddTrace(SHUTDOWN, DEBUGLEVEL, "Clean up on Exiting "); 
    
    // Catch any exceptions during cleanup. This could happen if
    // a thread was hard killed, leaving some object in an undefined 
    // state.
    
    try
    {
        if(g_pTExecuterImpl)
        {
            g_pTExecuterImpl->SetToCloseDown();    // Tell monitoring that closedown is imminent and clear up TestProces Object member data.

#ifndef CYRANO_TP_STANDALONE
            CORBA::release(TExecuter_ptr);
            g_pTExecuterImpl->_remove_ref();
            delete g_pTExecuterImpl;
            g_pTExecuterImpl = NULL;
            TExecuter_ptr = NULL;
#endif
        }
    }
#ifndef CYRANO_TP_STANDALONE
    catch(const CyrGlobalVariableException& e) 
    {
        if(logr)logr->AddTrace(SHUTDOWN, ERRORLEVEL, (const char*)e.what);
        return FALSE;
    }
    catch(const CORBA::Exception& e)
    {
        string msg = "Caught CORBA::" + CORBAExString(e) + " in clean-up";
        if(logr)logr->AddTrace(SHUTDOWN,ERRORLEVEL,msg.c_str());
        return FALSE;
    }
#endif
    catch(const CyrRepositoryException& e) 
    {
        string logstr = e.what() + ',' + e.where() + " Status: " + int_to_string(e.status()).c_str();
        if(logr)logr->AddTrace(SHUTDOWN, ERRORLEVEL, logstr.c_str());
        return FALSE;
    }
    catch(const exception &e)
    {
        string msg = string("Caught exception in clean-up :") + string(e.what());
        if(logr)logr->AddTrace(SHUTDOWN, ERRORLEVEL, msg.c_str());
        return FALSE;
    }
    catch(...)
    {
        if(logr)logr->AddTrace(SHUTDOWN, ERRORLEVEL, "Caught exception in clean-up ");
        return FALSE;
    }

    return TRUE;
} // End of ReleaseCORBA.

#endif // ndef __DLLSA__


// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

BOOL CMainExecuterStructure::RunThreadPoolController()
{
    if(logr) logr->AddTrace(STARTUP, TRACELEVEL, "Start Execution");

    // Read in atvf(tof) file to create the Script Rules, and commence execution.
    // Create rule for this apps file type

    string zOdlFile = m_zCurrentDir;
    if(!m_zCurrentDir.empty() && *m_zCurrentDir.rbegin() != '\\')
        zOdlFile += FN_PATHSEP;
    zOdlFile += ODLFILE;

	if(logr) logr->AddTrace(STARTUP, DEBUGLEVEL, ("Using odl file: " + zOdlFile).c_str());

    try
    {
        m_pRule = new TScriptRule(zOdlFile);    // Note: m_pRule is used solely by the VU controller. Could be created there?
    }
    catch(const mdbParseException &e)
    {
        TraceMsg("ERROR while creating the script rule: " + e.What());
        return FALSE;
    }
    catch(const mdbIOException &e)
    {
        TraceMsg(e.What() + ": " + e.ioname());
        return FALSE;
    } 

    if(logr) logr->AddTrace(DBGSTD, DEBUGLEVEL, "TScriptRule created.");

#ifndef CYRANO_TP_STANDALONE
    // Set the name to be registered, For test managers this is<nodename.project.test.testname.pid>

    //	  Create the instance of the test manager CORBA object
    g_pTExecuterImpl = new TestExecuterImpl(m_pServer, m_zStrProcessName, m_zUnderscoredHostIpAddress, m_dwProcessId, m_zExeName);
    if(g_pTExecuterImpl == NULL)
    {
        TraceMsg("ERROR: Could not create the test executer.");
        return FALSE;
    }

    TExecuter_ptr = g_pTExecuterImpl->_this();

    if(m_cmdLine.option_supplied("-d")) // If run in debug mode, register as such in order for the test manager to use the current debugging process instead of starting a new one.
    {
        if(RegisterCurrentDebugExecuter() == FALSE)
            return FALSE;
    }
#endif

    // Start the thread pool controller.

    _threadPoolController = m_pThreadPoolController = new CAsyncVirtualUserController(m_pRule, m_pTrace);

    if(m_pThreadPoolController == NULL)
    {
		if(logr) logr->AddTrace(DBGSTD, TRACELEVEL, "ERROR: could not create the thread pool controller.");
        TraceMsg("ERROR: could not create the thread pool controller.");
        return FALSE;
    }

    DWORD dwThreadPoolSize, dwNbConcurrentThreads, dwInitialVUCount, dwVUCountGrowBy, dwBrowserParallelism;
    EvaluateTPParams(dwThreadPoolSize, dwNbConcurrentThreads, dwInitialVUCount, dwVUCountGrowBy, dwBrowserParallelism);

    if(m_pThreadPoolController->Init(dwThreadPoolSize, dwNbConcurrentThreads, dwInitialVUCount, dwVUCountGrowBy, dwBrowserParallelism) != 0)
    {
		if(logr) logr->AddTrace(DBGSTD, TRACELEVEL, "ERROR: could not initialize the thread pool controller.");
        TraceMsg("ERROR: could not initialize the thread pool controller.");
        delete m_pThreadPoolController; _threadPoolController = m_pThreadPoolController = NULL;
        return FALSE;
    }

#ifndef __DLLSA__
    // Will bock on that semaphore until the test commander unlocks it after Repository (TECORBARepository) is initialised in 
    // TestExecuterImpl::Initialise.
    blocking_semaphore.lock();
#endif

    // Note that in the modeler mode, the VU controller will automatically start a user uppon EVENT_TYPE_THREADPOOL_START receival.
    // The threadpool is started either in current thread (standalone exe mode) or in a new thread (modeler's dll version).
#ifdef __DLLSA__
    if(StartThreadPoolHandlerThread() == FALSE)
    {
		if(logr) logr->AddTrace(DBGSTD, TRACELEVEL, "ERROR: could not start the thread pool thread.");
        TraceMsg("ERROR: could not start the thread pool thread.");
        delete m_pThreadPoolController; _threadPoolController = m_pThreadPoolController = NULL;
        return FALSE;
    }
#else
    ThreadPoolHandler(this);

    // Test commander requested to stop the executer.
    if(m_pThreadPoolController)
        m_pThreadPoolController->Stop();
#endif

    return TRUE;
}

BOOL CMainExecuterStructure::Complete()
{
#ifndef CYRANO_TP_STANDALONE
    // Send the stat files to the test commander.

    CStatistics::Release();

    LocalResultFiles_t localResultFiles;
 
    string FullName = m_zStatBaseFileName + ".stat";
    string Name     = filename_base(m_zStatBaseFileName) + ".stat";

    string FullURLName = m_zStatBaseFileName + ".urls";
    string URLName = filename_base(m_zStatBaseFileName) + ".urls";
 
    localResultFiles[Name] = FullName;
    localResultFiles[URLName] = FullURLName;
 
    Repository->SetLocalResultsFile(localResultFiles, true);
#endif

    // All done - perform any cleanup required

    BOOL bExecutionRes = FALSE;

	if(m_pThreadPoolController)
	{
        bExecutionRes = (_threadPoolController->GetFailedUsersCount() == 0);

	    m_pThreadPoolController->DropUsers();
		delete m_pThreadPoolController;
		_threadPoolController = m_pThreadPoolController = NULL;
	}

	if(m_pRule)
    {
        delete m_pRule;
        m_pRule = NULL;
    }

#ifndef __DLLSA__
    ReleaseCORBA();
#endif

    UnInit();

#ifdef __DLLSA__
    if(m_hWnd != NULL)
    {
        PostMessage(m_hWnd, WM_APP+3,
            (WPARAM)m_hThreadPoolHandlerThread,
            (LPARAM)(new std::string(bExecutionRes ? g_successMsg : g_failureMsg)));
    }
#endif

#ifdef USE_PERF_DATA
    ReleasePerfCounters();
#endif

    return TRUE;
}


// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

// If run in debug mode, register as such in order for the test manager to use the current debugging process instead of starting a new one.
BOOL CMainExecuterStructure::RegisterCurrentDebugExecuter()
{
#ifndef CYRANO_TP_STANDALONE
    // Test Manager registration in order to force our debug version to be used.
    string ObjectClass  = TestExecuter::CLASS_NAME;
    //TODO   string debug_reg    = "DEBUG." + zHostIpAddress + '.' + strip_extension(zExeName);
    string debug_reg    = "DEBUG_" + m_zUnderscoredHostIpAddress + '_' + strip_extension("TExecuter_htp.exe");
    string trcmsg       = "Debug requested, register as " + debug_reg;

    if(logr) logr->AddTrace(STARTUP, DEBUGLEVEL, trcmsg.c_str());

    CORBA::String_var dbg_reg = CORBA::string_dup(debug_reg.c_str());
    CORBA::String_var obj_reg = CORBA::string_dup(ObjectClass.c_str());

    assert(m_pServer);
    m_pServer->NameObject(TExecuter_ptr, obj_reg.inout(), dbg_reg.inout());
#endif

    return TRUE;
}


// ----------------------------------------------------------------------------
// Trace messages to the log if created.
// ----------------------------------------------------------------------------

void CMainExecuterStructure::TraceMsg(const string &zMsg) const
{
    if(m_pTrace)
    {
        GetTraceStream(m_pTrace) << zMsg << "\r\n";
        ReleaseTraceStream();
    }
}

#ifdef __DLLSA__

// ----------------------------------------------------------------------------
// Start the thread pool handler in a thread different from the Modeller's one.
// ----------------------------------------------------------------------------

BOOL CMainExecuterStructure::StartThreadPoolHandlerThread()
{
    m_hThreadPoolHandlerThread = BeginThreadEx(NULL, 0, ThreadPoolHandler, (PVOID)this, 0, NULL);
    return m_hThreadPoolHandlerThread != NULL;
}

#endif  // __DLLSA__

// ----------------------------------------------------------------------------
// Execute the thread pool controller. Will be started in a separate thread 
// when used in the modeller.
// ----------------------------------------------------------------------------

unsigned CMainExecuterStructure::ThreadPoolHandler(void *pVoidParams)
{
    BOOL bErr = FALSE;

    CMainExecuterStructure *pTestExecuter = (CMainExecuterStructure *)pVoidParams;
    CAsyncVirtualUserController *pThreadPoolController = pTestExecuter->GetThreadPoolController();
    try
    {
        if(pThreadPoolController->Run() != 0)
        {
			if(logr) logr->AddTrace(DBGSTD, TRACELEVEL, "ERROR: could not execute the threadpool.");
            bErr = TRUE;
        }
    }
    catch(const mdbRuntimeError &err)
    {
	    if(logr) logr->AddTrace(DBGSTD, TRACELEVEL, ("ERROR in Run: " + err.What()).c_str());
        bErr = TRUE;
    }

    pTestExecuter->Complete();  // If DLL, will post a message to the Modeler.

    if(bErr)
    {
        assert(0);
        // TODO: in case of error, cancel everything!
    }

    return 1;
}

void CMainExecuterStructure::Stop()
{
    if(m_pThreadPoolController)
    {
        // Causes the pThreadPoolController->Run() call in ThreadPoolHandler()
        // to return. The ThreadPoolHandler() function then calls pTestExecuter->Complete(),
        // which posts a message to the Modeler.
        m_pThreadPoolController->Stop();
    }
}

void CMainExecuterStructure::EvaluateTPParams(DWORD &dwThreadPoolSize, DWORD &dwNbConcurrentThreads, DWORD &dwInitialVUCount, DWORD &dwVUCountGrowBy, DWORD &dwBrowserParallelism)
{
    m_configFile.Get("BrowserParallelism",          (long&)dwBrowserParallelism,    4, "TEST");

#ifdef __DLLSA__
    dwThreadPoolSize = dwNbConcurrentThreads = dwInitialVUCount = dwVUCountGrowBy = 1;
#else
    m_configFile.Get("ThreadPoolSize",              (long&)dwThreadPoolSize,        0,      "THREAD POOL");
    m_configFile.Get("ThreadPoolConcurrentThreads", (long&)dwNbConcurrentThreads,   0,      "THREAD POOL");
    m_configFile.Get("InitialVirtualUserCount",     (long&)dwInitialVUCount,        1000,   "TEST");
    m_configFile.Get("VirtualUserGrowBy",           (long&)dwVUCountGrowBy,         20,     "TEST");

    if(dwNbConcurrentThreads == 0)
    {
        // Automatically compute the thread pool parameters.
        SYSTEM_INFO sI;
        GetSystemInfo(&sI);
        dwNbConcurrentThreads = sI.dwNumberOfProcessors;
    }

    if(dwThreadPoolSize  == 0)
    {
        dwThreadPoolSize = dwNbConcurrentThreads * 25;
    }
#endif
}