/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

/*
 * scl_chk_routines.c:
 * ===================
 *
 * This module contains the TPA 'check' action routines used by SCL.
 */

/*
 *  PSL Macros.
 */
#include <psl_macros.h>
#include <vtest_macros.h>

/*
 *  Standard ANSI header files.
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

/*
 *  Function Prototypes.
 */
#include <psl_rtndef.h>         /* PSL Routines         */
#include <trc_rtndef.h>         /* Tracing Routines     */
#include "scl_rtndef.h"         /* SCL Routines         */
#include "scl_def_tbl_rtndef.h"     /* SCL definitions routines */
#include "scl_env_tbl_rtndef.h"     /* SCL environment routines */
#include "scl_cod_tbl_rtndef.h"     /* SCL code  routines       */
#include "scl_sub_tbl_rtndef.h"     /* SCL subexpression routines   */

static void     CHK_SCRIPT_PRM  (struct tpaarg *argblk_ptr,
                                 char *nam_ptr,
                                 unsigned long namlen);

static void     CHK_THR_DEVTYPS (struct tpaarg *argblk_ptr);

static void     RESET_TOKEN (struct tpaarg *argblk_ptr);

/*
 *  Structure definitions.
 */
#include <shrmsg.h>         /* PSL Messages         */
#include <depprmdef.h>          /* Script dependency parameters */
#include <psl_tpaargdef.h>      /* TPARSE Argument Block    */
#include <psl_tpatabdef.h>      /* TPARSE Table         */
#include "scl_limits.h"         /* SCL limits           */
#include "scl_tof.h"            /* SCL TOF data         */
#include <scpdef.h>         /* SCP Header record        */
#include "scl_tlodef.h"         /* Test structures      */
#include "scl_tbldef.h"         /* SCL Table Definitions    */
#include <vtest_prmdef.h>       /* VTEST Parameters     */
#include "scl_cmd.h"            /* Command Line Data Structure  */
#include "scl_comdef.h"         /* Common Data          */
#include "scl_tpacom.h"         /* Tparse common Data       */
#include <psl_trc_prmdef.h>     /* Tracing parameters       */
#include <psl_vm_prmdef.h>      /* Linked list Parameters   */
#include <pslerr_prmdef.h>      /* PSL error routine Parameters */
#include <psldiag_prmdef.h>     /* PSL diagnostics parameters   */
#include <target_rtndef.h>      /* Target definitions       */
#include <tofprmdef.h>          /* TOF parameters       */

/************************************************************************/
/*                                  */
/*    CHECK_END_OF_FILE - Check for End of Source File          */
/*    ==================                        */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*  iret = CHECK_END_OF_FILE (argblk_ptr);              */
/*                                  */
/*    Where:                                */
/*                                  */
/*  iret       = Unsigned Long PSL Routine Return Status        */
/*  argblk_ptr = Pointer to TPARSE Argument Block Structure     */
/*                                  */
/*    This routine checks the End of File flag and returns success  */
/*    if it is set.                         */
/*                                  */
/************************************************************************/

unsigned long   CHECK_END_OF_FILE (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;

    if ((*com_ptr).eof_flagl != 0)
    {
        return PSL_NORMAL;
    }
    else
    {
        return 0;
    }

}

/************************************************************************/

/*    CHK_ARRAY_IDX - Check an array index is complete          */
/*    =============                         */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_ARRAY_IDX (argblk_ptr)               */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always PSL_NORMAL but message reported if error     */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr = Pointer to TPARSE Argument Block Structure */
/*    Outputs:  None                            */
/*                                  */
/*    This routine is called after parsing a ',' which is the array */
/*    index separator. It does the following:               */
/*                                  */
/*    1. Checks that the index expression is complete.          */
/*    2. Checks that  the number of indices in this reference does not  */
/*   exceed the number in the variable definition.          */
/*                                  */
/************************************************************************/
unsigned long CHK_ARRAY_IDX (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct   scl_com  *com_ptr;     /* Common Data          */
    struct   scl_tpa  *tpa_ptr;     /* TPARSE Data          */
    struct   symtbl   *var_ptr;     /* Pointer to symbol table  */
    struct   array_blk *array_blk_ptr;  /* Pointer to current array     */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    array_blk_ptr = (*tpa_ptr).array_blk_ptrps;
    var_ptr = (*array_blk_ptr).var_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking an array index for validity\n");
    }
#endif

    /*
     *    The current nesting level should equal that saved on the LL if the
     *    expression is valid. If not we will report an error.
     */
    if ((*array_blk_ptr).nesting_levell != (*tpa_ptr).expr_levell)
    {
        REPORT_ERROR (argblk_ptr, SCL_MISOPRDEL);
        /* Report missing operator/delimiter */
    }

    else
    {

        /*
         *    We have parsed a ',' character. This terminates previous indices so count
         *    this index by updating the value in the array block.
         */
        (*array_blk_ptr).array_indicesl += 1;

        /*
         *    ...and check to ensure that we do not exceed the number of dimensions
         *    in the array.
         */
        if ((*array_blk_ptr).array_indicesl == (*var_ptr).var_array_dimcntl)
        {
            REPORT_ERROR (argblk_ptr, SCL_INVNUMSUB);
        }               /* Report too many dimensions   */
    }

    /*
     *    ...and return success
     */
    return PSL_NORMAL;
}
/*
 * 
 */

/************************************************************************/

/*    CHK_ARRAY_REF - Check an array reference is complete      */
/*    =============                         */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_ARRAY_REF (argblk_ptr)               */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always PSL_NORMAL but message reported if error     */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr = Pointer to TPARSE Argument Block Structure */
/*    Outputs:  None                            */
/*                                  */
/*    This routine is called after parsing a ']' which is the delimiter */
/*    for an array reference. It does the following:            */
/*                                  */
/*    1. Appends the array end index delimiter to the default tof line  */
/*   buffer.                            */
/*    2. Checks that the number of indices in this reference matches    */
/*   the number in the variable definition.             */
/*    3. Decrements the ARRAY_LEVEL variable which is used to       */
/*   identify PRIMARY and ARRAY operands.               */
/*                                  */
/************************************************************************/

unsigned long   CHK_ARRAY_REF   (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com  *com_ptr;     /* Common Data          */
    struct   scl_tpa  *tpa_ptr;     /* TPARSE Data          */
    struct   symtbl   *var_ptr;     /* Pointer to symbol table  */
    struct   array_blk *array_blk_ptr;  /* Pointer to current array     */
    unsigned long     iret;         /* Status return variable.  */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    array_blk_ptr = (*tpa_ptr).array_blk_ptrps;
    var_ptr = (*array_blk_ptr).var_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking an array reference for validity\n");
    }
#endif

    /*
     *    The current nesting level should equal that saved on the LL if the
     *    expression is valid. If not we will report an error.
     */
    if ((*array_blk_ptr).nesting_levell != (*tpa_ptr).expr_levell)
    {
        REPORT_ERROR (argblk_ptr, SCL_MISOPRDEL);
        /* Report missing operator/delimiter */
    }

    else
    {

        /*
         *    ...and check to ensure that we do not exceed the number of dimensions
         *    in the array.
         */
        (*array_blk_ptr).array_indicesl += 1;
        /* Count the last dimension */
        if ((*array_blk_ptr).array_indicesl != (*var_ptr).var_array_dimcntl)
        {
            REPORT_ERROR (argblk_ptr, SCL_INVNUMSUB);
        }               /* Report too many dimensions   */
    }

    /*
     *    We may have a previous array block on the linked list and if so we
     *    should recover it and use it as the current one.
     */
    iret = PSL_VM_REMOVE ((*tpa_ptr).parsed_array_llps ,(char *) array_blk_ptr
                          ,NULL ,NULL ,PSL_VM_POS_END);
    /* Remove the OPD (at end)  */
    if (iret == PSL_RECNOTFOU)      /* If no more records...    */
    {
        (void) memset (array_blk_ptr ,'\0' ,sizeof (struct array_blk));
        /* Clear the array block.   */
    }

    else if (!(psl_test_status (iret))) /* ...otherwise other error */
    {
        PSL_SIGDIAG ((*tpa_ptr).diag_ptrps, 2, PSL_SYSERVERR
                     ,"PSL_VM_REMOVE (in CHK_ARRAY_REF)", iret);
    }

    /*
     *    Decrement the ARRAY_LEVEL variable to reduce the current array nesting
     *    level. This is incremented by a '[' character and zeroed at the end of a
     *    command. It is used to detect whether we are within an array when we
     *    add operands to the LL of parsed operands.
     */
    (*tpa_ptr).array_levell -= 1;

    /*
     *    ...and return success
     */
    return PSL_NORMAL;
}
/*
 * 
 */

/************************************************************************/

/*    CHK_BREAK  - Check BREAK followed by CASE,DEF_CASE or END SWICTH  */
/*    ==========                                */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_BREAK  (argblk_ptr)                  */
/*                                  */
/*    Type: Action routine                      */
/*    Status:                               */
/*          Returns 0 if no break parsed                */
/*      Returns 1 if break just parsed              */
/*                                  */
/*    Pointers: Unchanged                       */
/*    Inputs:   break_fgl                       */
/*    Outputs:  break_fgl reset to 0                    */
/*                                  */
/*    This routine is part of checking routines for SWITCH command.     */
/*                                          */
/*    When a BREAK is parsed INI_BREAK sets break_fgl   = 1             */
/*    If a break is not parsed this will be zero.                       */
/*                                          */
/*    The routine returns false if break_fgl not set ==> no break       */
/*    command was parsed and tparse continues as usual.                 */
/*                                                                      */
/*    If routine returns true- tparse then checks that a case statement */
/*    is parsed next.                           */
/*                                  */
/************************************************************************/

unsigned long   CHK_BREAK (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */

#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking CASE,ENDSW,DEF processed following BREAK\n");
    }
#endif

    /*
     *  If BREAK_FGL is 0 this means a BREAK statement was not parsed and
     *  so TPARSE can continue as usual.
     *
     *  If BREAK_FGL is 1 this means a BREAK statement was parsed and so
     *  the next statement MUST be a CASE/DEF_CASE or END SWITCH command.
     *  The parsing routine then checks that the next statement is
     *  valid.
     */
    if ((*tpa_ptr).break_fgl == 0)
    {
        return 0;                  /* Return failure               */
    }
    else
    {
        (*tpa_ptr).break_fgl = 0;       /* Reset switch indicator       */

        return PSL_NORMAL;      /* Return Success Status    */
    }
}

/************************************************************************/

/*    CHK_AFTER_TESTC - Check TEST followed by "reachable" command      */
/*    ===============                               */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_AFTER_TESTC  (argblk_ptr)                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:                               */
/*          Returns 0 if test parsed                */
/*      Returns 1 if test just parsed               */
/*                                  */
/*    Pointers: Unchanged                       */
/*    Inputs:   test_fgl                        */
/*    Outputs:  test_fgl reset to 0                 */
/*                                  */
/************************************************************************/
unsigned long   CHK_AFTER_TESTC (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */

#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking Keyword following PASS/FAIL\n");
    }
#endif

    /*
     *  If TEST_FGL is 0 this means a PASS/FAIL test-case statement was
     *  not parsed and so TPARSE can continue as usual.
     *
     *  If TEST_FGL is 1 this means a PASS/FAIL test-case statement was
     *  parsed and so the next statement may or may not be executed depending
     *  on what it is.
     */
    if ((*tpa_ptr).test_fgl == 0)
    {
        return 0;                  /* Return failure               */
    }
    else
    {
        (*tpa_ptr).test_fgl = 0;       /* Reset switch indicator       */

        return PSL_NORMAL;      /* Return Success Status    */
    }
}
/*
 * 
 */
/************************************************************************/

/*    CHK_ENDTEST - Check TEST followed by "reachable" command          */
/*    ===========                               */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_ENDTEST  (argblk_ptr)                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:                               */
/*          Returns 0 if test parsed                */
/*      Returns 1 if test just parsed               */
/*                                  */
/*    Pointers: Unchanged                       */
/*    Inputs:   test_fgl                        */
/*    Outputs:  test_fgl reset to 0                 */
/*                                  */
/************************************************************************/

unsigned long   CHK_ENDTEST (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */

#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking Keyword following PASS/FAIL\n");
    }
#endif
    (*tpa_ptr).test_fgl = 1;    /* Reset the flag           */
    return PSL_NORMAL;      /* Return Success Status    */
}
/*
 * 
 */

/************************************************************************/

/*    CHK_CASE_DEF  - Check that default case hasn't already been parsed*/
/*    ============                              */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_CASE_DEF  (argblk_ptr)               */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Returns FALSE if def_case already parsed        */
/*      Returns TRUE if no def_case parsed          */
/*    Pointers: Unchanged                       */
/*    Inputs:   (*sw_ptr).def_case_fgl                  */
/*    Outputs:  None                            */
/*                                  */
/*    This routine is part of checking routines for SWITCH command.     */
/*                                          */
/*    If there is a DEFAULT_CASE it has to be the last case at the      */
/*    current switch level.                                             */
/*                                  */
/*    i.e. there can be no other cases follwing a default case.     */
/*                                  */
/*    On parsing a default case the def_case_fgl flag is set this flag  */
/*    is checked at every case or default case and if already set there */
/*    is an error- error message from PARSE table.          */
/*                                  */
/************************************************************************/

unsigned long   CHK_CASE_DEF (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */
    struct   swstack *sw_ptr;       /* Switch stack                 */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    sw_ptr  = (*tpa_ptr).swstack_begps + (*tpa_ptr).sw_depthl -1;
    /* Point to current entry      */

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("\nChecking DEFAULT CASE is last case in switch");
        PSL_OUTPUT_TRACE ("\n1.CHK_CASE_DEF *DEF_CASE_FLG = %ld \n"
                          ,(*sw_ptr).def_case_fgl);
    }
#endif

    if ((*sw_ptr).def_case_fgl == 1)
    {
        (*tpa_ptr).sw_error_fgl = 1;    /* Set error flag       */
        return 0;                  /* Return Failure        */
    }
    else
    {
        return PSL_NORMAL;      /* Return Success Status    */
    }
}

/************************************************************************/

/*    CHK_SCRPRMCNT - Check Call Script Parameter Count         */
/*    =============                             */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_SCRPRMCNT  (argblk_ptr)              */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Zero        Script parameter count exceeded     */
/*      PSL_NORMAL  Script parameter count OK           */
/*    Pointers: Unchanged                       */
/*    Inputs:   tpa_ptr->call_prm_cntl - Script parameter count     */
/*    Outputs:  tpa_ptr->call_prm_cntl - Script parameter count     */
/*          incremented by one if parameter count OK.       */
/*                                  */
/*    This routine checks the call script parameter count. If the limit */
/*    on the number of call script parameters is exceeded return an */
/*    error status, otherwise increment the call script parameter   */
/*    count by one and return success.                  */
/*                                  */
/************************************************************************/
unsigned long CHK_SCRPRMCNT (struct tpaarg *argblk_ptr)
{
    /*
     *  Local data.
     */
    struct   scl_com *com_ptr;      /* Common Data */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) argblk_ptr->user_ptrp;
    tpa_ptr = com_ptr->tpa_ptrps;

    /*
     *  Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking CALL SCRIPT parameter count\n");
    }
#endif

    /*
     *  If the maximum call script parameter count is exceeded return an error status,
     *  otherwise increment the parameter count and return success.
     */
    if (tpa_ptr->call_prm_cntl >= MAX_PARAMETERS)
    {
        return 0;                  /* Return failure */
    }
    else
    {
        tpa_ptr->call_prm_cntl += 1;    /* Show 1 more parameter */
        return PSL_NORMAL;      /* Return Success Status */
    }
}


/************************************************************************/

/*    CHK_SCRPRMUNQ - Check Call Script Parameter Unique        */
/*    =============                         */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always PSL_NORMAL.                  */
/*    Pointers: Unchanged.                      */
/*    Inputs:   argblk_ptr = Tparse argument block pointer.     */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine checks whether the script parameter (identified  */
/*    by the last entry in the parsed operand linked list) is       */
/*    ambiguous, and if so reports an error.                */
/*                                  */
/************************************************************************/
unsigned long CHK_SCRPRMUNQ (struct tpaarg *argblk_ptr)
{
    /*
     *  Local data.
     */
    unsigned long   iret;           /* Status return variable */
    struct scl_com  *com_ptr;       /* SCL comman data */
    struct scl_tpa  *tpa_ptr;       /* SCL tparse data */
    struct   oph    *oph_ptr;       /* Operand header pointer */
    struct   opd    *opd_ptr;       /* Operand descriptor pointer */
    struct fattbl   *fat_ptr;       /* Pointer to File Access Table entry */
    struct symtbl   *var_ptr;       /* Pointer to Variables Symbol Table entry */
    char   *llb_ctx_ptr;       /* Linked list context */
    char   opdnam [TOFBUFLEN]; /* Operand name */
    unsigned long   opdlen;         /* Operand name length */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) argblk_ptr->user_ptrp;
    tpa_ptr = com_ptr->tpa_ptrps;

    /*
     *  Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE("Checking CALL SCRIPT parameter unique\n");
    }
#endif

    /*
     *  Get the name of the operand to check and its length - this will be the last 'primary'
     *  operand placed on the parsed operand linkd list. The operand may be a file, pipe or
     *  variable operand and variable operands may require multiple operands to describe
     *  them (e.g. arrays). Therefore we need to ensure that we pick up the 'primary' operand
     *  as this contains the name we need to check. Fortunately, the address of the
     *  primary operand is stored in the Tparse data structure so we can read it directly.
     */
    iret = PSL_VM_COPY ((*tpa_ptr).parsed_operand_llps, /* Get primary operand header */
                        (char *) &oph_ptr,
                        (*tpa_ptr).primary_operandpc,
                        &llb_ctx_ptr ,PSL_VM_MODE_NEXT);
    if  (!(psl_test_status (iret)))     /* ...report any errors */
    {
        PSL_SIGDIAG ((*tpa_ptr).diag_ptrps, 2, PSL_SYSERVERR,
                     "PSL_VM_COPY operand header  (in CHK_SCRPRMUNQ)", iret);
    }

    iret = PSL_VM_COPY ((*tpa_ptr).parsed_operand_llps, /* Get primary operand data */
                        (char *) &opd_ptr,
                        llb_ctx_ptr,
                        NULL ,PSL_VM_MODE_NEXT);
    if  (!(psl_test_status (iret)))     /* ...report any errors */
    {
        PSL_SIGDIAG ((*tpa_ptr).diag_ptrps, 2, PSL_SYSERVERR,
                     "PSL_VM_COPY operand data (in CHK_SCRPRMUNQ)", iret);
    }

    if (((*oph_ptr).addrl == ADDR_LOCAL)
            ||  ((*oph_ptr).addrl == ADDR_SCRIPT)
            ||  ((*oph_ptr).addrl == ADDR_THREAD)
            ||  ((*oph_ptr).addrl == ADDR_GLOBAL)
            ||  ((*oph_ptr).addrl == ADDR_TESTWIDE))
    {
        var_ptr = (*opd_ptr).uopd.var.symtbl_ptrps;
        (void) memcpy (opdnam, (*var_ptr).var_namec16, (*var_ptr).var_nlengthl);
        opdlen = (*var_ptr).var_nlengthl;
    }
    else if (((*oph_ptr).addrl == ADDR_FILE)
             ||   ((*oph_ptr).addrl == ADDR_PIPE))
    {
        fat_ptr = (*opd_ptr).uopd.fat.fattbl_ptrps;
        (void) strcpy (opdnam, (*fat_ptr).fat_namec20);
        opdlen = strlen ((*fat_ptr).fat_namec20);
    }
    else    /* Else not a file, pipe or variable operand */
    {
        return PSL_NORMAL;
    }

    /*
     *  Call a service routine to do the real work.
     */
    CHK_SCRIPT_PRM (argblk_ptr, opdnam, opdlen);

    return PSL_NORMAL;
}

/************************************************************************/

/*    CHK_SWITCH  - Check SWITCH followed by CASE           */
/*    ==========                                */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_SWITCH  (argblk_ptr)                 */
/*                                  */
/*    Type: Action routine                      */
/*    Status:                               */
/*          Returns 0 if no switch parsed               */
/*      Returns 1 if switch just parsed             */
/*                                  */
/*    Pointers: Unchanged                       */
/*    Inputs:   sw_startl                       */
/*    Outputs:  sw_startl reset to 0                    */
/*                                  */
/*    This routine is part of checking routines for SWITCH command.     */
/*                                          */
/*    When a SWITCH is parsed INI_SWITCH sets sw_startl = 1             */
/*    If a switch is not parsed this will be zero.                      */
/*                                          */
/*    The routine returns false if sw_startl not set ==> no switch      */
/*    command was parsed and tparse continues as usual.                 */
/*                                                                      */
/*    If routine returns true- tparse then checks that a case statement */
/*    is parsed next.                           */
/*                                  */
/************************************************************************/

unsigned long   CHK_SWITCH (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */

#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking CASE processed following SWITCH\n");
    }
#endif

    /*
     *  If SW_STARTL is 0 this means a SWITCH statement was not parsed and
     *  so TPARSE can continue as usual.
     *
     *  If SW_STARTL is 1 this means a SWITCH statement was parsed and so
     *  the next statement MUST be a CASE command. The parsing routine
     *  then checks that the next statement is indeed a case statement
     */
    if ((*tpa_ptr).sw_startl == 0)
    {
        return 0;                  /* Return failure               */
    }
    else
    {
        (*tpa_ptr).sw_startl = 0;       /* Reset switch indicator       */

        return PSL_NORMAL;      /* Return Success Status    */
    }
}

/************************************************************************/

/*    CHK_SW_INT  - Check CASE value same  data type as SWITCH variable */
/*    ===========                               */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_SW_DTYP_INT  (argblk_ptr)                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:                               */
/*      Returns 0 if switch variable not an integer     */
/*      Returns 1 if switch variable is integer         */
/*                                      */
/*    Pointers: Unchanged                       */
/*    Inputs:   (*sw_ptr).sw_dtpyl                  */
/*    Outputs:  None                            */
/*                                  */
/*    This routine is part of checking routines for SWITCH command.     */
/*                                          */
/*    It checks if the correct case value is a valid value for datatype */
/*    of switch variable.                       */
/*                                  */
/************************************************************************/

unsigned long   CHK_SW_INT (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */
    struct   swstack *sw_ptr;       /* Switch stack                 */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    sw_ptr  = (*tpa_ptr).swstack_begps + (*tpa_ptr).sw_depthl -1;
    /* Point to current entry      */
    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("\nChecking CASE value valid for dtype of SWITCH var. \n");
        PSL_OUTPUT_TRACE ("1.CHK_SW_INT* INT - check data type  %ld \n"
                          ,(*sw_ptr).sw_dtypl);
    }
#endif

    if ((*sw_ptr).sw_dtypl == DTYP_INTEGER)
    {
        return PSL_NORMAL;                      /* Integer - Return success */
    }
    else
    {
        return 0;               /* Not Integer -failure    */
    }
}

/************************************************************************/

/*    CHK_SW_CHAR -      Check CASE value same  data type as SWITCH     */
/*    ==============     variable                   */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_SW_DTYP_CHAR  (argblk_ptr)               */
/*                                  */
/*    Type: Action routine                      */
/*    Status:                               */
/*      Returns 0 if switch variable not char or int        */
/*      Returns 1 if switch variable is char            */
/*                                  */
/*    Pointers: Unchanged                       */
/*    Inputs:   (*sw_ptr).sw_dtpyl                  */
/*    Outputs:  None                            */
/*                                  */
/*    This routine is part of checking routines for SWITCH command.     */
/*                                          */
/*    It checks if the correct case value is a valid value for datatype */
/*    of switch variable.                       */
/*                                  */
/************************************************************************/

unsigned long   CHK_SW_CHAR (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */
    struct   swstack *sw_ptr;       /* Switch stack                 */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    sw_ptr  = (*tpa_ptr).swstack_begps + (*tpa_ptr).sw_depthl -1;
    /* Point to current entry      */
    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("\nChecking CASE value valid for dtype of SWITCH var. \n");
        PSL_OUTPUT_TRACE ("1.CHK_SW_CHAR* CHAR - check data type  %ld \n"
                          ,(*sw_ptr).sw_dtypl);
    }
#endif
    /*
     *  Return normal if switch data type is a character type.
     *  Else return failure.
     */
    if (((*sw_ptr).sw_dtypl == DTYP_CHARACTER)
            ||  ((*sw_ptr).sw_dtypl == DTYP_VP_SCREEN))
    {
        return PSL_NORMAL;                 /* Character - Return success    */
    }
    else
    {
        return 0;                     /* Not Char - Return failure  */
    }

}

/************************************************************************/

/*    CHK_SW_UNDEF  - Check CASE value same  data type as SWITCH        */
/*    ==============     variable                   */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_SW_UNDEF  (argblk_ptr)               */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Returns 0 if switch variable not undefined      */
/*      Returns 1 if switch variable is undefined       */
/*                                  */
/*    Pointers: Unchanged                       */
/*    Inputs:   (*sw_ptr).sw_dtpyl                  */
/*    Outputs:  None                            */
/*                                  */
/*    This routine is part of checking routines for SWITCH command.     */
/*                                          */
/*    It covers the option of having an undefined variables - which     */
/*    therefore has undefined data type. If this were not taken into    */
/*    errors occur when testing data type of switch varible.        */
/*                                  */
/************************************************************************/

unsigned long   CHK_SW_UNDEF (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */
    struct   swstack *sw_ptr;       /* Switch stack                 */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    sw_ptr  = (*tpa_ptr).swstack_begps + (*tpa_ptr).sw_depthl -1;
    /* Point to current entry      */
    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking CASE value valid for dtype of SWITCH  var. \n");
        PSL_OUTPUT_TRACE ("1.CHK_SW_UNDEF*  - check data type  %ld \n"
                          ,(*sw_ptr).sw_dtypl);
    }
#endif

    if ((*sw_ptr).sw_dtypl == 0)
    {
        return PSL_NORMAL;                      /* Undefined - Return success */
    }
    else
    {
        return 0;               /* Not Undefined -failure    */
    }
}

/************************************************************************/

/*    CHK_END_SW   - Check END SWITCH matches with SWITCH       */
/*    ==========                                */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_END_SW  (argblk_ptr)                 */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   sw_depthl                       */
/*    Outputs:  None                            */
/*                                  */
/*    This routine is called after parsing an END SWITCH statement.     */
/*    sw_depthl keeps a count of switch statements parsed. After an     */
/*    end_switch is parsed then the count is decremented. If the count  */
/*    is less than 0 this means there has been too many end_switches    */
/*                                  */
/*    At the end of a script sw_depthl should be zero and there is check*/
/*    for this in SCL_TPARSE.C / ACT_ROUTINES.c             */
/*                                  */
/************************************************************************/

unsigned long   CHK_END_SW (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking for END SWITCH matching  SWITCH statements\n");
    }
#endif

    /*
     * If the switch depth count is less than 0 this means there are too many
     * END SWITCH statements. Set error flag
     *
     * If switch depth count is greater than zero another END SWITCH can be
     * encoded.
     */
    if ((*tpa_ptr).sw_depthl <= 0)
    {
        REPORT_ERROR (argblk_ptr, SCL_INVENDSWITCH); /* Too many end switches */
        (*tpa_ptr).sw_error_fgl = 1;
    }
    else
    {
        (*tpa_ptr).sw_depthl -= 1;      /* Decrement switch counter     */
        /* Parsed a switch end command  */
    }
    return PSL_NORMAL;      /* Return Success Status */

}

/*
 * 
 */

/************************************************************************/
/*                                  */
/*    CHK_SEED_VALID  - Check to see if the SEED clause is valid.   */
/*    ==============                            */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_SEED_VALID  (argblk_ptr)             */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL if seed clause is valid          */
/*      0 - Otherwise                       */
/*    Pointers: Unchanged                       */
/*    Inputs:   TPARSE argument block                   */
/*    Outputs:  None                            */
/*                                  */
/*    This routine is called after parsing a RESET statement when the   */
/*    SEED clause has been parsed.                  */
/*                                  */
/*    Its purpose is to determine whether or not the SEED clause is     */
/*    valid for this variable. The seed clause is only aplicable for    */
/*    REPEATABLE RANDOM variables.                  */
/*                                  */
/*    Seeing as how the SEED (if it being validated) must be an integer */
/*    literal we can pick up the value directly from the TPA data.  */
/*                                  */
/************************************************************************/

unsigned long   CHK_SEED_VALID (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */
    struct   oph     *oph_ptr;      /* Pointer to OPH       */
    unsigned long    iret;          /* Status return variable   */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking if SEED option valid\n");
    }
#endif

    /*
     *    Pickup the PRIMARY operand parsed from the linked list of parsed operands
     */
    iret = PSL_VM_COPY ((*tpa_ptr).parsed_operand_llps
                        ,(char *) &oph_ptr ,NULL ,NULL ,PSL_VM_MODE_FIRST);
    if  (!(psl_test_status (iret)))     /* ...report any errors     */
    {
        PSL_SIGDIAG ((*tpa_ptr).diag_ptrps, 2, PSL_SYSERVERR
                     ,"PSL_VM_COPY (in CHK_SEED_VALID)", iret);
    }

    /*
     *    If the variable is a random variable then return success
     *    otherwise return failure.
     */
    if (((*oph_ptr).optl & VOPT_RANDOM) != 0)
    {                   /* If RANDOM option set     */
        return PSL_NORMAL;      /* ...return Success Status */
    }

    else                /* ...otherwise...      */
    {
        return 0;           /* ...return failure        */
    }
}

/************************************************************************/
/*                                  */
/*    CHK_CHAR_ATTRIBUTES - Check character attributes          */
/*    ===================                       */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_CHAR_ATTRIBUTES (argblk_ptr)             */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL  If attribute not specified      */
/*      Zero        If specified already            */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr = Pointer to TPARSE Argument Block Structure */
/*      ATTR_OPT_SPECIFIEDL - Options bitmask.          */
/*    Outputs:  ATTR_OPT_SPECIFIEDL updated to show option requested    */
/*                                  */
/*    This routine checks the character attributes to see if the    */
/*    specified attribute is already set and returns failure if it is.  */
/*                                  */
/************************************************************************/

unsigned long   CHK_CHAR_ATTRIBUTES (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking Character Attributes for duplication%c\n");
    }
#endif

    /*
     *    Check to see if the option is already set and return a failure status if
     *    it is.
     */
    if (((*tpa_ptr).attr_opt_specifiedl & (*argblk_ptr).paraml) != 0)
    {                   /* If option already specified  */
        return 0;           /* ...then return failure   */
    }

    else
    {                   /* ...else if not specified */
        (*tpa_ptr).attr_opt_specifiedl |= (*argblk_ptr).paraml;
        /* ...show specified now    */
        return PSL_NORMAL;      /* ...then return success   */
    }
}

/************************************************************************/
/*                                  */
/*    CHK_CHR_CMND_CODE - Check character command code          */
/*    =================                         */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_CHR_CMND_CODE (argblk_ptr)               */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Inputs:   argblk_ptr = Pointer to TPARSE Argument Block Structure */
/*    Pointers: Unchanged                       */
/*    Outputs:  None                            */
/*                                  */
/*    This routine checks that a hexadecimal character code value is    */
/*    within the range zero to FF. It puts the hexadecimal character    */
/*    command (i.e. ~<xx>) in tpa_ptr->string_chrc8 if the code is  */
/*    valid and reports a 'Character Code Out Of Range' error if it is  */
/*    not.                              */
/*                                  */
/************************************************************************/

unsigned long   CHK_CHR_CMND_CODE (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking Character Code %ld\n"
                          ,(*argblk_ptr).numberl);
    }
#endif

    /*
     *  If the number specified is not between 0 and 255 report an error, otherwise
     *  put the hexedecimal character command (format "~<xx>") into the character buffer.
     */
    if (((*argblk_ptr).numberl < 0) || ((*argblk_ptr).numberl > 255))
    {
        REPORT_ERROR (argblk_ptr, PSL_CODOUTRANGE);
    }

    else
    {
        (void) sprintf ((*tpa_ptr).string_chrc8, "%c<%0.2X>",
                        (*com_ptr).cmnd_chr_valuec256 [0], (*argblk_ptr).numberl);
    }
    return PSL_NORMAL;
}

/************************************************************************/
/*                                  */
/*    CHK_CHR_DELIM - Check character delimiter             */
/*    =============                         */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_CHR_DELIM (argblk_ptr)               */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL  If character is the character string    */
/*              delimiter               */
/*      Zero        Otherwise               */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr = Pointer to TPARSE Argument Block Structure */
/*    Outputs:  DELIMITER_CHR = Zero (if delimiter character)       */
/*                                  */
/*    This routine compares the current character with the character    */
/*    string delimiter. If it is the delimiter character the routine    */
/*    sets DELIMITER_CHR to zero and returns success, otherwise it  */
/*    returns failure.                          */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   CHK_CHR_DELIM (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking Character Delimiter %c against %c\n"
                          ,(*argblk_ptr).charb
                          ,(*tpa_ptr).delimiter_chrb);
    }
#endif

    if ((*argblk_ptr).charb == (*tpa_ptr).delimiter_chrb)
        /* If character is delimiter    */
    {
        (*tpa_ptr).delimiter_chrb = 0;  /* clear delimiter char */
        return PSL_NORMAL;      /* ...and return success    */
    }
    else
    {
        return 0;           /* ...else return failure   */
    }
}

/************************************************************************/
/*                                  */
/*    CHK_CMND_CHR - Check for command character            */
/*    =============                         */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_CMND_CHR (argblk_ptr)                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL If character is the defined Command Character*/
/*      Zero Otherwise                      */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr = Pointer to TPARSE Argument Block Structure */
/*    Outputs:  (*tpa_ptr).string_char = command character      */
/*                                  */
/*    This routine checks the current character and returns success if  */
/*    it is the currently defined Command Character and failure     */
/*    otherwise.                            */
/*                                  */
/************************************************************************/

unsigned long   CHK_CMND_CHR (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Check the CHAR in the argument block against the current command
     *    character and return success if they match, failure otherwise
     */
    if ((*argblk_ptr).charb == (*com_ptr).cmnd_chr_valuec256 [0])
    {
        return PSL_NORMAL;
    }

    else
    {
        return 0;
    }
}

/************************************************************************/
/*                                  */
/*    CHK_CMND_MODE - Check For Command Mode Active         */
/*    ===============                           */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_CMND_MODE (argblk_ptr)               */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL if NOT command mode              */
/*      0 - Otherwise                       */
/*    Pointers: Unchanged                       */
/*    Inputs:   SCPH record                     */
/*    Outputs:  None                            */
/*                                  */
/*    This routine checks to see if Command Mode is Active and if it is */
/*    returns an error status.                      */
/*                                  */
/************************************************************************/
unsigned long   CHK_CMND_MODE (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking for COMMAND MODE specified\n");
    }
#endif

    /*
     *    We are in 'command mode' unless another mode has explicitly been
     *    enabled. The only other mode currently is character-cell mode. If
     *    the character-cell mode bit is clear return success, otherwise
     *    return failure.
     */
    if (((*tpa_ptr).mode_flagl & MODE_CHR_CELL) == 0)
        /* If in command mode */
    {
        return 0;           /* ...then return failure */
    }

    else                /* ...else NOT in command mode */
    {
        return PSL_NORMAL;      /* Return success */
    }
}

/************************************************************************/
/*                                  */
/*    CHK_CMND_OPTS - Check the options of a command.           */
/*    =============                         */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_CMND_OPTS (argblk_ptr)               */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL  If valid command option         */
/*      Zero        If not                  */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr = Pointer to TPARSE Argument Block Structure */
/*      OPT_OPT_SPECIFIEDL - Options bitmask.           */
/*    Outputs:  (*tpa_ptr).cmnd_opt_specifiedl - updated to show option */
/*                          requested.  */
/*          (*tpa_ptr).this_cmnd_optl - set to show current option  */
/*                                  */
/*    This routine is called on parsing a command option. It examines   */
/*    the bitmask (*tpa_ptr).cmnd_opt_specifiedl to see if the command  */
/*    option specified by PSLTPA$L_PARAM(AP) has been set already. If   */
/*    'all is well' the routine returns a success status, otherwise an  */
/*    error status is returned.                     */
/*                                  */
/*    The purpose of this routine is to hold a bitmask of the options   */
/*    which have been specified for a command so that we can determine  */
/*    what has been specified when we come to encode the command and    */
/*    also so that we can detect when we have a conflicting/duplicate   */
/*    option. In addition a flag is set to indicate the current option  */
/*    being processed.                          */
/*                                  */
/************************************************************************/
unsigned long CHK_CMND_OPTS (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking COMMAND options bitmask\n");
    }
#endif

    /*
     *    Check to see if the option is already set and return a failure status if
     *    it is.
     */
    if (((*tpa_ptr).cmnd_opt_specifiedl & (*argblk_ptr).paraml) != 0)
    {                   /* If option already specified  */
        return 0;           /* ...then return failure   */
    }

    else
    {                   /* ...else if not specified */
        (*tpa_ptr).cmnd_opt_specifiedl |= (*argblk_ptr).paraml;
        /* ...show specified now */
        (*tpa_ptr).this_cmnd_optl = (*argblk_ptr).paraml;
        /* ...and current option */
        return PSL_NORMAL;      /* ...then return success */
    }
}

/************************************************************************/
/*                                  */
/*    CHK_CNV_OPTS - Check conversion options bitmask.          */
/*    ============                          */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_CNV_OPTS (argblk_ptr)                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL  If valid input command option       */
/*      Zero        If not                  */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr = Pointer to TPARSE Argument Block Structure */
/*      CNV_OPT_SPECIFIEDL - Options bitmask.           */
/*    Outputs:  CNV_OPT_SPECIFIEDL updated to show option requested */
/*                                  */
/*    This routine is called on parsing a convert option. It examines   */
/*    the bitmask CNV_OPT_SPECIFIED to see if the convert option    */
/*    specified by PSLTPA$L_PARAM(AP) has been set already. If 'all is  */
/*    well' the routine returns a success status, otherwise an error    */
/*    status is returned.                       */
/*                                  */
/************************************************************************/

unsigned long   CHK_CNV_OPTS (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking CONVERT options bitmask\n");
    }
#endif

    /*
     *    Check to see if the option is already set and return a failure status if
     *    it is.
     */
    if (((*tpa_ptr).cnv_opt_specifiedl & (*argblk_ptr).paraml) != 0)
    {                   /* If option already specified  */
        return 0;           /* ...then return failure   */
    }

    else
    {                   /* ...else if not specified */
        (*tpa_ptr).cnv_opt_specifiedl |= (*argblk_ptr).paraml;
        /* ...show specified now    */
        return PSL_NORMAL;      /* ...then return success   */
    }
}

/*======================================================================*/
/*                                  */
/*    CHK_CODE_LEN - Check Code Buffer Length               */
/*    =============                         */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_CODE_LEN (argblk_ptr)                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL  If Code Buffer has overflowed       */
/*          Zero Otherwise                      */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr = Pointer to TPARSE argument block.      */
/*    Outputs:  None                            */
/*                                  */
/*    This routine checks whether the object code buffer has        */
/*    overflowed. If it has, it reports an error and returns a      */
/*    success status, otherwise it returns a failure status.        */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   CHK_CODE_LEN (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Check to see if the code pointer points past the end of the code
     *    buffer and report an error if it does.
     */
    if ((*tpa_ptr).code_ptrpc > (*tpa_ptr).code_endpc)
    {
        /* If code overflows buffer */
        REPORT_ERROR (argblk_ptr, SCL_CODEOVERFLOW);
        /* Report code overflow     */
        return PSL_NORMAL;      /* ...and return success    */
    }

    return 0;               /* ...else return failure   */
}

/************************************************************************/

/*    CHK_COMPCOND_END - Check end of a complex conditional expression  */
/*    ================                          */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_COMPCOND_END (argblk_ptr)                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always success                      */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr = Pointer to TPARSE Argument Block Structure */
/*    Outputs:  None                            */
/*                                  */
/*    This routine is called at the end of parsing a complex        */
/*    conditional expression to ensure that we have the correct     */
/*    number of closing parentheses, i.e. we have not left the      */
/*    expression unterminated.                      */
/*                                  */
/************************************************************************/
unsigned long CHK_COMPCOND_END (struct tpaarg *argblk_ptr)
{
    /*
     *  Local data.
     */
    struct scl_com  *com_ptr;       /* Common Data */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) argblk_ptr->user_ptrp;
    tpa_ptr = com_ptr->tpa_ptrps;

    /*
     *  Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking for the end of a Complex Conditional clause\n");
    }
#endif

    /*
     *    If we are not at the top of the stack (level <> 0), then we cannot
     *    validly end the command and report an error.
     */
    if (tpa_ptr->cond_levell != 0)
    {
        REPORT_ERROR (argblk_ptr, SCL_MISOPRDEL);
    }

    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/

/*    CHK_EXPR_END - Check end of an expression.            */
/*    ============                          */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_EXPR_END (argblk_ptr)                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always success                      */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr = Pointer to TPARSE Argument Block Structure */
/*    Outputs:  None                            */
/*                                  */
/*    This routine is called at the end of parsing an expression    */
/*    to ensure that we have the correct number of closing      */
/*    parentheses, i.e. we have not left the expression unterminated.   */
/*                                  */
/************************************************************************/
unsigned long CHK_EXPR_END (struct tpaarg *argblk_ptr)
{
    /*
     *  Local data
     */
    struct   scl_com *com_ptr;      /* SCL Common data */
    struct   scl_tpa *tpa_ptr;      /* SCL Tparse ata */
    struct   array_blk *array_blk_ptr;  /* Pointer to current array */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) argblk_ptr->user_ptrp;
    tpa_ptr = com_ptr->tpa_ptrps;   /* SCL Tparse data */
    array_blk_ptr = (*tpa_ptr).array_blk_ptrps;

    /*
     *  Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Performing end of expression checks\n");
    }
#endif

    /*
     *  If we are not within an array subscript, check that we are back at the
     *  top, otherwise check that we are back at the level we started at.
     */
    if ((*tpa_ptr).array_levell == 0)
    {
        if (tpa_ptr->expr_levell != 0)
        {
            REPORT_ERROR (argblk_ptr, SCL_MISOPRDEL);
        }
    }
    else
    {
        if ((*tpa_ptr).expr_levell != (*array_blk_ptr).nesting_levell)
        {
            REPORT_ERROR (argblk_ptr, SCL_MISOPRDEL);
        }
    }

    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/
/*                                  */
/*    CHK_CONSTANT - Check character delimiter              */
/*    ============                          */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_CONSTANT (argblk_ptr)                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL  If character is the character string    */
/*              delimiter               */
/*      Zero        Otherwise               */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr = Pointer to TPARSE Argument Block Structure */
/*      VARIABLES_TABLE = Variables Symbol Table        */
/*      VAR_CNT = Variable count                */
/*    Outputs:  VAR_PTR = Variable's Symbol Table entry address     */
/*                                  */
/*    This routine checks the current token to see if it is a valid */
/*    constant dataname. If it is the Variables Symbol Table is     */
/*    scanned to see if the CONSTANT has been declared.         */
/*                                  */
/*    If the constant has been declared the address of its entry in */
/*    the Variables Symbol Table is returned in VAR_PTR, otherwise  */
/*    VAR_PTR is set to zero and a failure status is returned.      */
/*                                  */
/*    NOTE: A variable in this instance MUST BE A CONSTANT      */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   CHK_CONSTANT (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */
    struct   symtbl  *symtbl_ptr;       /* Pointer variable symbol table */
    unsigned long    isub;          /* Loop control         */
    long    comp_result;       /* Result of comparison     */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;   /* Common data          */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data          */
    symtbl_ptr = (*tpa_ptr).variable_begps;
    /* Variables symbol table   */

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking for a Defined Constant\n");
    }
#endif

    /*
     *    Scan the variables symbol table until we reach the end or find a match
     *    to this dataname.
     */
    for (isub=0; isub < (*tpa_ptr).var_cntl; isub++, symtbl_ptr++)
    {
        if (((*argblk_ptr).tokencntl == (*symtbl_ptr).var_nlengthl)
                &&  ((*symtbl_ptr).var_addrl == ADDR_CONSTANT))
            /* If the lengths are the same  */
            /* ...and it is a constant  */
        {
            comp_result = memcmp ((*argblk_ptr).token_ptrpc,
                                  (*symtbl_ptr).var_namec16,
                                  (*symtbl_ptr).var_nlengthl);
            /* Compare the 2 variables  */
            if (comp_result == 0) goto found_var;
            /* ...and branch if names match */
        }
    }

    /*
     *    We did not find a match for this CONSTANT. Reset the token and return
     *    an error status
     */
    RESET_TOKEN (argblk_ptr);       /* ...reset the token pointer   */
    return 0;               /* ...and return error status   */

    /*
     *    We found a match for the CONSTANT. Setup the symbol table address in the
     *    TPARSE data and return a success status.
     */
found_var:
    (*tpa_ptr).var_ptrps = symtbl_ptr;  /* Save symbol table entry address */
    return PSL_NORMAL;          /* ...and return success    */
}

/************************************************************************/
/*                                  */
/*    CHK_CTRL_CHR - Check for control character            */
/*    ============                          */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_CTRL_CHR (argblk_ptr)                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL If character is the defined Control Character*/
/*      Zero Otherwise                      */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr = Pointer to TPARSE Argument Block Structure */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine checks the current character and returns success if  */
/*    it is the currently defined Control Character and failure     */
/*    otherwise.                            */
/*                                  */
/************************************************************************/

unsigned long   CHK_CTRL_CHR (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Check the CHAR in the argument block against the current control
     *    character and return success if they match, failure otherwise
     */
    if ((*argblk_ptr).charb == (*com_ptr).ctrl_chr_valuec256 [0])
    {
        return PSL_NORMAL;
    }

    else
    {
        return 0;
    }
}

/************************************************************************/
/*                                  */
/*    CHK_CTRL_SPECIFIER - Check control specifier character        */
/*    ==================                        */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_CTRL_SPECIFIER (argblk_ptr)              */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr = Pointer to TPARSE Argument Block Structure */
/*    Outputs:  None                            */
/*                                  */
/*    This routine checks that the Control Specifier Character is a */
/*    character with an ASCII code in the range 40 to 5F (Hexadecimal). */
/*    If it is not, the character is underlined and an 'Invalid     */
/*    Control Specifier' error reported.                */
/*                                  */
/************************************************************************/

unsigned long   CHK_CTRL_SPECIFIER (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking for a Control Character\n");
    }
#endif

    /*
     *    Check to see if the parsed character is a control character ie between
     *    %X40 and %X5F. If it is not report an error, otherwise set up the
     *    control command (e.g. "^M") in the character string buffer and return
     *    success.
     */
    if (((*argblk_ptr).charb < 64) || ((*argblk_ptr).charb > 95))
        /* If control spec out of range */
    {
        REPORT_ERROR (argblk_ptr, PSL_INVCTRLSPEC);
        /* Pass to underline routine */
        return 0;           /* ...and return an error */
    }

    else
    {
        (*tpa_ptr).string_chrc8[0] = (*com_ptr).ctrl_chr_valuec256 [0];
        (*tpa_ptr).string_chrc8[1] = (*argblk_ptr).charb;
        (*tpa_ptr).string_chrc8[2] = '\0';
        return PSL_NORMAL;      /* ...and return success */
    }
}

/************************************************************************/
/*                                  */
/*    CHK_DEF_DTYP - Check definition data type.                */
/*    =============                             */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_DEF_DTYP (argblk_ptr)                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL  If definition is of character       */
/*              data type               */
/*      Zero        Otherwise               */
/*    Pointers: Unchanged                       */
/*    Inputs:   PSLTPA$L_PARAM(AP) = Data type to check for     */
/*      DEF_DTYPL = Variable definition data type       */
/*    Outputs:  None                            */
/*                                  */
/*    This routine returns success if the current definition is the     */
/*    same type as that supplied.                           */
/*                                  */
/************************************************************************/

unsigned long   CHK_DEF_DTYP (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking the Definition Datatype\n");
    }
#endif

    /*
     *    Check the current data type, returning failure if not the same as
     *    that specified.
     */
    if ((*tpa_ptr).def_dtypl != (*argblk_ptr).paraml)
    {
        return 0;           /* Not specified Return failure */
    }

    else
    {
        return PSL_NORMAL;      /* Specified return success */
    }
}

/************************************************************************/

/*    CHK_DELAY - Check STARTUP DELAY value in range            */
/*    =========                             */
/*                                          */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   scl_tpa.int_valuel = Delay value.           */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine checks whether the current integer value     */
/*    is in the valid range for a STARTUP DELAY clause. If it is not    */
/*    an error is reported.                     */
/*                                  */
/************************************************************************/
unsigned long CHK_DELAY (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Check the value is in range and report an error if it is not.
     */
    if (((*tpa_ptr).int_valuel < MIN_THR_DELAY)
            ||  ((*tpa_ptr).int_valuel > MAX_THR_DELAY))
    {
        REPORT_ERROR (argblk_ptr, SCL_BADSTART);
    }

    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/
/*                                  */
/*    CHK_DO_NEST - Check DO loop nesting level.                */
/*    ============                              */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_DO_NEST (argblk_ptr)             */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Alwaysd PSL_NORMAL                  */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr - Pointer to TPARSE argumennt block      */
/*      DO_CNTL - DO command nesting depth          */
/*    Outputs:  None                            */
/*                                  */
/*    This routine is called after parsing an ENDDO command to ensure   */
/*    that there is a corresponding DO command already parsed.      */
/*                                  */
/************************************************************************/

unsigned long   CHK_DO_NEST (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking DO Nesting level of %ld\n"
                          ,(*tpa_ptr).do_cntl);
    }
#endif

    /*
     *    If the current DO nesting level is 0 or less then there is no DO command
     *    to match up with this ENDDO. We must therefore report an error.
     */
    if ((*tpa_ptr).do_cntl <= 0)    /* If nesting depth <= 0    */
    {
        REPORT_ERROR (argblk_ptr, SCL_INVSTRUCT);
        /* Report Invalid DO/ENDDO structure */
    }
    return PSL_NORMAL;          /* Specified return success */
}

/************************************************************************/
/*                                  */
/*    CHK_END_SUBROUTINE - Ensure END SUBROUTINE valid at this point    */
/*    ==================                        */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_END_SUBROUTINE (argblk_ptr)              */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL if in a subroutine               */
/*      0 - Otherwise                       */
/*    Pointers: Unchanged                       */
/*    Inputs:   None.                           */
/*    Outputs:  (*tpa_ptr).subroutine_flagsl = Set to show NOT in a */
/*                          subroutine. */
/*                                  */
/*    This routine checks the subroutine flags to see whether or not    */
/*    an END SUBROUTINE clause is valid at the current time. This is    */
/*    determined by whether or not we are currently within a subroutine */
/*                                  */
/************************************************************************/

unsigned long   CHK_END_SUBROUTINE (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking for END SUBROUTINE valid\n");
    }
#endif

    /*
     *    Check to see if we are currently processing a subroutine. We return
     *    an error status if we are not but in all cases will clear the flag
     *    as this command will terminate any subroutines.
     */
    if (((*tpa_ptr).subroutine_flagsl & SUB_START) == 0)
    {                   /* Check to see if in subroutine   */
        return 0;           /* ...invalid as not in subroutine */
    }

    else
    {
        (*tpa_ptr).subroutine_flagsl &= (~SUB_START);
        /* ...valid so show not in subrtn now */
        return PSL_NORMAL;      /* ...and return success    */
    }
}

/************************************************************************/

/*    CHK_END_SUBTST - Ensure END SUB-TEST valid at this point      */
/*    ==============                            */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL if in a sub-test             */
/*      0 - Otherwise                       */
/*    Pointers: Unchanged.                      */
/*    Inputs:   None.                           */
/*    Outputs:  (*tpa_ptr).subtst_flagsl = Set to show NOT in a     */
/*                          sub-test.   */
/*                                  */
/*    This routine checks the sub-test flags to see whether or not  */
/*    an END SUB-TEST command is valid at the current point. This is    */
/*    determined by whether or not we are currently within a sub-test.  */
/*                                  */
/************************************************************************/
unsigned long CHK_END_SUBTST (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking for END SUB-TEST valid\n");
    }
#endif

    /*
     *    Check to see if we are currently processing a sub-test.
     */
    if (((*tpa_ptr).subtst_flagsl & SUBTST_START) == 0)
    {                   /* If END SUB-TEST NOT valid */
        return 0;
    }

    else                /* Else END SUB-TEST valid */
    {
        (*tpa_ptr).subtst_flagsl &= (~SUBTST_START);
        /* Flag NOT in a sub-test */
        return PSL_NORMAL;      /* End SUB-TEST is valid */
    }
}


/************************************************************************/

/*    CHK_SCRIPT_PRM - Check Validity of Named Script Parameter     */
/*    ==============                            */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always PSL_NORMAL                   */
/*    Pointers: Unchanged.                      */
/*    Inputs:   com_ptr = Pointer to SCL data               */
/*      nam_ptr = Pointer to Variable, File or Pipe name    */
/*            to check.                 */
/*      namlen = Length of name.                */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine checks whether the script parameter  - identified    */
/*    by the supplied name -  is ambiguous, and if so reports an    */
/*    error. A script parameter is ambiguous if the item (variable, */
/*    file or pipe) name appears in two or more of the Variables    */
/*    Symbol Table, File Access Table and Pipe Access Table. There is   */
/*    ambiguity because we cannot determine from the name which item    */
/*    is intended. In all other places where a file, pipe or variable   */
/*    name may be used the context determines the type of object.   */
/*                                  */
/************************************************************************/
static void CHK_SCRIPT_PRM (struct tpaarg *argblk_ptr,
                            char *nam_ptr,
                            unsigned long namlen)
{
    /*
     *  Local data.
     */
    struct   scl_com *com_ptr;      /* SCL data pointer */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data */
    struct   symtbl  *symtbl_ptr;       /* Pointer variable symbol table */
    struct   fattbl  *fattbl_ptr;       /* Pointer file/pipe symbol table */
    unsigned long    isub;          /* Loop control variable */
    long    comp_result;       /* Result of comparison */
    long    namcnt = 0;        /* Name count */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    symtbl_ptr = (*tpa_ptr).variable_begps;
    /* Variables symbol table */
    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking Script Parameter valid\n");
    }
#endif

    /*
     *  See if the parameter name is a variable name, if so increment name count.
     */
    for (isub=0; isub < (*tpa_ptr).var_cntl; isub++, symtbl_ptr++)
    {
        if (namlen == (*symtbl_ptr).var_nlengthl)
        {               /* If the lengths are the same  */
            comp_result = memcmp (nam_ptr,
                                  (*symtbl_ptr).var_namec16,
                                  namlen);
            /* Compare the 2 variables  */
            if (comp_result == 0)   /* ...if names match        */
            {
                if (((*symtbl_ptr).var_typel != DTYP_CONCHARACTER)
                        &&  ((*symtbl_ptr).var_typel != DTYP_CONINTEGER))
                {
                    namcnt++;
                    goto chk_file;  /* ...and if not a constant then jump */
                }
            }
        }
    }

    /*
     *  See if the parameter name is a file name, if so incrementname count.
     */
chk_file:
    ;
    for (isub=0, fattbl_ptr = (*tpa_ptr).fat_begps;
            isub < (*tpa_ptr).fat_cntl;
            isub++, fattbl_ptr++)
    {
        if (namlen == strlen ((*fattbl_ptr).fat_namec20))
        {               /* If the lengths are the same */
            comp_result = memcmp (nam_ptr,
                                  (*fattbl_ptr).fat_namec20,
                                  namlen);
            /* Compare the 2 file names */
            if (comp_result == 0)
            {
                namcnt++;
                goto chk_pipe;
            }
        }
    }

    /*
     *  If the parameter is NOT a pipe, check that there is no pipe of the same name.
     */
chk_pipe:
    ;
    for (isub=0, fattbl_ptr = (*tpa_ptr).pipe_begps;
            isub < (*tpa_ptr).pipe_cntl;
            isub++, fattbl_ptr++)
    {
        if (namlen == strlen ((*fattbl_ptr).fat_namec20))
        {               /* If the lengths are the same */
            comp_result = memcmp (nam_ptr,
                                  (*fattbl_ptr).fat_namec20,
                                  namlen);
            /* Compare the 2 pipes */
            if (comp_result == 0)
            {
                namcnt++;
                goto chk_namcnt;
            }
        }
    }

    /*
     *  Report an error if the parameter name is ambiguous.
     */
chk_namcnt:
    ;
    if (namcnt != 1)
    {
        REPORT_ERROR (argblk_ptr, SCL_AMBIGUOUSPRM);
    }

    return;
}

/************************************************************************/
/*                                  */
/*    CHK_ENV_CMND - Check for duplicated environment commands      */
/*    ============                          */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_ENV_CMND (argblk_ptr)                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr = Pointer to TPARSE Argument Block Structure */
/*    Outputs:  None                            */
/*                                  */
/*    This routine is called on parsing an environment command. It  */
/*    examines the bitmask ENV_OPT_SPECL to see if environment      */
/*    section command, identified by PSLTPA$L_PARAM(AP) has been    */
/*    specified already. If it has, the current token is underlined */
/*    and a 'Duplicate command' error reported.             */
/*                                  */
/************************************************************************/

unsigned long   CHK_ENV_CMND (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking for duplicate ENVIRONMENT commands\n");
    }
#endif

    /*
     *    Check to see if the bit mask passed as a parameter to this routine is
     *    set in the environment option mask.
     */
    if (((*tpa_ptr).env_opt_specl & (*argblk_ptr).paraml) != 0)
    {
        REPORT_ERROR (argblk_ptr, SCL_DUPCMND);
        /* Pass to underline routine    */
    }

    return PSL_NORMAL;          /* ...always return success */
}

/*======================================================================*/
/*                                  */
/*    CHK_EOF - Check for End-of-File                   */
/*    ========                              */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_EOF (argblk_ptr)                 */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL  If End of File              */
/*          Zero        Otherwise               */
/*    Pointers: Unchanged                       */
/*    Inputs:   TPARSE argument block                   */
/*    Outputs:  None                            */
/*                                  */
/*    This routine checks the End of File flag to see if the        */
/*    End of File has been reached. If the End of File flag is      */
/*    set and we are processing the top level script then this      */
/*    routine returns success. If the End of File flag is set and   */
/*    we are processing an INCLUDED script then we will close that file */
/*    and set the depth to be one less. If the End of File flag is  */
/*    clear then this routine will return failure.          */

/*    File index is decremented at end of include file. This is used    */
/*    for script debugging.                     */
/*----------------------------------------------------------------------*/

unsigned long   CHK_EOF (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    If not end of file then return failure
     */
    if ((*com_ptr).eof_flagl == 0)
    {
        return 0;           /* If not end of file return failure */
    }

    /*
     *    The end of file flag is set. Are we in an INCLUDE script and if we are
     *    close the include files.
     */
    if ((*com_ptr).include_depthl != 0)
    {
        (*com_ptr).include_depthl -= 1; /* Decrement include depth  */
        (*tpa_ptr).src_fil_indxl -= 1;  /* Decrement file index     */


        SCL_CLOSE_SOURCE (      /* Close the include file */
            com_ptr,
            INCLUDE_START_UNIT + (*com_ptr).include_depthl);

        (*com_ptr).eof_flagl = 0;   /* ...and clear EOF flag    */

        /*
         * 
         *
         *    Restore the source file index of the file from which the file which has
         *    just completed was included. We do this so that the index is always
         *    correct for the current script.
         */
        (*tpa_ptr).src_fil_indxl =
            (*com_ptr).src_fil_indxl10 [(*com_ptr).include_depthl];
        return 0;
    }

    return PSL_NORMAL;          /* Return success as EOF    */
}


/************************************************************************/
/*                                  */
/*    CHK_FKEY_TABLE - Scan the fkey table              */
/*    ====================================              */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL  If variable is a function key       */
/*    Zero  Otherwise                       */
/*    Pointers: Moved to the end of the variable            */
/*    Inputs:   Tparse argument block.                  */
/*    Outputs:  CHR_STRING - contains corresponding escape sequence */
/*          CHR_STRING_LENGTH contains the length of the above  */
/*                                      */
/*    This routine checks to see if the current token (parsed by    */
/*    PSLTPA_DATANAME is a valid function key name and returns the  */
/*    corresponding escape sequence if it is.               */
/*                                      */
/************************************************************************/

unsigned long   CHK_FKEY_TABLE (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */
    unsigned long    iret;          /* Status return        */

#ifdef WIN32
    char     emulator_fkeys [] = "VK_ AT_ DG_ HP_ IB_ AS_ IC_ PT_ ST_ TA_ PR_ WY_";
    /* Emulator FKEY introducers    */
    char     fkey[4];           /* First 4 characters of key    */
#endif /* WIN32 */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking for A Function Key Name\n");
    }
#endif

    /*
     * 
     *
     *    WINDOWS ONLY
     *    ============
     *
     *    We may jave a function key name which begins with one of the following:
     *
     *  VK_, AT_, DG_. HP_, IB_, AS_, IC_, PT_, ST_, TA_ ,PR_, WY_
     *
     *    These are all valid function keys which we do NOT process but instead 
     *    just pass through as if they are function keys but perform NO translation
     *    of the text. These FKEY's are used for terminal emulators.
     */
#ifdef WIN32
    (void) memcpy(fkey ,(*argblk_ptr).token_ptrpc ,3);
    /* Copy FKEY introducer     */
    fkey[3] = '\0';         /* ...and null terminate    */
    if ((strstr (emulator_fkeys ,fkey)) != NULL)
    {

        /*
         *    For this type of FKey input we need the enclosing '<'and '>' as they are 
         *    part of the sequence.
         */
        (*tpa_ptr).chr_stringc65536[0] = '<';
        (void) memcpy (&(*tpa_ptr).chr_stringc65536[1]
                       ,(*argblk_ptr).token_ptrpc
                       ,(*argblk_ptr).tokencntl);
        (*tpa_ptr).chr_stringc65536[(*argblk_ptr).tokencntl + 1] = '>';
        (*tpa_ptr).chr_strlenl = (*argblk_ptr).tokencntl + 2;
    }

    else
#endif  /* WIN32 ONLY */

    {

        /*
         *    Call a common routine to attempt to locate the specified
         *    function key within the internal function key table. If the key is found
         *    the the routine will return the corresponding escape sequence.
         */
        iret = PSL_DECODE_FKEY ((*argblk_ptr).token_ptrpc
                                ,(*argblk_ptr).tokencntl
                                ,(*tpa_ptr).chr_stringc65536
                                ,&(*tpa_ptr).chr_strlenl);
        /* Decode the function key  */
        if (!(psl_test_status (iret)))  /* ...if not a function key */
        {
            return 0;           /* ...and return failure */
        }
    }

    /*
     * 
     */
    return PSL_NORMAL;          /* return success as function key */
}

/************************************************************************/
/*                                  */
/*    CHK_FILE - Check if the current token is a FILE           */
/*    =========                             */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_FILE (argblk_ptr)                    */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Moved to end of 'timer name'                */
/*    Inputs:   (*tpa_ptr).fat_cntl - Count of defined files        */
/*    Outputs:  (*tpa_ptr).open_fididxl - FILE ID of located file   */
/*                                  */
/*    This routine is entered from a 'PSLTPA_DATANAME' transition, to   */
/*    determine whether the current token is the name of an External    */
/*    file.                             */
/*                                  */
/*    It scans the File Symbol Table to check if the current token  */
/*    is an existing file name                      */
/*                                  */
/************************************************************************/

unsigned long   CHK_FILE (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */
    struct   fattbl  *fattbl_ptr;       /* Pointer File symbol table    */
    unsigned long    isub;          /* Loop control         */
    unsigned long    namlen;        /* Name length          */
    long    comp_result;       /* Result of comparison     */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking for File Name in Table\n");
    }
#endif

    /*
     *    Scan the files symbol table until we reach the end or find a match
     *    to this dataname.
     */
    namlen = (*argblk_ptr).tokencntl;

    for (isub=0, fattbl_ptr = (*tpa_ptr).fat_begps;
            isub < (*tpa_ptr).fat_cntl;
            isub++, fattbl_ptr++)
    {
        if (namlen == strlen ((*fattbl_ptr).fat_namec20))
        {               /* If the lengths are the same  */
            comp_result = memcmp ((*argblk_ptr).token_ptrpc,
                                  (*fattbl_ptr).fat_namec20,
                                  namlen);
            /* Compare the 2 files      */
            if (comp_result == 0) goto found_file;
            /* ...and branch if names match */
        }
    }

    /*
     *    We have not found an entry for this file - return failure.
     */
    return 0;               /* return failure       */

    /*
     *    We found an entry for this file - return success
     *    We also return the index of the file noting that file id's start at 1 not
     *    at 0 as does the file access table !!!
     */
found_file:
    (*tpa_ptr).open_fididxl = isub + 1; /* Save ID of this file     */
    return PSL_NORMAL;          /* Return success       */
}

/*======================================================================*/
/*                                  */
/*    CHK_FILEID - Check File Identifier                */
/*    ===========                           */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_FILEID (argblk_ptr)                  */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL  If token is a label         */
/*          Zero        Otherwise               */
/*    Pointers: Moved to end of label                   */
/*    Inputs:   FATTBL = File Access Symbol Table           */
/*          FAT_CNT = Label Count                   */
/*    Outputs:  OPEN_FIDIDXL - Set to be current index          */
/*                                  */
/*    This routine checks to see if dataname specified as a FILEID is   */
/*    contained within the File ID table and if not adds record to the  */
/*    end of the table if there is room.                */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   CHK_FILEID (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */
    struct   fattbl  *fattbl_ptr;       /* Pointer File symbol table    */
    unsigned long    isub;          /* Loop control variable    */
    unsigned long    namlen;        /* Name length          */
    long    comp_result;       /* Result of comparison     */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking for a Valid FILE ID\n");
    }
#endif

    /*
     *    Scan the File Access symbol table until we reach the end or find a match
     *    to this dataname.
     *
     *    The index MUST be greater than 0 so we add one to the loop count.
     */
    namlen = (*argblk_ptr).tokencntl;

    for (isub=0, fattbl_ptr = (*tpa_ptr).fat_begps;
            isub < (*tpa_ptr).fat_cntl;
            isub++, fattbl_ptr++)
    {
        if (namlen == strlen ((*fattbl_ptr).fat_namec20))
        {               /* If the lengths are the same  */
            comp_result = memcmp ((*argblk_ptr).token_ptrpc,
                                  (*fattbl_ptr).fat_namec20,
                                  namlen);
            /* Compare the 2 files      */
            if (comp_result == 0)   /* ...if names match        */
            {
                (*tpa_ptr).open_fididxl = isub + 1;
                /* Save index of this file  */
                goto found_file;    /* ...and branch to found file  */
            }
        }
    }

    /*
     *    At this point we have determine that the file has not as yet been defined.
     *    Add an entry to the File Access symbol table to define it.
     */
    if (isub > MAX_FILES)       /* FID Symbol Table full ?  */
    {
        REPORT_ERROR (argblk_ptr, VTEST_MAXFILE);
        /* Report maximum files defined */
    }

    else

        /*
         *    Add a new File entry to the table at the current point
         */
    {
        (void) memcpy ((*fattbl_ptr).fat_namec20
                       ,(*argblk_ptr).token_ptrpc
                       ,(*argblk_ptr).tokencntl);
        /* Copy in file name        */
        (*fattbl_ptr).fat_namec20 [(*argblk_ptr).tokencntl] = '\0';
        /* ...and null-terminate    */
        (*fattbl_ptr).fat_refflgl = ACCESSED_FILE;
        /* No references yet        */
        (*tpa_ptr).open_fididxl = (*tpa_ptr).fat_cntl + 1;
        /* Save index of this file  */
        (*tpa_ptr).fat_cntl += 1;   /* ...and increment File count  */


        /*
         *    If required add it also to the EXTERNAL variables symbol table.
         */
        ADD_FILE_SYMBOL (com_ptr ,fattbl_ptr);
    }

    /*
     *    ...and return
     */
found_file:
    return PSL_NORMAL;          /* Return Success       */
}

/************************************************************************/

/*    CHK_HTTPREQ_END - Check end of an HTTP Request            */
/*    ===============                           */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always success                      */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr = Pointer to TPARSE Argument Block Structure */
/*    Outputs:  None                            */
/*                                  */
/*    This routine is called at the end of parsing an HTTP request  */
/*    to check for conflicting options that can only be checked once    */
/*    command parsing is complete. The only check currently made is */
/*    that if the "ON TIMEOUT GOTO tmolab" option has been specified,   */
/*    the "WITH TIMEOUT period" option has also been specified. If  */
/*    it hs not an error is reported.                   */
/*                                  */
/************************************************************************/
unsigned long CHK_HTTPREQ_END (struct tpaarg *argblk_ptr)
{
    /*
     *  Local data
     */
    struct   scl_com *com_ptr;      /* SCL Common data */
    struct   scl_tpa *tpa_ptr;      /* SCL Tparse ata */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) argblk_ptr->user_ptrp;
    tpa_ptr = com_ptr->tpa_ptrps;   /* SCL Tparse data */

    /*
     *  Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Performing end of HTTP Request checks\n");
    }
#endif

    /*
     *  If the "ON TIMEOUT GOTO tmolab" option has been specified, but
     *  the "WITH TIMEOUT period" option has NOT, report an error.
     */
    if (((tpa_ptr->cmnd_opt_specifiedl & CMND_OPT_TIMEOUT) != 0)
            &&  ((tpa_ptr->cmnd_opt_specifiedl & CMND_OPT_WAIT_PRD) == 0))
    {
        REPORT_ERROR (argblk_ptr, PSL_CNFDUPOPT);
    }

    return PSL_NORMAL;          /* Return success */
}


/*
 * 
 */
/************************************************************************/
/*                                      */
/*    CHK_OPEN_OPTS -Check compatible options               */
/*    ============                              */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs: (*tpa_ptr).open_fididxl - file index for access       */
/*        cmnd_optl - is binary flag set.               */
/*                                      */
/*    This routine is called on parsing COMMA in open data file command */
/*    or having parsed a binary option. The two are not compatible      */
/*                                  */
/*======================================================================*/

unsigned long   CHK_OPEN_OPTS (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */
    long    *file_ptr;     /* Pointer to file data table   */
    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    file_ptr = (*tpa_ptr).file_dtyppl;

    file_ptr +=  (*tpa_ptr).open_fididxl - 1;  /* Point to data type array */

    /*
     *    Check to see if the option is already set and return a failure status if
     *    it is.
     */
    if (((*tpa_ptr).cmnd_opt_specifiedl & OPEN_BINFILE)
            &&  ((*tpa_ptr).cmnd_opt_specifiedl & OPEN_COMMASEP))
    {                   /* If option already specified  */
        return 0;           /* ...then return failure   */
    }

    else
    {
        return PSL_NORMAL;      /* ...then return success   */
    }

}
/*
 * 
 */

/************************************************************************/
/*                                  */
/*    CHK_FORMAT_CHR - Check for Formatting Character           */
/*    ===============                           */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_FORMAT_CHR (argblk_ptr)              */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL  If character is a control character */
/*          Zero        Otherwise               */
/*    Pointers: Unchanged                       */
/*    Inputs:   PSLTPA$B_CHAR(AP) = Character to be checked     */
/*    Outputs:  None                            */
/*                                      */
/*    This routine checks whether the current character is a control    */
/*    character or a space. If it is it returns success, otherwise it   */
/*    returns failure.                          */
/*                                  */
/************************************************************************/

unsigned long   CHK_FORMAT_CHR (struct tpaarg *argblk_ptr)

{
    /*
     *    Check for a format character (space, FF, NL, CR, HT, VT)
     */
    if (isspace((*argblk_ptr).charb))   /* If a format character    */
    {
        return PSL_NORMAL;      /* ...then return success   */
    }
    else
    {
        return 0;           /* ...return failure as not format */
    }
}

/************************************************************************/
/*                                  */
/*    CHK_GRAPHIC_CHR - Check for Graphic Character         */
/*    ===============                           */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL  If graphic character            */
/*          Zero    Otherwise                   */
/*    Pointers: Unchanged                       */
/*    Inputs:   PSLTPA$B_CHAR(AP) = Character to be checked     */
/*    Outputs:  STRING_CHR = Current character (if graphic character)   */
/*                                  */
/*    This routine checks the current character, returning success and  */
/*    placing the current character in STRING_CHR if it is a graphic    */
/*    (i.e. printing) character and returning failure otherwise.    */
/*                                  */
/************************************************************************/

unsigned long   CHK_GRAPHIC_CHR (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    unsigned long offset;           /* Offset in line.      */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */
    int  iret;         /* Status return        */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Check to see if the current character is a printing character and return
     *    success if it is.
     */
    iret = isprint ((*argblk_ptr).charb);
    /* is character a printing char */
    if (iret)
    {

        /*
         *    The character is a printing character. Pick up the specified character
         *    from the source line as input and save it.
         */
        offset = (long) ((*argblk_ptr).token_ptrpc -
                         (*tpa_ptr).parsable_linec136);
        /* Current offset in line   */
        (*tpa_ptr).string_chrc8[0] = (*tpa_ptr).source_linec136 [offset];
        /* Save character as input  */
        (*tpa_ptr).string_chrc8[1] = '\0';
        return PSL_NORMAL;      /* ...and return success.   */
    }

    else
    {
        return 0;
    }
}

/************************************************************************/

/*    CHK_HTTP_MODE - Check For HTTP Mode Active            */
/*    =============                         */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   None.                           */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine checks to see if HTTP Mode is Active and if not  */
/*    reports an error that HTTP mode commands are not valid.       */
/*                                  */
/************************************************************************/
unsigned long CHK_HTTP_MODE (struct tpaarg *argblk_ptr)
{
    /*
     *  Local data.
     */
    struct scl_com  *com_ptr;       /* Common Data */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *  Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking for HTTP MODE specified\n");
    }
#endif

    /*
     *  Check to see if the HTTP mode bit is set in the mode flag and if it
     *  is not report an invalid mode.
     */
    if (((*tpa_ptr).mode_flagl & MODE_HTTP) == 0)
        /* If not mode HTTP */
    {
        REPORT_ERROR (argblk_ptr, SCL_HTTPMODE);
    }

    return PSL_NORMAL;          /* Always return success */
}


/************************************************************************/

/*    CHK_IFDEF_LEVEL - Validate the IFDEF nesting level        */
/*    ==============                                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always PSL_NORMAL                   */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr - TPARSE argument block pointer.     */
/*      IFDEF_LEVELL - Current IFDEF nesting level indicator    */
/*    Outputs:  None                            */
/*                                  */
/*    This routine is called when we encounter a #ELSE ,#ENDIF or   */
/*    #ELSEIF command to ensure that we have a valid #IFDEF structure   */
/*                                  */
/************************************************************************/

unsigned long   CHK_IFDEF_LEVEL (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking IFDEF Nesting level of %ld\n"
                          ,(*tpa_ptr).ifdef_levell  );
    }
#endif

    /*
     *    If the current IFDEF nesting level is 0 or less then there is no
     *    IFDEF command to match up with this #ENDIF/#ELSEIF.
     *    We must therefore report an error.
     */
    if ((*tpa_ptr).ifdef_levell == 0)   /* If nesting depth == 0    */
    {
        REPORT_ERROR (argblk_ptr, SCL_INVSTRUCT);
        /* Report Invalid structure     */
    }

    return PSL_NORMAL;          /* Specified return success */
}

/************************************************************************/
/*                                  */
/*    CHK_IF_NEST - Check IF nesting level.                     */
/*    ============                              */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_IF_NEST (argblk_ptr)             */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Alwaysd PSL_NORMAL                  */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr - Pointer to TPARSE argumennt block      */
/*      IFSTACK_LEVELL - IF command nesting depth       */
/*    Outputs:  None                            */
/*                                  */
/*    This routine is called after parsing an ENDIF/ELSEIF command to   */
/*    ensure that there is a corresponding IF command already parsed.   */
/*                                  */
/************************************************************************/

unsigned long   CHK_IF_NEST (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking IF Nesting level of %ld\n"
                          ,(*tpa_ptr).ifstack_levell  );
    }
#endif

    /*
     *    If the current IF nesting level is 0 or less then there is no IF command
     *    to match up with this ENDIF/ELSEIF. We must therefore report an error.
     */
    if ((*tpa_ptr).ifstack_levell <= 0) /* If nesting depth <= 0    */
    {
        REPORT_ERROR (argblk_ptr, SCL_INVSTRUCT);
        /* Report Invalid DO/ENDDO structure */
    }
    return PSL_NORMAL;          /* Specified return success */
}

/************************************************************************/

/*    CHK_IN_EXPR - Check in an expression              */
/*    ===========                           */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL  If we are in an expression, i.e.    */
/*                  within parentheses "(" ... ")"  */
/*      Zero        Otherwise               */
/*    Status:   Always success                      */
/*    Pointers: Unchanged                       */
/*    Inputs:   None.                               */
/*    Outputs:  None                            */
/*                                  */
/*    This routine is used to determine whether we are currently    */
/*    within a pair of parentheses within an expression (character, */
/*    integer or conditional). It was introduced to allow a comma   */
/*    (",") inside an expression to be differentiated from one      */
/*    outside. This allows TPARSE subexpressions for 'expressions'  */
/*    to differentiate expression-local option clauses and external */
/*    ones, e.g.:                           */
/*                                  */
/*    Expression-local:     (CUSNAM, Trailing spaces + "DEB")       */
/*    External:         Get Uri Uri on 2 Header GetHed, with... */
/*                                  */
/************************************************************************/
unsigned long CHK_IN_EXPR (struct tpaarg *argblk_ptr)
{
    /*
     *  Local data
     */
    struct   scl_com *com_ptr;      /* SCL Common data */
    struct   scl_tpa *tpa_ptr;      /* SCL Tparse ata */
    struct   array_blk *array_blk_ptr;  /* Pointer to current array */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) argblk_ptr->user_ptrp;
    tpa_ptr = com_ptr->tpa_ptrps;   /* SCL Tparse data */
    array_blk_ptr = (*tpa_ptr).array_blk_ptrps;

    /*
     *  Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking in expression\n");
    }
#endif

    /*
     *  1. If we are not within an array subscript, check that we are NOT at the
     *     top - if we are return failure.
     *  2. Else check that we are NOT back at the level we started at - if we
     *     are return failure.
     */
    if ((*tpa_ptr).array_levell == 0)
    {
        if (tpa_ptr->expr_levell == 0) return 0;
    }
    else
    {
        if ((*tpa_ptr).expr_levell == (*array_blk_ptr).nesting_levell) return 0;
    }

    return PSL_NORMAL;          /* Return success - we are in expression */
}


/************************************************************************/
/*                                  */
/*    CHK_INP_OPTS - Check input options bitmask.           */
/*    ============                          */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_INP_OPTS (argblk_ptr)                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL  If valid input command option       */
/*      Zero        If not                  */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr = Pointer to TPARSE Argument Block Structure */
/*      OPT_OPT_SPECIFIEDL - Options bitmask.           */
/*    Outputs:  OPT_OPT_SPECIFIEDL updated to show option requested */
/*                                  */
/*    This routine is called on parsing a variable option. It examines  */
/*    the bitmask OPT_OPT_SPECIFIED to see if the input command option  */
/*    specified by PSLTPA$L_PARAM(AP) has been set already. If 'all is  */
/*    well' the routine returns a success status, otherwise an error    */
/*    status is returned.                       */
/*                                  */
/************************************************************************/

unsigned long   CHK_INP_OPTS (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking the INPUT Command Options Bitmask\n");
    }
#endif

    /*
     *    Check to see if the option is already set and return a failure status if
     *    it is.
     */
    if (((*tpa_ptr).opt_opt_specifiedl & (*argblk_ptr).paraml) != 0)
    {                   /* If option already specified  */
        return 0;           /* ...then return failure   */
    }

    else
    {                   /* ...else if not specified */
        (*tpa_ptr).opt_opt_specifiedl |= (*argblk_ptr).paraml;
        /* ...show specified now    */
        return PSL_NORMAL;      /* ...then return success   */
    }
}

/*======================================================================*/

/*    CHK_IN_RECORD - Check to see if in a RECORD definition        */
/*    ==============                            */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_IN_RECORD (argblk_ptr)               */
/*                                          */
/*    Calling Format:                           */
/*        iret = CHK_IN_RECORD (argblk_ptr)             */
/*                                          */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL  If NOT in a record.             */
/*          Zero        If in a record              */
/*    Pointers: Unchanged                       */
/*    Inputs:                               */
/*    Outputs:                              */
/*                                  */
/*    This routine is called to determine whether or not we are within  */
/*    a RECORD in the definitions section of a script.          */
/*                                      */
/*======================================================================*/

unsigned long   CHK_IN_RECORD (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking for in a Record Definition\n");
    }
#endif

    /*
     *    Check the flag to determine whether or not we are within a record and
     *    return failure if we are, success if we are not.
     */
    if ((*tpa_ptr).def_recidxl != 0)    /* If we processing a record    */
    {
        return 0;           /* ...return failure        */
    }
    else                /* ...otherwise...      */
    {
        return PSL_NORMAL;      /* Return success       */
    }
}
/*
 * 
 */

/************************************************************************/

/*    CHK_INTERVAL - Check INTERVAL value in range          */
/*    ============                          */
/*                                          */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   scl_tpa.int_valuel = Number value.          */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine checks whether the current integer value     */
/*    is in the valid range for an INTERVAL clause. If it is not    */
/*    an error is reported.                     */
/*                                  */
/************************************************************************/
unsigned long CHK_INTERVAL (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Check the value is in range and report an error if it is not.
     */
    if (((*tpa_ptr).int_valuel < MIN_THR_INTERVAL)
            ||  ((*tpa_ptr).int_valuel > MAX_THR_INTERVAL))
    {
        REPORT_ERROR (argblk_ptr, SCL_BADINTERVAL);
    }

    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/
/*                                  */
/*    CHK_INT_RADIX - Check integer radix in range.         */
/*    =============                         */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_INT_RADIX (argblk_ptr)               */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL                      */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr = Pointer to TPARSE Argument Block Structure */
/*      (*tpa_ptr).int_valuel - Integer value of radix      */
/*    Outputs:  None                                                */
/*                                  */
/*   This routine validates the number passed in the radix operand to   */
/*   ensure that it is a valid radix (i.e. is within the range 2-36).   */
/*                                  */
/*   If the integer value is not valid an 'Radix Value Out Of Range'    */
/*   error is reported. In all cases a success status is returned.  */
/*                                  */
/************************************************************************/

unsigned long   CHK_INT_RADIX   (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Check the val;ue specified for the index which must be in the range
     *    2 to 36.
     */
    if (((*tpa_ptr).int_valuel < 2)
            ||  ((*tpa_ptr).int_valuel > 36))
    {                   /* If radix not between 2 adn 36 */
        REPORT_ERROR (argblk_ptr, SCL_INVRADIX);
        /* ...then report invalid radix */
    }
    return PSL_NORMAL;          /* ...and return success    */
}

/************************************************************************/
/*                                  */
/*    CHK_INT_VALUE - Check Integer Value in Range          */
/*    =============                         */
/*                                  */
/*  Type:    Action routine                     */
/*  Status:  Always succeeds (PSL_NORMAL)               */
/*  Pointers: Unchanged                     */
/*  Inputs:  PSLTPA$L_NUMBER(AP) = Integer value to be checked  */
/*       SIGN_CHR = Sign character (ASCII code zero, if none)   */
/*  Outputs: INT_VALUE = Signed integer value           */
/*       DEC_VALUE = Signed integer value           */
/*                                  */
/*    The components of the value are held in PSLTPA$L_NUMBER(AP)   */
/*    (which contains an unsigned number) and SIGN_CHR          */
/*    (which contains the sign character, or zero if no sign was    */
/*    specified).                           */
/*                                  */
/*    If the integer value valid it is stored in INT_VALUE, otherwise   */
/*    an 'Integer Value Out Of Range' error is reported. In both cases  */
/*    success status is returned.                   */
/*                                  */
/************************************************************************/

unsigned long   CHK_INT_VALUE (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking an Integer Value\n");
    }
#endif

    /*
     *    Determine from the any previously defined sign character whether this
     *    is a positive or negative value and save the appropriate value.
     */
    if (((*tpa_ptr).sign_chrb == 0) /* If we have no sign character */
            ||  ((*tpa_ptr).sign_chrb == '+'))  /* ...or explicitly '+'     */
    {
        (*tpa_ptr).dec_integerl = (*argblk_ptr).numberl;
        /* Store number as positive */
    }

    else                /* ...assume sign = negative    */
    {
        (*tpa_ptr).dec_integerl = -(*argblk_ptr).numberl;
        /* Store number as negative */
    }

    (*tpa_ptr).int_valuel = (*tpa_ptr).dec_integerl;
    /* ...and store in INT_VALUE as well */
    return PSL_NORMAL;
}

/*======================================================================*/
/*                                  */
/*    CHK_LAB_DEF - Check Label Definition              */
/*    ============                          */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_LAB_DEF (argblk_ptr)             */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL  If token is a label         */
/*          Zero        Otherwise               */
/*    Pointers: Moved to end of label                   */
/*    Inputs:   LABELS_TABLE = Labels Symbol Table          */
/*          LAB_CNT = Label Count                   */
/*    Outputs:  None                            */
/*                                  */
/*    This routine is entered from a 'PSLTPA_DATANAME' transition to    */
/*    check for a label command. It checks that the dataname is     */
/*    followed by a colon (indicating that we have a label). If not,    */
/*    the routine returns a failure status, otherwise it checks the     */
/*    string is a valid dataname (reporting             */
/*    an 'Invalid label' error if not) and scans the Labels Symbol  */
/*    Table to see if label has already been defined. It then performs  */
/*    one of three actions:                     */
/*                                  */
/*    1. If the label has already been defined a 'Duplicate label'  */
/*       error is reported.                     */
/*                                  */
/*    2. If the label has not been defined, or referenced, an entry */
/*       for it is added to the Labels Symbol Table.            */
/*                                  */
/*    3. If the label has not been defined, but has been referenced,    */
/*       the reference(s) is resolved and Symbol Table entry updated.   */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   CHK_LAB_DEF (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scph    *header_ptr;       /* POinter to header record.    */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */
    struct   labtbl  *labtbl_ptr;       /* Pointer label symbol table   */
    unsigned long    isub;          /* Loop control variable    */
    char    *wrk_ptr;      /* Work pointer to char     */
    unsigned long    label_len;     /* Length of label      */
    long    comp_result;       /* Result of comparison     */
    struct   scl_cmd *scl_cmd_ptr;      /* Pointer to Command Line Struct */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    scl_cmd_ptr = (*com_ptr).cmd_ptrps;
    header_ptr = (*tpa_ptr).header_begps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking for a Defined Label\n");
    }
#endif

    /*
     *    Labels MUST end with a colon. Check for this and add include it in the
     *    label. If a colon is not present then this is not a label and we must
     *    reset the token and string pointers to BEFORE the dataname.
     */
    wrk_ptr = (*argblk_ptr).string_ptrpc;
    /* Point to next character  */
    if (((*argblk_ptr).stringcntl == 0) /* If end of string         */
            ||  (*wrk_ptr != ':'))      /* ...or not ':' then not a label */
    {
        return 0;           /* ...and return error status   */
    }

    /*
     *    At this point we know we have a valid label. Count the colon.
     */
    label_len = (*argblk_ptr).tokencntl; /* Length of label - colon */
    (*argblk_ptr).stringcntl -= 1;  /* Decrement string count   */
    (*argblk_ptr).string_ptrpc += 1;    /* Increment string count   */
    (*argblk_ptr).tokencntl  += 1;  /* Increment token count    */

    /*
     *    Scan the labels symbol table until we reach the end or find a match
     *    to this dataname.
     */
    for (isub=0, labtbl_ptr = (*tpa_ptr).label_begps;
            isub < (*tpa_ptr).lab_cntl;
            isub++, labtbl_ptr++)
    {
        if (label_len == (*labtbl_ptr).lab_nlengthl)
        {               /* If the lengths are the same  */
            comp_result = memcmp ((*argblk_ptr).token_ptrpc,
                                  (*labtbl_ptr).lab_namec16,
                                  (*labtbl_ptr).lab_nlengthl);
            /* Compare the 2 labels     */
            if (comp_result == 0) goto found_label;
            /* ...and branch if names match */
        }
    }

    /*
     *  Drop through to here if we have a new label definition.
     *
     *  Add an entry to the Labels Symbol Table.
     */
    ADD_LABEL (argblk_ptr, (*argblk_ptr).token_ptrpc, label_len);

    /*
     *  Encode label 'command'.
     */
    ENC_LABEL (argblk_ptr, (*argblk_ptr).token_ptrpc, label_len);

    (*tpa_ptr).test_fgl = 0;

    return PSL_NORMAL;          /* Return success */

    /*
     *    We come here having found an entry for the current label
     *    within the Labels Symbol Table. This entry may have have been
     *    set up when a reference to an undefined label was encountered.
     *    If so, we need to update the Symbol Table entry and encode a
     *    Label command. If not, the label has already been defined so we
     *    need to report a 'Duplicate label' error.
     */
found_label:
    if ((*labtbl_ptr).lab_defn)     /* Label already defined ? */
    {
        REPORT_ERROR (argblk_ptr, SCL_DUPLAB);
    }

    else
    {
        ENC_LABEL (argblk_ptr, (*argblk_ptr).token_ptrpc, label_len);

        (*labtbl_ptr).lab_defn = PTRUE; /* Flag label as defined */
    }

    (*tpa_ptr).test_fgl = 0;

    return PSL_NORMAL;          /* Return success */
}

/*======================================================================*/
/*                                  */
/*    CHK_LABEL - Check for Label                   */
/*    ==========                            */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_LABEL (argblk_ptr)                   */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Moved to end of label name              */
/*    Inputs:   None                            */
/*    Outputs:  LABEL_NAME = Label name                 */
/*                                  */
/*    This routine saves the current token into global storage so that  */
/*    the label can be used later (as label 1)              */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   CHK_LABEL (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking for a Label (Saved into Data)\n");
    }
#endif

    /*
     *    Save the label name into global data for later reference.
     */
    (void) memcpy ((*tpa_ptr).label_namec16
                   ,(*argblk_ptr).token_ptrpc
                   ,(*argblk_ptr).tokencntl);
    (*tpa_ptr).label_lenl = (*argblk_ptr).tokencntl;
    return PSL_NORMAL;          /* ...and return success    */
}

/*======================================================================*/
/*                                  */
/*    CHK_LABEL_ONERROR - Check for On Error Label          */
/*    ============                          */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_LABEL_ONERROR (argblk_ptr)               */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Moved to end of label name              */
/*    Inputs:   None                            */
/*    Outputs:  ONERROR_LABELC16 = Second label name            */
/*                                  */
/*    This routine saves the current token into global storage so that  */
/*    the label can be used later (as label 2)              */
/*                                  */
/*    This routine is called to check an ON ERROR GOTO label.       */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   CHK_LABEL_ONERROR (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking the On Error Label (Saved into Data)\n");
    }
#endif

    /*
     *    Save the label name into global data for later reference.
     */
    (void) memcpy ((*tpa_ptr).onerror_labelc16
                   ,(*argblk_ptr).token_ptrpc
                   ,(*argblk_ptr).tokencntl);
    (*tpa_ptr).onerror_lenl = (*argblk_ptr).tokencntl;

    return PSL_NORMAL;          /* ...and return success    */
}


/*======================================================================*/
/*                                  */
/*    CHK_LABEL_TIMEOUT - Check for On Timeout Label            */
/*    =================                         */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_LABEL_TIMEOUT (argblk_ptr)               */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Moved to end of label name              */
/*    Inputs:   None                            */
/*    Outputs:  ONTIMEOUT_LABELC20 = Second label name          */
/*                                  */
/*    This routine saves the current token into global storage so that  */
/*    the label can be used later (as label 2)              */
/*                                  */
/*    This routine is called to check an ON TIMEOUT GOTO label.     */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   CHK_LABEL_TIMEOUT (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking the On Timeout Label (Saved into Data)\n");
    }
#endif

    /*
     *    Save the label name into global data for later reference.
     */
    (void) memcpy ((*tpa_ptr).ontimeout_labelc20
                   ,(*argblk_ptr).token_ptrpc
                   ,(*argblk_ptr).tokencntl);
    (*tpa_ptr).ontimeout_labelc20[(*argblk_ptr).tokencntl] = '\0';

    return PSL_NORMAL;          /* ...and return success    */
}


/************************************************************************/

/*    CHK_LOAD_ACCOUNT - Check for the LOAD ACCOUNTING command valid    */
/*    ================                          */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_LOAD_ACCOUNT (argblk_ptr)                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Zero (Always)                       */
/*    Pointers: Unchanged                       */
/*    Inputs:   None                            */
/*    Outputs:  None                            */
/*                                  */
/*    This routine checks to see if the LOAD ACCOUNTING command is      */
/*    valid on this platform or not.                    */
/*                                  */
/*    This command is not applicable on UNIX or WIN32           */
/*                                  */
/************************************************************************/

unsigned long   CHK_LOAD_ACCOUNT    (struct tpaarg *argblk_ptr)

{
    return 0;               /* Not Valid on UNIX or WIN32 */
}


/************************************************************************/

/*    CHK_MODE - Check that supplied mode does not conflict     */
/*    ========                                  */
/*                                          */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL  If mode conflict.           */
/*      Zero        Otherwise.              */
/*    Pointers: Unchanged.                      */
/*    Inputs:   tpaarg.paraml - Mode option specified.          */
/*    Outputs:  scph_ptr->scph_mode_b - Mode mask updated with      */
/*          supplied mode if no conflict.           */
/*                                  */
/*    This routine is called on parsing the Environment section     */
/*    mode statement to check the mode flags to see if the      */
/*    mode option specified by tpaarg.paraml has been set       */
/*    already. If it has not been set already the routine returns a */
/*    success status and sets the mode flag, otherwise an error     */
/*    status is returned.                       */
/*                                  */
/************************************************************************/
unsigned long   CHK_MODE (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data.
     */
    struct   scl_com *com_ptr;      /* Common Data */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data */
    struct   scph *scph_ptr;        /* SCPH pointer */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) argblk_ptr->user_ptrp;
    tpa_ptr = com_ptr->tpa_ptrps;
    scph_ptr = tpa_ptr->header_begps;   /* Pick up script header pointer */

    /*
     *   Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("\nChecking for incompatible MODE");
    }
#endif

    /*
     *  Disallow duplicates.
     */
    if ((argblk_ptr->paraml & scph_ptr->scph_mode_b) != 0)
    {
        return 0;           /* Mode conflict */
    }
    else                /* ...otherwise valid mode */
    {
        scph_ptr->scph_mode_b |= argblk_ptr->paraml;
        /* Set mode */
        return PSL_NORMAL;      /* Return Success Status */
    }
}


/************************************************************************/

/*    CHK_MODES_END - Check that modes statement has no conflicts   */
/*    =============                             */
/*                                          */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL).               */
/*    Pointers: Unchanged.                      */
/*    Inputs:   scph_ptr->scph_mode_b - Specified modes mask.       */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine is called after parsing of the Environment section   */
/*    mode statement is complete to perform all mode flag checks that   */
/*    cannot be performed earlier. Currently, we check that:        */
/*                                  */
/*    1. CHARACTER-CELL mode has been specified if SCREEN mode has been */
/*       specified.                         */
/*    2. If HTTP mode is specified, then neither CHARACTER-CELL nor */
/*       SCREEN mode is specified.                  */
/*                                  */
/*    If either of these checks fails an error is reported.     */
/*                                  */
/************************************************************************/
unsigned long CHK_MODES_END (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data.
     */
    struct   scl_com *com_ptr;      /* Common Data */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data */
    struct   scph *scph_ptr;        /* SCPH pointer */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) argblk_ptr->user_ptrp;
    tpa_ptr = com_ptr->tpa_ptrps;
    scph_ptr = tpa_ptr->header_begps;   /* Pick up script header pointer */

    /*
     *   Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("\nChecking for incompatible MODES");
    }
#endif

    /*
     *  If SCREEN mode has been specified, check that CHARACTER_CELL mode
     *  has also been specified - if not return 'conflict'.
     */
    if (((scph_ptr->scph_mode_b & MODE_SCREEN) != 0) &&
            ((scph_ptr->scph_mode_b & MODE_CHR_CELL) == 0))
    {
        REPORT_ERROR (argblk_ptr, SCL_SCRMODINV);
    }

    /*
     *  If HTTP mode has been specified, check that CHARACTER-CELL
     *  mode and SCREEN mode have not - if not return conflict.
     */
    if (((scph_ptr->scph_mode_b & MODE_HTTP) != 0)      /* If HTTP mode */
            && (((scph_ptr->scph_mode_b & MODE_SCREEN) != 0)      /* ...and (SCREEN mode */
                ||  ((scph_ptr->scph_mode_b & MODE_CHR_CELL) != 0)))  /* ...or CHARACTER-CELL mode) */
    {
        REPORT_ERROR (argblk_ptr, SCL_HTTPMODECNF);
    }

    return PSL_NORMAL;          /* Return success status */
}


/*======================================================================*/

/*    CHK_NODE - Check node name                    */
/*    ========                              */
/*                                          */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   None                            */
/*    Outputs:  scl_tpa.chr_stringc65536 = Node name.           */
/*      scl_tpa.chr_strlenl = Node name length.         */
/*                                  */
/*    This routine checks whether the current string - stored in    */
/*    tpa_ptr->chr_stringc65536 - is a valid node name, i.e. it is a    */
/*    character string and is not too long. If it is a valid node   */
/*    name it is saved, otherwise an error is reported.         */
/*                                  */
/*----------------------------------------------------------------------*/
unsigned long CHK_NODE (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data.
     */
    struct   scl_com *com_ptr;      /* Common Data */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking for valid node name\n");
    }
#endif

    /*
     *    Check that the token value is not too long.
     */
    if ((*tpa_ptr).chr_strlenl  > LEN_EXE_NODE)
    {
        REPORT_ERROR (argblk_ptr, SCL_NODETOOLONG);
        goto end;
    }
end:
    ;
    return PSL_NORMAL;          /* ...and return success */
}


/************************************************************************/

/*    CHK_NUMBER - Check NUMBER value in range              */
/*    ==========                            */
/*                                          */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   scl_tpa.int_valuel = Number value.          */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine checks whether the current integer value     */
/*    is in the valid range for a NUMBER clause. If it is not       */
/*    an error is reported.                     */
/*                                  */
/************************************************************************/
unsigned long CHK_NUMBER (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Check the value is in range and report an error if it is not.
     */
    if (((*tpa_ptr).int_valuel < MIN_THR_NUMBER)
            ||  ((*tpa_ptr).int_valuel > MAX_THR_NUMBER))
    {
        REPORT_ERROR (argblk_ptr, SCL_BADNUMBER);
    }

    return PSL_NORMAL;          /* Return success */
}



/*======================================================================*/
/*                                  */
/*    CHK_PIPEID - Check Pipe Identifier                */
/*    ==========                            */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_PIPEID (argblk_ptr)                  */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL  If token is a valid pipe id     */
/*          PSL_ABORT   Otherwise               */
/*    Pointers: Moved to end of Pipe Id                 */
/*    Inputs:   PIPE_BEGPS = Pointer to Pipe Symbol Table       */
/*          PIPE_CNTL  = Symbol Count               */
/*    Outputs:  OPEN_FIDIDXL - Set to be current index          */
/*                                  */
/*    This routine checks to see if dataname specified as a PIPEID is   */
/*    contained within the Pipe ID table.               */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   CHK_PIPEID (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */
    struct   fattbl  *fattbl_ptr;       /* Pointer to Pipe symbol table */
    unsigned long    isub;          /* Loop control variable    */
    unsigned long    namlen;        /* Name length          */
    long    comp_result;       /* Result of comparison     */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking for a Valid PIPE ID\n");
    }
#endif

    /*
     *    Scan the Pipe Access symbol table until we reach the end or find a match
     *    to this dataname.
     *
     *    The index MUST be greater than 0 so we add one to the loop count.
     */
    namlen = (*argblk_ptr).tokencntl;

    for (isub=0, fattbl_ptr = (*tpa_ptr).pipe_begps;
            isub < (*tpa_ptr).pipe_cntl;
            isub++, fattbl_ptr++)
    {
        if (namlen == strlen ((*fattbl_ptr).fat_namec20))
        {               /* If the lengths are the same  */
            comp_result = memcmp ((*argblk_ptr).token_ptrpc,
                                  (*fattbl_ptr).fat_namec20,
                                  namlen);
            /* Compare the 2 pipes      */
            if (comp_result == 0)   /* ...if names match        */
            {
                (*tpa_ptr).open_fididxl = isub + 1;
                /* Save index of this pipe  */
                goto found_pipe;    /* ...and branch to found pipe  */
            }
        }
    }

    /*
     *    At this point we have determined that the pipe has not been defined.
     *    For pipes this is illegal, so return failure.
     */
    return PSL_ABORT;           /* Return Failure       */
    /*
     *    ...and return
     */
found_pipe:
    return PSL_NORMAL;          /* Return Success       */
}


/************************************************************************/
/*                                  */
/*    CHK_PIPE_SYNC - Check if Current Pipe is Synchronous      */
/*    =============                             */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_PIPE_SYNC (argblk_ptr)               */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL if current pipe is NOT synchronous       */
/*      PSL_ABORT otherwise                 */
/*    Pointers: Unchanged                       */
/*    Inputs:   OPEN_FIDIDXL = Current Pipe id              */
/*    Outputs:  None                            */
/*                                  */
/*    This routine checks to see if the current pipe is synchronous */
/*    and, if it is, returns a failure status.              */
/*                                  */
/************************************************************************/

unsigned long   CHK_PIPE_SYNC (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */
    struct   fattbl  *fat_ptr;      /* Pipe Access Table pointer    */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking PIPE to see if SYNCHRONOUS\n");
    }
#endif

    /*
     *    Check to see if the current pipe is synchronous. If it is
     *    then return a failure status
     */
    fat_ptr = (*tpa_ptr).pipe_begps + (*tpa_ptr).open_fididxl - 1;
    /* Point to current entry in table */
    if (((*fat_ptr).fat_refflgl & PIPATT_SYNC) != 0)
        /* ...if pipe is synchronous    */
    {
        return PSL_ABORT;       /* ...then return failure   */
    }

    else                /* ...else if not command mode  */
    {
        return PSL_NORMAL;      /* Return success       */
    }
}


/************************************************************************/

/*    CHK_POST_CMND - Check Current HTTP Request is a POST      */
/*    =============                         */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL if current HTTP request is a POST        */
/*      PSL_ABORT otherwise                 */
/*    Pointers: Unchanged                       */
/*    Inputs:   None.                           */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine checks to see if the current HTTP request is a   */
/*    POST and if not returns a failure status.             */
/*                                  */
/************************************************************************/
unsigned long CHK_POST_CMND (struct tpaarg *argblk_ptr)
{
    /*
     *  Local data.
     */
    struct scl_com  *com_ptr;       /* Common Data */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *  Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking for HTTP POST REQUEST specified\n");
    }
#endif

    /*
     *  Return success if HTTP request type is POST and failure otherwise.
     */
    if ((*tpa_ptr).http_reqtypl == TOF_REQTYP_POST)
    {
        return PSL_NORMAL;
    }
    else
    {
        return PSL_ABORT;
    }
}


/************************************************************************/
/*                                  */
/*    CHK_REC_COMPLETE  - Check un terminated record            */
/*    =================                         */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_REC_COMPLETE  (argblk_ptr)               */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   DEF_RECIDXL Record Imdex                */
/*    Outputs:  None                            */
/*                                  */
/*    This routine is called after parsing a RECORD statement.      */
/*                                  */
/*    Its purpose is to determine whether or not we already have a  */
/*    record being processed and if so report an error.         */
/*                                  */
/*    We always return with success as we may as well continue parsing  */
/*    the record as it might otherwise be valid.            */
/*                                  */
/*    We also set the index of the entry in the variables table and set */
/*    the variable length to zero.                  */
/*                                  */
/************************************************************************/

unsigned long   CHK_REC_COMPLETE (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking for a RECORD being processed\n");
    }
#endif

    if ((*tpa_ptr).def_recidxl != 0)
    {
        REPORT_ERROR (argblk_ptr, SCL_NORECNEST);
        /* Pass to underline routine      */
    }

    else
    {
        (*tpa_ptr).shrvar_rec_offl = 0;
        /* Set record global offset */
        /* Used to find record size */

        (*tpa_ptr).def_recidxl = (*tpa_ptr).var_cntl;
        /* Save index of record variable   */
        (*tpa_ptr).def_sizel = 0;   /* Clear size              */
    }

    return PSL_NORMAL;          /* Return Success Status       */
}


/*
 * 
 */
/************************************************************************/
/*                                  */
/*    CHK_REC_UNTERM  - Check un terminated record          */
/*    =================                         */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_REC_UNTERM  (argblk_ptr)             */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   DEF_RECIDXL Record Imdex                */
/*    Outputs:  None                            */
/*                                  */
/*    This routine is called after parsing ~CODE statement.     */
/*                                  */
/*    It checks if all record definitions complete              */
/*                                  */
/************************************************************************/

unsigned long   CHK_REC_UNTERM (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking for a RECORD being processed\n");
    }
#endif

    if ((*tpa_ptr).def_recidxl != 0)
    {
        REPORT_ERROR (argblk_ptr, SCL_MISENDREC);
        /* Pass to underline routine      */
        return 0;           /* Return Failure Status       */
    }

    return PSL_NORMAL;          /* Return Success Status       */
}
/*
 * 
 */

/************************************************************************/

/*    CHK_RETURN_VALID - Check that RETURN command is valid     */
/*    ================                          */
/*                                          */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL  If RETURN command is valid here     */
/*      Zero        Otherwise.              */
/*    Pointers: Unchanged.                      */
/*    Inputs:   None.                           */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine checks whether the RETURN command is valid here. */
/*                                  */
/************************************************************************/
unsigned long CHK_RETURN_VALID (struct tpaarg *argblk_ptr)
{
    /*
     *  Local data.
     */
    struct   scl_com    *com_ptr;       /* Common data */
    struct   scl_tpa    *tpa_ptr;       /* TPARSE data */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *  Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking if RETURN command valid\n");
    }
#endif

    /*
     *  The RETURN command is only valid within a subroutine or a sub-test.
     */
    if (((*tpa_ptr).subrtn_cntl == 0) &&
            ((!(*com_ptr).type_testn) ||
             ((*tpa_ptr).subtst_cntl == 0)))
    {
        return 0;           /* RETURN command NOT valid */
    }
    else
    {
        return PSL_NORMAL;      /* RETURN command valid */
    }
}


/************************************************************************/

/*    CHK_SCL_SECTION - Check script section being parsed       */
/*    ===============                           */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_SCK_SECTION (argblk_ptr)             */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL If SCL scetion flag matches the value    */
/*             passed in the TPARSE parameter field     */
/*      0      Otherwise                    */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr = Pointer to TPARSE Argument Block Structure */
/*    Outputs:  None                            */
/*                                  */
/*    This routine checks whether the SCL section flag (indicating  */
/*    which script section is currently being parsed) matches the   */
/*    the value passed in TPARSE argument parameter field. If it    */
/*    does the routine returns a success status, otherwise an error */
/*    status is returned.                       */
/*                                  */
/************************************************************************/

unsigned long   CHK_SCL_SECTION (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE (
            "Checking for SCL section flag value of %ld; actual value is %ld\n",
            (*argblk_ptr).paraml, (*tpa_ptr).sclsec_flagl);
    }
#endif

    /*
     *    Check whether the SCL section flag matches.
     */
    if ((*tpa_ptr).sclsec_flagl == (*argblk_ptr).paraml)
    {
        return PSL_NORMAL;      /* Return success on match */
    }
    else
    {
        return 0;           /* Return failure */
    }
}


/************************************************************************/
/*                                  */
/*    CHK_SCREEN_DEPTH - Check screen depth in range            */
/*    ================                          */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_SCREEN_DEPTH (argblk_ptr)                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr = Pointer to TPARSE Argument Block Structure */
/*    Outputs:  None                            */
/*                                  */
/*    This routine saves the maximum number of lines in the VP$SCREEN   */
/*    screen image in the script object ('tof') file header data    */
/*    buffer after ensuring that the value is in range.         */
/*                                  */
/************************************************************************/
unsigned long CHK_SCREEN_DEPTH (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data */
    struct scl_tofh *tofh_ptr;      /* Pointer to 'tof' header data */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    tofh_ptr = (*tpa_ptr).tofh_begps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking Screen Depth of %ld in range\n"
                          ,(*argblk_ptr).numberl);
    }
#endif

    /*
     *  If the supplied number is out of range report an error, otherwise store
     *  the value in the script object (tof') file header data buffer.
     */
    if (((*argblk_ptr).numberl < MIN_SLINE)
            ||  ((*argblk_ptr).numberl > MAX_SLINE))
    {
        REPORT_ERROR (argblk_ptr, SCL_VPSLINOUTRANGE);
    }

    else
    {
        (*tofh_ptr).tofh_sline_w = (short) (*argblk_ptr).numberl;
    }

    /*
     *   ...and return success
     */
    return PSL_NORMAL;          /* ...always return success */
}

/************************************************************************/
/*                                  */
/*    CHK_SCREEN_MODE - Check For Screen Mode Active            */
/*    ===============                           */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   SCPH record                     */
/*    Outputs:  None                            */
/*                                  */
/*    This routine checks to see if Screen Mode is Active and if not    */
/*    reports an error that Screen mode commands are not valid.     */
/*                                  */
/************************************************************************/

unsigned long   CHK_SCREEN_MODE (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking for SCREEN MODE specified\n");
    }
#endif

    /*
     *    Check to see if the screen mode bit is set in the mode flag and if it
     *    is then return success. If it is not then report an invalid mode.
     */
    if (((*tpa_ptr).mode_flagl & MODE_SCREEN) == 0)
        /* If not mode screen       */
    {
        REPORT_ERROR (argblk_ptr, SCL_NOTSCREENMODE);
        /* Pass to underline routine    */
    }

    return PSL_NORMAL;          /* Always return success    */
}

/*======================================================================*/
/*                                  */
/*    CHK_SCRIPT_OPTS - Check Script Options                */
/*    ================                          */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_SCRIPT_OPTS (argblk_ptr)             */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   SCL_OPTL = Script Options               */
/*    Outputs:  None                            */
/*                                  */
/*    This routine checks to see if a Script Option has already been    */
/*    specified. It reports a 'Duplicate Keyword' error if it has.  */
/*                                  */
/*======================================================================*/

unsigned long   CHK_SCRIPT_OPTS (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking Script Options\n");
    }
#endif

    if (((*tpa_ptr).scl_optl & (*argblk_ptr).paraml) != 0)
        /* If option already specified  */
    {
        REPORT_ERROR (argblk_ptr, PSL_CNFDUPOPT);
        /* Pass to underline routine    */
    }

    return PSL_NORMAL;          /* ...and return success regardless */
}

/************************************************************************/
/*                                  */
/*    CHK_SCREEN_WIDTH - Check screen width in range            */
/*    ================                          */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_SCREEN_WIDTH (argblk_ptr)                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr = Pointer to TPARSE Argument Block Structure */
/*    Outputs:  None                            */
/*                                  */
/*    This routine saves the maximum number of columns in the VP$SCREEN */
/*    screen image in the script object ('tof') file header data    */
/*    buffer after ensuring that the value is in range.         */
/*                                  */
/************************************************************************/
unsigned long CHK_SCREEN_WIDTH (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common data */
    struct scl_tpa  *tpa_ptr;       /* TPARSE data */
    struct scl_tofh *tofh_ptr;      /* Pointer to 'tof' header data */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    tofh_ptr = (*tpa_ptr).tofh_begps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking Screen Width of %ld in Range\n"
                          ,(*argblk_ptr).numberl);
    }
#endif

    /*
     *  If the supplied number is out of range report an error, otherwise store
     *  the value in the script object (tof') file header data buffer.
     */
    if (((*argblk_ptr).numberl < MIN_SCOLS)
            ||  ((*argblk_ptr).numberl > MAX_SCOLS))
    {
        REPORT_ERROR (argblk_ptr, SCL_VPSCOLOUTRANGE);
    }

    else
    {
        (*tofh_ptr).tofh_scols_w = (short) (*argblk_ptr).numberl;
    }

    /*
     *   ...and return success
     */
    return PSL_NORMAL;          /* ...always return success */
}

/************************************************************************/

/*    CHK_SYNCHRONIZE_END - Check end of a SYNCHRONIZE REQUEST Request  */
/*    ===================                       */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always success                      */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr = Pointer to TPARSE Argument Block Structure */
/*    Outputs:  None                            */
/*                                  */
/*    This routine is called at the end of parsing a SYNCHRONIZE    */
/*    REQUEST request to check for conflicting options that can only    */
/*    be checked once command parsing is complete. The only check   */
/*    currently made is that if the "ON TIMEOUT GOTO tmolab" option */
/*    has been specified, the "WITH TIMEOUT period" option has also */
/*    been specified. If it hs not an error is reported.        */
/*                                  */
/************************************************************************/
unsigned long CHK_SYNCHRONIZE_END (struct tpaarg *argblk_ptr)
{
    /*
     *  Local data
     */
    struct   scl_com *com_ptr;      /* SCL Common data */
    struct   scl_tpa *tpa_ptr;      /* SCL Tparse ata */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) argblk_ptr->user_ptrp;
    tpa_ptr = com_ptr->tpa_ptrps;   /* SCL Tparse data */

    /*
     *  Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Performing end of SYNCHRONIZE REQUEST Request checks\n");
    }
#endif

    /*
     *  If the "ON TIMEOUT GOTO tmolab" option has been specified, but
     *  the "WITH TIMEOUT period" option has NOT, report an error.
     */
    if (((tpa_ptr->cmnd_opt_specifiedl & CMND_OPT_TIMEOUT) != 0)
            &&  ((tpa_ptr->cmnd_opt_specifiedl & CMND_OPT_WAIT_PRD) == 0))
    {
        REPORT_ERROR (argblk_ptr, PSL_CNFDUPOPT);
    }

    return PSL_NORMAL;          /* Return success */
}

/*======================================================================*/

/*    CHK_EXE_SCRIPT - Check execute script name is a dataname          */
/*    ==============                            */
/*                                          */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   None                            */
/*    Outputs:  scl_tpa.chr_stringc65536 = Node name.           */
/*      scl_tpa.chr_strlenl = Node name length.         */
/*                                  */
/*    This routine checks whether the script name - stored in       */
/*    tpa_ptr->chr_stringc65536 - is an OpenSTA dataname. If it is a  */
/*    valid dataname it is saved and an entry added to the script   */
/*    dependency list, otherwise an error is reported.          */
/*                                  */
/*----------------------------------------------------------------------*/
unsigned long CHK_EXE_SCRIPT (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data.
     */
    struct   scl_com *com_ptr;      /* Common Data */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data */
    struct   oph     *oph_ptr;      /* Operand header pointer */
    struct   opd     *opd_ptr;      /* Operand descriptor pointer */
    struct   tdf     *tdf_ptr;      /* Test thread data pointer */
    struct tstthrtbl
                *tstthr_ptr;        /* Work test thread pointer */
    unsigned long    iret;          /* Status return */
    PSL_U_WORD_T status;       /* Routine status return */
    PSL_U_WORD_T vallen;       /* Character value ength */
    char depnam [DEPBUFLEN];   /* Dependency name */
    char depdat [DEPBUFLEN];   /* Dependency type specific data */
    char    *llb_ctx;      /* Linked list context */
    int     toftyp = DEP_TOFTYP_SCRIPT;
    /* Depdendency TOF type flag - assume 'script' */
    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
{
        PSL_OUTPUT_TRACE ("Checking for valid script name\n");
    }
#endif

    /*
     *  Check that the script name is a valid CYRANO dataname.
     */
    vallen = (*tpa_ptr).chr_strlenl;
    CHK_LONG_DATANAME ((*tpa_ptr).chr_stringc65536, &vallen ,&status);

    if (status != 0)
    {
        REPORT_ERROR (argblk_ptr, SCL_BADSCRIPT);
        goto end;
    }

    /*
     *  Copy the operand containing the script name from the parsed operand stack.
     */
    iret = PSL_VM_COPY ((*tpa_ptr).parsed_operand_llps ,(char *) &oph_ptr
                        ,NULL ,&llb_ctx ,PSL_VM_MODE_FIRST);
    /* Remove the OPH pointer first */
    if  (!(psl_test_status (iret))) /* ...report any errors */
    {
        PSL_SIGDIAG ((*tpa_ptr).diag_ptrps, 2, PSL_SYSERVERR
                     ,"PSL_VM_COPY OPH (in ADD_DEP_EXESCR)", iret);
    }

    iret = PSL_VM_COPY ((*tpa_ptr).parsed_operand_llps ,(char *) &opd_ptr
                        ,llb_ctx ,NULL ,PSL_VM_MODE_NEXT);
    /* ...then remove the OPD ptr */
    if  (!(psl_test_status (iret))) /* ...report any errors */
    {
        PSL_SIGDIAG ((*tpa_ptr).diag_ptrps, 2, PSL_SYSERVERR
                     ,"PSL_VM_COPY OPD (in ADD_DEP_EXESCR)", iret);
    }

    /*
     *  Add script name to dependency list, if required.
     */
    if ((*oph_ptr).addrl == ADDR_TSTTHR)
    {
        (void) memcpy (depnam, (*tpa_ptr).chr_stringc65536, vallen);
        depnam [vallen] = '\0';
        tstthr_ptr = (*opd_ptr).uopd.tstthr.tstthrtbl_ptrps;
        tdf_ptr = (*tstthr_ptr).tstthr_tdf_ptrps;

        /*
         *    Pick up the command options as this should tell us whether we have an excute script or an 
         *    execute TOF which will need reflecting in the generated dependancy file. Scripts will be 0
         *    where as TOF will be 1.
         */
        if (((*tpa_ptr).cmnd_opt_specifiedl & CMND_OPT_TOF) != 0)
        {
            toftyp = DEP_TOFTYP_TOF;
        }

        /*
         *    ...and construct the dependancy object from this information.
         */
        (void) sprintf (depdat, "%d%c%d", (*tdf_ptr).tdf_targetl, DEP_SEPCHR, toftyp);
        ADD_DEP (com_ptr, depnam, DEP_TYP_EXESCR, depdat);
    }

end:
    ;
    return PSL_NORMAL;          /* ...and return success */
}


/************************************************************************/
/*                                  */
/*    CHK_START_SUBROUTINE - Ensure SUBROUTINE valid at this point  */
/*    ====================                      */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_START_SUBROUTINE (argblk_ptr)            */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL if NOT already in a subroutine or sub-test   */
/*      0 - Otherwise                       */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr = Pointer to TPARSE Argument Block Structure */
/*    Outputs:  None                            */
/*                                  */
/*    This routine checks the subroutine flags to see whether or not    */
/*    we are within a subroutine. If we are an error status is returned */
/*                                  */
/************************************************************************/

unsigned long   CHK_START_SUBROUTINE (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking for SUBROUTINE command valid\n");
    }
#endif

    /*
     *    Check to see if we are currently processing a subroutine or sub-test.
     *    We return an error status if we are, otherwise we reset the SUBROUTINE
     *    flags and return success.
     */
    if ((((*tpa_ptr).subroutine_flagsl & SUB_START) != 0) ||
            (((*tpa_ptr).subtst_flagsl & SUBTST_START) != 0))
    {                   /* Processing subroutine/sub-test? */
        return 0;           /* ...if so, SUBROUTINE invalid */
    }

    else
    {
        (*tpa_ptr).subroutine_flagsl |= SUB_START;
        /* ...valid so show in subrtn now */
        (*tpa_ptr).subprm_cntl = 0; /* Clear subroutine               */
        /* parameter count.       */
        return PSL_NORMAL;      /* ...and return success      */
    }
}

/************************************************************************/

/*    CHK_START_SUBTST - Ensure SUB-TEST valid at this point        */
/*    ================                          */
/*                                          */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL if NOT already in a subroutine or sub-test   */
/*      0 - Otherwise                       */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr = Pointer to TPARSE Argument Block Structure */
/*    Outputs:  (*tpa_ptr).subtst_flagsl = Show now in a sub-test.  */
/*                                  */
/*    Check whether a SUB-TEST statement is valid at this point. Check  */
/*    to see if we are currently processing a subroutine or sub-test.   */
/*    We return an error status if we are, otherwise we reset the   */
/*    SUB-TEST flags and return success.                */
/*                                  */
/************************************************************************/
unsigned long CHK_START_SUBTST (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking for SUB-TEST command valid\n");
    }
#endif

    /*
     *    Check to see if we are currently processing a subroutine or sub-test.
     *    We return an error status if we are, otherwise we reset the SUB-TEST
     *    flags and return success.
     */
    if ((((*tpa_ptr).subroutine_flagsl & SUB_START) != 0) ||
            (((*tpa_ptr).subtst_flagsl & SUBTST_START) != 0))
    {                   /* Processing subroutine/sub-test? */
        return 0;           /* ...if so, SUB-TEST invalid */
    }

    else
    {
        (*tpa_ptr).subtst_flagsl |= SUBTST_START;
        /* ...valid so show in sub-test now */
        return PSL_NORMAL;      /* ...and return success */
    }
}


/************************************************************************/

/*    CHK_SUBTST - Check for declared sub-test              */
/*    ==========                            */
/*                                          */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   None.                           */
/*    Outputs:  (*tsttbl_ptr).tst_definedn = PTRUE sub-test defined */
/*      (*tpa_ptr).tst_ptrps = Sub-test Symbol Table address    */
/*                                  */
/*    This routine scans the Tests Symbol Table to determine if the     */
/*    current token (which has already been proved to be a dataname)    */
/*    is in fact a declared sub-test AND has NOT been defined.      */
/*                                  */
/*    If the sub-test has been declared and has NOT been defined    */
/*    its name is saved and a flag is set to show that it       */
/*    has now been defined.                     */
/*                                  */
/************************************************************************/
unsigned long CHK_SUBTST (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data */
    struct   tsttbl  *tsttbl_ptr;       /* Pointer to Tests Symbol Table */
    unsigned long    isub;          /* Loop control variable */
    long    comp_result;       /* Result of comparison */
    long    namlen;        /* Name length work variable */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking for a declared SUB-TEST\n");
    }
#endif

    /*
     *    Scan the Tests Symbol Table until we reach the end or find a name match.
     *
     *    The index MUST be greater than zero so we add one to the loop count.
     */
    for (isub=0, tsttbl_ptr = (*tpa_ptr).tst_begps;
            isub < (*tpa_ptr).tst_cntl;
            isub++, tsttbl_ptr++)
    {
        namlen = strlen ((*tsttbl_ptr).tst_namec68);

        if ((*argblk_ptr).tokencntl == namlen)
        {               /* If the lengths are the same */
            comp_result = memcmp ((*argblk_ptr).token_ptrpc,
                                  (*tsttbl_ptr).tst_namec68,
                                  namlen);
            /* Compare the names */
            if (comp_result == 0) goto found_tst;
        }
    }

    /*
     *    At this point we have determined that the test has not been declared.
     *    For tests this is illegal, so return failure.
     */
    return PSL_ABORT;           /* Return failure */

    /*
     *    At this point we have determined that the test has been declared.
     *    Now check that its has been declared as a SUB-TEST (not a TEST) and that
     *    has not already been defined. If it has report an error, otherwise save
     *    the sub-test name and flag that it has now been defined.
     */
found_tst:
    if (!((*tsttbl_ptr).tst_subtstn))       /* If declared as TEST not SUB-TEST */
    {
        REPORT_ERROR (argblk_ptr, SCL_TSTNOTSUBTST);
    }
    else if ((*tsttbl_ptr).tst_definedn)    /* If already defined SUB-TEST */
    {
        REPORT_ERROR (argblk_ptr, SCL_DUPSUBTST);
    }
    else
    {
        (void) memcpy ((*tpa_ptr).subtst_namec20,
                       (*argblk_ptr).token_ptrpc,
                       (*argblk_ptr).tokencntl);
        (*tpa_ptr).subtst_namec20 [(*argblk_ptr).tokencntl] = '\0';

        (*tsttbl_ptr).tst_definedn = PTRUE;
        (*tpa_ptr).tst_ptrps = tsttbl_ptr;
    }
    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/
/*                                  */
/*    CHK_SUBRTN_DEF - Check for Subroutine already defined     */
/*    ==============                            */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_SUBRTN_DEF (argblk_ptr)              */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   N/A                         */
/*    Pointers: Moved to end of timer name              */
/*    Inputs:   argblk_ptr - Pointer to TPARSE data         */
/*    Outputs:  None                                                */
/*                                  */
/************************************************************************/
void CHK_SUBRTN_DEF (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */
    struct   subtbl  *subtbl_ptr;       /* Subroutine symbol table  */
    unsigned long    isub;          /* Loop control variable    */
    long    comp_result;       /* Result of comparison     */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking for Subroutine Name Defined \n");
    }
#endif

    /*
     *    Scan the subroutines symbol table until we reach the end or find a match
     *    to this dataname.
     */
    for (isub=0, subtbl_ptr = (*tpa_ptr).subroutine_begps;
            isub < (*tpa_ptr).subrtn_cntl;
            isub++, subtbl_ptr++)
    {
        comp_result = strcmp ((*tpa_ptr).subroutine_namec20,
                              (*subtbl_ptr).sub_namec20);
        /* Compare the 2 subroutines    */
        if (comp_result == 0) goto found_subrtn;
        /* ...and branch if names match */
    }

    /*
     *    At this point we have located a subroutine which has not as yet been
     *    defined. Add an entry to the labels symbol table to define it.
     */
    if (isub >= MAX_SUBROUTINES)    /* Subroutine Symbol Table full ? */
    {
        REPORT_ERROR (argblk_ptr, SCL_MAXSUB);
        /* Report maximum subroutines defined */
    }

    else

        /*
         *    Add a new label entry to the table at the current point
         */
    {
        (void) strcpy ((*subtbl_ptr).sub_namec20
                       ,(*tpa_ptr).subroutine_namec20);
        /* Copy name into table     */
        (*subtbl_ptr).sub_defn = PTRUE;
        /* Flag subroutine defined */
        (*subtbl_ptr).sub_param_cntl = (*tpa_ptr).subprm_cntl;
        /* Save parameter count     */
        (*tpa_ptr).subrtn_cntl += 1;    /* Increment count of subroutines */
    }

    return;             /* Return           */

    /*
     *    We come here having found an entry for the current subroutine
     *    within the Subroutine Symbol Table. This entry may have have been
     *    set up when a reference to an undefined subroutine was encountered.
     *    If so, we need to resolve all references to this subroutine and update
     *    the Symbol Table entry. If not, the subroutine has already been defined
     *    so we need to report a 'Duplicate subroutine' error.
     */
found_subrtn:
    if ((*subtbl_ptr).sub_defn)     /* Subroutine already defined ? */
    {
        REPORT_ERROR (argblk_ptr, SCL_DUPSUB);
        /* Report Duplicate subrtn defined */
    }

    /*
     * Check the parameter count is same as previously defined subroutine
     * Report an error if it does not match.
     */

    else if ((*tpa_ptr).subprm_cntl != (*subtbl_ptr).sub_param_cntl)
    {
        REPORT_ERROR (argblk_ptr, SCL_INVSUBPRM);
    }

    /*
     *  Flag subroutine defined.
     */
    (*subtbl_ptr).sub_defn = PTRUE;

    return;             /* ...and return success    */
}

/************************************************************************/

/*    CHK_TEST - Check for declared Test                */
/*    ========                              */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_TEST (argblk_ptr)                    */
/*                                          */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL If declared test             */
/*          PSL_ABORT  Otherwise                    */
/*    Pointers: Unchanged.                      */
/*    Inputs:   None.                           */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine scans the Tests Symbol Table to determine if the     */
/*    current token (which has already been proved to be a dataname)    */
/*    is in fact a declared test.                   */
/*                                  */
/*    If the test has been declared the address of its entry in     */
/*    the Tests Symbol Table is returned in tst_ptrps, otherwise    */
/*    tst_ptrps is set to zero and a failure status is returned.    */
/*                                  */
/************************************************************************/
unsigned long CHK_TEST (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data */
    struct   tsttbl  *tsttbl_ptr;       /* Pointer to Tests Symbol Table */
    unsigned long    isub;          /* Loop control variable */
    long    comp_result;       /* Result of comparison */
    long    namlen;        /* Name length work variable */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking for a declared Test\n");
    }
#endif

    /*
     *    Scan the Tests Symbol Table until we reach the end or find a name match.
     *
     *    The index MUST be greater than zero so we add one to the loop count.
     */
    for (isub=0, tsttbl_ptr = (*tpa_ptr).tst_begps;
            isub < (*tpa_ptr).tst_cntl;
            isub++, tsttbl_ptr++)
    {
        namlen = strlen ((*tsttbl_ptr).tst_namec68);

        if ((*argblk_ptr).tokencntl == namlen)
        {               /* If the lengths are the same */
            comp_result = memcmp ((*argblk_ptr).token_ptrpc,
                                  (*tsttbl_ptr).tst_namec68,
                                  namlen);
            /* Compare the names */
            if (comp_result == 0) goto found_tst;
        }
    }

    /*
     *    At this point we have determined that the test has not been defined.
     *    For tests this is illegal, so return failure.
     */
    return PSL_ABORT;           /* Return failure */

    /*
     *    At this point we have determined that the test has been defined.
     *    Save the symbol table entry address and return success.
     */
found_tst:
    (*tpa_ptr).tst_ptrps = tsttbl_ptr;  /* Save symbol table entry address */
    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/

/*    CHK_TEST_THREAD - Check for declared Test Thread          */
/*    ===============                           */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_TEST_THREAD (argblk_ptr)             */
/*                                          */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL If declared test             */
/*          PSL_ABORT  Otherwise                    */
/*    Pointers: Unchanged.                      */
/*    Inputs:   None.                           */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine scans the Test Threads Symbol Table to determine */
/*    if the current token (which has already been proved to be a   */
/*    dataname) is in fact a declared test thread.          */
/*                                  */
/*    If the test thread has been declared the address of its entry in  */
/*    the Test Threads Symbol Table is returned in tstthr_ptrps,    */
/*    otherwise tstthr_ptrps is set to zero and a failure status is */
/*    returned.                             */
/*                                  */
/************************************************************************/
unsigned long CHK_TEST_THREAD (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data */
    struct   tstthrtbl
                *tstthrtbl_ptr;    /* Pointer to Test Threads Symbol */
    /* Table */
    unsigned long    isub;          /* Loop control variable */
    long    comp_result;       /* Result of comparison */
    long    namlen;        /* Name length work variable */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
{
        PSL_OUTPUT_TRACE ("Checking for a declared Test Thread\n");
    }
#endif

    /*
     *    Scan the Test Threads Symbol Table until we reach the end or find a
     *    name match.
     *
     *    The index MUST be greater than zero so we add one to the loop count.
     */
    for (isub=0, tstthrtbl_ptr = (*tpa_ptr).tstthr_begps;
            isub < (*tpa_ptr).tstthr_cntl;
            isub++, tstthrtbl_ptr++)
    {
        namlen = strlen ((*tstthrtbl_ptr).tstthr_namec68);

        if ((*argblk_ptr).tokencntl == namlen)
        {               /* If the lengths are the same */
            comp_result = memcmp ((*argblk_ptr).token_ptrpc,
                                  (*tstthrtbl_ptr).tstthr_namec68,
                                  namlen);
            /* Compare the names */
            if (comp_result == 0) goto found_tstthr;
        }
    }

    /*
     *    At this point we have determined that the test thread has not been
     *    defined. For test threads this is illegal, so return failure.
     */
    return PSL_ABORT;           /* Return failure */

    /*
     *    At this point we have determined that the test thread has been defined.
     *    Save the symbol table entry address and return success.
     */
found_tstthr:
    (*tpa_ptr).tstthr_ptrps = tstthrtbl_ptr;
    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/

/*    CHK_THREAD - Check for declared Thread                */
/*    ==========                            */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_THREAD (argblk_ptr)                  */
/*                                          */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL If declared Thread script thread     */
/*          PSL_ABORT  Otherwise                    */
/*    Pointers: Unchanged.                      */
/*    Inputs:   None.                           */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine scans the Threads Symbol Table to determine      */
/*    if the current token (which has already been proved to be a   */
/*    dataname) is in fact a declared thread.               */
/*                                  */
/*    If the thread has been declared the address of its entry in   */
/*    the Threads Symbol Table is returned in thr_ptrps,        */
/*    otherwise thr_ptrps is set to zero and a failure status is    */
/*    returned.                             */
/*                                  */
/************************************************************************/
unsigned long CHK_THREAD (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data */
    struct   thrtbl *thrtbl_ptr;        /* Pointer to Threads Symbol Table */
    unsigned long    isub;          /* Loop control variable */
    long    comp_result;       /* Result of comparison */
    long    namlen;        /* Name length work variable */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking for a declared Thread\n");
    }
#endif

    /*
     *    Scan the Threads Symbol Table until we reach the end or find a
     *    name match.
     *
     *    The index MUST be greater than zero so we add one to the loop count.
     */
    for (isub=0, thrtbl_ptr = (*tpa_ptr).thr_begps;
            isub < (*tpa_ptr).thr_cntl;
            isub++, thrtbl_ptr++)
    {
        namlen = strlen ((*thrtbl_ptr).thr_namec20);

        if ((*argblk_ptr).tokencntl == namlen)
        {               /* If the lengths are the same */
            comp_result = memcmp ((*argblk_ptr).token_ptrpc,
                                  (*thrtbl_ptr).thr_namec20,
                                  namlen);
            /* Compare the names */
            if (comp_result == 0) goto found_thr;
        }
    }

    /*
     *    At this point we have determined that the thread has not been
     *    defined. For threads this is illegal, so return failure.
     */
    return PSL_ABORT;           /* Return failure */

    /*
     *    At this point we have determined that the thread has been defined.
     *    Save the symbol table entry address and return success.
     */
found_thr:
    (*tpa_ptr).thr_ptrps = thrtbl_ptr;
    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/

/*    CHK_THR_DEF - Perform end of thread definition checks     */
/*    ===========                           */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*     iret = CHK_THR_DEF (argblk_ptr);             */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL (always)                 */
/*    Pointers: Unchanged.                      */
/*    Inputs:   None.                           */
/*                                  */
/*    This routine performs all validation necessary on a complete  */
/*    thread definition and reports any errors.             */
/*                                  */
/************************************************************************/
unsigned long CHK_THR_DEF (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   scl_com    *com_ptr;       /* Pointer to Common Data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE Data */
    struct   tdf        *tdf_ptr;       /* Pointer to Terminal Definition */
    struct   thrtbl     *thr_ptr;       /* Pointer to thread definition */
    char       thr_name [20];  /* Local copy of thread name */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    thr_ptr = (*tpa_ptr).def_thr_wrkps;
    tdf_ptr = (*thr_ptr).thr_tdf_ptrps;

    /*
     *  Get local copy of thread name for ease of reference.
     */
    (void) memcpy (thr_name, (*tpa_ptr).def_namec64, (*tpa_ptr).def_nlengthl);
    thr_name [(*tpa_ptr).def_nlengthl] = '\0';

    /*
     *    The majority of the validation which this routine performs is not
     *    applicable when editing the $DEFAULT test - we allow duplicate/conflicting
     *    options in the $DEFAULT test.
     */
    if (strcmp (thr_name, "$DEFAULT") != 0)
    {                   /* If NOT the $DEFAULT test */

        /*
         *  The Username may only be specified if the Login Type is
         *  AUTO, AUTOIM, AUTODLY, TELNET or RLOGIN.
         */
        if ((*tpa_ptr).opt_flagn64 [OPT_THR_USRNAM] &&  /* If USERNAME and */
                (*tpa_ptr).opt_flagn64 [OPT_THR_LOGIN]) /* LOGIN TYPE specified */
        {
            if (((*tdf_ptr).tdf_login_typel != TOF_THRLOG_AUTO) &&
                    ((*tdf_ptr).tdf_login_typel != TOF_THRLOG_AUTOIM) &&
                    ((*tdf_ptr).tdf_login_typel != TOF_THRLOG_AUTODELAY) &&
                    ((*tdf_ptr).tdf_login_typel != TOF_THRLOG_TELNET) &&
                    ((*tdf_ptr).tdf_login_typel != TOF_THRLOG_RLOGIN))
            {
                PSL_PUTDIAG ((*com_ptr).diag_ptrps, 1, SCL_THROPTCONF,
                             "Username", "Login Type", thr_name,
                             (*tpa_ptr).opt_linel64 [OPT_THR_USRNAM],
                             (*tpa_ptr).opt_linel64 [OPT_THR_LOGIN],
                             (*com_ptr).scl_namec256);
                PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
                (*tpa_ptr).error_countl += 1;
            }
        }

        /*
         *  The Character Size may NOT be specified if the Device Type is VT52 or VT100.
         */
        if ((*tpa_ptr).opt_flagn64 [OPT_THR_CHRSIZ] &&  /* If CHAR SIZE and */
                (*tpa_ptr).opt_flagn64 [OPT_THR_DEVTYP])    /* DEVICE TYPE specified */
        {
            if (((*tdf_ptr).tdf_device_typel == DEVICE_TYPE_VT52)
                    || ((*tdf_ptr).tdf_device_typel == DEVICE_TYPE_VT100))
            {
                PSL_PUTDIAG ((*com_ptr).diag_ptrps, 1, SCL_THROPTCONF,
                             "Character Size", "Device Type", thr_name,
                             (*tpa_ptr).opt_linel64 [OPT_THR_CHRSIZ],
                             (*tpa_ptr).opt_linel64 [OPT_THR_DEVTYP],
                             (*com_ptr).scl_namec256);
                PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
                (*tpa_ptr).error_countl += 1;
            }
        }

        /*
         *  The Interval may only be specified if the Number is > 1.
         */
        if ((*tpa_ptr).opt_flagn64 [OPT_THR_INT] && /* If INTERVAL and */
                (*tpa_ptr).opt_flagn64 [OPT_THR_NUMBER])    /* NUMBER specified */
        {
            if ((*tdf_ptr).tdf_numberl <= 1)
            {               /* If NUMBER <= 1 */
                PSL_PUTDIAG ((*com_ptr).diag_ptrps, 1, SCL_THROPTCONF,
                             "Interval", "Number", thr_name,
                             (*tpa_ptr).opt_linel64 [OPT_THR_INT],
                             (*tpa_ptr).opt_linel64 [OPT_THR_NUMBER],
                             (*com_ptr).scl_namec256);
                PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
                (*tpa_ptr).error_countl += 1;
            }
        }

        /*
         *  Recording & Trace Filenames must contain "^T" if Number > 1
         */
        if ((*tpa_ptr).opt_flagn64 [OPT_THR_RECFIL] &&  /* If RECORD FILE and */
                (*tpa_ptr).opt_flagn64 [OPT_THR_NUMBER])    /* NUMBER specified */
        {
            if ((*tdf_ptr).tdf_numberl > 1)
            {

                /*
                 *    Ensure that the RECORD FILENAME contains the "^T" value.
                 */
                if ((strstr ((*tdf_ptr).tdf_recordfilec52, "^T") == NULL) &&
                        (strstr ((*tdf_ptr).tdf_recordfilec52, "^t") == NULL))
                {
                    PSL_PUTDIAG ((*com_ptr).diag_ptrps, 2, SCL_THROPTCONF,
                                 "Recording Filename", "Number", thr_name,
                                 (*tpa_ptr).opt_linel64 [OPT_THR_RECFIL],
                                 (*tpa_ptr).opt_linel64 [OPT_THR_NUMBER],
                                 (*com_ptr).scl_namec256, SCL_NOTUNIQUE),
                    PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
                    (*tpa_ptr).error_countl += 1;
                }
            }
        }

        if ((*tpa_ptr).opt_flagn64 [OPT_THR_TRCFIL] &&  /* If TRACE FILE and */
                (*tpa_ptr).opt_flagn64 [OPT_THR_NUMBER])    /* NUMBER specified */
        {
            if ((*tdf_ptr).tdf_numberl > 1)
            {

                /*
                 *      Ensure that the TRACE FILENAME contains the "^T" value.
                 */
                if ((strstr ((*tdf_ptr).tdf_tracefilec52, "^T") == NULL) &&
                        (strstr ((*tdf_ptr).tdf_tracefilec52, "^t") == NULL))
                {
                    PSL_PUTDIAG ((*com_ptr).diag_ptrps, 2, SCL_THROPTCONF,
                                 "Trace Filename", "Number", thr_name,
                                 (*tpa_ptr).opt_linel64 [OPT_THR_TRCFIL],
                                 (*tpa_ptr).opt_linel64 [OPT_THR_NUMBER],
                                 (*com_ptr).scl_namec256, SCL_NOTUNIQUE),
                    PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
                    (*tpa_ptr).error_countl += 1;
                }
            }
        }

        /*
         *     If SOCKET CONNECTION has been specified then we must also 
         *     have specified TERMINAL TYPE of Remote TCP/IP
         */
        if ((*tpa_ptr).opt_flagn64 [OPT_THR_SOCKET])    /* If SOCKET CONNECTION */
        {
            if ((((*tdf_ptr).tdf_default_flags1l & TDF_THRDF1_TERMINAL_TYPE) != 0)
                    /* If terminal type defaulted   */
                    ||  ((*tdf_ptr).tdf_terminal_typel != TOF_THRTTP_REMTCPIP))
                /* ...or is NOT Remote TCP/IP   */
            {
                PSL_PUTDIAG ((*com_ptr).diag_ptrps, 1, SCL_THROPTCONF
                             ,"Socket Connection"
                             ,"Non-Remote TCP/IP Thread"
                             ,thr_name
                             ,(*tpa_ptr).opt_linel64 [OPT_THR_SOCKET]
                             ,(*tpa_ptr).opt_linel64 [OPT_THR_TRMTYP]
                             ,(*com_ptr).scl_namec256);
                PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
                (*tpa_ptr).error_countl += 1;
            }
        }

        /*
         *     PORT NUMBER may only be specified if we have specified SOCKET CONNECTION
         *     as ENABLED or we have specified EMULATOR as the terminal type
         */
        if ((*tpa_ptr).opt_flagn64 [OPT_THR_PORT])  /* If PORT NUMBER specified */
        {
            if (((!((*tpa_ptr).opt_flagn64 [OPT_THR_SOCKET]))
                    ||  ((*tdf_ptr).tdf_use_tcp_socketl == TOF_GENSTA_DISABLED))
                    &&  ((*tdf_ptr).tdf_terminal_typel != TOF_THRTTP_EMULATOR))
            {
                PSL_PUTDIAG ((*com_ptr).diag_ptrps, 1, SCL_THROPTCONF
                             ,"Port Number"
                             ,"Socket Connection Disabled"
                             ,thr_name
                             ,(*tpa_ptr).opt_linel64 [OPT_THR_PORT]
                             ,(*tpa_ptr).opt_linel64 [OPT_THR_SOCKET]
                             ,(*com_ptr).scl_namec256);
                PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
                (*tpa_ptr).error_countl += 1;
            }
        }

        /*
         *    RAW Login Type may only be specified if SOCKET CONNECTION
         */
        if ((*tdf_ptr).tdf_login_typel == LOGIN_TYPE_RAW)
        {
            if ((!((*tpa_ptr).opt_flagn64 [OPT_THR_SOCKET]))
                    ||  ((*tdf_ptr).tdf_use_tcp_socketl == TOF_GENSTA_DISABLED))
                /* ...or not using socket connection  */
            {
                PSL_PUTDIAG ((*com_ptr).diag_ptrps, 1, SCL_THROPTCONF
                             ,"Login Type"
                             ,"Socket Connection Disabled"
                             ,thr_name
                             ,(*tpa_ptr).opt_linel64 [OPT_THR_LOGIN]
                             ,(*tpa_ptr).opt_linel64 [OPT_THR_SOCKET]
                             ,(*com_ptr).scl_namec256);
                PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
                (*tpa_ptr).error_countl += 1;
            }
        }

        /*
         *  If the Terminal Type was specified perform further validation.
         */
        if ((*tpa_ptr).opt_flagn64 [OPT_THR_TRMTYP])
        {
            CHK_THR_DEVTYPS (argblk_ptr);
        }
    }

    return PSL_NORMAL;
}


/************************************************************************/

/*  CHK_THR_DEVTYPS - Perform device type checks            */
/*  ===============                         */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     CHK_THR_DEVTYPS (argblk_ptr);                */
/*                                  */
/*  Where:                              */
/*                                  */
/*     argblk_ptr = Pointer to TPARSE Argument Block Structure  */
/*                                  */
/*  This routine checks to see if fields have been specified that   */
/*  rely on a particular Terminal Type.                             */
/*                                  */
/*  It is assumed that the TYPE field has been specified also.  */
/*                                  */
/************************************************************************/
static void CHK_THR_DEVTYPS (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE Data */
    struct   scl_com    *com_ptr;       /* Pointer to Common Data */
    struct   tdf        *tdf_ptr;       /* Pointer to Terminal Definition */
    struct   thrtbl     *thr_ptr;       /* Pointer to thread definition */
    char       thr_name [20];  /* Local copy of thread name */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    thr_ptr = (*tpa_ptr).def_thr_wrkps;
    tdf_ptr = (*thr_ptr).thr_tdf_ptrps;

    /*
     *  Get local copy of thread name for ease of reference.
     */
    (void) memcpy (thr_name, (*tpa_ptr).def_namec64, (*tpa_ptr).def_nlengthl);
    thr_name [(*tpa_ptr).def_nlengthl] = '\0';

    /*
     *  Missing Internet Address?
     */
    if ((!((*tpa_ptr).opt_flagn64 [OPT_THR_INET])) &&
            (((*tdf_ptr).tdf_terminal_typel == TOF_THRTTP_REMTCPIP)
             || ((*tdf_ptr).tdf_terminal_typel == TOF_THRTTP_EMULATOR)))
    {
        PSL_PUTDIAG ((*com_ptr).diag_ptrps, 1, SCL_MISINTERNET,
                     thr_name,
                     (*com_ptr).scl_namec256);
        PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
        (*tpa_ptr).error_countl += 1;
    }

    /*
     *  Missing Service Name?
     */
    if ((!((*tpa_ptr).opt_flagn64 [OPT_THR_SRV])) &&
            ((*tdf_ptr).tdf_terminal_typel == TOF_THRTTP_REMLAT))
    {
        PSL_PUTDIAG ((*com_ptr).diag_ptrps, 1, SCL_MISSERVICE,
                     thr_name,
                     (*com_ptr).scl_namec256);
        PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
        (*tpa_ptr).error_countl += 1;
    }

    /*
     *  Check Answerback
     */
    if (((*tpa_ptr).opt_flagn64 [OPT_THR_ANSWER]) &&
            ((*tdf_ptr).tdf_terminal_typel != TOF_THRTTP_PSEUDO))
    {
        PSL_PUTDIAG ((*com_ptr).diag_ptrps, 1, SCL_THROPTCONF,
                     "Answerback", "Terminal Type",
                     thr_name,
                     (*tpa_ptr).opt_linel64 [OPT_THR_ANSWER],
                     (*tpa_ptr).opt_linel64 [OPT_THR_TRMTYP],
                     (*com_ptr).scl_namec256);
        PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
        (*tpa_ptr).error_countl += 1;
    }

    /*
     *  Check Auto Synchronization
     */
    if (((*tpa_ptr).opt_flagn64 [OPT_THR_AUTSYN]) &&
            ((*tdf_ptr).tdf_terminal_typel == TOF_THRTTP_COMMAND))
    {
        PSL_PUTDIAG ((*com_ptr).diag_ptrps, 1, SCL_THROPTCONF,
                     "Auto Synchronization", "Terminal Type",
                     thr_name,
                     (*tpa_ptr).opt_linel64 [OPT_THR_AUTSYN],
                     (*tpa_ptr).opt_linel64 [OPT_THR_TRMTYP],
                     (*com_ptr).scl_namec256);
        PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
        (*tpa_ptr).error_countl += 1;
    }

    /*
     *  Check Baud Rate
     */
    if (((*tpa_ptr).opt_flagn64 [OPT_THR_BAUD]) &&
            (((*tdf_ptr).tdf_terminal_typel != TOF_THRTTP_PSEUDO) &&
             ((*tdf_ptr).tdf_terminal_typel != TOF_THRTTP_REMLAT) &&
             ((*tdf_ptr).tdf_terminal_typel != TOF_THRTTP_REMTCPIP)))
    {
        PSL_PUTDIAG ((*com_ptr).diag_ptrps, 1, SCL_THROPTCONF,
                     "Baud Rate", "Terminal Type",
                     thr_name,
                     (*tpa_ptr).opt_linel64 [OPT_THR_BAUD],
                     (*tpa_ptr).opt_linel64 [OPT_THR_TRMTYP],
                     (*com_ptr).scl_namec256);
        PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
        (*tpa_ptr).error_countl += 1;
    }

    /*
     *  Check Character Size
     */
    if (((*tpa_ptr).opt_flagn64 [OPT_THR_CHRSIZ]) &&
            (((*tdf_ptr).tdf_terminal_typel != TOF_THRTTP_PSEUDO) &&
             ((*tdf_ptr).tdf_terminal_typel != TOF_THRTTP_REMLAT) &&
             ((*tdf_ptr).tdf_terminal_typel != TOF_THRTTP_REMTCPIP)))
    {
        PSL_PUTDIAG ((*com_ptr).diag_ptrps, 1, SCL_THROPTCONF,
                     "Character Size", "Terminal Type",
                     thr_name,
                     (*tpa_ptr).opt_linel64 [OPT_THR_CHRSIZ],
                     (*tpa_ptr).opt_linel64 [OPT_THR_TRMTYP],
                     (*com_ptr).scl_namec256);
        PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
        (*tpa_ptr).error_countl += 1;
    }

    /*
     *  Check Device Type
     */
    if (((*tpa_ptr).opt_flagn64 [OPT_THR_DEVTYP]) &&
            (((*tdf_ptr).tdf_terminal_typel != TOF_THRTTP_PSEUDO) &&
             ((*tdf_ptr).tdf_terminal_typel != TOF_THRTTP_REMLAT) &&
             ((*tdf_ptr).tdf_terminal_typel != TOF_THRTTP_REMTCPIP)))
    {
        PSL_PUTDIAG ((*com_ptr).diag_ptrps, 1, SCL_THROPTCONF,
                     "Device Type", "Terminal Type",
                     thr_name,
                     (*tpa_ptr).opt_linel64 [OPT_THR_DEVTYP],
                     (*tpa_ptr).opt_linel64 [OPT_THR_TRMTYP],
                     (*com_ptr).scl_namec256);
        PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
        (*tpa_ptr).error_countl += 1;
    }

    /*
     *  Check Function-Key Control
     */
    if (((*tpa_ptr).opt_flagn64 [OPT_THR_FKEY]) &&
            ((*tdf_ptr).tdf_terminal_typel == TOF_THRTTP_COMMAND))
    {
        PSL_PUTDIAG ((*com_ptr).diag_ptrps, 1, SCL_THROPTCONF,
                     "Function-Key Control", "Terminal Type",
                     thr_name,
                     (*tpa_ptr).opt_linel64 [OPT_THR_FKEY],
                     (*tpa_ptr).opt_linel64 [OPT_THR_TRMTYP],
                     (*com_ptr).scl_namec256);
        PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
        (*tpa_ptr).error_countl += 1;
    }

    /*
     *  Check Hangup on Detach
     */
    if (((*tpa_ptr).opt_flagn64 [OPT_THR_HNGDET]) &&
            (((*tdf_ptr).tdf_terminal_typel == TOF_THRTTP_PSEUDO) ||
             ((*tdf_ptr).tdf_terminal_typel == TOF_THRTTP_REMLAT) ||
             ((*tdf_ptr).tdf_terminal_typel == TOF_THRTTP_REMTCPIP) ||
             ((*tdf_ptr).tdf_terminal_typel == TOF_THRTTP_COMMAND)))
    {
        PSL_PUTDIAG ((*com_ptr).diag_ptrps, 1, SCL_THROPTCONF,
                     "Hangup on Detach", "Terminal Type",
                     thr_name,
                     (*tpa_ptr).opt_linel64 [OPT_THR_HNGDET],
                     (*tpa_ptr).opt_linel64 [OPT_THR_TRMTYP],
                     (*com_ptr).scl_namec256);
        PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
        (*tpa_ptr).error_countl += 1;
    }

    /*
     *  Check TCP/IP Address
     */
    if (((*tpa_ptr).opt_flagn64 [OPT_THR_INET]) &&
            (((*tdf_ptr).tdf_terminal_typel != TOF_THRTTP_REMTCPIP)
             && ((*tdf_ptr).tdf_terminal_typel != TOF_THRTTP_EMULATOR)))
    {
        PSL_PUTDIAG ((*com_ptr).diag_ptrps, 1, SCL_THROPTCONF,
                     "TCP/IP Address", "Terminal Type",
                     thr_name,
                     (*tpa_ptr).opt_linel64 [OPT_THR_INET],
                     (*tpa_ptr).opt_linel64 [OPT_THR_TRMTYP],
                     (*com_ptr).scl_namec256);
        PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
        (*tpa_ptr).error_countl += 1;
    }

    /*
     *  Check Input Timing
     */
    if (((*tpa_ptr).opt_flagn64 [OPT_THR_INPTIM]) &&
            ((*tdf_ptr).tdf_terminal_typel == TOF_THRTTP_COMMAND))
    {
        PSL_PUTDIAG ((*com_ptr).diag_ptrps, 1, SCL_THROPTCONF,
                     "Input Timing", "Terminal Type",
                     thr_name,
                     (*tpa_ptr).opt_linel64 [OPT_THR_INPTIM],
                     (*tpa_ptr).opt_linel64 [OPT_THR_TRMTYP],
                     (*com_ptr).scl_namec256);
        PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
        (*tpa_ptr).error_countl += 1;
    }

    /*
     *  Check Enable IOM
     */
    if (((*tpa_ptr).opt_flagn64 [OPT_THR_IOM]) &&
            ((*tdf_ptr).tdf_terminal_typel == TOF_THRTTP_COMMAND))
    {
        PSL_PUTDIAG ((*com_ptr).diag_ptrps, 1, SCL_THROPTCONF,
                     "Enable IOM", "Terminal Type",
                     thr_name,
                     (*tpa_ptr).opt_linel64 [OPT_THR_IOM],
                     (*tpa_ptr).opt_linel64 [OPT_THR_TRMTYP],
                     (*com_ptr).scl_namec256);
        PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
        (*tpa_ptr).error_countl += 1;
    }

    /*
     *  Check Login Type
     */
    if (((*tpa_ptr).opt_flagn64 [OPT_THR_LOGIN]) &&
            (((*tdf_ptr).tdf_terminal_typel == TOF_THRTTP_COMMAND) ||
             ((((*tdf_ptr).tdf_login_typel == LOGIN_TYPE_AUTO)
               || ((*tdf_ptr).tdf_login_typel == LOGIN_TYPE_AUTOIM)
               || ((*tdf_ptr).tdf_login_typel == LOGIN_TYPE_AUTODELAY)) &&
              (((*tdf_ptr).tdf_terminal_typel == TOF_THRTTP_REMLAT)
               || ((*tdf_ptr).tdf_terminal_typel == TOF_THRTTP_EMULATOR)
               || ((*tdf_ptr).tdf_terminal_typel == TOF_THRTTP_REMTCPIP)))))
    {
        PSL_PUTDIAG ((*com_ptr).diag_ptrps, 1, SCL_THROPTCONF,
                     "Login Type", "Terminal Type",
                     thr_name,
                     (*tpa_ptr).opt_linel64 [OPT_THR_LOGIN],
                     (*tpa_ptr).opt_linel64 [OPT_THR_TRMTYP],
                     (*com_ptr).scl_namec256);
        PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
        (*tpa_ptr).error_countl += 1;
    }

    /*
     *  Check Node Name
     */
    if (((*tpa_ptr).opt_flagn64 [OPT_THR_NODE]) &&
            (((*tdf_ptr).tdf_terminal_typel != TOF_THRTTP_DECTERM)
             && ((*tdf_ptr).tdf_terminal_typel != TOF_THRTTP_REMLAT)))
    {
        PSL_PUTDIAG ((*com_ptr).diag_ptrps, 1, SCL_THROPTCONF,
                     "Node Name", "Terminal Type",
                     thr_name,
                     (*tpa_ptr).opt_linel64 [OPT_THR_NODE],
                     (*tpa_ptr).opt_linel64 [OPT_THR_TRMTYP],
                     (*com_ptr).scl_namec256);
        PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
        (*tpa_ptr).error_countl += 1;
    }

    /*
     *  Check Number
     */
    if (((*tpa_ptr).opt_flagn64 [OPT_THR_NUMBER]) &&
            ((*tdf_ptr).tdf_terminal_typel == TOF_THRTTP_PHYS))
    {
        PSL_PUTDIAG ((*com_ptr).diag_ptrps, 1, SCL_THROPTCONF,
                     "Number", "Terminal Type",
                     thr_name,
                     (*tpa_ptr).opt_linel64 [OPT_THR_NUMBER],
                     (*tpa_ptr).opt_linel64 [OPT_THR_TRMTYP],
                     (*com_ptr).scl_namec256);
        PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
        (*tpa_ptr).error_countl += 1;
    }

    /*
     *  Check Page
     */
    if (((*tpa_ptr).opt_flagn64 [OPT_THR_PAGE]) &&
            (((*tdf_ptr).tdf_terminal_typel == TOF_THRTTP_PHYS)
             || ((*tdf_ptr).tdf_terminal_typel == TOF_THRTTP_COMMAND)))
    {
        PSL_PUTDIAG ((*com_ptr).diag_ptrps, 1, SCL_THROPTCONF,
                     "Page", "Terminal Type",
                     thr_name,
                     (*tpa_ptr).opt_linel64 [OPT_THR_PAGE],
                     (*tpa_ptr).opt_linel64 [OPT_THR_TRMTYP],
                     (*com_ptr).scl_namec256);
        PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
        (*tpa_ptr).error_countl += 1;
    }

    /*
     *  Check Physical Device
     */
    if (((*tpa_ptr).opt_flagn64 [OPT_THR_PHYDEV]) &&
            ((*tdf_ptr).tdf_terminal_typel != TOF_THRTTP_PHYS))
    {
        PSL_PUTDIAG ((*com_ptr).diag_ptrps, 1, SCL_THROPTCONF,
                     "Physical Device", "Terminal Type",
                     thr_name,
                     (*tpa_ptr).opt_linel64 [OPT_THR_PHYDEV],
                     (*tpa_ptr).opt_linel64 [OPT_THR_TRMTYP],
                     (*com_ptr).scl_namec256);
        PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
        (*tpa_ptr).error_countl += 1;
    }

    /*
     *  Check Record
     */
    if (((*tpa_ptr).opt_flagn64 [OPT_THR_RECORD]) &&
            ((*tdf_ptr).tdf_terminal_typel == TOF_THRTTP_COMMAND))
    {
        PSL_PUTDIAG ((*com_ptr).diag_ptrps, 1, SCL_THROPTCONF,
                     "Record", "Terminal Type",
                     thr_name,
                     (*tpa_ptr).opt_linel64 [OPT_THR_RECORD],
                     (*tpa_ptr).opt_linel64 [OPT_THR_TRMTYP],
                     (*com_ptr).scl_namec256);
        PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
        (*tpa_ptr).error_countl += 1;
    }

    /*
     *  Check Service Name
     */
    if (((*tpa_ptr).opt_flagn64 [OPT_THR_SRV]) &&
            ((*tdf_ptr).tdf_terminal_typel != TOF_THRTTP_REMLAT))
    {
        PSL_PUTDIAG ((*com_ptr).diag_ptrps, 1, SCL_THROPTCONF,
                     "Service Name", "Terminal Type",
                     thr_name,
                     (*tpa_ptr).opt_linel64 [OPT_THR_SRV],
                     (*tpa_ptr).opt_linel64 [OPT_THR_TRMTYP],
                     (*com_ptr).scl_namec256);
        PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
        (*tpa_ptr).error_countl += 1;
    }

    /*
     *  Check Stop Record
     */
    if (((*tpa_ptr).opt_flagn64 [OPT_THR_RECSTP]) &&
            (((*tdf_ptr).tdf_terminal_typel == TOF_THRTTP_COMMAND)
             || ((*tdf_ptr).tdf_terminal_typel == TOF_THRTTP_PSEUDO)
             || ((*tdf_ptr).tdf_terminal_typel == TOF_THRTTP_REMLAT)
             || ((*tdf_ptr).tdf_terminal_typel == TOF_THRTTP_REMTCPIP)))
    {
        PSL_PUTDIAG ((*com_ptr).diag_ptrps, 1, SCL_THROPTCONF,
                     "Record Stop", "Terminal Type",
                     thr_name,
                     (*tpa_ptr).opt_linel64 [OPT_THR_RECSTP],
                     (*tpa_ptr).opt_linel64 [OPT_THR_TRMTYP],
                     (*com_ptr).scl_namec256);
        PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
        (*tpa_ptr).error_countl += 1;
    }

    /*
     *  Check Terminal ID
     */
    if (((*tpa_ptr).opt_flagn64 [OPT_THR_TRMID]) &&
            (((*tdf_ptr).tdf_terminal_typel != TOF_THRTTP_PSEUDO) &&
             ((*tdf_ptr).tdf_terminal_typel != TOF_THRTTP_REMLAT) &&
             ((*tdf_ptr).tdf_terminal_typel != TOF_THRTTP_REMTCPIP)))
    {
        PSL_PUTDIAG ((*com_ptr).diag_ptrps, 1, SCL_THROPTCONF,
                     "Terminal ID", "Terminal Type",
                     thr_name,
                     (*tpa_ptr).opt_linel64 [OPT_THR_TRMID],
                     (*tpa_ptr).opt_linel64 [OPT_THR_TRMTYP],
                     (*com_ptr).scl_namec256);
        PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
        (*tpa_ptr).error_countl += 1;
    }

    /*
     *  Check Typeahead
     */
    if (((*tpa_ptr).opt_flagn64 [OPT_THR_TYPAHD]) &&
            ((*tdf_ptr).tdf_terminal_typel == TOF_THRTTP_COMMAND))
    {
        PSL_PUTDIAG ((*com_ptr).diag_ptrps, 1, SCL_THROPTCONF,
                     "Typeahead", "Terminal Type",
                     thr_name,
                     (*tpa_ptr).opt_linel64 [OPT_THR_TYPAHD],
                     (*tpa_ptr).opt_linel64 [OPT_THR_TRMTYP],
                     (*com_ptr).scl_namec256);
        PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
        (*tpa_ptr).error_countl += 1;
    }

    /*
     *  Check Width
     */
    if (((*tpa_ptr).opt_flagn64 [OPT_THR_WIDTH]) &&
            (((*tdf_ptr).tdf_terminal_typel == TOF_THRTTP_PHYS)
             || ((*tdf_ptr).tdf_terminal_typel == TOF_THRTTP_COMMAND)))
    {
        PSL_PUTDIAG ((*com_ptr).diag_ptrps, 1, SCL_THROPTCONF,
                     "Width", "Terminal Type",
                     thr_name,
                     (*tpa_ptr).opt_linel64 [OPT_THR_WIDTH],
                     (*tpa_ptr).opt_linel64 [OPT_THR_TRMTYP],
                     (*com_ptr).scl_namec256);
        PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
        (*tpa_ptr).error_countl += 1;
    }

    /*
     *  Check Emulator type/Model
     */
    if ((*tdf_ptr).tdf_terminal_typel == TOF_THRTTP_EMULATOR)
    {
        if (!((*tpa_ptr).opt_flagn64 [OPT_THR_EMULTYP]))
        {
            PSL_PUTDIAG ((*com_ptr).diag_ptrps, 1, SCL_THROPTCONF,
                         "Terminal Type Emulator", "No Emulator Type",
                         thr_name,
                         (*tpa_ptr).opt_linel64 [OPT_THR_EMULTYP],
                         (*tpa_ptr).opt_linel64 [OPT_THR_TRMTYP],
                         (*com_ptr).scl_namec256);
            PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
            (*tpa_ptr).error_countl += 1;
        }
        else
        {
            if ((strcmp ((*tdf_ptr).tdf_emulator_typec16 ,"IBM3270") == 0)
                    ||  (strcmp ((*tdf_ptr).tdf_emulator_typec16 ,"IBM5250") == 0))
            {
                if (!((*tpa_ptr).opt_flagn64 [OPT_THR_EMULMOD]))
                {
                    PSL_PUTDIAG ((*com_ptr).diag_ptrps, 1, SCL_THROPTCONF,
                                 "Emulator Type", "No Emulator Model",
                                 thr_name,
                                 (*tpa_ptr).opt_linel64 [OPT_THR_EMULTYP],
                                 (*tpa_ptr).opt_linel64 [OPT_THR_EMULMOD],
                                 (*com_ptr).scl_namec256);
                    PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
                    (*tpa_ptr).error_countl += 1;
                }
            }

            else if ((*tpa_ptr).opt_flagn64 [OPT_THR_EMULMOD])
            {
                PSL_PUTDIAG ((*com_ptr).diag_ptrps, 1, SCL_THROPTCONF,
                             "Emulator Type", "Emulator Model",
                             thr_name,
                             (*tpa_ptr).opt_linel64 [OPT_THR_EMULTYP],
                             (*tpa_ptr).opt_linel64 [OPT_THR_EMULMOD],
                             (*com_ptr).scl_namec256);
                PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
                (*tpa_ptr).error_countl += 1;
            }
        }
    }

    /*
     *    Now check for emulator type/model being specified when the terminal type
     *    is not emulator.
     */
    if (((*tpa_ptr).opt_flagn64 [OPT_THR_EMULTYP]) &&
            ((*tdf_ptr).tdf_terminal_typel != TOF_THRTTP_EMULATOR))
    {
        PSL_PUTDIAG ((*com_ptr).diag_ptrps, 1, SCL_THROPTCONF,
                     "Emulator Type", "Terminal Type",
                     thr_name,
                     (*tpa_ptr).opt_linel64 [OPT_THR_EMULTYP],
                     (*tpa_ptr).opt_linel64 [OPT_THR_TRMTYP],
                     (*com_ptr).scl_namec256);
        PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
        (*tpa_ptr).error_countl += 1;
    }

    if (((*tpa_ptr).opt_flagn64 [OPT_THR_EMULMOD]) &&
            ((*tdf_ptr).tdf_terminal_typel != TOF_THRTTP_EMULATOR))
    {
        PSL_PUTDIAG ((*com_ptr).diag_ptrps, 1, SCL_THROPTCONF,
                     "Emulator Model", "Terminal Type",
                     thr_name,
                     (*tpa_ptr).opt_linel64 [OPT_THR_EMULMOD],
                     (*tpa_ptr).opt_linel64 [OPT_THR_TRMTYP],
                     (*com_ptr).scl_namec256);
        PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
        (*tpa_ptr).error_countl += 1;
    }

    return;
}


/************************************************************************/

/*    CHK_THR_NOTDDEFAULT - Check thread NOT $DEFAULT thread        */
/*    ===================                       */
/*                                          */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL  If current thread NOT $DEFAULT thread.  */
/*      Zero        Otherwise.              */
/*    Pointers: Unchanged.                      */
/*    Inputs:   None.                           */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine checks whether the current thread is the $DEFAULT    */
/*    thread.                               */
/*                                  */
/************************************************************************/
unsigned long CHK_THR_NOTDDEFAULT (struct tpaarg *argblk_ptr)
{
    /*
     *  Local data.
     */
    struct   scl_com    *com_ptr;       /* Common data */
    struct   scl_tpa    *tpa_ptr;       /* TPARSE data */
    struct   thrtbl    *thr_ptr;        /* Pointer to thread definition */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    thr_ptr = (*tpa_ptr).def_thr_wrkps;

    /*
     *  Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking thread NOT $DEFAULT thread\n");
    }
#endif

    /*
     *  If thread name is "$DEFAULT" return failure status.
     */
    if (strcmp ((*thr_ptr).thr_namec20, "$DEFAULT") == 0)
    {
        return 0;
    }
    else
    {
        return PSL_NORMAL;
    }
}


/************************************************************************/

/*    CHK_THR_OPTS - Check thread definition options            */
/*    ============                          */
/*                                          */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL  If valid thread definition option   */
/*      Zero        Otherwise               */
/*    Pointers: Unchanged.                      */
/*    Outputs:  scl_tpa.thr_opt_flagn64 = New option flag set if not    */
/*          set already.                    */
/*                                  */
/*    Inputs:   tpaarg.paraml - Thread option specified.        */
/*    Inputs:   argblk_ptr - Pointer to TPARSE argument block structure */
/*      scl_tpa.thr_opt_flagn64 - Boolean array indicating  */
/*          thread options currently set.           */
/*    Outputs:  scl_tpa.thr_opt_flagn64 - Updated to show option    */
/*                          requested.  */
/*                                  */
/*    This routine is called on parsing a thread definition option. It  */
/*    examines the scl_tpa.thr_opt_flagsn64 array to see if the     */
/*    thread option specified by tpaarg.paraml has been set     */
/*    already. If it has not been set already the routine returns a */
/*    success status, otherwise an error status is returned.        */
/*                                  */
/*    The purpose of this routine is to maintain a boolean array    */
/*    indicating the options which have been specified for a thread.    */
/*    This allows conflicting and duplicate options to be detected. It  */
/*    also means that when the definition is encoded we know what   */
/*    options have been specified.                  */
/*                                  */
/************************************************************************/
unsigned long CHK_THR_OPTS (struct tpaarg *argblk_ptr)
{
    /*
     *  Local data.
     */
    struct scl_com  *com_ptr;       /* Common data */
    struct scl_tpa  *tpa_ptr;       /* TPARSE data */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *  Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking THREAD DEFINITION options bitmask\n");
    }
#endif

    /*
     *  If the option is NOT already set then set it, record the corresponding
     *  source line number and return success, otherwise return failure.
     */
    if (!(*tpa_ptr).opt_flagn64 [(*argblk_ptr).paraml])
        /* If option NOT already specified */
    {
        (*tpa_ptr).opt_flagn64 [(*argblk_ptr).paraml] = PTRUE;
        (*tpa_ptr).opt_linel64 [(*argblk_ptr).paraml] = (*tpa_ptr).line_ref_numberl;
        return PSL_NORMAL;      /* Return success */
    }
    else                /* Else option already specified */
    {
        return 0;           /* Return failure */
    }
}


/************************************************************************/

/*    CHK_THR_TARGET - Check thread 'target' option         */
/*    ==============                                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL  If valid target             */
/*          Zero        Otherwise                   */
/*    Pointers: Unchanged.                      */
/*    Inputs:   None.                           */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine checks whether the current token is a valid      */
/*    target. If so it sets the thread 'target type' field in the   */
/*    thread definition work structure.                 */
/*                                  */
/************************************************************************/
unsigned long CHK_THR_TARGET (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   scl_com    *com_ptr;       /* Common data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE data */
    struct   thrtbl     *thr_ptr;       /* Pointer to thread definition */
    struct   tdf        *tdf_ptr;       /* Pointer to terminal definition */
    unsigned long       targetid;       /* Target id */
    char       targetnam[TOFBUFLEN];
    /* Target name */
    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    thr_ptr = (*tpa_ptr).def_thr_wrkps;
    tdf_ptr = (*thr_ptr).thr_tdf_ptrps;

    /*
     *  Check that the current token corresponds to a valid target name.
     *  If it does set field in thread definition work record.
     */
    (void) memcpy (targetnam, (*argblk_ptr).token_ptrpc, (*argblk_ptr).tokencntl);
    targetnam[(*argblk_ptr).tokencntl] = '\0';

    targetid = Tgt_GetTargetID (targetnam);

    if (targetid != TARGET_UNKNOWN)
    {
        (*tdf_ptr).tdf_targetl = targetid;
        return PSL_NORMAL;
    }
    else
    {
        return 0;
    }
}


/************************************************************************/
/*                                  */
/*    CHK_TIMER - Check for Timer                   */
/*    ==========                            */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_TIMER (argblk_ptr)                   */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL  If defined timer            */
/*    Zero  Otherwise                       */
/*    Pointers: Moved to end of timer name              */
/*    Inputs:   OBJ_TIMERS = Timers Symbol Table            */
/*      TIM_CNT = Timer count                   */
/*    Outputs:  TIM_PTR = Timer's Symbol Table entry address        */
/*                                  */
/*    This routine scans the timers symbol table to determine if the    */
/*    current token (which has already been proved to be a dataname)    */
/*    is in fact a timer.                       */
/*                                  */
/*    If the timer has been declared the address of its entry in    */
/*    the Timers Symbol Table is returned in TIM_PTR, otherwise     */
/*    TIM_PTR is set to zero and a failure status is returned.      */
/*                                  */
/************************************************************************/

unsigned long   CHK_TIMER (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */
    struct   timtbl  *timtbl_ptr;       /* Pointer timer symbol table   */
    unsigned long    isub;          /* Loop control variable    */
    unsigned long    namlen;        /* Timer name length        */
    long    comp_result;       /* Result of comparison     */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking for Timer Name Defined\n");
    }
#endif

    /*
     *    Scan the timers symbol table until we reach the end or find a match
     *    to this dataname.
     */
    namlen = (*argblk_ptr).tokencntl;

    for (isub=0, timtbl_ptr = (*tpa_ptr).timer_begps;
            isub < (*tpa_ptr).tim_cntl; isub++, timtbl_ptr++)
    {
        if (namlen == strlen ((*timtbl_ptr).tim_namec68))
            /* If the lengths are the same  */
        {
            comp_result = memcmp ((*argblk_ptr).token_ptrpc,
                                  (*timtbl_ptr).tim_namec68,
                                  namlen);
            /* Compare the 2 timers     */
            if (comp_result == 0) goto found_timer;
            /* ...and branch if names match */
        }
    }

    /*
     *    We did not find a match for this timer so return an error status
     *    after resetting the string pointer/string count/token count to before
     *    the dataname.
     */
    RESET_TOKEN (argblk_ptr);       /* ...reset the token pointer   */
    return 0;               /* ...and return error status   */

    /*
     *    We found a match for the timer. Setup the symbol table address in the
     *    TPARSE data and return a success status.
     */
found_timer:
    (*tpa_ptr).tim_ptrps = timtbl_ptr;  /* Save symbol table entry address */
    return PSL_NORMAL;          /* ...and return success    */
}

/************************************************************************/

/*    CHK_TSTTHR_OPTS - Check test thread definition options        */
/*    ===============                           */
/*                                          */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL  If valid test thread definition option  */
/*      Zero        Otherwise               */
/*    Pointers: Unchanged.                      */
/*    Inputs:   tpaarg.paraml - Test option specified.          */
/*      scl_tpa.thr_opt_flagn64 - Boolean array indicating  */
/*          thread options currently set.           */
/*    Outputs:  scl_tpa.thr_opt_flagn64 - Updated to show option    */
/*                          requested.  */
/*                                  */
/*    This routine is called on parsing a test thread definition    */
/*    option. It examines the scl_tpa.thr_opt_flagsn64 array to see */
/*    if the test thread option specified by tpaarg.paraml has been set */
/*    already. If it has not been set already the routine returns a */
/*    success status, otherwise an error status is returned.        */
/*                                  */
/*    The purpose of this routine is to maintain a boolean array    */
/*    indicating the options which have been specified for a test   */
/*    thread. This allows conflicting and duplicate options to be   */
/*    detected. It also means that when the definition is encoded   */
/*    we know what options have been specified.             */
/*                                  */
/************************************************************************/
unsigned long CHK_TSTTHR_OPTS (struct tpaarg *argblk_ptr)
{
    /*
     *  Local data.
     */
    struct scl_com  *com_ptr;       /* Common data */
    struct scl_tpa  *tpa_ptr;       /* TPARSE data */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *  Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking TEST THREAD DEFINITION options bitmask\n");
    }
#endif

    /*
     *  If the option is NOT already set then set it, record the corresponding
     *  source line number and return success, otherwise return failure.
     */
    if (!(*tpa_ptr).opt_flagn64 [(*argblk_ptr).paraml])
        /* If option NOT already specified */
    {
        (*tpa_ptr).opt_flagn64 [(*argblk_ptr).paraml] = PTRUE;
        (*tpa_ptr).opt_linel64 [(*argblk_ptr).paraml] = (*tpa_ptr).line_ref_numberl;
        return PSL_NORMAL;      /* Return success */
    }
    else                /* Else option already specified */
    {
        return 0;           /* Return failure */
    }
}


/************************************************************************/

/*    CHK_TSTTHR_TARGET - Check test thread 'target' option     */
/*    =================                             */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL  If valid target             */
/*          Zero        Otherwise                   */
/*    Pointers: Unchanged.                      */
/*    Inputs:   None.                           */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine checks whether the current token is a valid      */
/*    target. If so it sets the thread 'target type' field in the   */
/*    test thread definition work structure.                */
/*                                  */
/************************************************************************/
unsigned long CHK_TSTTHR_TARGET (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   scl_com    *com_ptr;       /* Common data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE data */
    struct   tstthrtbl  *tstthr_ptr;    /* Pointer to test thread definition */
    struct   tdf        *tdf_ptr;       /* Pointer to terminal definition */
    char       targetnam[TOFBUFLEN];
    /* Target name */
    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    tstthr_ptr = (*tpa_ptr).def_tstthr_wrkps;
    tdf_ptr = (*tstthr_ptr).tstthr_tdf_ptrps;

    /*
     *  Set the target name in the test thread definition work area and if the
     *  name is a valid target name return success, otherwise return failure.
     */
    (void) memcpy (targetnam, (*argblk_ptr).token_ptrpc, (*argblk_ptr).tokencntl);
    targetnam[(*argblk_ptr).tokencntl] = '\0';

    (*tdf_ptr).tdf_targetl = Tgt_GetTargetID (targetnam);

    if ((*tdf_ptr).tdf_targetl != TARGET_UNKNOWN)
    {
        return PSL_NORMAL;
    }
    else
    {
        return 0;
    }
}


/************************************************************************/

/*    CHK_TST_OPTS - Check test definition options          */
/*    ============                          */
/*                                          */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL  If valid test definition option     */
/*      Zero        Otherwise               */
/*    Pointers: Unchanged.                      */
/*    Inputs:   tpaarg.paraml - Test option specified.          */
/*      scl_tpa.thr_opt_flagn64 - Boolean array indicating  */
/*          thread options currently set.           */
/*    Outputs:  scl_tpa.thr_opt_flagn64 - Updated to show option    */
/*                          requested.  */
/*                                  */
/*    This routine is called on parsing a test definition option. It    */
/*    examines the scl_tpa.thr_opt_flagsn64 array to see if the     */
/*    test option specified by tpaarg.paraml has been set       */
/*    already. If it has not been set already the routine returns a */
/*    success status, otherwise an error status is returned.        */
/*                                  */
/*    The purpose of this routine is to maintain a boolean array    */
/*    indicating the options which have been specified for a test.  */
/*    This allows conflicting and duplicate options to be detected. It  */
/*    also means that when the definition is encoded we know what   */
/*    options have been specified.                  */
/*                                  */
/************************************************************************/
unsigned long CHK_TST_OPTS (struct tpaarg *argblk_ptr)
{
    /*
     *  Local data.
     */
    struct scl_com  *com_ptr;       /* Common data */
    struct scl_tpa  *tpa_ptr;       /* TPARSE data */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *  Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking TEST DEFINITION options bitmask\n");
    }
#endif

    /*
     *  If the option is NOT already set then set it, record the corresponding
     *  source line number and return success, otherwise return failure.
     */
    if (!(*tpa_ptr).opt_flagn64 [(*argblk_ptr).paraml])
        /* If option NOT already specified */
    {
        (*tpa_ptr).opt_flagn64 [(*argblk_ptr).paraml] = PTRUE;
        (*tpa_ptr).opt_linel64 [(*argblk_ptr).paraml] = (*tpa_ptr).line_ref_numberl;
        return PSL_NORMAL;      /* Return success */
    }
    else                /* Else option already specified */
    {
        return 0;           /* Return failure */
    }
}


/************************************************************************/

/*    CHK_TYPE_SCRIPT - Check script type script            */
/*    ===============                           */
/*                                          */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL  If /TYPE=SCRIPT (explicit or default)   */
/*      Zero        Otherwise               */
/*    Pointers: Unchanged.                      */
/*    Inputs:   SCP_COM record.                     */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine checks whether /TYPE=SCRIPT was specified (or    */
/*    defaulted).                           */
/*                                  */
/************************************************************************/
unsigned long CHK_TYPE_SCRIPT (struct tpaarg *argblk_ptr)
{
    /*
     *  Local data.
     */
    struct scl_com  *com_ptr;       /* Common Data */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;

    /*
     *  Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking if /TYPE=SCRIPT specified\n");
    }
#endif

    /*
     *  Check if /TYPE=SCRIPT specified.
     */
    if ((*com_ptr).type_scriptn)    /* If /TYPE=SCRIPT */
    {
        return PSL_NORMAL;      /* ...return success */
    }
    else                /* ...else */
    {
        return 0;           /* ...return failure */
    }
}


/************************************************************************/

/*    CHK_TYPE_TEST - Check script type test                */
/*    =============                         */
/*                                          */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL  If /TYPE=TEST               */
/*      Zero        Otherwise               */
/*    Pointers: Unchanged.                      */
/*    Inputs:   SCP_COM record.                     */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine checks whether /TYPE=TEST was specified (or      */
/*    defaulted).                           */
/*                                  */
/************************************************************************/
unsigned long   CHK_TYPE_TEST (struct tpaarg *argblk_ptr)
{
    /*
     *  Local data.
     */
    struct scl_com  *com_ptr;       /* Common Data */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;

    /*
     *  Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking if /TYPE=TEST specified\n");
    }
#endif

    /*
     *  Check if /TYPE=TEST specified.
     */
    if ((*com_ptr).type_testn)      /* If /TYPE=TEST */
    {
        return PSL_NORMAL;      /* ...return success */
    }
    else                /* ...else */
    {
        return 0;           /* ...return failure */
    }
}

/************************************************************************/

/*    CHK_TYPE_THREAD - Check script type thread            */
/*    ===============                           */
/*                                          */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL  If /TYPE=THREAD             */
/*      Zero        Otherwise               */
/*    Pointers: Unchanged.                      */
/*    Inputs:   SCP_COM record.                     */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine checks whether /TYPE=THREAD was specified (or    */
/*    defaulted).                           */
/*                                  */
/************************************************************************/
unsigned long   CHK_TYPE_THREAD (struct tpaarg *argblk_ptr)
{
    /*
     *  Local data.
     */
    struct scl_com  *com_ptr;       /* Common Data */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;

    /*
     *  Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking if /TYPE=THREAD specified\n");
    }
#endif

    /*
     *  Check if /TYPE=THREAD specified.
     */
    if ((*com_ptr).type_threadn)    /* If /TYPE=THREAD */
    {
        return PSL_NORMAL;      /* ...return success */
    }
    else                /* ...else */
    {
        return 0;           /* ...return failure */
    }
}


/************************************************************************/

/*    CHK_VAR_ARRAY - Check to see if the current variable is an array  */
/*    =============                         */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_VAR_ARRAY (argblk_ptr)               */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL if an array, 0 otherwise         */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr = Pointer to TPARSE Argument Block Structure */
/*      VAR_PTR = Variable's Symbol Table entry address     */
/*    Outputs:  None                            */
/*                                  */
/*    This routine returns success if the integer variable specified    */
/*    is an array variable and 0 otherwise.             */
/*                                  */
/************************************************************************/

unsigned long   CHK_VAR_ARRAY   (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */
    struct symtbl   *symtbl_ptr;        /* Pointer to entry symbol table */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    symtbl_ptr = (*tpa_ptr).var_ptrps;  /* Get pointer to variable  */

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking If array Variable Modifier \n");
    }
#endif

    /*
     *    Check to see if this variable is flagged as being an array or not by
     *    checking the symbol table entry.
     */
    if ((*symtbl_ptr).var_array_dimcntl == 0)
    {
        return 0;
    }
    else
    {
        return PSL_NORMAL;
    }
}
/*
 * 
 */

/************************************************************************/
/*                                  */
/*    CHK_VAR_CNT - Check Variable Modifier             */
/*    ============                          */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_VAR_CNT (argblk_ptr)                 */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Character modifier to be checked            */
/*    Outputs:  None                            */
/*                                  */
/*    This routine checks that the character modifier value is      */
/*    within the range 1 to the maximum char size reports a     */
/*    'Character Modifier Out Of Range' error if it is not.     */
/*                                  */
/************************************************************************/

unsigned long   CHK_VAR_CNT (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking Variable Modifier %ld\n"
                          ,(*tpa_ptr).int_valuel);
    }
#endif

    if (((*argblk_ptr).numberl == 0)    /* If character size out of range */
            ||  ((*argblk_ptr).numberl > MAX_VARIABLE_SIZE))

    {
        REPORT_ERROR (argblk_ptr, SCL_MODOUTRANGE);
        /* Pass to underline routine    */
    }

    else
    {
        (*tpa_ptr).def_sizel = (*argblk_ptr).numberl;
    }

    return PSL_NORMAL;
}

/************************************************************************/
/*                                  */
/*    CHK_VAR_DTYP - Check Variable Data Type               */
/*    ============                          */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_VAR_DTYP (argblk_ptr)                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL  If variable data type matches       */
/*          specified data type                 */
/*      Zero Otherwise                      */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr = Pointer to TPARSE Argument Block Structure */
/*      VAR_PTR = Variable's Symbol Table entry address     */
/*    Outputs:  None                            */
/*                                  */
/*    This routine returns success if the variable, whose Symbol Table  */
/*    entry address is held in VAR_PTR on entry, has the data type  */
/*    passed in PSLTPA$L_PARAM(AP), otherwise it returns failure.   */
/*                                  */
/*    If the entry has an invalid data type then we remove it from  */
/*    the list of parsed op[erands as it is either an error (in which   */
/*    case the command will not be encoded anyway) or the command has   */
/*    parsed integer and character operands separately as they go down  */
/*    different paths and has chosen the wrong type first.      */
/*                                  */
/************************************************************************/

unsigned long   CHK_VAR_DTYP (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */
    struct   symtbl  *symtbl_ptr;       /* Pointer to entry symbol table */
    unsigned long    iret;          /* Status return variable   */

    struct   oph    *oph_ptr;       /* Operand header pointer   */
    struct   opd    *opd_ptr;       /* Operand descriptor pointer   */


    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    symtbl_ptr = (*tpa_ptr).var_ptrps;  /* Get pointer to variable  */

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking Variable Datatype\n");
    }
#endif

    /*
     *    Check to see if the variable is in the symbol table and if so whether
     *    the data type is the same as that specified in the TPARSE argumnet block
     */
    if (((*tpa_ptr).var_ptrps == NULL)  /* If variable NOT in table */
            ||  ((*symtbl_ptr).var_typel == 0)  /* ...or data type not specified */
            ||  ((*symtbl_ptr).var_typel == (*argblk_ptr).paraml))
        /* ...or data types match   */
    {
        return PSL_NORMAL;      /* ...return success        */
    }

    else
    {

        /*
         *    This operand is of a different type to that required so remove it from
         *    the LL of parsed operands. Note we only remove the last operand parsed
         *    and it consists of both an OPH and OPD pointer.
         */
        iret = PSL_VM_REMOVE ((*tpa_ptr).parsed_operand_llps ,(char *) &opd_ptr
                              ,NULL ,NULL ,PSL_VM_POS_END);
        /* Remove the OPD (at end)  */
        if  (!(psl_test_status (iret))) /* ...report any errors     */
        {
            PSL_SIGDIAG ((*tpa_ptr).diag_ptrps, 2, PSL_SYSERVERR
                         ,"PSL_VM_REMOVE opd (in CHK_VAR_DTYP)", iret);
        }

        iret = PSL_VM_REMOVE ((*tpa_ptr).parsed_operand_llps ,(char *) &oph_ptr
                              ,NULL ,NULL ,PSL_VM_POS_END);
        /* ...and then the OPH      */
        if  (!(psl_test_status (iret))) /* ...report any errors     */
        {
            PSL_SIGDIAG ((*tpa_ptr).diag_ptrps, 2, PSL_SYSERVERR
                         ,"PSL_VM_REMOVE oph (in CHK_VAR_DTYP)", iret);
        }

        /* Free memory associated with the operand */
        FREE_OPH_OPD (oph_ptr, opd_ptr);

        return 0;           /* return failure       */
    }
}

/************************************************************************/

/*    CHK_CVAR_DTYP - Check Variable Data Type              */
/*    ============                          */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_CVAR_DTYP (argblk_ptr)               */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL  If variable data type matches       */
/*          specified data type                 */
/*      Zero Otherwise                      */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr = Pointer to TPARSE Argument Block Structure */
/*      VAR_PTR = Variable's Symbol Table entry address     */
/*    Outputs:  None                            */
/*                                  */
/*    This routine has been added for checking if the data type of a    */
/*    variable is of character type - either :              */
/*                                  */
/*      DTYP_RECORD or DTYP_CHARACTER               */
/*                                  */
/*    As both can not be tested at the same time, this script checks    */
/*    for a data type without removing it from LList of operands if     */
/*    it is of an incorrect type. The CHK_VAR_DTYP is the routine that  */
/*    should follow this one in any TPARSE checking data type routines. */
/*                                  */
/*    This routine returns success if the variable, whose Symbol Table  */
/*    entry address is held in VAR_PTR on entry, has the data type  */
/*    passed in PSLTPA$L_PARAM(AP), otherwise it returns failure.   */
/*                                  */
/*    If the entry has an invalid data type then test fails.        */
/*    This is only called once in !CHAR_VARIABLE when checking for      */
/*    character or record data types.                   */
/*    Record data types are treated exactly the same as characters      */
/*                                  */
/************************************************************************/

unsigned long   CHK_CVAR_DTYP (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */
    struct   symtbl  *symtbl_ptr;       /* Pointer to entry symbol table */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    symtbl_ptr = (*tpa_ptr).var_ptrps;  /* Get pointer to variable  */

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking Variable Datatype\n");
    }
#endif

    /*
     *    Check to see if the variable is in the symbol table and if so whether
     *    the data type is the same as that specified in the TPARSE argumnet block
     *    This is firat check for variable - may be record.
     */
    if (((*tpa_ptr).var_ptrps == NULL)  /* If variable NOT in table */
            ||  ((*symtbl_ptr).var_typel == 0)  /* ...or data type not specified */
            ||  ((*symtbl_ptr).var_typel == (*argblk_ptr).paraml))
        /* ...or data types match   */
    {
        return PSL_NORMAL;      /* ...return success        */
    }

    else
    {
        return 0;           /* return failure       */
    }
}
/*
 * 
 */


/************************************************************************/
/*                                  */
/*    CHK_VARIABLE - Check for Variable                 */
/*    ============                          */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL If defined variable              */
/*      Zero Otherwise                      */
/*    Pointers: Moved to end of variable name               */
/*    Inputs:   VARIABLES_TABLE = Variables Symbol Table        */
/*      VAR_CNT = Variable count                */
/*    Outputs:  VAR_PTR = Variable's Symbol Table entry address     */
/*                                  */
/*    This routine scans the variable symbol table to determine if the  */
/*    current token (which has already been proved to be a dataname)    */
/*    is in fact a variable.                        */
/*                                  */
/*    If the variable has been declared the address of its entry in */
/*    the Variables Symbol Table is returned in VAR_PTR, otherwise  */
/*    VAR_PTR is set to zero and a failure status is returned.      */
/*                                  */
/*    NOTE: A variable in this instance may not ne a constant       */
/*                                  */
/************************************************************************/

unsigned long   CHK_VARIABLE (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */
    struct   symtbl  *symtbl_ptr;       /* Pointer variable symbol table */
    unsigned long    isub;          /* Loop control variable    */
    long    comp_result;       /* Result of comparison     */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    symtbl_ptr = (*tpa_ptr).variable_begps;
    /* Variables symbol table   */

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking for Variable Name Defined\n");
    }
#endif

    /*
     *    Scan the variables symbol table until we reach the end or find a match
     *    to this dataname. Not that we do not count a constant as being a match
     */
    for (isub=0; isub < (*tpa_ptr).var_cntl; isub++, symtbl_ptr++)
    {
        if ((*argblk_ptr).tokencntl == (*symtbl_ptr).var_nlengthl)
        {               /* If the lengths are the same  */
            comp_result = memcmp ((*argblk_ptr).token_ptrpc,
                                  (*symtbl_ptr).var_namec16,
                                  (*symtbl_ptr).var_nlengthl);
            /* Compare the 2 variables  */
            if (comp_result == 0)   /* ...if names match        */
            {

                if (((*symtbl_ptr).var_typel != DTYP_CONCHARACTER)
                        &&  ((*symtbl_ptr).var_typel != DTYP_CONINTEGER))
                {
                    goto found_var; /* ...and if not a constant then */
                    /* ...jump as found the variable */
                }
            }
        }
    }

    /*
     *    We did not find a match for this variable so return an error status
     *    after resetting the string pointer/string count/token count to before
     *    the dataname.
     */
    RESET_TOKEN (argblk_ptr);       /* ...reset the token pointer   */
    return 0;               /* ...and return error status   */

    /*
     *    We found a match for the variable. Setup the symbol table address in the
     *    TPARSE data and return a success status.
     */
found_var:
    (*tpa_ptr).var_ptrps = symtbl_ptr;  /* Save symbol table entry address */
    return PSL_NORMAL;          /* ...and return success    */
}

/************************************************************************/

/*    CHK_VARIANT_REQUIRED - Check Script Variant required      */
/*    ====================                      */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_variant_REQUIRED (argblk_ptr)            */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL if variant in list of required       */
/*      0 otherwise                     */
/*    Pointers: Unchanged                       */
/*    Inputs:   VAR_CNT = Variable count                */
/*    Outputs:  IFDEF stack level marked as processed if variant reqd   */
/*      IFDEF_DISCARD_LEVEL = Current level         */
/*                                  */
/*    This routine is called when we have parsed a #IFDEF variant or    */
/*    a #IFNDEF variant to determine whether the variant specified  */
/*    has been specified on the command line. If it has then the code   */
/*    variant is required and we return success otherwise we return 0.  */
/*                                  */
/*    If the variant is required we will mark this IFDEF level as   */
/*    processed. This will prevent subsequent sections of the IFDEF */
/*    being processed (e.g. subsequent ELSE or ELIF's)          */
/*                                  */
/*    We also save the index of this level so that we can determine */
/*    when to stop throwing away the source lines           */
/*                                  */
/*    If an operating system variant or platform variant is encountered */
/*    a flag is set in the TOF header record data structure.        */
/*                                  */
/************************************************************************/

unsigned long   CHK_VARIANT_REQUIRED (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com  *com_ptr;     /* Common Data          */
    struct   scl_tpa  *tpa_ptr;     /* TPARSE Data          */
    struct   scl_cmd  *scl_cmd_ptr;     /* SCL Command Line Data    */
    struct   scl_tofh *tofh_ptr;        /* 'tof' header record data pointer */
    unsigned long     isub;         /* Loop contro variable     */
    char     **value_pptr;     /* Pointer to pointer to value  */
    char     *value_ptr;       /* Pointer to value     */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    scl_cmd_ptr = (*com_ptr).cmd_ptrps;
    tofh_ptr = (*tpa_ptr).tofh_begps;   /* 'tof' header record data */

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking Variant Name in Table\n");
    }
#endif
    /*
     *    If this variant is an operating system variant set the
     *    'operating system specific code present' flag.
     */
    if ((strncmp ((*tpa_ptr).variant_namec20, "%UNIX",5) ==0) ||
            (strncmp ((*tpa_ptr).variant_namec20, "%VMS",4) == 0) ||
            (strncmp ((*tpa_ptr).variant_namec20, "%WIN32",6) == 0))
    {
        (*tofh_ptr).tofh_osspcl = TOF_GENSTA_TRUE;
    }

    /*
     *    If this variant is a platform variant set the
     *    'platform specific code present' flag.
     */
    if ((strncmp ((*tpa_ptr).variant_namec20, "%VAX",4) == 0)   ||
            (strncmp ((*tpa_ptr).variant_namec20, "%ALPHA",6) == 0) ||
            (strncmp ((*tpa_ptr).variant_namec20, "%HP",3) == 0)    ||
            (strncmp ((*tpa_ptr).variant_namec20, "%IBM",4) == 0)   ||
            (strncmp ((*tpa_ptr).variant_namec20, "%NCR",4) == 0)   ||
            (strncmp ((*tpa_ptr).variant_namec20, "%OSF",4) == 0)   ||
            (strncmp ((*tpa_ptr).variant_namec20, "%SUN",4) == 0)   ||
            (strncmp ((*tpa_ptr).variant_namec20, "%NTI",4) == 0))
    {
        (*tofh_ptr).tofh_pltspcl = TOF_GENSTA_TRUE;
    }

    /*
     *    Loop round the list of variants specified to see if this is one of them
     */
    value_pptr = (*scl_cmd_ptr).variant_value_ptrppc;
    /* Get pointer to pointer to 1st val */
    for (isub = 0;
            isub < (*scl_cmd_ptr).variant_value_countl;
            isub++ ,value_pptr++)
    {
        value_ptr = *value_pptr;    /* Pointer to command line variant */

        if (strcmp (value_ptr ,(*tpa_ptr).variant_namec20) == 0) goto success;
    }

    /*
     *    No match found - We will be discarding ALL source lines at this level
     *    until we find an ELSE, ELIF or ENDIF at this level. To enable us to do
     *    that we must save the level now.
     */
    (*tpa_ptr).ifdef_discard_levell = (*tpa_ptr).ifdef_levell - 1;
    /* Save LAST used level index   */
    return 0;

    /*
     *    SUCCESS
     *    -------
     *
     *    We branch here when we have found a match to the variant - return success
     */
success:
    ;

    return PSL_NORMAL;
}

/*======================================================================*/
/*                                  */
/*    CHK_VAR_OPT - Check Variable Option               */
/*    ============                          */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_VAR_OPT (argblk_ptr)                 */
/*                                          */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL If specified variable option set         */
/*          Zero       If not                   */
/*    Pointers: Unchanged                       */
/*    Inputs:   PSLTPA$L_PARAM(AP) = Option 'bit' to be checked     */
/*              VAR_PTR = Variable's Symbol Table entry address     */
/*    Outputs:  None                            */
/*                                  */
/*    This routine returns success if the variable, whose Symbol Table  */
/*    entry address is held in VAR_PTR on entry, has the option     */
/*    passed in PSLTPA$L_PARAM(AP) set, otherwise it returns failure.   */
/*                                  */
/*======================================================================*/

unsigned long   CHK_VAR_OPT  (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */
    struct   symtbl  *var_ptr;      /* Pointer to variable      */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    var_ptr = (*tpa_ptr).var_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking Variable Options Bitmask\n");
    }
#endif

    /*
     *    Check to see if the required option has been specified.
     */
    if (((*var_ptr).var_optl & (*argblk_ptr).paraml) != 0)
        /* If option specified      */
    {
        return PSL_NORMAL;      /* ...then return success   */
    }

    else                /* ...otherwise         */
    {
        return 0;           /* Return an error status   */
    }
}

/*======================================================================*/
/*                                  */
/*    CHK_VAR_OPTS - Check Variable Option              */
/*    ============                          */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_VAR_OPTS (argblk_ptr)                */
/*                                          */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL  If valid variable option        */
/*          Zero        If not                  */
/*    Pointers: Unchanged                       */
/*    Inputs:   PSLTPA$L_PARAM(AP) = Option 'bit' to be checked     */
/*          DEF_OPTL = Options specified bitmask            */
/*          DEF_OPT_SPECIFIED = Bitmask indicating variable     */
/*                      options currently set.      */
/*    Outputs:  DEF_OPT_SPECIFIED = 'PSLTPA$L_PARAM(AP)' bit set    */
/*                      if not set already.         */
/*                                  */
/*    This routine is called on parsing a variable option. It checks    */
/*    that:                             */
/*                                  */
/*    1. If this is a RANDOM or REPEATABLE RANDOM option, a value   */
/*   list or range has been defined.                */
/*    2. If this is a FILE option, no value list or range have been */
/*   specified AND the variable is NOT an array variable AND    */
/*   is NOT part of a record AND none of the script parameter   */
/*   options have been specified.                   */
/*    3. If this is a script parameter option, the variable is NOT a    */
/*   file variable.                         */
/*    4. The option has NOT been set already.               */
/*                                  */
/*    If 'all is well' the routine returns a success status,        */
/*    otherwise an error status is returned.                */
/*                                      */
/*======================================================================*/

unsigned long   CHK_VAR_OPTS  (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking Variable Options Valid    \n");
    }
#endif

    /*
     *    If this is a random option, check that a value range/list is defined.
     */
    if ((*argblk_ptr).paraml == VOPT_RANDOM)
    {
        if (((*tpa_ptr).def_optl & (VOPT_LIST | VOPT_RANGE)) == 0)
        {
            return 0;           /* ...return an error status    */
        }
    }

    /*
     *    If this is a FILE option, check that:
     *    1. There is no value range/list defined.
     *    2. This is not an array variable.
     *    3. This variable is not part of a record definition.
     *    4. None of the script parameter options have been specified.
     */
    if ((*argblk_ptr).paraml == VOPT_FILE)
    {
        if ((((*tpa_ptr).def_optl & (VOPT_LIST | VOPT_RANGE)) != 0)
                ||  ((*tpa_ptr).array_dimcntl != 0)
                ||  ((*tpa_ptr).def_recidxl != 0)
                ||  (((*tpa_ptr).def_optl & (VOPT_ECHO | VOPT_PROMPT | VOPT_UPPERCASE)) != 0))
        {
            return 0;           /* ...return an error status */
        }
    }

    /*
     *    If this is a script parameter option, check that the variable is NOT a file
     *    variable.
     */
    if (((*argblk_ptr).paraml & (VOPT_ECHO | VOPT_PROMPT | VOPT_UPPERCASE)) != 0)
    {
        if (((*tpa_ptr).def_optl & VOPT_FILE) != 0)
        {
            return 0;           /* ...return an error status */
        }
    }

    /*
     *    Now ensure that the option has not already been specified.
     */
    if (((*tpa_ptr).def_opt_specifiedl & (*argblk_ptr).paraml) != 0)
        /* If option already specified  */
    {
        return 0;           /* Return an error status   */
    }

    else                /* ...otherwise         */
    {
        (*tpa_ptr).def_opt_specifiedl |= (*argblk_ptr).paraml;
        /* ...show option specified now */
    }

    return PSL_NORMAL;          /* ...and return success regardless */
}

/************************************************************************/
/*                                  */
/*    CHK_VAR_SCOPE - Check variabel scope              */
/*    ==============                            */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_VAR_SCOPE (argblk_ptr)               */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL  If valid variable option        */
/*      Zero        If not                  */
/*    Pointers: Unchanged                       */
/*    Inputs:   PSLTPA$L_PARAM(AP) = Specified addressing mode      */
/*      DEF_OPT_SPECIFIED = Bitmask indicating variable     */
/*          options currently set.              */
/*    Outputs:  DEF_OPT_SPECIFIED = VOPT_SCOPE bit set if not       */
/*          set already.                    */
/*                                  */
/*    This routine is called on parsing a variable scope option. It */
/*    examines the bitmask DEF_OPT_SPECIFIED to see if a variable   */
/*    scope option has been specified already, returning a failure  */
/*    status if it has. If the scope option is THREAD, GLOBAL or    */
/*    TESTWIDE the routine also checks that no value list/range has */
/*    been specified since these arte invalid for thses types of    */
/*    variable.                             */
/*                                  */

/*    Checks scope of variables in records - that they are the same as  */
/*    that specified in record header.                                  */
/*                                                                      */
/************************************************************************/

unsigned long   CHK_VAR_SCOPE (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;


    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking for SCOPE option specified\n");
    }
#endif

    /*
     *    Check to see if we have already specified the SCOPE and report an error
     *    if we have.
     */
    if (((*tpa_ptr).def_opt_specifiedl & VOPT_SCOPE) != 0)
    {
        return 0;           /* Return failure - scope set   */
    }

    (*tpa_ptr).def_opt_specifiedl =
        ((*tpa_ptr).def_opt_specifiedl | VOPT_SCOPE);
    /* Show scope specified now */
    /*
     *    If we are defining a shareable variable (i.e. a thread, global or testwide
     *    variable) do not allow value lists or ranges.
     */

    if ((((*argblk_ptr).paraml == ADDR_GLOBAL)
            /* If this is a GLOBAL variable */
            ||   ((*argblk_ptr).paraml == ADDR_THREAD)
            /* ...or a THREAD variable  */
            ||   ((*argblk_ptr).paraml == ADDR_TESTWIDE))
            /* ...or a TESTWIDE variable    */
            &&  (((*tpa_ptr).def_optl & (VOPT_LIST | VOPT_RANGE)) != 0))
        /* ...and Value list/range spec */
    {
        return 0;            /* Return an error status   */
    }

    /*
     * 
     *
     *  Code to check for errors when specifying record variables scope.
     *  (If in record header no checks necessary.)
     */

    if (((*tpa_ptr).def_recidxl != 0)        /* If processing a record      */
            && ((*tpa_ptr).in_rec_headl == 0)       /*  And not in record header   */
            && ((*argblk_ptr).paraml != (*tpa_ptr).rec_addrl))
        /* Error if scope not same as  */
        /*  record header scope        */
    {
        REPORT_ERROR (argblk_ptr, SCL_INVRECSCOPE);
    }
    else
    {

        /*
         * Setting addresses
         */

        if (((*tpa_ptr).def_recidxl != 0)           /* If in a record */
                &&  ((*tpa_ptr).in_rec_headl != 0))          /* And in record header */
        {
            (*tpa_ptr).rec_addrl = (*argblk_ptr).paraml;
            (*tpa_ptr).def_addrl= (*tpa_ptr).rec_addrl;
            /* Save addresses*/
        }


        if (((*tpa_ptr).def_recidxl != 0)       /* If in record body */
                &&  ((*tpa_ptr).in_rec_headl == 0))     /* But not in header */
        {
            (*tpa_ptr).def_addrl= (*tpa_ptr).rec_addrl;
            /* Default is record scope*/
        }
        else                                       /* Not specifying records */
        {
            (*tpa_ptr).def_addrl = (*argblk_ptr).paraml;
            /* Save addressing mode*/
        }

    }
    /*
     * 
     */
    return PSL_NORMAL;         /* ...and return success    */
}

/************************************************************************/
/*                                  */
/*    CHK_VPO_MODIFIER - Check VP$OUTPUT Modifier in range      */
/*    =================                         */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_VPO_MODIFIER (argblk_ptr)                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   PSLTPA$L_NUMBER(AP) = Character modifier to be checked  */
/*    Outputs:  INT_VALUE                       */
/*                                  */
/*    This routine checks that the character modifier value passed in   */
/*    PSLTPA$L_NUMBER(AP) is within the range 1 to 512 (decimal). It    */
/*    reports a 'Character Modifier Out Of Range' error if it is not.   */
/*                                  */
/************************************************************************/

unsigned long   CHK_VPO_MODIFIER (struct tpaarg *argblk_ptr)

{
    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking VP$OUTPUT Modifier %ld\n"
                          ,(*argblk_ptr).numberl);
    }
#endif

    /*
     *    Ensure that the value entered is in the range 1 to 512
     */
    if (((*argblk_ptr).numberl == 0)    /* If number not between 1 and 512 */
            ||  ((*argblk_ptr).numberl > 512))
    {
        REPORT_ERROR (argblk_ptr, SCL_VPOOUTRANGE);
        /* Report invalid VP$OUTPUT mod */
    }

    return PSL_NORMAL;
}

/************************************************************************/
/*                                  */
/*    CHK_VPS_MODIFIER - Check VP$SCREEN Modifier in range      */
/*    =================                         */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_VPS_MODIFIER (argblk_ptr)                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   INT_VALUEL - Modifier value to be tested        */
/*      PSLTPA$L_NUMBER(AP) = Address of OPD for value      */
/*    Outputs:  None                            */
/*                                  */
/*    This routine checks that the character modifier value passed in   */
/*    INT_VALUEL is within the permitted range for a VP$SCREEN      */
/*    modifier. It reports a 'Character Modifier Out Of Range' error    */
/*    if it is not.                         */
/*                                  */
/************************************************************************/

unsigned long   CHK_VPS_MODIFIER (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct  scl_com  *com_ptr;      /* Common Data */
    struct  scl_tpa  *tpa_ptr;      /* TPARSE Data */
    struct  scl_tofh *tofh_ptr;     /* Pointer to 'tof' header data */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;   /* Pointer to common data   */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* Pointer to TPARSE data   */
    tofh_ptr = (*tpa_ptr).tofh_begps; /* Pointer to 'tof' header data */

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking VP$SCREEN Modifier %ld\n"
                          ,(*tpa_ptr).int_valuel);
    }
#endif

    /*
     *    A value of zero is invalid in all cases.
     */
    if ((*tpa_ptr).int_valuel <= 0) goto out_range;
    /* If number <= 0 report error  */

    /*
     *    Check to see which VP$SCREEN modifier we are specifying. We can tell this
     *    by looking at the passed parameter which will be in the range 1-4
     */
    if ((*argblk_ptr).paraml == 1)  /* If first modifier (line 1)   */
    {
        if ((*tpa_ptr).int_valuel > (long) (*tofh_ptr).tofh_sline_w)
        {
            goto out_range;     /* If > max line value then error */
        }
    }

    else if ((*argblk_ptr).paraml == 2) /* If second modifier (line 2)  */
    {

        /*
         *    This operand is the second line number and so we check to see if it is
         *    greater than the maximum line number permitted
         */
        if ((*tpa_ptr).int_valuel > (long) (*tofh_ptr).tofh_sline_w)
            /* If > max line value      */
        {
            goto out_range;     /* ...Report out of range   */
        }
    }

    else if ((*argblk_ptr).paraml == 3) /* If third  modifier (column 1) */
    {
        if ((*tpa_ptr).int_valuel > (long) (*tofh_ptr).tofh_scols_w)
        {
            goto out_range;     /* If > max columns value then error */
        }
    }

    else if ((*argblk_ptr).paraml == 4) /* If fourth modifier (column 2) */
    {

        /*
         *    This operand is the second column number and so we check to see if it is
         *    greater than the maximum column number permitted.
         */
        if ((*tpa_ptr).int_valuel > (*tofh_ptr).tofh_scols_w)
            /* If > max column value    */
        {
            goto out_range;     /* ...Report out of range   */
        }
    }

    return PSL_NORMAL;          /* All o.k. return success  */

    /*
     *    The supplied modifier is outside the permitted range. Report this.
     */
out_range:
    REPORT_ERROR (argblk_ptr, VTEST_INVVPSMOD);
    /* Report invalid VP$SCREEN mod */
    return PSL_NORMAL;          /* ...but return success    */
}

/************************************************************************/
/*                                  */
/*    RESET_TOKEN - Reset TPARSE token pointers             */
/*    ============                          */
/*                                  */
/*    Type: Service routine                     */
/*    Status:   Not applicable                      */
/*    Pointers: Reset to indicate one character token           */
/*    Inputs:   TPARSE Argument block.                  */
/*    Outputs:  None                            */
/*                                  */
/*    This routine is used to reset the token and string fields in  */
/*    the TPARSE argument block. It is called after it has been     */
/*    decided to backout any token expansion that has been performed.   */
/*    The original token must be a single character token.      */
/*                                  */
/*    Tests show that this routine may not be required, since TPARSE    */
/*    may correctly reset the pointers. However, I am not sure that */
/*    this is always true and feel that it is a lot safer to undo   */
/*    the token expansion explicitly.                   */
/*                                  */
/*----------------------------------------------------------------------*/

static void RESET_TOKEN (struct tpaarg *argblk_ptr)

{
    (*argblk_ptr).string_ptrpc = (*argblk_ptr).token_ptrpc;
    /* Reset STRING pointer     */
    (*argblk_ptr).string_ptrpc++;   /* Increment string pointer */
    (*argblk_ptr).stringcntl += (*argblk_ptr).tokencntl;
    /* Reset string count       */
    (*argblk_ptr).stringcntl--;     /* Decrement string counter */
    (*argblk_ptr).tokencntl = 1;    /* Reset token counter      */
    return;
}

/*======================================================================*/

/*    CHK_SUB_PARAM  - Check subroutine Parameter count         */
/*    ================                          */
/*                                  */
/*    Calling Format:                           */
/*  iret = CHK_SUB_PARAM (argblk_ptr)               */
/*                                  */
/*    Type: TPARSE action routine                   */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr - Pointer to the tparse argument block.  */
/*    Outputs:  Subroutine Parameter count incremented          */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   CHK_SUB_PARAM (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Do not encode anything if we have had an error in the command.
     */
    if ((*tpa_ptr).error_in_commandn) goto end;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("\nChecking subroutine parameter count\n");
    }
#endif

    /*
     *    Check to determine if we have already coded the maximum number of
     *    parameters and report an error if we have.
     */
    if ((*tpa_ptr).subprm_cntl >= MAX_PARAMETERS)
    {                   /* No more parameters allowed ? */
        REPORT_ERROR (argblk_ptr, SCL_MAXSUBPRM);
    }

    else
    {
        (*tpa_ptr).subprm_cntl += 1;/* Show 1 more parameter    */
    }

end:
    ;
    return PSL_NORMAL;          /* ...and return success    */
}
/*
 * 
 */

/************************************************************************/
/*                                  */
/*    CHK_TESTCASE - Check For a TEST-CASE active           */
/*    ===============                           */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_TESTCASE (argblk_ptr)                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL if OK                    */
/*      0 - Otherwise                       */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk.paraml                       */
/*    Outputs:  None                            */
/*                                  */
/*    This routine checks to see if a test-case is Active and if it is  */
/*    returns an success status.                    */
/*                                  */
/************************************************************************/

unsigned long   CHK_TESTCASE (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking for TEST-CASE active\n");
    }
#endif

    /*
     *    If this is a START TEST-CASE command, and a test-case is already active
     *    return a failure status.
     */
    if (argblk_ptr->paraml == CMND_START_TESTCASE)
    {
        if ((tpa_ptr->testcase_flagsl & TCAS_ACTIVE_TC) != 0) return 0;
    }
    /*
     * 
     *    We do not validate the position of test case statements.
     *    Instead we leave an errors of this kind to be captured by the runtime
     *    component as we are unable to second gues the exact script execution
     *    path and therefore cannot be certain that test case statements are/not
     *    being defined outside of a test case.
     * 
     */
    else
    {

        /*
         *  Reset the goto flag.  This is used to determine when the optional
         *      GOTO clause has been specified.
         */
        tpa_ptr->testcase_flagsl &= ~TCAS_GOTO_TC;
        tpa_ptr->test_fgl = 0;
    }
    /*
     *    Everything seems to be OK so return success.
     */
    return PSL_NORMAL;
}


/************************************************************************/

/*    CHK_CALL_SCRIPT - Check to see if the current variable is an arra */
/*    Calling Format:                           */
/*  iret = CHK_CALL_SCRIPT (argblk_ptr)             */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL if OK                    */
/*      0 - Otherwise                       */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk.paraml                       */
/*    Outputs:  None                            */
/*                                  */
/*    This routine checks to see if a call script is being parsed       */
/*    and if it is returns an failure status. This causes TPARSE to     */
/*    avoid problems of arrays and call scripts.            */
/*                                  */
/************************************************************************/

unsigned long   CHK_CALL_SCRIPT (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking to see if in call script command\n");
    }
#endif


    if ((*tpa_ptr).script_flagl == 0)
    {
        return PSL_NORMAL;
    }
    else
    {
        (*tpa_ptr).script_flagl = 0;        /* Reset call script flag */
        return 0;
    }

}
/*
 * 
 */


/************************************************************************/
/*                                  */
/*    CHK_MAP_RECORD - Check the record specified for the MAP command   */
/*    =================                         */
/*                                          */
/*    Calling Format:                           */
/*  iret = CHK_MAP_RECORD (argblk_ptr)              */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr = Pointer to TPARSE Argument Block Structure */
/*    Outputs:  None                            */
/*                                  */
/*    This routine checks the record specified for the MAP command to   */
/*    ensure that it is valid. The record MUST have a character variable*/
/*    as its first field that must be no more than 64 characters long   */
/*                                  */
/************************************************************************/

unsigned long   CHK_MAP_RECORD  (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */
    struct scph *scp_ptr;       /* Pointer to SCP header record */
    struct   symtbl  *symtbl_ptr;       /* Pointer variable symbol table */
    unsigned long    isub;          /* Loop control variable    */
    long    comp_result;       /* Result of comparison     */
    long    *pFormatArray;     /* Format Array for Record  */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    scp_ptr = (*tpa_ptr).header_begps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking Validity of MAP record\n");
    }
#endif

    /*
     *    We first need to locate the symbol table entry for this record as this
     *    contains the index of the record with which we can scan the record table.
     */
    symtbl_ptr = (*tpa_ptr).variable_begps;
    /* Variables symbol table   */

    /*
     *    Scan the variables symbol table until we reach the end or find a match
     *    to this dataname. Not that we do not count a constant as being a match
     */
    for (isub=0; isub < (*tpa_ptr).var_cntl; isub++, symtbl_ptr++)
    {
        if ((*argblk_ptr).tokencntl == (*symtbl_ptr).var_nlengthl)
        {               /* If the lengths are the same  */
            comp_result = memcmp ((*argblk_ptr).token_ptrpc,
                                  (*symtbl_ptr).var_namec16,
                                  (*symtbl_ptr).var_nlengthl);
            /* Compare the 2 variables  */
            if (comp_result == 0)   /* ...if names match        */
            {

                if (((*symtbl_ptr).var_typel != DTYP_CONCHARACTER)
                        &&  ((*symtbl_ptr).var_typel != DTYP_CONINTEGER))
                {
                    goto found_var; /* ...and if not a constant then */
                    /* ...jump as found the variable */
                }
            }
        }
    }


    /*
     *    Now that we have found the symbol table entry for this record we must 
     *    locate its format array so that we can check the size/type of the first
     *    variable in the record.
     */
found_var:
    ;
    pFormatArray = (*symtbl_ptr).var_start_rec_arraypl;
    if ((*pFormatArray <= 0) || (*pFormatArray > 64))
    {
        REPORT_ERROR (argblk_ptr, SCL_INVMAPRECORD);
        /* Pass to underline routine    */
    }


    /*
     *   ...and return success
     */
    return PSL_NORMAL;          /* ...always return success */
}

